/* eslint-disable prefer-const */
import { useDebounce } from "lib/utils";
import { ApiError } from "pages/api/_service";
import {
    AllowedReturn,
    SearchQuery,
    SearchResults,
} from "pages/api/admin/search";
import { Dispatch, SetStateAction, useEffect, useState } from "react";
import { Alert, Button, Form, InputGroup, Table } from "react-bootstrap";

import {
    date_str,
    fetchSearchResults,
    fetchUsersGroups,
    UsersGroups,
} from "./utils";

import { GroupData, UserData } from "database/services/service.typings";

import styles from "./admin_panel.module.scss";

/** Group editor used in admin panel
 * Allows to edit the group of a user and to add/remove users from a group
 *
 * Two searchable lists with ticks
 *
 * @returns
 */
export function GroupManagementEditor() {
    // Set by id
    const [selectedGroup, setSelectedGroup] = useState<GroupData>(undefined);
    const [selectedUsers, setSelectedUsers] = useState<UserData[]>([]);

    // List of many to many relationships
    const [usersGroups, setUsersGroups] =
        useState<Array<UsersGroups>>(undefined);

    return (
        <div className={styles.groupManagementEditor}>
            <Selector
                type="group"
                select="single"
                selected={[selectedGroup]}
                setSelected={(g) => {
                    setSelectedGroup(g[0]);
                }}
            />
            <Selector
                type="user"
                selected={selectedUsers}
                setSelected={setSelectedUsers}
            />
            <GroupUsersManagement
                usersGroups={usersGroups}
                setUsersGroups={setUsersGroups}
                selectedGroup={selectedGroup}
                selectedUsers={selectedUsers}
            />
            <Apply usersGroups={usersGroups} />
        </div>
    );
}

/** Executes the selected query and update on the server
 * @param query
 * @returns
 */
async function applyUsersGroups(users_groups: UsersGroups[]): Promise<boolean> {
    return await fetch("/api/admin/set_users_groups", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify({ users_groups: users_groups }),
    }).then(async (res) => {
        if (res.ok) {
            return (await res.json()) as boolean;
        }

        const error = (await res.json()) as ApiError;
        throw new Error(error.details);
    });
}

function Apply({ usersGroups }) {
    const [success, setSuccess] = useState({
        success: false,
        error: false,
        message: "",
    });
    const [scale, setScale] = useState<string>("1");

    if (!usersGroups) {
        return null;
    }
    //Close button clears the results
    const close = () => {
        setSuccess((prev) => {
            return { ...prev, success: false, error: false };
        });
    };
    const close_btn = (
        <button className={"btn btn-light " + styles.close_btn} onClick={close}>
            <i className="bi bi-x"></i>
        </button>
    );

    return (
        <div className={styles.apply_btn}>
            <Button
                variant="primary"
                onClick={() => {
                    applyUsersGroups(usersGroups)
                        .then((res) => {
                            setSuccess({
                                success: res,
                                error: false,
                                message: "Permissions applied successfully",
                            });
                            setScale("1.03");
                            setTimeout(() => {
                                setScale("1");
                            }, 300);
                        })
                        .catch((err: string) => {
                            setSuccess({
                                success: false,
                                error: true,
                                message: err,
                            });
                            setScale("1.03");
                            setTimeout(() => {
                                setScale("1");
                            }, 300);
                        });
                }}
            >
                Apply
            </Button>
            <Alert
                variant="success"
                show={success.success}
                style={{ scale: scale, animation: "scale 2s ease-in-out 2s" }}
            >
                {close_btn}
                Permissions applied!
            </Alert>
            <Alert
                variant="danger"
                show={success.error}
                style={{ scale: scale, animation: "scale 2s ease-in-out 2s" }}
            >
                {close_btn}
                {String(success.message)}
            </Alert>
        </div>
    );
}

function GroupUsersManagement({
    usersGroups,
    setUsersGroups,
    selectedGroup,
    selectedUsers,
}: {
    usersGroups: Array<UsersGroups>;
    setUsersGroups: Dispatch<SetStateAction<Array<UsersGroups>>>;
    selectedGroup: GroupData;
    selectedUsers: UserData[];
}) {
    useEffect(() => {
        if (selectedGroup && selectedUsers && selectedUsers.length > 0) {
            // Get the usersGroups
            const user_ids = selectedUsers.map((u) => u.id);
            fetchUsersGroups([selectedGroup.id], user_ids).then((ug) => {
                setUsersGroups(ug);
            });
        } else {
            setUsersGroups(undefined);
        }
    }, [selectedGroup, selectedUsers, setUsersGroups]);

    let ret;
    if (!selectedGroup || !selectedUsers || selectedUsers.length == 0) {
        ret = (
            <p>
                Please select a group and at least one user to manage the
                relationship between them.
            </p>
        );
    } else if (!usersGroups) {
        ret = <p>Loading...</p>;
    } else {
        ret = (
            <UserGroupTable
                usersGroups={usersGroups}
                setUsersGroups={setUsersGroups}
                selectedGroup={selectedGroup}
                selectedUsers={selectedUsers}
            />
        );
    }

    return (
        <div className={styles.searchForm + " " + styles.groupUsersManagement}>
            <h3>Group Membership:</h3>
            {ret}
        </div>
    );
}

function UserGroupTable({
    selectedGroup,
    selectedUsers,
    usersGroups,
    setUsersGroups,
}: {
    selectedGroup: GroupData;
    selectedUsers: UserData[];
    usersGroups: Array<UsersGroups>;
    setUsersGroups: Dispatch<SetStateAction<Array<UsersGroups>>>;
}): JSX.Element {
    const rows = [];

    for (const user of selectedUsers) {
        const inGroup = usersGroups.some(
            (ug) =>
                ug.user_id == user.id &&
                ug.group_id == selectedGroup.id &&
                ug.member
        );

        //Get user
        rows.push(
            <tr key={user.id}>
                <td>{user.id}</td>
                <td>{user.email}</td>
                <td>
                    <Form.Check
                        type="switch"
                        checked={inGroup}
                        onChange={() => {
                            // Check if user and group is in usersGroups
                            const index = usersGroups.findIndex(
                                (ug) =>
                                    ug.user_id == user.id &&
                                    ug.group_id == selectedGroup.id
                            );
                            if (index == -1) {
                                // Add new entry
                                setUsersGroups([
                                    ...usersGroups,
                                    {
                                        user_id: user.id,
                                        group_id: selectedGroup.id,
                                        member: !inGroup,
                                    },
                                ]);
                            } else {
                                // Update entry
                                const newUsersGroups = [...usersGroups];
                                newUsersGroups[index].member = !inGroup;
                                setUsersGroups(newUsersGroups);
                            }
                        }}
                    />
                </td>
            </tr>
        );
    }

    return (
        <Table striped bordered hover>
            <thead>
                <tr>
                    <th>User Id</th>
                    <th>User email</th>
                    <th>Member of {selectedGroup.name}?</th>
                </tr>
            </thead>
            <tbody>{rows}</tbody>
        </Table>
    );
}

/** Allows to select on or multiple users
 */
function Selector({
    selected,
    setSelected,
    type,
    select = "multiple",
}: {
    selected: AllowedReturn[];
    setSelected: Dispatch<SetStateAction<AllowedReturn[]>>;
    type: "user" | "group";
    select?: "multiple" | "single";
}) {
    const [results, setResults] = useState<SearchResults>(undefined);
    const [searchState, setSearchState] = useState<SearchQuery>({
        query: "",
        type: type,
    });

    const debounced_query = useDebounce(searchState.query, 500);
    useEffect(() => {
        if (debounced_query) {
            fetchSearchResults(debounced_query, searchState.type).then(
                setResults
            );
        }
    }, [debounced_query, searchState.type]);

    const handleSearch = async (e) => {
        setSearchState((prev) => ({ ...prev, query: e.target.value }));
    };

    const h3 = type === "user" ? "User" : "Group";
    let headers;
    if (type === "user") {
        headers = ["Id", "Email", "Validated", "Created", "Updated"];
    } else {
        headers = ["Id", "Name", "Description", "Created", "Updated"];
    }

    return (
        <div className={styles.searchForm}>
            <h3>{h3}:</h3>
            <InputGroup>
                <Form.Control
                    type="text"
                    placeholder="Search..."
                    aria-label="search"
                    style={{ width: "70%" }}
                    value={searchState.query}
                    onChange={handleSearch}
                />
                <InputGroup.Text>{h3}</InputGroup.Text>
            </InputGroup>
            {select == "single" ? (
                <TableWithSingleSelect
                    headers={headers}
                    rows={results?.data}
                    selected={selected[0]}
                    setSelected={(an: AllowedReturn) => setSelected([an])}
                />
            ) : (
                <TableWithMultiSelect
                    headers={headers}
                    rows={results?.data}
                    selected={selected}
                    addRemoveSelected={(item: AllowedReturn) => {
                        // Add or remove
                        if (selected.some((i) => i.id === item.id)) {
                            setSelected(
                                selected.filter((i) => i.id !== item.id)
                            );
                        } else {
                            setSelected([...selected, item]);
                        }
                    }}
                ></TableWithMultiSelect>
            )}
        </div>
    );
}

function TableWithSingleSelect({
    headers,
    rows,
    selected,
    setSelected,
}: {
    headers: string[];
    rows: Array<AllowedReturn>;
    selected: AllowedReturn;
    setSelected: Dispatch<SetStateAction<AllowedReturn>>;
}) {
    if (!rows) {
        return null;
    }

    const selectedId = selected?.id;

    const body = rows.map((entity) => {
        const r = [];
        for (let [key, value] of Object.entries(entity)) {
            if (["created", "last_updated"].includes(key)) {
                value = date_str(value);
            }
            if (key == "configs") {
                continue;
            }
            if (value === null) {
                value = "";
            }
            r.push(<td key={key}>{String(value)}</td>);
        }

        return (
            <tr
                key={entity.id}
                className={selectedId == entity.id ? "selected" : ""}
                onClick={() => {
                    if (selectedId == entity.id) {
                        setSelected(undefined);
                    } else {
                        setSelected(entity);
                    }
                }}
            >
                {r}
            </tr>
        );
    });

    return (
        <div className={styles.searchResults}>
            <Table striped bordered hover>
                <thead>
                    <tr>
                        {headers.map((header) => (
                            <th key={header}>{header}</th>
                        ))}
                    </tr>
                </thead>
                <tbody>{body}</tbody>
            </Table>
        </div>
    );
}

function TableWithMultiSelect({
    headers,
    rows,
    selected,
    addRemoveSelected,
}: {
    headers: string[];
    rows: Array<AllowedReturn>;
    selected: AllowedReturn[];
    addRemoveSelected: (id: AllowedReturn) => void;
}) {
    if (!rows) {
        return null;
    }

    const selectedIds = selected.map((s) => s.id);

    const body = rows.map((entity) => {
        const r = [];
        for (let [key, value] of Object.entries(entity)) {
            if (["created", "last_updated"].includes(key)) {
                value = date_str(value);
            }
            if (key == "configs") {
                continue;
            }
            if (value === null) {
                value = "";
            }
            r.push(<td key={key}>{String(value)}</td>);
        }

        r.push(
            <td key={"cb"}>
                <Form.Check
                    type="checkbox"
                    checked={selectedIds.includes(entity.id)}
                    onChange={() => addRemoveSelected(entity)}
                ></Form.Check>
            </td>
        );

        return <tr key={entity.id}>{r}</tr>;
    });

    return (
        <div className={styles.searchResults}>
            <Table striped bordered>
                <thead>
                    <tr>
                        {headers.map((h, i) => (
                            <th key={i}>{h}</th>
                        ))}
                        <th>
                            <i className="icon bi-check2-square"></i>
                        </th>
                    </tr>
                </thead>
                <tbody>{body}</tbody>
            </Table>
        </div>
    );
}
