import {
    createContext,
    Dispatch,
    SetStateAction,
    useMemo,
    useState,
} from "react";

import { BrowserContextMenu, ContextMenuState } from "./BrowserContextMenu";
import { BrowserNavbar } from "./BrowserNavbar";
import { BrowserToolbar } from "./BrowserToolbar";
import { FileView } from "./FileView";

import styles from "./BooksBrowser.module.scss";

/** Wrapper that exports a full book browser with all the components.
 */
export function FullBooksBrowser({
    rootFolderId,
    fileMap,
}: {
    rootFolderId: FileID;
    fileMap: FileMap;
}) {
    return (
        <BooksBrowser rootFolderId={rootFolderId} fileMap={fileMap}>
            <BrowserNavbar />
            <BrowserToolbar />
            <FileView />
            <BrowserContextMenu />
        </BooksBrowser>
    );
}

interface BrowserContext {
    // Root folder and file map
    rootFolderId: FileID;
    fileMap: FileMap;

    // Current folder
    currentPath: FileData[];
    currentFolder: FileData;
    currentFolderId: FileID;
    setCurrentFolderId: Dispatch<SetStateAction<FileID>>;
    ascendToParentFolder: () => void;
    moveFileToFolder: (fileId: FileID, folderId: FileID) => void;

    // Order of the books
    order: Order;
    setOrder: Dispatch<SetStateAction<Order>>;
    fileMapOrder: FileID[];
    nHidden: number;

    // Current view
    view: View;
    setView: Dispatch<SetStateAction<View>>;

    // Context menu
    contextMenuState: ContextMenuState;
    setContextMenuState: Dispatch<SetStateAction<ContextMenuState>>;
    onContextMenuFile: (
        e: React.MouseEvent<HTMLDivElement, MouseEvent>,
        file: FileData
    ) => void;

    deleteFolder: (folderId: FileID) => void;
}

export const BrowserContext = createContext<BrowserContext>(undefined);

/** BooksBrowser is a context provider for the general setup of the book browser.
 * It provides the root folder id, the file map, and the current folder id.
 * It also provides a function to set the current folder id and a way to ascend
 * to the parent folder.
 */
export function BooksBrowser({
    rootFolderId,
    fileMap,
    children,
}: {
    rootFolderId: FileID;
    fileMap: FileMap;
    children: React.ReactNode;
}) {
    /** Create parent and childrenIds if they don't exist! Also create thubmnail link
     */

    fileMap = useMemo(() => {
        const newFileMap = { ...fileMap };
        const files = Object.values(newFileMap);
        for (const file of files) {
            if (file.childrenIds) {
                file.childrenIds.forEach((childId) => {
                    const child = newFileMap[childId];
                    if (!child.parentId) {
                        child.parentId = file.id;
                    }
                });
            } else {
                file.childrenIds = [];
            }

            if (file.parentId) {
                const parent = newFileMap[file.parentId];
                if (
                    !parent.childrenIds ||
                    !parent.childrenIds.includes(file.id)
                ) {
                    parent.childrenIds = [file.id];
                }
            }
            // Parse all dates (undefined parses to invalid date...)
            file.last_updated = file.last_updated
                ? new Date(file.last_updated)
                : undefined;
            file.created = file.created ? new Date(file.created) : undefined;

            // Create thumbnail link
            if (!file.isDir && !file.thumbnail) {
                file.thumbnail = `/api/books/${file.id}/cover_preview`;
            }
        }

        // We do this afterwards because we want to dates to be set already
        for (const file of files) {
            // Last updated of folders is the last updated of the most recently
            // updated file in the folder.
            if (file.isDir) {
                let last_updated = file.created;
                for (const childId of file.childrenIds) {
                    const child = newFileMap[childId];
                    if (child.last_updated > last_updated) {
                        last_updated = child.last_updated;
                    }
                }
                file.last_updated = last_updated;
            }
        }

        return newFileMap;
    }, [fileMap]);

    /** The current folder id is the id of the folder that is currently
     * being displayed.
     */
    const [currentFolderId, setCurrentFolderId] = useState(rootFolderId);
    const currentFolder = fileMap[currentFolderId];
    const currentPath = useMemo(() => {
        const path: FileData[] = [];
        let currentId = currentFolderId;
        while (currentId) {
            path.push(fileMap[currentId]);
            if (currentId === rootFolderId) {
                break;
            }
            currentId = fileMap[currentId].parentId;
        }
        return path.reverse();
    }, [currentFolderId, fileMap, rootFolderId]);

    // ascendToParentFolder ascends to the parent folder of the current folder.
    const ascendToParentFolder = () => {
        const currentFolder = fileMap[currentFolderId];
        if (currentFolder.parentId) {
            setCurrentFolderId(currentFolder.parentId);
        }
    };

    /** Sort the books depending on the state
     */
    const [order, setOrder] = useState<Order>({
        by: "last_updated",
        asc: false,
        foldersFirst: true,
        showFinished: true,
        search: "",
    });
    const { fileMapOrder, nHidden } = useMemo(() => {
        let fileOrder = Object.values(fileMap);

        // Filter out all files not in current folder
        fileOrder = fileOrder.filter(
            (file) =>
                file.parentId === currentFolderId ||
                currentFolder.childrenIds.includes(file.id)
        );

        // Filter out folders and finished books if necessary
        if (!order.showFinished) {
            fileOrder = fileOrder.filter((file) => !file.finished);
        }
        let folders = [];
        if (order.foldersFirst) {
            folders = fileOrder.filter((file) => file.isDir);
            fileOrder = fileOrder.filter((file) => !file.isDir);
        }

        // Filter out all files that don't match the search
        let nHidden = 0;
        if (order.search !== "") {
            nHidden = fileOrder.length;
            fileOrder = fileOrder.filter((file) =>
                file.name.toLowerCase().includes(order.search.toLowerCase())
            );
            nHidden -= fileOrder.length;
        }

        fileOrder = fileOrder.sort(compareFunctionFactory(order));

        // Insert folders at the beginning if necessary
        if (order.foldersFirst) {
            folders = folders.sort(compareFunctionFactory(order));
            fileOrder = [...folders, ...fileOrder];
        }

        return {
            fileMapOrder: fileOrder.map((file) => file.id),
            nHidden,
        };
    }, [currentFolder.childrenIds, currentFolderId, fileMap, order]);

    /** Current view (list,grid)
     */
    const [view, setView] = useState<View>("list");

    const [, update] = useState({});

    const moveFileToFolder = (fileId: FileID, folderId: FileID) => {
        const file = fileMap[fileId];
        const oldParent = fileMap[file.parentId];
        const newParent = fileMap[folderId];

        // Remove file from old parent
        oldParent.childrenIds = oldParent.childrenIds.filter(
            (childId) => childId !== fileId
        );

        // Add file to new parent
        newParent.childrenIds.push(fileId);
        file.parentId = folderId;

        // Force update
        update({});

        // Update on server
        fetch(`/api/folders/move`, {
            method: "POST",
            headers: {
                "Content-Type": "application/json",
            },
            body: JSON.stringify({
                type: file.isDir ? "folder" : "book",
                itemID: fileId.replace("folder_", ""),
                moveToID: folderId.replace("folder_", ""),
            }),
        }).then(async (res) => {
            const data = await res.json();
            if (!res.ok) {
                // Revert changes
                oldParent.childrenIds.push(fileId);
                newParent.childrenIds = newParent.childrenIds.filter(
                    (childId) => childId !== fileId
                );
                file.parentId = oldParent.id;
                update({});
                alert("Error moving file" + data.error);
            }
        });
    };

    const deleteFolder = (folderId: FileID) => {
        const folder = fileMap[folderId];
        const parent = fileMap[folder.parentId];

        // Move all children to parent
        for (const childId of folder.childrenIds) {
            const child = fileMap[childId];
            parent.childrenIds.push(childId);
            child.parentId = parent.id;
        }
        // Remove folder from parent
        parent.childrenIds = parent.childrenIds.filter(
            (childId) => childId !== folderId
        );

        // Remove folder from fileMap
        delete fileMap[folderId];

        // Update on server
        fetch(`/api/folders/delete`, {
            method: "POST",
            headers: {
                "Content-Type": "application/json",
            },
            body: JSON.stringify({
                folderID: folderId.replace("folder_", ""),
            }),
        }).then(async (res) => {
            const data = await res.json();
            if (!res.ok) {
                // Revert changes
                fileMap[folderId] = folder;
                if (parent) {
                    for (const childId of folder.childrenIds) {
                        moveFileToFolder(childId, folder.id);
                    }
                }
                alert("Error deleting folder" + data.error);
            }
            update({});
        });
    };

    /** Context menu for the files
     *
     */
    const [contextMenuState, setContextMenuState] =
        useState<ContextMenuState>(null);
    const onContextMenuFile = (
        e: React.MouseEvent<HTMLDivElement, MouseEvent>,
        file: FileData
    ) => {
        e.preventDefault();

        // TODO figure out way to hide context menu when clicking outside
        setContextMenuState({
            file,
            x: e.clientX,
            y: e.clientY,
            visible: true,
        });
    };

    const browserContext: BrowserContext = {
        rootFolderId,
        fileMap,

        currentPath,
        currentFolder,
        currentFolderId,
        setCurrentFolderId,
        ascendToParentFolder,
        moveFileToFolder,

        order,
        setOrder,
        fileMapOrder,
        nHidden,

        view,
        setView,

        contextMenuState,
        setContextMenuState,
        onContextMenuFile,
        deleteFolder,
    };

    return (
        <BrowserContext.Provider value={browserContext}>
            <div className={styles.BooksBrowser}>{children}</div>
        </BrowserContext.Provider>
    );
}

function compareFunctionFactory(
    order: Order
): (a: FileData, b: FileData) => number {
    return (a: FileData, b: FileData) => {
        if (order.by == "name") {
            // Using localeCompare to sort by name
            return order.asc
                ? b.name.localeCompare(a.name, undefined, {
                      ignorePunctuation: true,
                      sensitivity: "base",
                  })
                : a.name.localeCompare(b.name, undefined, {
                      ignorePunctuation: true,
                      sensitivity: "base",
                  });
        } else if (order.by == "created") {
            if (a.created < b.created) {
                return order.asc ? 1 : -1;
            } else if (a.created > b.created) {
                return order.asc ? -1 : 1;
            } else {
                return 0;
            }
        } else if (order.by == "last_updated") {
            if (a.last_updated < b.last_updated) {
                return order.asc ? -1 : 1;
            } else if (a.last_updated > b.last_updated) {
                return order.asc ? 1 : -1;
            } else {
                return 0;
            }
        } else if (order.by == "numPages") {
            if (a.numPages < b.numPages) {
                return order.asc ? 1 : -1;
            } else if (a.numPages > b.numPages) {
                return order.asc ? -1 : 1;
            } else {
                return 0;
            }
        }
    };
}
