import { fetcher, res_parse_json } from "lib/fetcher";
import { useRouter } from "next/router";
import { AllowedTypes } from "pages/api/admin/search";
import { PermissionACLDataResolved } from "pages/api/books/[id]/edit/get_acl";
import { useContext, useEffect, useRef, useState } from "react";
import { Alert, Button, Form, InputGroup, Table } from "react-bootstrap";
import useSWR from "swr";

import { EditBookContext } from ".";

import styles from "./AccessEditor.module.scss";

export const AccessEditor = () => {
    // Get owner useSWR(book.owner_user_id || book.owner_group_id);
    const { book, user, owner } = useContext(EditBookContext);

    const [acl, setAcl] = useState<PermissionACLDataResolved[]>([]);
    const { data } = useSWR<PermissionACLDataResolved[]>(
        `/api/books/${book.id}/edit/get_acl`,
        fetcher
    );

    // Update data if it changes
    useEffect(() => {
        if (data) {
            // Return unique entities from prev and new data
            setAcl((prev) => {
                const prev_filtered = prev.filter(
                    (prevEntity) =>
                        !data.find(
                            (dataEntity) =>
                                dataEntity.entity_id === prevEntity.entity_id &&
                                dataEntity.entity_type ===
                                    prevEntity.entity_type
                        )
                );
                return [...prev_filtered, ...data];
            });
        }
    }, [data, setAcl]);

    // Add entity to acl if it doesn't already exist
    const addToAcl = (entity: PermissionACLDataResolved) => {
        if (
            !acl.find(
                (a) =>
                    a.entity_id === entity.entity_id &&
                    a.entity_type === entity.entity_type
            )
        ) {
            setAcl([...acl, entity]);
        }
    };

    // Construct the rows for the table
    const rows = [];
    if (data && owner && acl.length > 0) {
        // Set owner at top
        const owner_perms = acl.find(
            (a) => a.entity_id == owner.id && a.entity_type == owner.entity_type
        );
        // Set all p to true
        owner_perms.pRead = true;
        owner_perms.pWrite = true;
        owner_perms.pDelete = true;
        owner_perms.pACL = true;
        rows.push(
            <EntityRow
                key={`owner`}
                entity={{
                    id: owner_perms.entity_id,
                    type: owner_perms.entity_type,
                    name: owner_perms.entity_name,
                }}
                permission={owner_perms}
                disabled
                note={
                    "Owner" +
                    (owner_perms.entity_id == user.id &&
                    owner_perms.entity_type == "user"
                        ? " & You"
                        : "")
                }
            />
        );

        // Set everything else below
        acl.filter((a) => {
            const isOwner =
                a.entity_id == owner.id && a.entity_type == owner.entity_type;
            return !isOwner;
        }).forEach((a) => {
            rows.push(
                <EntityRow
                    key={`${a.entity_type}-${a.entity_id}`}
                    entity={{
                        id: a.entity_id,
                        type: a.entity_type,
                        name: a.entity_name,
                    }}
                    permission={a}
                    setPermission={async (p) => {
                        // Set local state
                        setAcl((prev) => {
                            const newAcl = prev.map((a) => {
                                if (
                                    a.entity_id == p.entity_id &&
                                    a.entity_type == p.entity_type
                                ) {
                                    return p;
                                }
                                return a;
                            });
                            return newAcl;
                        });
                        // Update on server
                        await fetch(`/api/books/${book.id}/edit/set_acl`, {
                            method: "POST",
                            headers: {
                                "Content-Type": "application/json",
                            },
                            body: JSON.stringify({ acl: p }),
                        });
                    }}
                    disabled={a.entity_id == user.id}
                    note={
                        a.entity_id == user.id && a.entity_type == "user"
                            ? "You"
                            : undefined
                    }
                />
            );
        });
    }

    return (
        <div className={styles.container}>
            <h2>Access Control List</h2>
            <p>
                Use the forms below to manage the access to the book. Add an
                user with their email or group with their name via the search
                field to edit their permissions. By default the table shows all
                users and groups that have access to the book.
            </p>
            <AddEmailGroup addCB={addToAcl} />
            <Table striped bordered hover>
                <colgroup>
                    <col />
                    <col />
                    <col className={styles.col_switch} />
                    <col className={styles.col_switch} />
                    <col className={styles.col_switch} />
                    <col className={styles.col_switch} />
                </colgroup>
                <thead>
                    <tr>
                        <th>Name/Email</th>
                        <th>Type</th>
                        <th className={styles.th_switch}>pRead</th>
                        <th className={styles.th_switch}>pWrite</th>
                        <th className={styles.th_switch}>pDelete</th>
                        <th className={styles.th_switch}>pACL</th>
                    </tr>
                </thead>
                <tbody>{rows}</tbody>
            </Table>
        </div>
    );
};

function AddEmailGroup({
    addCB,
}: {
    addCB: (entity: PermissionACLDataResolved) => void;
}) {
    // Information to display feedback
    const [success, setSuccess] = useState({
        success: false,
        error: false,
        message: "",
    });
    function close() {
        setSuccess({
            success: false,
            error: false,
            message: "",
        });
    }
    const close_btn = (
        <button className={"btn btn-light " + styles.close_btn} onClick={close}>
            <i className="bi bi-x"></i>
        </button>
    );

    const [bg, setBg] = useState(undefined);
    useEffect(() => {
        //Flash feedback message if shown
        if (success.success || success.error) {
            setBg(success.error ? "#9b332b" : "green");
            setTimeout(() => {
                setBg(undefined);
            }, 200);
        }
    }, [success, setBg]);

    // Query
    const input = useRef(null);
    const book_id = useRouter().query.id as string;

    async function onSubmit(e) {
        e.preventDefault();
        e.stopPropagation();
        //Submit search
        if (input.current.value == "") {
            return;
        }

        await fetch(`/api/books/${book_id}/edit/get_acl_entity`, {
            method: "POST",
            headers: {
                "Content-Type": "application/json",
            },
            body: JSON.stringify({
                entity_name: input.current.value,
            }),
        })
            .then(res_parse_json)
            .then((result) => {
                if (result.error) {
                    setSuccess({
                        success: false,
                        error: true,
                        message: result.error,
                    });
                } else {
                    setSuccess({
                        success: true,
                        error: false,
                        message: "Added user/group",
                    });
                    input.current.value = "";
                    addCB(result);
                }
            })
            .catch((e) => {
                console.error(e);
                return;
            });
    }

    return (
        <>
            <form onSubmit={onSubmit}>
                <InputGroup className="p-3">
                    <Form.Control
                        type="text"
                        ref={input}
                        placeholder="Email or group name"
                        aria-label="search"
                        style={{ width: "70%" }}
                    />
                    <Button variant="outline-secondary" type="submit">
                        Add
                    </Button>
                </InputGroup>
            </form>
            <Alert
                variant="success"
                transition={false}
                show={success.success}
                style={{
                    borderColor: bg,
                }}
            >
                {close_btn}
                Added to acl list. Edit the new permissions below!
            </Alert>

            <Alert
                variant="danger"
                transition={false}
                show={success.error}
                style={{
                    borderColor: bg,
                }}
            >
                {close_btn}
                {success.message}
            </Alert>
        </>
    );
}

interface Entity {
    id: number;
    type: AllowedTypes;
    name: string;
}

function EntityRow({
    entity,
    permission,
    setPermission = () => {
        return;
    },
    disabled = false,
    note = undefined,
}: {
    entity: Entity;
    permission: PermissionACLDataResolved;
    setPermission?: (p: PermissionACLDataResolved) => void;
    disabled?: boolean;
    note?: string;
}) {
    return (
        <tr>
            <td>
                {entity.name}{" "}
                {note ? (
                    <span style={{ fontVariant: "super" }}>({note})</span>
                ) : null}
            </td>
            <td>{entity.type}</td>
            <td>
                <Form.Check
                    type="switch"
                    className={styles.switch}
                    checked={permission.pRead}
                    onChange={(e) => {
                        const newPermission = { ...permission };
                        newPermission.pRead = e.target.checked;
                        setPermission(newPermission);
                    }}
                    disabled={disabled}
                />
            </td>
            <td>
                <Form.Check
                    type="switch"
                    className={styles.switch}
                    checked={permission.pWrite}
                    onChange={(e) => {
                        const newPermission = { ...permission };
                        newPermission.pWrite = e.target.checked;
                        setPermission(newPermission);
                    }}
                    disabled={disabled}
                />
            </td>
            <td>
                <Form.Check
                    type="switch"
                    className={styles.switch}
                    checked={permission.pDelete}
                    onChange={(e) => {
                        const newPermission = { ...permission };
                        newPermission.pDelete = e.target.checked;
                        setPermission(newPermission);
                    }}
                    disabled={disabled}
                />
            </td>
            <td>
                <Form.Check
                    type="switch"
                    className={styles.switch}
                    checked={permission.pACL}
                    onChange={(e) => {
                        const newPermission = { ...permission };
                        newPermission.pACL = e.target.checked;
                        setPermission(newPermission);
                    }}
                    disabled={disabled}
                />
            </td>
        </tr>
    );
}
