import { fetcher } from "lib/fetcher";
import { ApiGetToken } from "pages/api/books/[id]/edit/tokens";
import { ChangeEvent, useContext, useState } from "react";
import { Button, Form, InputGroup, Modal, Table } from "react-bootstrap";
import useSWR from "swr";

import { EditBookContext } from ".";

import styles from "./AccessEditor.module.scss";

export default function AccessTokens() {
    const { book } = useContext(EditBookContext);
    const [description, setDescription] = useState("");

    const [showModal, setShowModal] = useState(false);
    const [addedToken, setAddedToken] = useState<
        ApiGetToken & { token: string }
    >(null);

    const { data, mutate } = useSWR<ApiGetToken[]>(
        `/api/books/${book.id}/edit/tokens`,
        fetcher
    );

    const rows = data
        ?.filter((token) => token.type == "api")
        .map((token) => (
            <tr key={token.id}>
                <td>{token.description}</td>
                <td>{new Date(token.created_at).toDateString()}</td>
                <td>{token.email}</td>
                <td className={styles.remove}>
                    <button
                        className="btn btn-outline-secondary"
                        onClick={() => {
                            deleteToken(book.id, token.id).then((r) => {
                                if (r.ok) {
                                    mutate();
                                }
                            });
                        }}
                    >
                        {trash_svg}
                    </button>
                </td>
            </tr>
        ));

    return (
        <div className={styles.container}>
            <h2>API Tokens</h2>
            <p>
                Tokens are a powerful way to authenticate with the system,
                especially in cases where you don&apos;t want to manually log in
                every time. This feature is particularly useful for automated
                insertion of snippets, as it allows you to upload code to your
                book programmatically. Each token is associated with a specific
                book, and allows for full read and write access to that book. By
                using tokens, you can automate workflows and streamline your
                development process. For more information on how to use tokens
                effectively, see the TODO section in our documentation.
            </p>

            <form
                onSubmit={(e) => {
                    e.preventDefault();
                    createNewToken(book.id, description).then((r) => {
                        setAddedToken(r);
                        setShowModal(true);
                        setDescription("");
                    });
                }}
            >
                <InputGroup className="p-3">
                    <Form.Control
                        type="text"
                        placeholder="Description"
                        style={{ width: "70%" }}
                        onInput={(e: ChangeEvent<HTMLTextAreaElement>) => {
                            setDescription(e.target.value);
                        }}
                        value={description}
                    />
                    <Button variant="outline-secondary" type="submit">
                        Create token
                    </Button>
                </InputGroup>
            </form>
            {addedToken && (
                <TokenModal
                    show={showModal}
                    addedToken={addedToken}
                    onHide={() => {
                        setShowModal(false);
                        setAddedToken(null);
                        mutate();
                    }}
                />
            )}
            {rows?.length > 0 && (
                <Table striped bordered hover>
                    <colgroup>
                        <col className={styles.col_description} />
                        <col />
                        <col />
                        <col className={styles.col_remove} />
                    </colgroup>
                    <thead>
                        <tr>
                            <th>Description</th>
                            <th>Created at</th>
                            <th>Created by</th>
                            <th></th>
                        </tr>
                    </thead>
                    <tbody>{rows}</tbody>
                </Table>
            )}
        </div>
    );
}

export const trash_svg = (
    <svg
        xmlns="http://www.w3.org/2000/svg"
        className="icon"
        width="24"
        height="24"
        viewBox="0 0 24 24"
        strokeWidth="2"
        stroke="currentColor"
        fill="none"
        strokeLinecap="round"
        strokeLinejoin="round"
    >
        <path stroke="none" d="M0 0h24v24H0z" fill="none"></path>
        <path d="M4 7l16 0"></path>
        <path d="M10 11l0 6"></path>
        <path d="M14 11l0 6"></path>
        <path d="M5 7l1 12a2 2 0 0 0 2 2h8a2 2 0 0 0 2 -2l1 -12"></path>
        <path d="M9 7v-3a1 1 0 0 1 1 -1h4a1 1 0 0 1 1 1v3"></path>
    </svg>
);

function TokenModal({ show, onHide, addedToken }) {
    const [isCopied, setIsCopied] = useState(false);

    function handleCopyClick() {
        navigator.clipboard.writeText(addedToken.token);
        setIsCopied(true);
    }

    return (
        <Modal show={show} onHide={onHide} centered backdrop="static">
            <Modal.Header closeButton>
                <Modal.Title>Token created!</Modal.Title>
            </Modal.Header>
            <Modal.Body>
                <p>
                    Your token has been created. Please keep it safe, as it will
                    not be shown again. You can revoke it at any time by
                    clicking the trash can icon next to it in the list!
                </p>
                <InputGroup className="mb-3">
                    <InputGroup.Text>Token</InputGroup.Text>
                    <Form.Control
                        type="text"
                        value={addedToken.token}
                        readOnly
                    />
                    <Button
                        variant="outline-secondary"
                        onClick={handleCopyClick}
                        disabled={isCopied}
                    >
                        {isCopied ? "Copied!" : "Copy"}
                    </Button>
                </InputGroup>
            </Modal.Body>
            <Modal.Footer>
                <Button variant="secondary" onClick={onHide}>
                    Close
                </Button>
            </Modal.Footer>
        </Modal>
    );
}

export const createNewToken = async (
    book_id,
    description,
    type = "api",
    expireIn = undefined
) => {
    const payload = {
        description,
        type,
    };
    if (expireIn) {
        payload["expireIn"] = expireIn;
    }

    return fetch(`/api/books/${book_id}/edit/tokens`, {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify(payload),
    }).then((r) => {
        if (r.ok) {
            return r.json();
        } else {
            throw new Error("Failed to create token!");
        }
    });
};

export const deleteToken = async (book_id, token_id) => {
    return fetch(`/api/books/${book_id}/edit/tokens`, {
        method: "DELETE",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify({
            token_id: token_id,
        }),
    });
};
