import { fetcher } from "lib/fetcher";
import { ApiError } from "pages/api/_service";
import { BackgroundType } from "pages/api/books/available_background_types";
import { useCallback, useContext, useEffect, useRef, useState } from "react";
import { Alert, Button, Form, InputGroup } from "react-bootstrap";
import useSWR from "swr";

import { EditBookContext } from ".";

import { render_background_by_id } from "controller/backgrounds/render_background_by_id";

import Loading from "components/utils/Loading";

import styles from "./AccessEditor.module.scss";

export const BackgroundTypeEditor = () => {
    const { book } = useContext(EditBookContext);
    return (
        <div className={styles.container}>
            <h2>Background Type</h2>
            <p>
                Do you miss the good old checkered notebooks? Or do you prefer a
                more modern look? Don&lsquo;t worry, you can change the
                background type of your book at any time!
            </p>
            <BackgroundTypeSelector book_id={book.id} />
        </div>
    );
};

function BackgroundTypeSelector({ book_id }) {
    // Get all available background types
    const {
        data: availableTypes,
        error,
        isValidating,
    } = useSWR<BackgroundType[]>(
        "/api/books/available_background_types",
        fetcher
    );

    //Get default background type
    const {
        data: selected,
        mutate: mutateSelected,
        error: error_s,
    } = useSWR<{ background_type_id: number }>(
        `/api/books/${book_id}/edit/background_types`,
        fetcher
    );

    // Local state for the dropdown (index)
    const [iSelectedLocal, setiSelectedLocal] = useState<number>(0);

    async function select() {
        // fetch
        await fetch(`/api/books/${book_id}/edit/background_types`, {
            method: "POST",
            headers: {
                "Content-Type": "application/json",
            },
            body: JSON.stringify({
                background_type_id: availableTypes[iSelectedLocal].id,
            }),
        })
            .then(async (res) => {
                if (res.ok) {
                    const success = (await res.json()) as boolean;
                    mutateSelected();
                    return success;
                } else {
                    const error = (await res.json()) as ApiError;
                    throw new Error(error.details);
                }
            })
            .catch((e) => {
                console.error(e);
            });
    }
    if (error || error_s) {
        return (
            <Alert variant="danger">Loading failed, please contact us!</Alert>
        );
    }

    if (!availableTypes || !selected || isValidating) {
        return <Loading />;
    }

    return (
        <>
            <InputGroup className="p-3">
                <InputGroup.Text>Background Type</InputGroup.Text>
                <Form.Select
                    aria-label="select_background_type"
                    value={iSelectedLocal}
                    onChange={(e) =>
                        setiSelectedLocal(parseInt(e.target.value))
                    }
                >
                    {availableTypes.map((type, i) => (
                        <option key={type.id} value={i}>
                            {type.name}{" "}
                            {type.id == selected.background_type_id
                                ? "(current)"
                                : ""}
                        </option>
                    ))}
                </Form.Select>
                <Button variant="outline-secondary" onClick={select}>
                    Select
                </Button>
            </InputGroup>
            <div className={styles.wrapper}>
                <p className={styles.description}>
                    {availableTypes[iSelectedLocal].description}
                </p>
                <BackgroundPreview type={availableTypes[iSelectedLocal]} />
            </div>
        </>
    );
}

function BackgroundPreview({ type }) {
    // Shows the background as preview in a canvas
    const container = useRef<HTMLDivElement>();
    const canvasRef = useRef<HTMLCanvasElement>();

    const draw = useCallback(() => {
        // Check if size updated
        canvasRef.current.width = container.current.clientWidth;
        canvasRef.current.height = container.current.clientHeight;

        // Redraw
        const ctx = canvasRef.current.getContext("2d");
        ctx.clearRect(0, 0, ctx.canvas.width, ctx.canvas.height);
        render_background_by_id(type.id, ctx, false);
    }, [type]);

    // Draw loop on width or height changes
    useEffect(() => {
        if (container.current) {
            const resizeObserver = new ResizeObserver(draw);
            resizeObserver.observe(container.current);

            return () => {
                resizeObserver.disconnect();
            };
        }
    }, [container, draw]);

    useEffect(() => {
        //Render canvas
        if (canvasRef) {
            draw();
        }
    }, [draw, canvasRef]);

    return (
        <div ref={container} className={styles.preview + " m-3"}>
            <canvas ref={canvasRef} />
        </div>
    );
}
