/* eslint-disable @next/next/no-img-element */
//import Image from "next/image";
import Link from "next/link";
import React, { useContext, useState } from "react";
import { Alert, Button, Form, InputGroup } from "react-bootstrap";

import { EditBookContext } from ".";

import styles from "./BookInfo.module.scss";

export const BookInfo = () => {
    const { error, setError } = useContext(EditBookContext);

    const close_btn = (
        <button
            className={"btn btn-light " + styles.close_btn}
            onClick={() => {
                setError(null);
            }}
        >
            <i className="bi bi-x"></i>
        </button>
    );

    return (
        <div className={styles.container}>
            <h2>Book overview</h2>
            <p>
                You can view and edit the general information about this book
                here. Every user or group with the <span>pEdit</span> permission
                is allowed to edit any of this information.
            </p>
            {error ? (
                <Alert variant="danger" transition={false}>
                    {close_btn}
                    {error.details}
                </Alert>
            ) : null}
            <div className={styles.content}>
                <TitleEditor />
                <Preview />
                <CommentEditor />
                <Created />
                <LastUpdated />
                <Finished />
                <Download_Utils />
            </div>
        </div>
    );
};

const TitleEditor = () => {
    const { book, updateBook } = useContext(EditBookContext);
    const [title, setTitle] = useState(book.title);

    const onSubmit = (e) => {
        e.preventDefault();
        updateBook({ title: title });
    };

    return (
        <div className={styles.title_editor}>
            <h3>Title:</h3>
            <Form onSubmit={onSubmit}>
                <InputGroup>
                    <Form.Control
                        type="text"
                        value={title}
                        onChange={(e) => setTitle(e.target.value)}
                        aria-label="Book title"
                    />
                    <Button variant="outline-secondary" type="submit">
                        Update
                    </Button>
                </InputGroup>
            </Form>
        </div>
    );
};

const CommentEditor = () => {
    const { book, updateBook } = useContext(EditBookContext);
    const [comment, setComment] = useState(book.comment);

    const onSubmit = (e) => {
        e.preventDefault();
        updateBook({ comment: comment });
    };
    return (
        <div className={styles.comment_editor}>
            <h3>Comment:</h3>
            <Form onSubmit={onSubmit}>
                <InputGroup>
                    <Form.Control
                        as="textarea"
                        value={comment || ""}
                        onChange={(e) => setComment(e.target.value)}
                        aria-label="Book comment"
                    />
                    <Button variant="outline-secondary" type="submit">
                        Update
                    </Button>
                </InputGroup>
            </Form>
        </div>
    );
};

const Created = () => {
    const { book } = useContext(EditBookContext);

    return (
        <div className={styles.created}>
            <h3>Created:</h3>
            <div className={styles.date}>
                {Intl.DateTimeFormat("en-GB", { dateStyle: "full" }).format(
                    new Date(book.created)
                )}
            </div>
        </div>
    );
};

const LastUpdated = () => {
    const { book } = useContext(EditBookContext);

    return (
        <div className={styles.last_updated}>
            <h3>Last modified:</h3>
            <div className={styles.date}>
                {Intl.DateTimeFormat("en-GB", {
                    dateStyle: "full",
                }).format(new Date(book.last_updated))}
            </div>
        </div>
    );
};

const Finished = () => {
    const { book, finished, updateFinished } = useContext(EditBookContext);

    return (
        <div className={styles.finished}>
            <h3>Finished:</h3>
            <p>
                You can mark this book as finished. This allows you to disable
                all editing features. This is reversible at any time.
            </p>

            {book.finished ? (
                <p className={styles.date}>
                    This book was marked as finished on: <br />
                    {Intl.DateTimeFormat("en-GB", {
                        dateStyle: "full",
                    }).format(new Date(book.finished))}
                </p>
            ) : (
                <p className={styles.date}>
                    This book is not marked as finished.
                </p>
            )}
            <Form>
                <Form.Check
                    type="switch"
                    label="Mark as finished"
                    id="finished-switch"
                    checked={finished ? true : false}
                    onChange={(e) => {
                        updateFinished(e.target.checked);
                    }}
                />
            </Form>
        </div>
    );
};

const Preview = () => {
    const { book } = useContext(EditBookContext);

    return (
        <div className={styles.preview}>
            <Link href={`/books/${book.id}`}>
                <img
                    src={"/api/books/" + book.id + "/cover_preview"}
                    alt="No cover page set!"
                    width={240}
                    height={350}
                />
            </Link>
        </div>
    );
};

const Download_Utils = () => {
    const { book } = useContext(EditBookContext);

    // Parse book as json file
    // TODO: Add download api endpoint for more complete metadata
    const download = () => {
        const data = JSON.stringify(
            {
                id: book.id,
                title: book.title,
                comment: book.comment,
                created: book.created,
                last_updated: book.last_updated,
                finished: book.finished,
            },
            null,
            4
        );
        const blob = new Blob([data], { type: "application/json" });
        const url = URL.createObjectURL(blob);
        const link = document.createElement("a");
        link.download = book.title + ".json";
        link.setAttribute("href", url);
        link.setAttribute("download", book.title + ".json");
        link.click();
    };

    return (
        <div className={styles.download_utils}>
            <h3>Metadata:</h3>
            <p>
                Here you can download all the metadata that is stored in the
                database for this book. This does not include the content!
            </p>
            <Button variant="outline-secondary" onClick={download}>
                Download
            </Button>
        </div>
    );
};
