import { fetcher } from "lib/fetcher";
import ms, { StringValue } from "ms";
import { ApiGetToken } from "pages/api/books/[id]/edit/tokens";
import { ChangeEvent, useContext, useMemo, useState } from "react";
import { Button, Form, InputGroup, Table } from "react-bootstrap";
import useSWR from "swr";

import { EditBookContext } from ".";
import { createNewToken, deleteToken, trash_svg } from "./AccessTokens";

import styles from "./AccessEditor.module.scss";

export default function ShareLinks() {
    const { book } = useContext(EditBookContext);
    const [description, setDescription] = useState("");
    const [expiration, setExpiration] = useState("1d");

    const valid_expiration = useMemo(() => isValidMs(expiration), [expiration]);

    const { data, mutate } = useSWR<ApiGetToken[]>(
        `/api/books/${book.id}/edit/tokens`,
        fetcher
    );

    const rows = data
        ?.filter((token) => token.type == "ui")
        .map((token) => {
            const url = `/books/${token.book_id}?token=${token.token}`;
            const base_url = `${window.location.protocol}//${window.location.host}`;

            return (
                <tr key={token.id}>
                    <td>{token.description}</td>
                    <td
                        className={styles.link}
                        onClick={async (e) => {
                            //Check if clipboard is already the link
                            const elem = e.target as HTMLElement;
                            copyToClipboard(e, base_url + url);
                            elem.dataset.copied = "true";
                            elem.style.color = "var(--bs-primary)";

                            setTimeout(() => {
                                elem.style.removeProperty("color");
                            }, 1000);
                        }}
                    >
                        {base_url + url}
                    </td>
                    <td>{new Date(token.created_at).toDateString()}</td>
                    <td>{new Date(token.expires_at).toDateString()}</td>
                    <td className={styles.remove}>
                        <button
                            className="btn btn-outline-secondary"
                            onClick={() => {
                                deleteToken(book.id, token.id).then((r) => {
                                    if (r.ok) {
                                        mutate();
                                    }
                                });
                            }}
                        >
                            {trash_svg}
                        </button>
                    </td>
                </tr>
            );
        });

    return (
        <div className={styles.container}>
            <h2>Share Links</h2>
            <p>
                Share links allow you to share your book with others without the
                need for them to create an account. They can be used to share
                your book with anyone who has the link. You can also set an
                expiration date for the link and revoke it at any time.
            </p>
            <p>This can only be used to grant read access!</p>
            <form
                onSubmit={(e) => {
                    e.preventDefault();
                    createNewToken(book.id, description, "ui", expiration).then(
                        () => {
                            setDescription("");
                            mutate();
                        }
                    );
                }}
            >
                <InputGroup className="p-3" hasValidation>
                    <Form.Control
                        type="text"
                        placeholder="Description"
                        style={{ width: "60%" }}
                        onInput={(e: ChangeEvent<HTMLTextAreaElement>) => {
                            setDescription(e.target.value);
                        }}
                        value={description}
                    />
                    <Form.Control
                        type="text"
                        placeholder="Expires In"
                        style={{ width: "20%" }}
                        onInput={(e: ChangeEvent<HTMLTextAreaElement>) => {
                            // Check if valid using ms
                            setExpiration(e.target.value);
                        }}
                        value={expiration}
                        isValid={valid_expiration}
                        isInvalid={!valid_expiration}
                        required
                    />
                    <Button variant="outline-secondary" type="submit">
                        Create link
                    </Button>
                    <Form.Control.Feedback type="invalid">
                        Invalid expiration time use ms format i.e. 1d, 1h, 1m 1
                        week etc.
                    </Form.Control.Feedback>
                </InputGroup>
            </form>

            {rows?.length > 0 && (
                <Table striped bordered hover>
                    <colgroup>
                        <col className={styles.col_description} />
                        <col className={styles.col_link} />
                        <col />
                        <col />
                        <col className={styles.col_remove} />
                    </colgroup>
                    <thead>
                        <tr>
                            <th>Description</th>
                            <th>Link</th>
                            <th>Created at</th>
                            <th>Expires at</th>
                            <th></th>
                        </tr>
                    </thead>
                    <tbody>{rows}</tbody>
                </Table>
            )}
        </div>
    );
}

const isValidMs = (value: string) => {
    console.log(value);
    try {
        const r = ms(value as StringValue);
        return r !== undefined;
    } catch (e) {
        console.log(e);
        return false;
    }
};

function copyToClipboard(e: React.MouseEvent, text: string) {
    e.preventDefault();
    navigator.clipboard.writeText(text);
}
