import Link from "next/link";
import { BookOwner } from "pages/api/books/[id]/edit/get_owner";
import { createContext, useContext, useState } from "react";

import { AccessEditor } from "./AccessEditor";
import { BookInfo } from "./BookInfo";
import { TransferOwner } from "./OwnershipEditor";

import { BookData } from "controller/book";
import { UserData } from "database/services/service.typings";

/**This index file exposes all components from the files
 * for easier importing.
 */
export { AccessEditor } from "./AccessEditor";
export { BackgroundTypeEditor } from "./BackgroundTypeEditor";
export { TransferOwner } from "./OwnershipEditor";

import { fetcher } from "lib/fetcher";
import { useRouter } from "next/router";
import { ApiError } from "pages/api/_service";
import { BookPermission } from "pages/books";
import useSWR from "swr";

import AccessTokens from "./AccessTokens";
import { BackgroundTypeEditor } from "./BackgroundTypeEditor";
import ShareLinks from "./ShareLinks";

import styles from "./Edit.module.scss";

/** Renders the edit book view
 *  i.e. a sidebar with all the edit components
 *  and the selected edit tool.
 */
export default function EditBook({
    book,
    perms,
    user,
    userIsOwner,
}: {
    book: BookData;
    perms: BookPermission;
    user: UserData;
    userIsOwner: boolean;
}) {
    const router = useRouter();
    const [selectedTool, setSelectedTool] = useState<Tools>("info");
    const [error, setError] = useState<ApiError | null>(null);

    // Call this function whenever you want to
    // refresh props!
    const refreshData = () => {
        router.replace(router.asPath);
    };

    // Update a key in the book, if null is passed
    // the key is deleted.
    const updateBook = (newBook: Partial<BookData>) => {
        // Perform sever side update
        fetch(`/api/books/${book.id}/edit/update`, {
            method: "POST",
            headers: {
                "Content-Type": "application/json",
            },
            body: JSON.stringify(newBook),
        }).then((r) => {
            // Do nothing if successful else show error
            if (r.ok) {
                refreshData();
                setError(null);
            } else {
                r.json().then((r) => {
                    setError(r);
                });
            }
        });
    };
    const { data: owner } = useSWR<BookOwner>(
        `/api/books/${book.id}/edit/get_owner`,
        fetcher
    );

    const updateFinished = (finished: boolean) => {
        if (finished) {
            updateBook({ finished: new Date() });
        } else {
            updateBook({ finished: null });
        }
    };

    const ret: EditBookContext = {
        book,
        updateBook,
        error,
        setError,
        user,
        perms,
        userIsOwner,
        selectedTool,
        setSelectedTool,
        finished: book.finished || false,
        updateFinished,
        owner,
    };

    return (
        <EditBookContext.Provider value={ret}>
            <div className={styles.container}>
                <div className={styles.sidebar}>
                    <ul className={"nav nav-pills " + styles.tabs}>
                        <Tab tool="info" label="Info" />
                        {perms.pWrite ? (
                            <Tab tool="background" label="Background" />
                        ) : null}
                        {perms.pACL ? (
                            <Tab tool="access" label="Access" />
                        ) : null}
                        <BackButton />
                    </ul>
                </div>
                <div className={styles.content}>
                    <div
                        className={
                            styles.wrapper +
                            " col-12 col-lg-10 col-xxl-8 d-grid"
                        }
                    >
                        {selectedTool === "info" && <BookInfo />}
                        {selectedTool === "background" && (
                            <BackgroundTypeEditor />
                        )}

                        {selectedTool === "access" && (
                            <>
                                <AccessEditor />
                                <ShareLinks />
                            </>
                        )}
                        {selectedTool === "access" && perms.pACL && (
                            <AccessTokens />
                        )}
                        {selectedTool === "access" && userIsOwner && (
                            <TransferOwner />
                        )}
                    </div>
                </div>
            </div>
        </EditBookContext.Provider>
    );
}

function Tab({
    tool,
    label,
    disabled = false,
}: {
    tool: Tools;
    label: string;
    disabled?: boolean;
}) {
    const { selectedTool, setSelectedTool } = useContext(EditBookContext);
    return (
        <li
            className={
                styles.tab +
                " " +
                (selectedTool === tool ? styles.active : "") +
                (disabled ? " disabled" : "")
            }
            onClick={() => setSelectedTool(tool)}
        >
            {label}
        </li>
    );
}

/** Context provider for
 * the edit book view.
 */

type Tools = "info" | "access" | "background" | "collaborators";

interface EditBookContext {
    book: BookData;
    updateBook: (newBook: Partial<BookData>) => void;
    error: ApiError | null;
    setError: (error: ApiError | null) => void;
    user: UserData;
    userIsOwner: boolean;
    perms: BookPermission;
    selectedTool: Tools;
    setSelectedTool: (tool: Tools) => void;
    finished: false | Date;
    updateFinished: (finished: boolean) => void;
    owner: BookOwner;
}

export const EditBookContext = createContext<EditBookContext>(undefined);

function BackButton() {
    return (
        <div className={styles.backButton}>
            <Link href="/books">
                <i className="icon bi-arrow-left d-flex"></i> Back
            </Link>
        </div>
    );
}
