import dynamic from "next/dynamic";
import Image, { StaticImageData } from "next/image";
import { useEffect, useState } from "react";

import styles from "./BackgroundImage.module.scss";

// Load all images from background folder
function importAll(r) {
    return r.keys().map(r);
}

type Transition = "fade" | "none";

export default dynamic(() => Promise.resolve(BackgroundImage), { ssr: false });

function BackgroundImage({
    images = undefined,
    transition = "fade",
    image_change_time = 10000,
    transition_time = 8000,
}: {
    images?: StaticImageData[];
    transition?: Transition;
    image_change_time?: number;
    transition_time?: number;
}) {
    if (!images) {
        images = importAll(
            require.context("@img/background", false, /\.(png|jpe?g|svg)$/)
        );
    }

    const [image, setImage] = useState(
        images[Math.floor(Math.random() * images.length)]
    );

    /** Transition image */
    useEffect(() => {
        const interval = setInterval(() => {
            const rng = Math.floor(Math.random() * images.length);
            if (images[rng] !== image) {
                setImage(images[rng]);
            }
        }, image_change_time);
        return () => clearInterval(interval);
    }, [image_change_time, image, images]);

    return (
        <div className={styles.background}>
            <div className={styles.wrapper}>
                <ImageWithTransition
                    src={image}
                    transition={transition}
                    transition_time={transition_time}
                    alt="Background Image"
                    layout="fill"
                    objectFit="cover"
                    objectPosition="center"
                    priority
                />
            </div>
            <label className={styles.caption}>
                Image generated using{" "}
                <a href="https://openai.com/dall-e-2/">DALL·E 2</a>
            </label>
        </div>
    );
}

function ImageWithTransition({
    src,
    transition,
    transition_time,
    ...props
}: {
    src: StaticImageData;
    transition: Transition;
    transition_time: number;
    [key: string]: unknown;
}) {
    const [stack, setStack] = useState([]);

    // Store src in stack and update on src change
    useEffect(() => {
        setStack((stack) => {
            if (stack === undefined) {
                return [src];
            } else {
                return [...stack, src];
            }
        });
        setTimeout(() => {
            setStack((stack) => {
                if (stack.length > 1) {
                    return [...stack.slice(1)];
                } else {
                    return stack;
                }
            });
        }, transition_time);
    }, [src, transition_time]);

    const images = [];
    for (let i = 0; i < stack.length; i++) {
        let opacity = 1;
        if (stack.length > 1 && i == 0) {
            opacity = 0; // This invokes css transition set to 6s atm
        }
        if (transition == "none") {
            opacity = 1;
        }
        images.push(
            <Image
                key={stack[i].default.src}
                src={stack[i]}
                alt="Background"
                style={{ opacity, zIndex: -i }}
                {...props}
            />
        );
    }
    return <>{images}</>;
}
