import BrandImage from "@img/logo.png";
import useUser, { useIsAdmin } from "lib/useUser";
import Image from "next/image";
import Link from "next/link";
import { useRouter } from "next/router";

import { admin_svg, books_svg, logout_svg, person_svg } from "./Sidebar";

import { logoutUser } from "components/utils/Logout";

import styles from "./Navbar.module.scss";

interface NavbarProps {
    children?: JSX.Element[] | JSX.Element;
    dropdown_children?: JSX.Element[] | JSX.Element;
    checkUser?: boolean;
}

/** Navbar
 * Possible to pass custom children which are added
 * between brand and menu. Or into the dropdown menu.
 *
 * @param {NavbarProps} props
 */
export default function Navbar(prop: NavbarProps) {
    return (
        <nav
            id={styles.snipNavbar}
            className="navbar bg-primary navbar-expand text-light navbar-dark"
        >
            <div id="snipNavbar-container" className="container-fluid">
                <Logo />
                <BrandName />
                {prop.children}
                <NavbarDropdown>{prop.dropdown_children}</NavbarDropdown>
            </div>
        </nav>
    );
}

function Logo() {
    return (
        <Link href="/" id={styles.logo} className="d-flex">
            <Image
                src={BrandImage}
                width={40}
                height={40}
                alt="logo"
                style={{ objectFit: "contain" }}
            />
        </Link>
    );
}
function BrandName() {
    return (
        <div className="me-auto">
            <Link href="/" className="navbar-brand">
                Snip
            </Link>
        </div>
    );
}

interface NavbarDropdownProps {
    children?: JSX.Element[] | JSX.Element;
    dropdown_children?: JSX.Element[] | JSX.Element;
}

/** Dropdown menu with logout and profile button */
function NavbarDropdown(props: NavbarDropdownProps) {
    const { mutateUser, isLoggedIn } = useUser();
    const { isAdmin } = useIsAdmin();
    const router = useRouter();

    // Render profile and logout button if user is logged in
    // else login form
    let elements = null;
    if (isLoggedIn) {
        elements = (
            <>
                <li>
                    <Link href="/">
                        <button>
                            Books
                            {books_svg}
                        </button>
                    </Link>
                </li>
                <li>
                    <Link href="/account">
                        <button>
                            Account
                            {person_svg}
                        </button>
                    </Link>
                </li>
                {isAdmin ? (
                    <li>
                        <Link href="/admin">
                            <button>
                                Admin Panel
                                {admin_svg}
                            </button>
                        </Link>
                    </li>
                ) : null}
                <hr />
                <li>
                    <button
                        onClick={async () => {
                            await logoutUser();
                            mutateUser(null);
                            router.push("/");
                        }}
                    >
                        Logout
                        {logout_svg}
                    </button>
                </li>
            </>
        );
    } else if (router.pathname == "/account/signup") {
        elements = (
            <li>
                <Link href="/account/login">
                    <button>
                        Login
                        <i className="icon bi-person"></i>
                    </button>
                </Link>
            </li>
        );
    } else if (router.pathname == "/account/login") {
        elements = (
            <li>
                <Link href="/signup">
                    <button>
                        SignUp
                        <i className="bi bi-person-plus"></i>
                    </button>
                </Link>
            </li>
        );
    } else if (
        router.pathname == "/account/forgotPassword" ||
        router.pathname == "/" ||
        router.pathname == "/frontpage"
    ) {
        elements = (
            <>
                <li>
                    <Link href="/account/login">
                        <button>
                            Login
                            <i className="icon bi-person"></i>
                        </button>
                    </Link>
                </li>
                <li>
                    <Link href="/signup">
                        <button>
                            SignUp
                            <i className="bi bi-person-plus"></i>
                        </button>
                    </Link>
                </li>
            </>
        );
    }

    return (
        <div id={styles.dropdown} className="nav-item dropdown">
            <a
                id={styles.dropdown_toggle}
                className="nav-link"
                role="button"
                data-bs-toggle="dropdown"
                aria-expanded="false"
            >
                <i className="bi bi-list"></i>
            </a>
            <ul
                id={styles.dropdown_menu}
                className="dropdown-menu dropdown-menu-end"
                aria-labelledby={styles.dropdown_toggle}
            >
                {props.children}
                {elements}
            </ul>
        </div>
    );
}
