import useUser, { useIsAdmin } from "lib/useUser";
import Link from "next/link";
import { useRouter } from "next/router";
import { HTMLAttributes, ReactNode } from "react";

import { logoutUser } from "components/utils/Logout";

import styles from "./DefaultLayout.module.scss";

export default function Sidebar() {
    const router = useRouter();
    const { isAdmin } = useIsAdmin();

    return (
        <nav className={styles.sidebar}>
            <ul className={"nav nav-pills " + styles.tabs}>
                <Tab_Link
                    href="/books"
                    active={router.pathname === "/books"}
                    img={books_svg}
                >
                    Books
                </Tab_Link>

                <Tab_Link
                    href="/account"
                    active={router.pathname === "/account"}
                    img={person_svg}
                >
                    Account
                </Tab_Link>
                <Tab_Link
                    href="/upload"
                    active={router.pathname === "/upload"}
                    img={upload_svg}
                >
                    Upload
                </Tab_Link>
                {isAdmin && (
                    <Tab_Link
                        href="/admin"
                        active={router.pathname === "/admin"}
                        img={admin_svg}
                    >
                        Admin
                    </Tab_Link>
                )}
                <LogoutButton />
            </ul>
        </nav>
    );
}

function LogoutButton() {
    const { mutateUser } = useUser();
    const router = useRouter();
    return (
        <Tab_Link
            align="left"
            onClick={async () => {
                await logoutUser();
                mutateUser(null);
                router.push("/");
            }}
            img={logout_svg}
            special
        >
            Logout
        </Tab_Link>
    );
}

type Props = {
    children: ReactNode;
    href?: string;
    active?: boolean;
    special?: boolean;
    align?: "left" | "right";
    img?: ReactNode;
} & HTMLAttributes<HTMLDivElement>;

function Tab_Link({
    children,
    href = undefined,
    active = false,
    special = false,
    align = "right",
    img = undefined,
    ...props
}: Props) {
    const content = (
        <div
            className={styles.tab}
            data-align={align}
            data-active={active}
            data-special={special}
            {...props}
        >
            {img}
            <span>{children}</span>
        </div>
    );

    return href ? <Link href={href}>{content}</Link> : content;
}

export const books_svg = (
    <svg
        xmlns="http://www.w3.org/2000/svg"
        className="icon"
        width="24"
        height="24"
        viewBox="0 0 24 24"
        strokeWidth="2"
        stroke="currentColor"
        fill="none"
        strokeLinecap="round"
        strokeLinejoin="round"
    >
        <path stroke="none" d="M0 0h24v24H0z" fill="none"></path>
        <path d="M5 4m0 1a1 1 0 0 1 1 -1h2a1 1 0 0 1 1 1v14a1 1 0 0 1 -1 1h-2a1 1 0 0 1 -1 -1z"></path>
        <path d="M9 4m0 1a1 1 0 0 1 1 -1h2a1 1 0 0 1 1 1v14a1 1 0 0 1 -1 1h-2a1 1 0 0 1 -1 -1z"></path>
        <path d="M5 8h4"></path>
        <path d="M9 16h4"></path>
        <path d="M13.803 4.56l2.184 -.53c.562 -.135 1.133 .19 1.282 .732l3.695 13.418a1.02 1.02 0 0 1 -.634 1.219l-.133 .041l-2.184 .53c-.562 .135 -1.133 -.19 -1.282 -.732l-3.695 -13.418a1.02 1.02 0 0 1 .634 -1.219l.133 -.041z"></path>
        <path d="M14 9l4 -1"></path>
        <path d="M16 16l3.923 -.98"></path>
    </svg>
);

export const logout_svg = (
    <svg
        xmlns="http://www.w3.org/2000/svg"
        className="icon"
        width="24"
        height="24"
        viewBox="0 0 24 24"
        strokeWidth="2"
        stroke="currentColor"
        fill="none"
        strokeLinecap="round"
        strokeLinejoin="round"
    >
        <path stroke="none" d="M0 0h24v24H0z" fill="none"></path>
        <path d="M14 8v-2a2 2 0 0 0 -2 -2h-7a2 2 0 0 0 -2 2v12a2 2 0 0 0 2 2h7a2 2 0 0 0 2 -2v-2"></path>
        <path d="M7 12h14l-3 -3m0 6l3 -3"></path>
    </svg>
);

export const person_svg = (
    <svg
        xmlns="http://www.w3.org/2000/svg"
        className="icon"
        width="24"
        height="24"
        viewBox="0 0 24 24"
        strokeWidth="2"
        stroke="currentColor"
        fill="none"
        strokeLinecap="round"
        strokeLinejoin="round"
    >
        <path stroke="none" d="M0 0h24v24H0z" fill="none"></path>
        <path d="M8 7a4 4 0 1 0 8 0a4 4 0 0 0 -8 0"></path>
        <path d="M6 21v-2a4 4 0 0 1 4 -4h4a4 4 0 0 1 4 4v2"></path>
    </svg>
);

export const admin_svg = (
    <svg
        xmlns="http://www.w3.org/2000/svg"
        className="icon icon-tabler icon-tabler-alert-triangle"
        width="24"
        height="24"
        viewBox="0 0 24 24"
        strokeWidth="2"
        stroke="currentColor"
        fill="none"
        strokeLinecap="round"
        strokeLinejoin="round"
    >
        <path stroke="none" d="M0 0h24v24H0z" fill="none"></path>
        <path d="M10.24 3.957l-8.422 14.06a1.989 1.989 0 0 0 1.7 2.983h16.845a1.989 1.989 0 0 0 1.7 -2.983l-8.423 -14.06a1.989 1.989 0 0 0 -3.4 0z"></path>
        <path d="M12 9v4"></path>
        <path d="M12 17h.01"></path>
    </svg>
);

export const upload_svg = (
    <svg
        xmlns="http://www.w3.org/2000/svg"
        className="icon icon-tabler icon-tabler-upload"
        width="24"
        height="24"
        viewBox="0 0 24 24"
        strokeWidth="2"
        stroke="currentColor"
        fill="none"
        strokeLinecap="round"
        strokeLinejoin="round"
    >
        <path stroke="none" d="M0 0h24v24H0z" fill="none"></path>
        <path d="M4 17v2a2 2 0 0 0 2 2h12a2 2 0 0 0 2 -2v-2"></path>
        <path d="M7 9l5 -5l5 5"></path>
        <path d="M12 4l0 12"></path>
    </svg>
);
