import { useState } from "react";

import { CommentInput } from "./NewBookForm";

import styles from "./CreateFolderForm.module.scss";

/** Post request to create a new folder
 * @param name Name of the folder
 * @param comment Comment for the folder
 * @param parent_id ID of the parent folder (optional)
 */
async function create_new_folder(
    name: string,
    comment: string = undefined,
    parent_id: string = undefined
) {
    const response = await fetch("/api/folders/new", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify({ name, comment, parent_id }),
    });
    const data = await response.json();
    if (response.ok) {
        return data;
    }
    throw new Error(data.error);
}

/** CreateFolder component
 * Primarily used in the books browser but can also be used in other places.
 * @param callback_submit Callback function to be called when the form is submitted
 * @param callback_cancel Callback function to be called when the form is cancelled
 */
export function CreateFolderForm({
    callback_submit,
    callback_cancel = undefined,
}: {
    callback_submit: () => void;
    callback_cancel?: () => void;
}) {
    const [error, setError] = useState<string>(undefined);

    const resetForm = () => {
        (document.getElementById("newBookForm") as HTMLFormElement).reset();
        setError(undefined);
    };

    const handleSubmit = async (event) => {
        event.preventDefault();
        await create_new_folder(event.target.name.value)
            .then(() => {
                resetForm();
                if (callback_submit) {
                    callback_submit();
                }
            })
            .catch((error) => {
                setError(error.message);
            });
    };

    const handleCancel = () => {
        // Reset form
        resetForm();
        if (callback_cancel) {
            callback_cancel();
        }
    };

    return (
        <>
            <div className={styles.header}>Create a folder</div>
            <hr />

            {error && <div className="alert alert-danger">{error}</div>}
            <form onSubmit={handleSubmit} id="newBookForm">
                <NameInput />
                <CommentInput type="folder" />
                <div className={styles.submit}>
                    {callback_cancel && (
                        <button
                            className="btn btn-outline-secondary w-100"
                            type="button"
                            onClick={handleCancel}
                        >
                            Cancel
                        </button>
                    )}
                    <button
                        className="btn btn-primary w-100"
                        type="submit"
                        form="newBookForm"
                    >
                        Create folder
                    </button>
                </div>
            </form>
        </>
    );
}

const NameInput = () => {
    return (
        <div className={styles.item}>
            <div className={styles.inputField}>
                <input
                    className="w-100 form-control form__input"
                    type="text"
                    id="name"
                    name="name"
                    placeholder=" "
                    autoComplete="off"
                    required
                />
                <label htmlFor="name">Folder name</label>
            </div>
            <span className={styles.description}>
                The name of the folder you want to create. This can be changed
                later.
            </span>
        </div>
    );
};
