import useUser from "lib/useUser";
import Link from "next/link";
import { useRouter } from "next/router";
import { useEffect, useState } from "react";

import styles from "./Form.module.scss";

// eslint-disable-next-line @typescript-eslint/no-unused-vars
function validateEmail(email) {
    const res =
        /^(([^<>()[\]\\.,;:\s@"]+(\.[^<>()[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/;
    return res.test(String(email).toLowerCase());
}

async function onFormSubmit(event) {
    event.preventDefault();

    const email = event.target.email.value;
    const password = event.target.password.value;

    const response = await fetch("/api/account/login", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify({ email, password }),
    });

    const data = await response.json();

    // response.ok is true when res.status is 2xx
    // https://developer.mozilla.org/en-US/docs/Web/API/Response/ok
    if (response.ok) {
        return data;
    }

    throw new Error(response.statusText);
}

export default function LoginFormFullPage() {
    return (
        <div className={styles.wrapper}>
            <LoginForm />
        </div>
    );
}

export function LoginForm() {
    // here we just check if user is already logged in and redirect to index page
    const { mutateUser } = useUser();
    useRedirect("/books");
    const [error, setError] = useState(null);

    let errorDiv = null;
    if (error) {
        errorDiv = (
            <div className="alert alert-danger" role="alert">
                <strong>Login failed!</strong> Please check your credentials.
            </div>
        );
    }
    return (
        <div className={styles.form}>
            <form
                id="loginForm"
                onSubmit={async (event) => {
                    await onFormSubmit(event)
                        .then((user) => {
                            mutateUser(user);
                        })
                        .catch((errorCode) => {
                            setError(errorCode);
                            setTimeout(() => {
                                setError(null);
                            }, 10000);
                        });
                }}
            >
                <fieldset>
                    <div className={styles.header}>
                        Please login to your account
                    </div>
                    {errorDiv}
                    <div>
                        <div
                            className={styles.inputField + " form-outline mb-3"}
                        >
                            <input
                                className={"w-100 form-control form__input"}
                                type="email"
                                name="email"
                                id="email"
                                placeholder=" "
                                autoComplete="email"
                                required
                            />
                            <label htmlFor="email">Email</label>
                        </div>
                    </div>
                    <div>
                        <div
                            className={styles.inputField + " form-outline mb-3"}
                        >
                            <input
                                className="w-100 form-control form__input"
                                type="password"
                                name="password"
                                id="password"
                                autoComplete="current-password"
                                placeholder=" "
                            />
                            <label htmlFor="password">Password</label>
                        </div>
                    </div>
                    <div className="d-flex align-items-center text-center mb-3 flex-column">
                        <button
                            className="btn btn-primary btn-block fa-lg gradient-custom-2 mb-2"
                            style={{ width: "90%" }}
                            type="submit"
                        >
                            Log in
                        </button>
                        <button
                            className="btn btn-outline-dark btn-block fa-lg mb-2"
                            style={{ width: "90%" }}
                            type="button"
                            disabled
                        >
                            Log in with SSO
                        </button>
                        <Link
                            className="text-muted"
                            href="/account/forgotPassword"
                        >
                            Forgot password?
                        </Link>
                    </div>
                    <div className="d-flex align-items-center justify-content-center">
                        <p className="mb-0 me-2">Don&apos;t have an account?</p>
                        <Link href="/signup">
                            <button
                                type="button"
                                className="btn btn-outline-dark"
                            >
                                Create new
                            </button>
                        </Link>
                    </div>
                </fieldset>
            </form>
        </div>
    );
}

/** Redirect to the page the user was on before logging in
 * this reads the query parameter redirect and redirects
 */
export const useRedirect = (default_location) => {
    const { isLoggedIn } = useUser();
    const router = useRouter();
    useEffect(() => {
        if (isLoggedIn) {
            if (router.query.redirect) {
                router.push(router.query.redirect.toString());
            }
            router.push(default_location);
        }
    }, [default_location, isLoggedIn, router]);
};
