import { useEffect, useRef, useState } from "react";

import styles from "./NewBookForm.module.scss";

/** Post request to create a new book
 * @param title Title of the book
 * @param comment Comment for the book
 * @param collaborators List of collaborators
 */
async function create_new_book(
    title: string,
    comment: string = undefined,
    collaborators: string[] = undefined
) {
    const response = await fetch("/api/books/new", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify({ title, collaborators, comment }),
    });
    const data = await response.json();
    if (response.ok) {
        return data;
    }
    throw new Error(data);
}

/** NewBookForm component
 * Primarily used in the books browser but can also be used in other places.
 * @param callback_submit Callback function to be called when the form is submitted
 * @param callback_cancel Callback function to be called when the form is cancelled
 */
export function NewBookForm({
    callback_submit,
    callback_cancel = undefined,
}: {
    callback_submit: () => void;
    callback_cancel?: () => void;
}) {
    const [emails, setEmails] = useState<string[]>([]);
    const [error, setError] = useState<string>(undefined);

    const resetForm = () => {
        (document.getElementById("newBookForm") as HTMLFormElement).reset();
        setEmails([]);
        setError(undefined);
    };

    const handleSubmit = async (event) => {
        event.preventDefault();
        await create_new_book(
            event.target.title.value,
            event.target.comment.value,
            emails
        )
            .then(() => {
                resetForm();
                if (callback_submit) {
                    callback_submit();
                }
            })
            .catch((err) => {
                setError(err.message);
            });
    };

    const handleCancel = () => {
        // Reset form
        resetForm();
        if (callback_cancel) {
            callback_cancel();
        }
    };

    return (
        <>
            <div className={styles.header}>Create a new book</div>
            <hr />
            {error && <div className="alert alert-danger">{error}</div>}
            <form onSubmit={handleSubmit} id="newBookForm">
                <TitleInput />
                <CommentInput />
                <CollaboratorsInput emails={emails} setEmails={setEmails} />
                <div className={styles.submit}>
                    {callback_cancel && (
                        <button
                            className="btn btn-outline-secondary w-100"
                            type="button"
                            onClick={handleCancel}
                        >
                            Cancel
                        </button>
                    )}
                    <button
                        className="btn btn-primary w-100"
                        type="submit"
                        form="newBookForm"
                    >
                        Create book
                    </button>
                </div>
            </form>
        </>
    );
}

const TitleInput = () => {
    return (
        <div className={styles.item}>
            <div className={styles.inputField}>
                <input
                    className="w-100 form-control form__input"
                    type="text"
                    id="title"
                    name="title"
                    placeholder=" "
                    autoComplete="off"
                    required
                />
                <label htmlFor="title">Book title</label>
            </div>
            <span className={styles.description}>
                The title of the book is shown in the book browser and on the
                books page. It can be changed later but should be set to
                something meaningful.
            </span>
        </div>
    );
};

export const CommentInput = ({ type = "book" }: { type?: string }) => {
    const textareaRef = useRef<HTMLTextAreaElement>(null);

    useEffect(() => {
        if (textareaRef.current) {
            textareaRef.current.style.height = "0px";
            const scrollHeight = textareaRef.current.scrollHeight;
            textareaRef.current.style.height = scrollHeight + "px";

            textareaRef.current.addEventListener("input", (event) => {
                const target = event.target as HTMLTextAreaElement;
                target.style.height = "0px";
                const scrollHeight = target.scrollHeight;
                target.style.height = scrollHeight + "px";
            });
        }
    }, []);

    return (
        <div className={styles.item}>
            <div className={styles.inputField}>
                <textarea
                    ref={textareaRef}
                    className="w-100 form-control form__input"
                    id="comment"
                    name="comment"
                    placeholder=" "
                    autoComplete="off"
                />
                <label htmlFor="comment">Comment</label>
            </div>
            <span className={styles.description}>
                The comment can give more information about the {type}. It can
                be used to find the {type} more easily. It is not required but
                recommended.
            </span>
        </div>
    );
};

const CollaboratorsInput = ({
    emails,
    setEmails,
}: {
    emails: string[];
    setEmails: (emails: string[]) => void;
}) => {
    const keyHandler = (event) => {
        if (event.key === "Enter") {
            event.preventDefault();
            if (event.target.value === "") {
                return;
            }
            setEmails([...emails, event.target.value]);
            event.target.value = "";
        }

        if (event.key === "Backspace" && event.target.value === "") {
            event.preventDefault();
            setEmails(emails.slice(0, emails.length - 1));
        }
    };

    const active = emails.length > 0;

    return (
        <div className={styles.item}>
            <div className={styles.collaborators}>
                <div
                    className={
                        styles.inputField + " w-100 form-control form__input"
                    }
                    data-active={active}
                >
                    {emails.map((email, index) => (
                        <Collaborator
                            key={index}
                            email={email}
                            onClick={() => {
                                setEmails(emails.filter((_, i) => i !== index));
                            }}
                        />
                    ))}
                    <input
                        className="form__input"
                        type="text"
                        id="collaborators"
                        name="collaborators"
                        placeholder=" "
                        autoComplete="off"
                        onKeyDown={keyHandler}
                    />
                    <label htmlFor="collaborators">Collaborators</label>
                </div>
            </div>
            <span className={styles.description}>
                You can add collaborators via <strong>e-mail</strong>. This
                automatically sends them an invite link. This can also be done
                from the books setting page.
            </span>
        </div>
    );
};

const Collaborator = ({ email, onClick }) => {
    // check if email is valid using regex
    const emailRegex = /^[^\s@]+@[^\s@]+\.[^\s@]+$/;
    const isValid = emailRegex.test(email);

    return (
        <div className={styles.collaborator} data-isvalid={isValid}>
            <div className="d-flex">
                <span>{email}</span>
                <i className="bi bi-x" onClick={onClick}></i>
            </div>
        </div>
    );
};
