import useUser from "lib/useUser";
import Link from "next/link";
import { useState } from "react";

import { ConfirmPasswordInput, EmailInput, PasswordInput } from "./Form";
import { useRedirect } from "./LoginForm";
import { validateEmail, validatePW } from "./validation";

import { UserData } from "database/services/service.typings";

import styles from "./Form.module.scss";

async function submitForm(event) {
    const email = event.target.email.value;
    const password = event.target.password.value;

    const response = await fetch("/api/account/signup", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify({ email, password }),
    });
    const data = await response.json();

    // response.ok is true when res.status is 2xx
    // https://developer.mozilla.org/en-US/docs/Web/API/Response/ok
    if (response.ok) {
        return data;
    }

    throw new Error(data);
}

export interface Validation {
    email: "unset" | "valid" | "invalid";
    password: "unset" | "valid" | "invalid";
    confirmPassword: "unset" | "valid" | "invalid";
}

export default function SignupFormFullPage() {
    return (
        <div className={styles.wrapper}>
            <SignupForm />
        </div>
    );
}

export function SignupForm() {
    // here we just check if user is already logged in and redirect to index page
    const { mutateUser } = useUser();
    useRedirect("/");

    const [valid, setValid] = useState<Validation>({
        email: "unset",
        password: "unset",
        confirmPassword: "unset",
    });

    /** Validate the form on the client side
     * and set the state of the form accordingly
     */
    const validateFormClientSide = (event) => {
        const pw_valid = validatePW(event.currentTarget.password.value);
        const pws_match =
            event.currentTarget.password.value ===
            event.currentTarget.confirm_password.value;
        const email_valid = validateEmail(event.currentTarget.email.value);

        const n_valid = valid;
        if (!pw_valid) {
            n_valid.password = "invalid";
        } else {
            n_valid.password = "valid";
        }
        if (!pws_match) {
            n_valid.confirmPassword = "invalid";
        } else {
            if (pw_valid) {
                n_valid.confirmPassword = "valid";
            } else {
                n_valid.confirmPassword = "unset";
            }
        }
        if (!email_valid) {
            n_valid.email = "invalid";
        } else {
            n_valid.email = "valid";
        }

        //Update state
        setValid({ ...n_valid });

        // If all fields are valid, return true
        if (
            n_valid.email === "valid" &&
            n_valid.password === "valid" &&
            n_valid.confirmPassword === "valid"
        ) {
            return true;
        } else {
            return false;
        }
    };

    return (
        <div className={styles.form}>
            <form
                id="signupForm"
                method="POST"
                onSubmit={async (event) => {
                    event.preventDefault();
                    // Validate form on the client side
                    const clientSideValid = validateFormClientSide(event);
                    if (!clientSideValid) {
                        return;
                    }

                    // Submit form on the server side
                    // This returns a user object if successful
                    // else it returns an 400 and validation object
                    await submitForm(event)
                        .then((user: UserData) => {
                            mutateUser(user);
                        })
                        .catch((error: Error | Validation) => {
                            if (error instanceof Error) {
                                setValid({ ...valid, email: "invalid" });
                            }
                            if ((error as Validation).email) {
                                setValid(error as Validation);
                            }
                        });
                }}
                //onKeyUp={validateFormClientSide}
            >
                <fieldset>
                    <div className={styles.header}>Create an account</div>

                    <EmailInput
                        isValid={valid.email == "valid"}
                        isInvalid={valid.email == "invalid"}
                    />
                    <PasswordInput
                        isValid={valid.password == "valid"}
                        isInvalid={valid.password == "invalid"}
                    />
                    <ConfirmPasswordInput
                        isValid={valid.confirmPassword == "valid"}
                        isInvalid={valid.confirmPassword == "invalid"}
                    />

                    <div className="d-flex align-items-center text-center mb-3 flex-column">
                        <button
                            className="btn btn-primary btn-block fa-lg gradient-custom-2 mb-2"
                            style={{ width: "90%" }}
                            type="submit"
                        >
                            Sign up
                        </button>
                    </div>
                    <div className="d-flex align-items-center justify-content-center">
                        <p className="mb-0 me-2">Already have an account?</p>
                        <Link href="/login">
                            <button
                                type="button"
                                className="btn btn-outline-dark"
                            >
                                Login
                            </button>
                        </Link>
                    </div>
                </fieldset>
            </form>
        </div>
    );
}
