import useUser from "lib/useUser";
import { VerifyPayload } from "pages/api/authentication/verify";
import { useEffect, useState } from "react";

import styles from "./Form.module.scss";

export function VerifyForm({
    onSuccess,
    link_error,
}: {
    onSuccess: (payload: VerifyPayload) => void;
    link_error?: string;
}) {
    const { user } = useUser();
    const [error, _setError] = useState(null);

    const setError = (error: string, link = false) => {
        let html;
        if (link) {
            html = (
                <>
                    <strong>Link verification failed: </strong>
                    {error.replace("Token", "Code")}
                </>
            );
        } else {
            html = (
                <>
                    <strong>Verification failed: </strong>
                    {error.replace("Token", "Code")}
                </>
            );
        }
        _setError(html);
        setTimeout(() => {
            _setError(null);
        }, 15000);
    };

    useEffect(() => {
        if (link_error) {
            setError(link_error, true);
        }
    }, [link_error]);

    let errorDiv = null;
    if (error) {
        errorDiv = (
            <div className="alert alert-danger" role="alert">
                {error}
            </div>
        );
    }

    // Check if user is already verified
    if (user && user.emailValidated) {
        onSuccess({ success: true });
    }

    return (
        <div className={styles.form}>
            <form
                onSubmit={async (event) => {
                    const payload: VerifyPayload = await onFormSubmit(event);
                    if (payload.success) {
                        onSuccess(payload);
                    } else {
                        setError(payload.error || "Unknown error");
                    }
                }}
            >
                <div className={styles.header}>
                    To verify your email please paste the code we sent you in
                    the field below.
                </div>
                {errorDiv}
                <div>
                    <div className={styles.inputField + " form-outline mb-3"}>
                        <input
                            className={"w-100 form-control form__input"}
                            type="text"
                            name="token"
                            id="token"
                            placeholder=" "
                            autoComplete="off"
                            required
                        />
                        <label htmlFor="token">Code</label>
                    </div>
                    <div className="d-flex align-items-center text-center flex-column mb-3">
                        <button
                            className="btn btn-primary btn-block fa-lg gradient-custom-2 "
                            style={{ width: "90%" }}
                            type="submit"
                        >
                            Verify
                        </button>
                    </div>
                </div>
            </form>
            {user ? <ResendButton email={user.email} /> : null}
        </div>
    );
}

async function onFormSubmit(event) {
    event.preventDefault();
    const token = event.target.token.value;

    const response = await fetch("/api/authentication/verify", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify({ token }),
    });

    const data = await response.json();

    if (response.ok) {
        return data;
    }
    throw new Error(response.statusText);
}

function ResendButton({ email }: { email: string }) {
    const [resendSuccess, setResendSuccess] = useState(false);

    if (resendSuccess) {
        return <EnvelopeAnimation email={email} />;
    }

    return (
        <div className="d-flex align-items-center justify-content-center flex-column">
            <p className="mb-2 me-2">
                <strong>Didn&apos;t receive an email?</strong>
            </p>
            <button
                className="btn btn-outline-dark"
                style={{ width: "90%" }}
                onClick={async (event) => {
                    event.preventDefault();
                    const success = await resendEmail().catch(() => {
                        return false;
                    });
                    setResendSuccess(success);
                }}
                disabled={resendSuccess}
            >
                Resend verification code to <i>{email}</i>
            </button>
        </div>
    );
}

function EnvelopeAnimation({ email }) {
    const icons = [
        <i key="env" className="bi bi-envelope"></i>,
        <i key="op" className="bi bi-envelope-open"></i>,
        <i key="pap" className="bi bi-envelope-paper"></i>,
        <i key="env" className="bi bi-envelope"></i>,
    ];

    const [currentFrame, setCurrentFrame] = useState(0);

    useEffect(() => {
        const interval = setInterval(() => {
            setCurrentFrame((currentFrame + 1) % icons.length);
        }, 500);
        return () => clearInterval(interval);
    }, [currentFrame, icons.length]);

    return (
        <div className="d-flex align-items-center justify-content-center flex-column mt-3">
            <p className="mb-2 me-2 text-center">
                <strong>Verification code sent </strong>to <i>{email}</i>!
            </p>
            <div className="d-flex align-items-center justify-content-center flex-column">
                {icons[currentFrame]}
            </div>
        </div>
    );
}

async function resendEmail() {
    const response = await fetch("/api/authentication/resendVerifyEmail", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
    });
    const data = await response.json();
    if (response.ok) {
        return data;
    }
    throw new Error(response.statusText);
}
