import { useEffect, useRef, useState } from "react";

import styles from "./Resizer.module.scss";

/** Resizer component
 *  This component is used to resize two adjacent html elements
 *  to match the size of their parents.
 */
export default function Resizer({ min_previous = 8, max_previous = 100 }) {
    const resizerRef = useRef<HTMLDivElement>(null);
    const previousElementRef = useRef<HTMLElement>(null);
    const nextElementRef = useRef<HTMLElement>(null);
    const parentElementRef = useRef<HTMLElement>(null);

    const [mirrorImg, setMirrorImg] = useState(true);

    // Get the elements surrounding the resizer
    useEffect(() => {
        previousElementRef.current = resizerRef.current
            .previousElementSibling as HTMLElement;
        nextElementRef.current = resizerRef.current
            .nextElementSibling as HTMLElement;
        parentElementRef.current = resizerRef.current
            .parentElement as HTMLElement;

        // Set parent width to initial value ~ 2 pages
        previousElementRef.current.style.width = "350px";
        nextElementRef.current.style.width = "calc(100% - 10px - 350px)";
    }, [resizerRef]);

    // The current/mousdown position of the mouse and the previous element
    const pointer = { x: 0, y: 0 };
    let prevWidth = 0;

    function pointerDownHandler(e: React.PointerEvent) {
        // Recheck the elements
        previousElementRef.current = resizerRef.current
            .previousElementSibling as HTMLElement;
        nextElementRef.current = resizerRef.current
            .nextElementSibling as HTMLElement;
        parentElementRef.current = resizerRef.current
            .parentElement as HTMLElement;

        console.debug("pointerDownHandler");
        console.debug(e);
        console.debug(e.nativeEvent);

        pointer.x = e.clientX;
        pointer.y = e.clientY;
        prevWidth = previousElementRef.current.getBoundingClientRect().width;

        // Attach the listeners to `document`
        document.addEventListener("pointermove", pointerMoveHandler);
        document.addEventListener("pointerup", pointerUpHandler);
    }

    function pointerMoveHandler(e: PointerEvent) {
        // How far the mouse has been moved
        const dx = e.pageX - pointer.x;
        //const dy = e.pageY - mouse.y;
        const width_parent =
            parentElementRef.current.getBoundingClientRect().width;

        let newPrevWidth = Math.min(
            Math.max(((prevWidth + dx) * 100) / width_parent, min_previous),
            max_previous
        );
        if (newPrevWidth == min_previous) {
            newPrevWidth = 0;
            previousElementRef.current.classList.add("d-none");
        }
        if (newPrevWidth > min_previous) {
            previousElementRef.current.style.display = "flex";
            previousElementRef.current.classList.remove("d-none");
        }

        previousElementRef.current.style.width = `${newPrevWidth}%`;
        nextElementRef.current.style.width = `calc(${
            100 - newPrevWidth
        }% - 12px)`; //12px resizer size

        resizerRef.current.style.cursor = "col-resize";
        document.body.style.cursor = "col-resize";

        previousElementRef.current.style.userSelect = "none";
        previousElementRef.current.style.pointerEvents = "none";

        nextElementRef.current.style.userSelect = "none";
        nextElementRef.current.style.pointerEvents = "none";

        nextElementRef.current.style.userSelect = "none";
        nextElementRef.current.style.pointerEvents = "none";
    }

    function check_image_mirror() {
        //Mirror images if needed
        if (
            parseFloat(previousElementRef.current.style.width.slice(0, -1)) >=
            min_previous
        ) {
            setMirrorImg(true);
        } else if (
            parseFloat(previousElementRef.current.style.width.slice(0, -1)) <
                min_previous ||
            parseFloat(previousElementRef.current.style.width.slice(0, -1)) == 0
        ) {
            setMirrorImg(false);
        }
    }

    function pointerUpHandler() {
        resizerRef.current.style.removeProperty("cursor");
        document.body.style.removeProperty("cursor");

        previousElementRef.current.style.removeProperty("user-select");
        previousElementRef.current.style.removeProperty("pointer-events");

        nextElementRef.current.style.removeProperty("user-select");
        nextElementRef.current.style.removeProperty("pointer-events");

        // Remove the handlers
        document.removeEventListener("pointermove", pointerMoveHandler);
        document.removeEventListener("pointerup", pointerUpHandler);
        check_image_mirror();
    }

    const [preDBWidth, setPreDBWidth] = useState(0);
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    function doubleClickHandler(e: React.MouseEvent) {
        if (!previousElementRef.current.classList.contains("d-none")) {
            setPreDBWidth(
                (previousElementRef.current.getBoundingClientRect().width *
                    100) /
                    parentElementRef.current.getBoundingClientRect().width
            );
            previousElementRef.current.classList.add("d-none");
            previousElementRef.current.style.width = `0%`;
            nextElementRef.current.style.width = `100%`;
        } else {
            previousElementRef.current.classList.remove("d-none");
            if (preDBWidth < min_previous) {
                previousElementRef.current.style.width = "20%";
            } else {
                previousElementRef.current.style.width = preDBWidth + "%";
                nextElementRef.current.style.width = 100 - preDBWidth + "%";
            }
        }
        check_image_mirror();
        const sel = window.getSelection
            ? window.getSelection()
            : document.getSelection();
        if (sel) {
            if (sel.removeAllRanges) {
                sel.removeAllRanges();
            } else if (sel.empty) {
                sel.empty();
            }
        }
    }

    let cls = null;
    if (mirrorImg) {
        cls = styles.icon_flipped;
    }
    return (
        <div
            onPointerDown={pointerDownHandler}
            onDoubleClick={doubleClickHandler}
            ref={resizerRef}
            className={styles.resizer}
        >
            <div className={"img-container " + cls}>
                <i className="bi bi-caret-right-fill"></i>
                <i className="bi bi-caret-right-fill"></i>
                <i className="bi bi-caret-right-fill"></i>
            </div>
        </div>
    );
}
