import { debounce } from "lib/utils";
import {
    createContext,
    type MutableRefObject,
    useCallback,
    useContext,
    useEffect,
    useRef,
    useState,
} from "react";

import { ToolBarContext } from "./toolbar";
import { getContext } from "./tools";
import { MovePageToolContext } from "./tools/movePage.tool";
import { _centerElement } from "./tools/movePage/resizable";
import { Overlay } from "./tools/typing";
import { OverlayWorker, useOverlayWorker } from "./worker/overlayWorker";
import { PageWorker, usePageWorker } from "./worker/pageWorker";

import { PageData } from "controller/page";

import styles from "./page.module.scss";

/** Just renders a page no functions added
 * - Render using webworker
 * - for additional function see tools folder
 */
export default function SinglePageViewer({
    idx = 0,
    divProps,
}: {
    idx?: number;
    divProps?: React.HTMLAttributes<HTMLDivElement>;
}) {
    // Elements on the page shared via context
    const {
        mainCanvasRef,
        overlayCanvasRef,
        draggableRef,
        wrapperRef,
        containerRef,
        setRef,
        pageDataRef,
        visible,
        workers,
        overlayWorkers,
    } = useContext(PagesContext);

    /** The active tool might add an overlay
     * This is added to the virtual dom. Overlays
     * are optional and not all tools have an overlay
     */
    const { activeTool } = useContext(ToolBarContext);
    const context = useContext(getContext(activeTool));
    let overlay = null;
    let overlay_page_idx = null;
    if (context && "overlay" in context && context.overlay) {
        const o = context.overlay as Overlay;
        overlay = o.dom;
        overlay_page_idx = o.page_idx;
    }
    /**
     * The fitElement function is needed by default thus if we cant
     * find the context we will create a dummy function
     */
    const { centerPage } = useContext(MovePageToolContext) || {
        centerPage: () => {
            /*_fitElement(
                draggableRef.current.get(idx),
                containerRef.current.get(idx)
            );
            */
            _centerElement(
                draggableRef.current.get(idx),
                wrapperRef.current.get(idx),
                containerRef.current.get(idx)
            );
        },
    };

    /** Initial render
     * - Worker is created
     * - Page is rendered
     * - Page is resized to fit the draggable
     * - Overlay is resized to fit the draggable
     */
    useEffect(() => {
        if (
            !visible ||
            !workers ||
            !workers[idx] ||
            !overlayWorkers ||
            !draggableRef.current ||
            !pageDataRef.current
        )
            return;

        /** Initial setup only run once
         *
         */
        console.log("Initial setup all");
        const worker = workers[idx];
        const overlayWorker = overlayWorkers[idx];
        const draggable = draggableRef.current.get(idx);
        const pageData = pageDataRef.current.get(idx);
        worker.resize(draggable.clientWidth, draggable.clientHeight);
        overlayWorker.resize(draggable.clientWidth, draggable.clientHeight);
        worker.setupPage(pageData);
        overlayWorker.setupPage(pageData);
        centerPage(idx);

        /** Setup Rerender on resize
         * - Debounce to prevent multiple renders on rapid resize
         */
        const debounce_render = debounce((width, height) => {
            overlayWorker.resize(width, height);
            worker.resize(width, height);
            overlayWorker.render();
            worker.renderPage();
        }, 100);

        const onResize = () => {
            const draggable = draggableRef.current.get(idx);
            const width = draggable.clientWidth;
            const height = draggable.clientHeight;
            debounce_render(width, height);
        };

        const resizeObserver = new ResizeObserver(onResize);
        resizeObserver.observe(draggable);
        return () => {
            resizeObserver.disconnect();
        };
    }, [
        draggableRef,
        centerPage,
        idx,
        overlayWorkers,
        pageDataRef,
        workers,
        visible,
    ]);

    return (
        <>
            <div
                className={styles.container}
                ref={setRef(containerRef, idx)}
                {...divProps}
            >
                <div
                    className={styles.canvas_wrapper}
                    ref={setRef(wrapperRef, idx)}
                    style={{ visibility: visible ? "visible" : "hidden" }}
                >
                    {idx == overlay_page_idx ? overlay : null}
                    <div
                        className={styles.canvas_draggable}
                        ref={setRef(draggableRef, idx)}
                    >
                        <canvas
                            ref={setRef(overlayCanvasRef, idx)}
                            style={{ zIndex: 1 }}
                        />
                        <canvas ref={setRef(mainCanvasRef, idx)} />
                    </div>
                </div>
            </div>
        </>
    );
}

export function MultiPageViewer({
    divProps,
}: {
    divProps?: React.HTMLAttributes<HTMLDivElement>;
}) {
    const { numPages } = useContext(PagesContext);

    return (
        <>
            {numPages == 0 ? (
                <div className={styles.placeholder}>No page selected</div>
            ) : null}
            {Array(numPages)
                .fill(0)
                .map((_, i) => (
                    <SinglePageViewer key={i} divProps={divProps} idx={i} />
                ))}
        </>
    );
}

// Create a mapping type to allow for easier indexing
export type EleMap<T> = Map<number, T>;
export type SetEleRef = <T>(
    ref: MutableRefObject<EleMap<T>>,
    idx: number
) => (el: T) => void;

interface PagesContext {
    /** References to all elements in the viewer
     */
    draggableRef: MutableRefObject<EleMap<HTMLDivElement>>;
    wrapperRef: MutableRefObject<EleMap<HTMLDivElement>>;
    containerRef: MutableRefObject<EleMap<HTMLDivElement>>;
    mainCanvasRef: MutableRefObject<EleMap<HTMLCanvasElement>>;
    overlayCanvasRef: MutableRefObject<EleMap<HTMLCanvasElement>>;
    pageDataRef: MutableRefObject<EleMap<PageData>>;
    setRef: SetEleRef;

    /** Worker to render the page
     */
    workers: PageWorker[];
    overlayWorkers: OverlayWorker[];
    addPageData: (pageData: PageData, idx: number) => void;
    setPageData: (pagesData: PageData[]) => void;
    numPages: number;

    /** Events
     */
    visible: boolean;
}

export const PagesContext = createContext<PagesContext>(undefined);

/** To make it easier to pass the elements
 * to the tools we will use a context provider
 */
export function PagesContextProvider({
    children,
}: {
    children: React.ReactNode;
}) {
    /** Define the references as arrays as we
     * allow multiple pages to be rendered
     */
    const mainCanvasRef = useRef<EleMap<HTMLCanvasElement>>(null),
        overlayCanvasRef = useRef<EleMap<HTMLCanvasElement>>(null),
        draggableRef = useRef<EleMap<HTMLDivElement>>(null),
        wrapperRef = useRef<EleMap<HTMLDivElement>>(null),
        containerRef = useRef<EleMap<HTMLDivElement>>(null),
        pageDataRef = useRef<EleMap<PageData>>(null);

    function setRef<T>(ref: MutableRefObject<EleMap<T>>, idx: number) {
        return (element: T) => {
            if (!element) return;
            if (!ref.current) ref.current = new Map<number, T>();
            ref.current.set(idx, element);
        };
    }

    /** Create worker to outsource page and overlay rendering
     */
    const [numPages, setNumPages] = useState(0);
    const { workers } = usePageWorker(mainCanvasRef, numPages);
    const { overlayWorkers } = useOverlayWorker(overlayCanvasRef, numPages);

    // Enables to set the page data after the component is mounted
    // i.e. check if a page is selected
    const [visible, setVisible] = useState(false);

    const updateNumPages = useCallback(() => {
        setNumPages(pageDataRef.current.size);
    }, []);

    const updateHash = useCallback(() => {
        // Set the location hash to the page number
        const l = pageDataRef.current.size;
        let hash = "";
        let i = 0;
        pageDataRef.current.forEach((pD) => {
            hash += "page=" + (pD.page_number + 1);
            if (i < l - 1) {
                hash += "&";
            }
            i++;
        });
        window.location.hash = hash;
    }, []);

    /** Set the page data and render the page(s) using the worker
     */
    const addPageData = useCallback(
        (pageData: PageData, idx: number) => {
            if (!pageDataRef.current) pageDataRef.current = new Map();

            // Check if pageData is already set
            if (
                pageDataRef.current.has(idx) &&
                pageDataRef.current.get(idx).id == pageData.id
            ) {
                return;
            }
            pageDataRef.current.set(idx, pageData);

            // Update number of pages and url hash
            updateNumPages();
            updateHash();

            // TODO: check
            setVisible(true);
            //events.current.emit("pageChange", pageData, idx);
        },
        [updateHash, updateNumPages]
    );

    const setPageData = useCallback(
        (pagesData: PageData[]) => {
            if (!pageDataRef.current) pageDataRef.current = new Map();
            // Get size
            const l = pagesData.length;

            // Add all the page data
            pagesData.forEach((pD, idx) => {
                pageDataRef.current.set(idx, pD);
            });
            let s = pageDataRef.current.size;

            // Check length and remove if necessary
            if (s > l) {
                for (let i = l; i < s; i++) {
                    pageDataRef.current.delete(i);
                }
            }

            // Update number of pages and url hash
            updateNumPages();
            updateHash();

            // Render the pages
            s = pageDataRef.current.size;
            if (!workers || workers.length < s) return;
            for (let i = 0; i < s; i++) {
                const pD = pageDataRef.current.get(i);
                if (pD) {
                    overlayWorkers[i].setupPage(pD);
                    workers[i].setupPage(pD);
                }
            }
            setVisible(true);
        },
        [overlayWorkers, updateHash, updateNumPages, workers]
    );

    useEffect(() => {
        // Rerender all every time the workers change
        if (
            !workers ||
            !pageDataRef.current ||
            pageDataRef.current.size < workers.length
        )
            return;
        const pageData = pageDataRef.current;
        pageData.forEach((pD, idx) => {
            workers[idx].setupPage(pD);
        });
    }, [workers]);

    const pageRefs: PagesContext = {
        // Refs
        draggableRef,
        wrapperRef,
        containerRef,
        mainCanvasRef,
        overlayCanvasRef,
        pageDataRef,
        setRef,

        // Callbacks
        setPageData,
        addPageData,
        numPages: numPages,

        // Events
        visible,

        // Workers
        workers,
        overlayWorkers: overlayWorkers,
    };
    return (
        <PagesContext.Provider value={pageRefs}>
            {children}
        </PagesContext.Provider>
    );
}
