import { fetcher } from "lib/fetcher";
import { useDebounce } from "lib/utils";
import {
    createContext,
    Dispatch,
    SetStateAction,
    useCallback,
    useContext,
    useEffect,
    useState,
} from "react";
import useSWR from "swr";

import DownloadBtn from "./downloadBtn";
import FullScreenButton from "./fullScreenBtn";
import LayoutSwitchBtn from "./layoutSwitchBtn";
import { PagePreviews } from "./pagePreviews";
import { PageSwitchBtns } from "./pageSwitchBtns";
import { UnplacedSnipPreviews } from "./unplacedSnipPreviews";

import { Book } from "controller/book";
import { BaseSnip } from "snips/general/base";

import Loading from "components/utils/Loading";

import styles from "./sidebar.module.scss";

export interface SidebarContext {
    activeTab: string;
    setActiveTab: (tab: string) => void;
    searchTermPages: string;
    setSearchTermPages: (term: string) => void;
    searchTermSnips: string;
    setSearchTermSnips: (term: string) => void;

    activePageId: number[];
    setActivePageId: Dispatch<SetStateAction<number[]>>;

    activeSnipId: number;
    setActiveSnipId: Dispatch<SetStateAction<number>>;

    snips: BaseSnip[];
    setSnips: Dispatch<SetStateAction<BaseSnip[]>>;
    activeSnip: BaseSnip;
}
export const SidebarContext = createContext<SidebarContext>(undefined);

export function SidebarContextProvider({
    children,
}: {
    children: React.ReactNode;
}) {
    /** Logic for the search state
     * here we use debounce logic to delay inputs
     * -> better for performance and less requests
     */
    const [searchTermPages, setSearchTermPages] = useState("");
    const [searchTermSnips, setSearchTermSnips] = useState("");

    // Debounce search term so that it only gives us latest value ...
    // ... if searchTerm has not been updated within last 500ms.
    // The goal is to only have the API call fire when user stops typing ...
    // ... so that we aren't hitting our API rapidly.
    const debouncedSearchTermPages = useDebounce(searchTermPages, 500);
    const debouncedSearchTermSnips = useDebounce(searchTermSnips, 500);

    // State and setter for currently active tab
    const [activeTab, setActiveTab] = useState<string>("pages");

    // State for active page
    const [activePageId, setActivePageId] = useState<number[]>([]);

    // State for unplaced snips, we need to keep track of the active one
    // and all the others
    const [snips, setSnips] = useState<BaseSnip[]>(null);
    const [activeSnipId, setActiveSnipId] = useState<number>(null);
    let activeSnip = undefined;
    if (snips && activeSnipId) {
        activeSnip = snips.find((s) => s.id === activeSnipId);
    }

    const ret: SidebarContext = {
        activeTab,
        setActiveTab,
        searchTermPages: debouncedSearchTermPages,
        setSearchTermPages,
        searchTermSnips: debouncedSearchTermSnips,
        setSearchTermSnips,
        activePageId,
        setActivePageId,
        activeSnipId,
        setActiveSnipId,
        snips,
        setSnips,
        activeSnip,
    };

    return (
        <SidebarContext.Provider value={ret}>
            {children}
        </SidebarContext.Provider>
    );
}

export function Sidebar({
    book,
    showSnipsToPlace,
    tokenAlert = false,
}: {
    book: Book;
    showSnipsToPlace: boolean;
    tokenAlert?: boolean;
}) {
    const {
        searchTermPages,
        setSearchTermPages,
        searchTermSnips,
        activeTab,
        setActiveTab,
    } = useContext(SidebarContext);

    // Get all pages that match the search term
    // Using an api request, since we also want to filter
    // the containing snips, we dont want to transmit all the data
    // to the client.
    const id = book?.id;

    const { data: filter, isValidating } = useSWR(() => {
        if (activeTab === "pages") {
            return searchTermPages
                ? `/api/pages/search?id=${id}&search=${searchTermPages}`
                : undefined;
        }
        return searchTermSnips
            ? `/api/snips/search?id=${id}&search=${searchTermSnips}`
            : undefined;
    }, fetcher);

    // Check if book is read only i.e. finished
    const alerts = [];
    if (book?.finished) {
        alerts.push(
            <div
                className={styles.readOnly + " alert alert-danger"}
                key={"finished"}
            >
                This book is finished! It can no longer be edited. All changes
                are only local and will not be saved.
            </div>
        );
    }

    if (tokenAlert) {
        alerts.push(
            <div
                className={styles.readOnly + " alert alert-danger"}
                key={"token"}
            >
                You are using a temporary access token you shall only view this
                book.
            </div>
        );
    }

    // Page previews for the pages tab
    const [previews, setPreviews] = useState<string[]>(null);

    const updatePreview = useCallback(
        (page_id: number, img: string) => {
            console.log("Update preview");
            const page = book.pages.find((page) => page.id === page_id);
            if (page) {
                setPreviews((src) => {
                    const newSrc = [...src];
                    newSrc[page.page_number] = img;
                    return newSrc;
                });
            }
        },
        [book.pages]
    );
    useEffect(() => {
        book.events.addListener("updatePreview", updatePreview);
        return () => {
            book.events.removeListener("updatePreview", updatePreview);
        };
    }, [book, updatePreview]);
    useEffect(() => {
        book.getPreviews().then((src) => {
            setPreviews(src);
        });
    }, [book]);

    return (
        <div className={styles.sideBar}>
            <ul className="nav nav-tabs nav-fill" role="tablist">
                <li className="nav-item" role="presentation">
                    <button
                        className="nav-link"
                        onClick={() => {
                            setActiveTab("pages");
                        }}
                        data-active={activeTab == "pages"}
                    >
                        Pages
                    </button>
                </li>
                {showSnipsToPlace ? (
                    <li className="nav-item" role="presentation">
                        <button
                            className="nav-link"
                            onClick={() => {
                                setActiveTab("snips");
                            }}
                            data-active={activeTab == "snips"}
                        >
                            Snips to place
                        </button>
                    </li>
                ) : null}
            </ul>
            {alerts}
            <div className={styles.topNavigation}>
                <SearchBar
                    isValidating={isValidating}
                    onChange={(e) => {
                        console.log("Search", e.target.value);
                        setSearchTermPages(e.target.value);
                    }}
                />
                <PageSwitchBtns book={book} />
            </div>
            <TabPages
                active={activeTab === "pages"}
                previews={previews}
                book={book}
                filter={filter}
            />
            {showSnipsToPlace ? (
                <TabSnips
                    active={activeTab === "snips"}
                    book={book}
                    filter={filter}
                />
            ) : null}
            <div className={styles.extraFunctions}>
                <DownloadBtn book={book} />
                <FullScreenButton />
                <LayoutSwitchBtn book={book} />
            </div>
        </div>
    );
}

function TabPages({
    active,
    previews,
    book,
    filter,
}: {
    active: boolean;
    previews: string[];
    book: Book;
    filter: number[];
}) {
    if (!previews) {
        return (
            <div
                className={styles.tabContent}
                style={{
                    display: "flex",
                    justifyContent: "center",
                }}
            >
                <Loading />
            </div>
        );
    }

    return (
        <div
            className={styles.tabContent}
            style={{ display: active ? "block" : "none" }}
        >
            <PagePreviews book={book} filter_ids={filter} previews={previews} />
        </div>
    );
}

function TabSnips({
    active,
    book,
    filter,
}: {
    active: boolean;
    book: Book;
    filter: number[];
}) {
    return (
        <div
            className={styles.tabContent}
            style={{ display: active ? "block" : "none" }}
        >
            <UnplacedSnipPreviews book={book} filter_ids={filter} />
        </div>
    );
}

/** Simple searchbar only containing a search input
 * also some feedback for the user if a query is
 * currently being processed i.e. isValidating
 *  */
function SearchBar({ onChange = undefined, isValidating = false }) {
    return (
        <div className={styles.searchBar}>
            <input type="text" placeholder="Search" onChange={onChange} />
            {isValidating ? (
                <div className="loadingSpinner">
                    <span
                        className="spinner-grow spinner-grow-sm"
                        role="status"
                    ></span>
                </div>
            ) : null}
        </div>
    );
}
