/* eslint-disable @next/next/no-img-element */
import {
    MouseEvent,
    MouseEventHandler,
    useContext,
    useEffect,
    useRef,
} from "react";

import { PagesContext } from "../page";
import { SidebarContext } from ".";

import { Book } from "controller/book";
import { PageData } from "controller/page";

import styles from "./sidebar.module.scss";

export function PagePreviews({
    book,
    filter_ids,
    previews,
}: {
    book: Book;
    filter_ids: number[];
    previews: string[];
}) {
    const { setPageData } = useContext(PagesContext);
    //For each page context

    // Currently active page in preview
    const { activePageId, setActivePageId } = useContext(SidebarContext);

    // If the page has a hash in the url, scroll to that page
    // match page-number
    useEffect(() => {
        const hash = window.location.hash;
        //Get all matches url#page=[any number positive or negative]
        if (hash && hash.match(/page/g)) {
            const pages = hash
                .match(/page=-?\d+/g)
                .map((p) => parseInt(p.split("=")[1]))
                .sort((a, b) => a - b)
                .map((p) => {
                    if (p == -1) {
                        return book.pages.at(-1);
                    }
                    return book.pages.at(p - 1);
                })
                .filter((p) => p !== undefined);

            if (pages) {
                setActivePageId(pages.map((p) => p.id));
                setPageData(pages);
            }
        }
    }, [book.pages, setActivePageId, setPageData]);

    /** If a modifier key is pressed we allow to
     * select multiple pages
     */
    const onClickHandler = (
        event: MouseEvent<HTMLDivElement>,
        page: PageData
    ) => {
        event.preventDefault();
        // check if ctrl or shift is pressed

        if (event.ctrlKey || event.shiftKey) {
            // add page to active pages up to max 2
            setActivePageId((prev) => {
                if (prev.length < 2) {
                    if (!prev.includes(page.id)) {
                        prev.push(page.id);
                    } else {
                        prev = prev.filter((p) => p !== page.id);
                    }
                    prev = [...prev];
                }
                return prev;
            });
        } else {
            // set page as active
            setActivePageId((prev) => {
                if (prev.includes(page.id) && prev.length === 1) {
                    return prev;
                } else if (prev.includes(page.id) && prev.length > 1) {
                    return prev.filter((p) => p !== page.id);
                }
                prev.pop();
                prev.unshift(page.id);
                return [...prev];
            });
        }
    };

    // Set the page data if the active page id changes
    useEffect(() => {
        if (activePageId) {
            const pages = activePageId
                .map((id) => book.pages.find((p) => p.id === id))
                .filter((p) => p !== undefined);
            setPageData(pages);
        }
    }, [activePageId, book.pages, setPageData]);

    // subscribe to page insert event
    const content = book.pages.map((page, i) => {
        return (
            <PagePreview
                key={i}
                page={page}
                onClick={(e) => onClickHandler(e, page)}
                src={previews[i]}
                active={activePageId?.includes(page.id)}
                hidden={filter_ids ? !filter_ids.includes(page.id) : false}
            />
        );
    });

    return <div className={styles.pagePreviewContainer}>{content}</div>;
}

function PagePreview({
    page,
    onClick,
    active,
    src,
    hidden = false,
}: {
    page: PageData;
    onClick: MouseEventHandler<HTMLDivElement>;
    active: boolean;
    src: string;
    hidden?: boolean;
}) {
    // Scroll to page preview when active
    const ref = useRef<HTMLImageElement>(null);

    //Scroll into view
    useEffect(() => {
        if (active && !isInViewport(ref.current)) {
            ref.current.scrollIntoView({
                behavior: "smooth",
                block: "end",
                inline: "nearest",
            });
        }
    }, [active]);

    return (
        <div
            className={styles.pagePreview}
            data-active={active}
            data-hidden={!active ? hidden : false}
            onClick={onClick}
        >
            <img
                ref={ref}
                src={src}
                width="100%"
                height="100%"
                alt={page.id.toString()}
            />
            <div className={styles.number}>
                {(page.page_number + 1).toString()}
            </div>
        </div>
    );
}

function isInViewport(element) {
    const rect = element.getBoundingClientRect();
    return (
        rect.top >= 0 &&
        rect.left >= 0 &&
        rect.bottom <=
            (window.innerHeight || document.documentElement.clientHeight) &&
        rect.right <=
            (window.innerWidth || document.documentElement.clientWidth)
    );
}
