import { useCallback, useContext, useEffect, useState } from "react";

import { PagesContext } from "../page";
import { SidebarContext } from ".";

import { Book } from "controller/book";
import { PageData } from "controller/page";

import styles from "./sidebar.module.scss";

export function PageSwitchBtns({ book }: { book: Book }) {
    const [isLocked, setIsLocked] = useState(false);

    // Add event listener for page lock
    useEffect(() => {
        if (book?.finished) {
            setIsLocked(true);
        }
    }, [book]);

    const { nextPage, prevPage, newPage, boundaries } = usePageSwitcher(book);

    return (
        <div className={styles.extraBtns}>
            <button onClick={prevPage} disabled={boundaries.first}>
                <File_left
                    width="25px"
                    height="30px"
                    alt="p"
                    fill={boundaries.first ? "#aca8a9" : "var(--bs-dark)"}
                />
            </button>
            <button onClick={nextPage} disabled={boundaries.last}>
                <File_right
                    width="25px"
                    height="30px"
                    alt="n"
                    fill={boundaries.last ? "#aca8a9" : "var(--bs-dark)"}
                />
            </button>
            <button onClick={newPage} disabled={isLocked}>
                <File_plus
                    width="25px"
                    height="30px"
                    alt="+"
                    fill={isLocked ? "#aca8a9" : "var(--bs-dark)"}
                />
            </button>
        </div>
    );
}

/** Helper hook to encapsulate the page switch function as they take quite some space
 *
 */
function usePageSwitcher(book: Book) {
    const { pageDataRef } = useContext(PagesContext);
    const { activePageId, setActivePageId } = useContext(SidebarContext);

    // check for last/first page selected
    const [boundaries, setBoudaries] = useState<{
        first: boolean;
        last: boolean;
    }>({ first: false, last: false });

    const updateBoudariesCheck = useCallback(
        (pagesData: PageData | PageData[]) => {
            if (!Array.isArray(pagesData)) {
                pagesData = [pagesData];
            }
            const page_numbers = pagesData.map((p) => p.page_number);
            const first = page_numbers.some((n) => n <= 0);
            const last = page_numbers.some((n) => n >= book.pages.length - 1);
            setBoudaries({ first, last });
        },
        [book.pages.length]
    );

    const incPage = useCallback(
        (inc: number) => {
            const pagesData = pageDataRef.current;
            // Check if next pages are valid
            const next_page_nums = [];
            pagesData.forEach((p) => {
                next_page_nums.push(p.page_number + inc);
            });

            // Get next pages
            const nextPages = next_page_nums
                .map((n) => {
                    return book.pages.at(n);
                })
                .filter((p) => p !== undefined);

            const nextPageIds = nextPages.map((p) => p.id);

            setActivePageId(nextPageIds);
        },
        [book.pages, pageDataRef, setActivePageId]
    );

    async function newPage() {
        const nP = await book.createPage();
        setActivePageId((prevIds) => {
            prevIds.pop();
            prevIds.unshift(nP.id);
            return [...prevIds];
        });
    }

    /** Initial check for boundaries */
    useEffect(() => {
        const pages = activePageId
            .map((id) => book.pages.filter((p) => p.id === id))
            .flat();
        updateBoudariesCheck(pages);
    }, [activePageId, book.pages, updateBoudariesCheck]);

    return {
        nextPage: () => incPage(1),
        prevPage: () => incPage(-1),
        newPage,
        boundaries: boundaries,
    };
}

const File_left = (props) => {
    return (
        <svg
            xmlns="http://www.w3.org/2000/svg"
            width="16"
            height="16"
            fill="currentColor"
            viewBox="0 0 16 16"
            {...props}
        >
            <path d="M2 2a2 2 0 0 1 2-2h8a2 2 0 0 1 2 2v12a2 2 0 0 1-2 2H4a2 2 0 0 1-2-2V2zm10-1H4a1 1 0 0 0-1 1v12a1 1 0 0 0 1 1h8a1 1 0 0 0 1-1V2a1 1 0 0 0-1-1z" />
            <g transform="matrix(-1,0,0,1,44.338982,-0.40677966)">
                <path
                    id="path822"
                    d="m 34.024977,7.9067797 c -0.666667,0 -0.666667,1 0,1 h 3.420118 l -2.147,2.1460013 c -0.472,0.472 0.236,1.18 0.708,0.708 l 3,-3.0000013 c 0.195858,-0.195364 0.195858,-0.512636 0,-0.708 l -3,-2.9999998 c -0.472,-0.4719998 -1.18,0.2360002 -0.708,0.708 l 2.147,2.1459998 z"
                />
            </g>
        </svg>
    );
};

const File_right = (props) => {
    return (
        <svg
            xmlns="http://www.w3.org/2000/svg"
            width="16"
            height="16"
            fill="currentColor"
            viewBox="0 0 16 16"
            {...props}
        >
            <path d="M2 2a2 2 0 0 1 2-2h8a2 2 0 0 1 2 2v12a2 2 0 0 1-2 2H4a2 2 0 0 1-2-2V2zm10-1H4a1 1 0 0 0-1 1v12a1 1 0 0 0 1 1h8a1 1 0 0 0 1-1V2a1 1 0 0 0-1-1z" />
            <g transform="translate(-28.338983,-0.40677966)">
                <path d="m 34.024977,7.9067797 c -0.666667,0 -0.666667,1 0,1 h 3.420118 l -2.147,2.1460013 c -0.472,0.472 0.236,1.18 0.708,0.708 l 3,-3.0000013 c 0.195858,-0.195364 0.195858,-0.512636 0,-0.708 l -3,-2.9999998 c -0.472,-0.4719998 -1.18,0.2360002 -0.708,0.708 l 2.147,2.1459998 z" />
            </g>
        </svg>
    );
};

const File_plus = (props) => {
    return (
        <svg
            xmlns="http://www.w3.org/2000/svg"
            width="16"
            height="16"
            fill="currentColor"
            viewBox="0 0 16 16"
            {...props}
        >
            <path d="M8.5 6a.5.5 0 0 0-1 0v1.5H6a.5.5 0 0 0 0 1h1.5V10a.5.5 0 0 0 1 0V8.5H10a.5.5 0 0 0 0-1H8.5V6z" />
            <path d="M2 2a2 2 0 0 1 2-2h8a2 2 0 0 1 2 2v12a2 2 0 0 1-2 2H4a2 2 0 0 1-2-2V2zm10-1H4a1 1 0 0 0-1 1v12a1 1 0 0 0 1 1h8a1 1 0 0 0 1-1V2a1 1 0 0 0-1-1z" />
        </svg>
    );
};
