import { useContext, useEffect, useRef } from "react";

import { SidebarContext } from ".";
import UploadForm from "./upload";

import { Book } from "controller/book";
import { BaseSnip } from "snips/general/base";
import { get_snip_from_data } from "snips/get_snip_from_data";

import Loading from "components/utils/Loading";

import styles from "./sidebar.module.scss";

export function UnplacedSnipPreviews({
    book,
    filter_ids,
}: {
    book: Book;
    filter_ids: number[];
}) {
    // Currently active snip if placement is active
    const { activeSnipId, setActiveSnipId, snips, setSnips } =
        useContext(SidebarContext);

    useEffect(() => {
        //Get snips
        book.getUnplacedSnips().then((snips) => {
            setSnips(snips.map(get_snip_from_data));
        });

        // new unplaced snip added (can be both from this client or another)
        const insertedIntoQueue = (snip) => {
            setSnips((snips) => {
                return [...snips, get_snip_from_data(snip)];
            });
        };

        // snip was placed (can be both from this client or another)
        const removeUnplacedSnip = (page_id, snip_id) => {
            console.log("snip placed", snip_id);
            setActiveSnipId((activeSnipId) =>
                activeSnipId === snip_id ? null : activeSnipId
            );
            setSnips((snips) => {
                return snips.filter((snip) => snip.id !== snip_id);
            });
        };

        book.events.addListener("book:insertedIntoQueue", insertedIntoQueue);
        book.events.addListener("snip:placed", removeUnplacedSnip);
        return () => {
            book.events.removeListener(
                "book:insertedIntoQueue",
                insertedIntoQueue
            );
            book.events.removeListener("snip:placed", removeUnplacedSnip);
        };
    }, [book, setActiveSnipId, setSnips]);

    if (!snips) return <Loading />;

    // Sort by time
    const snips_decending = snips.sort(function (a, b) {
        return b.created.getTime() - a.created.getTime();
    });

    const content = snips_decending.map((snip, i) => {
        return (
            <SnipPreview
                key={i}
                snip={snip}
                onClick={() => {
                    if (activeSnipId === snip.id) {
                        setActiveSnipId(null);
                    } else setActiveSnipId(snip.id);
                }}
                active={snip.id === activeSnipId}
                hidden={filter_ids ? !filter_ids.includes(snip.id) : false}
            />
        );
    });

    // insert Upload button at start
    content.unshift(
        <UploadForm
            key="upload"
            book={book}
            setActiveSnipId={setActiveSnipId}
        />
    );

    return <div className={styles.previewContainer}>{content}</div>;
}

/** Renders a single snip to a preview canvas
 *  this canvas is quite small 140x200
 *
 * @param param0
 * @returns
 */
function SnipPreview({
    snip,
    onClick,
    active,
    hidden,
}: {
    snip: BaseSnip;
    onClick: () => void;
    active: boolean;
    hidden: boolean;
}) {
    const canvasRef = useRef<HTMLCanvasElement>(null);

    //Render snip
    useEffect(() => {
        if (canvasRef.current) {
            const ctx = canvasRef.current.getContext("2d");
            if (ctx) {
                const trans = ctx.getTransform().inverse();
                const width = trans.a * 140;
                const height = trans.d * 200;

                ctx.clearRect(0, 0, width, height);
                //Scale canvas to fit snip
                const scale = Math.min(140 / snip.width, 200 / snip.height);
                ctx.setTransform(scale, 0, 0, scale, 0, 0);
                snip.render(ctx);
            }
        }
    }, [snip]);

    return (
        <div
            className={styles.snipPreview}
            onClick={onClick}
            data-active={active}
            data-hidden={hidden}
        >
            <canvas ref={canvasRef} width={140} height={200} />
            <div className={styles.type}>{snip.type}</div>
            <div className={styles.timestamp}>
                {to_local_date_string(snip.created)}
            </div>
        </div>
    );
}

const to_local_date_string = (date: Date) => {
    // If the date is today, only show the time
    const timeZone = Intl.DateTimeFormat().resolvedOptions().timeZone;

    if (date.toDateString() === new Date().toDateString()) {
        return new Intl.DateTimeFormat("en-US", {
            hour: "numeric",
            minute: "numeric",
            hour12: false,
        }).format(date);
    } else {
        return new Intl.DateTimeFormat("en-US", {
            year: "numeric",
            month: "short",
            day: "numeric",
            hour12: false,
            timeZone,
        }).format(date);
    }
};
