import {
    createContext,
    Dispatch,
    SetStateAction,
    useContext,
    useEffect,
    useState,
} from "react";
import { usePopper } from "react-popper";

import { SidebarContext } from "../sidebar";
import { getToolComponents } from "../tools";
import { type Tool } from "../tools/typing";

import styles from "./toolbar.module.scss";

interface ToolBarContext {
    activeTool: Tool;
    setActiveTool: (tool: Tool) => void;
    activeSubMenu: Tool;
    setActiveSubMenu: (tool: Tool) => void;
    showInfo: boolean;
    setShowInfo: Dispatch<SetStateAction<boolean>>;
}
export const ToolBarContext = createContext<ToolBarContext>(undefined);

export function ToolBarProvider({
    children,
    initialActiveTool = undefined,
}: {
    children: React.ReactNode;
    initialActiveTool?: Tool;
}) {
    // Instead of just storing the active tool, we could also store the history of tools
    // this would allow us to go back to the previous tool if a tool is deactivated
    const [history, _setHistory] = useState<Tool[]>([initialActiveTool]);

    //Constrain two values in history
    const setHistory = (newHistory: SetStateAction<Tool[]>) => {
        if (typeof newHistory === "function") {
            newHistory = newHistory(history);
        }
        if (newHistory.length > 2) {
            newHistory = newHistory.slice(-2);
        }
        _setHistory(newHistory);
    };

    // The currently active tool is independent of the current show SubMenu
    const [activeSubMenu, _setActiveSubMenu] =
        useState<Tool>(initialActiveTool);

    const { activeSnipId, setActiveSnipId } = useContext(SidebarContext);

    // If info should be shown
    // TODO: user config
    const [showInfo, setShowInfo] = useState<boolean>(false);

    const setActiveTool = (tool: Tool) => {
        // The snip editor is a bit special if it disables we also want
        // to disable the active snip
        if (tool != "snipEditor") {
            if (activeSnipId) {
                setActiveSnipId(undefined);
            }
        }

        if (history.at(-1) === tool) {
            // Remove the tool from the history if it is set twice in a row
            setHistory((prev) => {
                prev.pop();
                prev = prev.filter((t) => t !== "snipEditor");
                return [...prev];
            });
        } else {
            // Add the tool to the history
            setHistory((prev) => [...prev, tool]);
        }
    };

    const setActiveSubMenu = (tool: Tool) => {
        _setActiveSubMenu((prev) => (prev === tool ? undefined : tool));
    };

    const ret: ToolBarContext = {
        activeTool: history.at(-1),
        setActiveTool,
        activeSubMenu,
        setActiveSubMenu,
        showInfo,
        setShowInfo,
    };

    return (
        <ToolBarContext.Provider value={ret}>
            {children}
        </ToolBarContext.Provider>
    );
}

/** Factory to create Toolbar component given a list
 * of tools to show in the toolbar.
 */
export function createToolbar(tools: Tool[]) {
    const toolButtons: JSX.Element[] = [];
    const toolSubMenus = [];
    const toolInfos = [];
    tools.forEach((tool) => {
        const { Button, SubMenu, Info } = getToolComponents(tool);
        toolButtons.push(<Button key={tool + "btn"} />);
        toolSubMenus.push(<SubMenu key={tool + "submenu"} />);
        toolInfos.push(<Info key={tool + "info"} />);
    });

    return function Toolbar() {
        const { activeSubMenu, showInfo, setShowInfo, activeTool } =
            useContext(ToolBarContext);
        const activeSubMenuIndex = tools.indexOf(activeSubMenu);
        const activeToolIndex = tools.indexOf(activeTool);

        /** Popper for submenu and info
         */
        const [anchorEl, setAnchorEl] = useState<HTMLElement>(null);
        const [subMenuEl, setSubMenuEl] = useState<HTMLElement>(null);

        // Offset for the submenu (depending on the current tool shown)
        const [offset, setOffset] = useState<[number, number]>([0, 6]);

        const { styles: subMenuStyles, attributes: subMenuAttributes } =
            usePopper(anchorEl, subMenuEl, {
                placement: "right-start",
                modifiers: [
                    {
                        name: "offset",
                        options: {
                            offset,
                        },
                    },
                ],
            });

        // Set offset depending on the current tool
        useEffect(() => {
            if (activeSubMenuIndex != -1) {
                setOffset([54 * activeSubMenuIndex, 6]);
            }
        }, [activeSubMenuIndex]);

        return (
            <>
                <div className={styles.toolbar} ref={setAnchorEl}>
                    {toolButtons}
                    <div
                        ref={setSubMenuEl}
                        style={subMenuStyles.popper}
                        {...subMenuAttributes.popper}
                    >
                        {
                            <SubMenu isShown={activeSubMenu ? true : false}>
                                {toolSubMenus[activeSubMenuIndex]}
                            </SubMenu>
                        }
                    </div>
                </div>
                <div
                    className={styles.info}
                    data-shown={showInfo ? true : false}
                    onClick={() => setShowInfo((prev: boolean) => !prev)}
                >
                    <i className={"bi bi-info " + styles.InfoIcon}></i>
                    {activeTool ? toolInfos[activeToolIndex] : <ToolBarInfo />}
                </div>
            </>
        );
    };
}

function SubMenu({ children, isShown }) {
    return (
        <div className={styles.subMenu} data-shown={isShown}>
            {children}
        </div>
    );
}

/** simple help screen for the toolbar
 */
function ToolBarInfo() {
    return (
        <>
            <div className={styles.header}>Toolbar</div>
            <p>
                Click on a tool to activate it. Click on it again to deactivate
                it.
            </p>
            <p>
                Some tools have sub-menus. Right click on the tool to open the
                submenu.
            </p>
        </>
    );
}
