import useUser from "lib/useUser";
import {
    createContext,
    useCallback,
    useContext,
    useEffect,
    useState,
} from "react";

import { ToolBarContext } from "../toolbar";
import { useDrawing } from "./doodle/drawing";
import {
    ColorSelector,
    SizeSelector,
    SmoothingSelector,
} from "./doodle/submenu";
import { PenEmpty, PenFilled } from "./doodle/svgs";
import { DoodleToolContextI } from "./typing";

import stylesTB from "../toolbar/toolbar.module.scss";

export const DoodleToolContext = createContext<DoodleToolContextI>(undefined);

export function DoodleToolContextProvider({ children }) {
    const { activeTool, setActiveTool, activeSubMenu, setActiveSubMenu } =
        useContext(ToolBarContext);
    const { config, isLoadingConfig, mutateConfig } = useUser();
    const active = activeTool === "doodle";
    const toggleActive = useCallback(
        () => setActiveTool("doodle"),
        [setActiveTool]
    );
    const active_sm = activeSubMenu === "doodle";
    const toggleSubMenu = () => setActiveSubMenu("doodle");

    //Colors
    const [currentColor, _setCurrentColor] = useState("#0000ff");
    const [colorList, setColorList] = useState([
        "#000000",
        "#ff0000",
        "#00ff00",
        "#0000ff",
    ]);
    const setCurrentColor = useCallback(
        (color) => {
            _setCurrentColor(color);
            mutateConfig({ pen_color: color });
        },
        [mutateConfig]
    );

    // Sizes
    const [currentSize, _setCurrentSize] = useState(4);
    const setCurrentSize = useCallback(
        (currentSize) => {
            //Handle max an min size
            const minmax = Math.min(Math.max(currentSize, 1), 25);
            _setCurrentSize(minmax);
            mutateConfig({ pen_size: minmax });
        },
        [mutateConfig]
    );

    // Smoothing
    const [currentSmoothing, _setCurrentSmoothing] = useState(0);
    const setCurrentSmoothing = useCallback(
        (currentSmoothing) => {
            //Handle max an min size
            const minmax = Math.min(Math.max(currentSmoothing, 0), 1);
            _setCurrentSmoothing(minmax);
            mutateConfig({ pen_smoothing: minmax });
        },
        [mutateConfig]
    );

    // Drawing hook
    useDrawing(active, currentSize, currentColor, currentSmoothing);

    useEffect(() => {
        // TEMP:
        const cb = (e: KeyboardEvent) => {
            const skip_tags = ["INPUT", "TEXTAREA"];
            if (skip_tags.includes(document.activeElement.tagName)) {
                return;
            }
            if (e.key === "p") {
                toggleActive();
            }
            // Use i and o to change color
            if (e.key === "i") {
                setCurrentColor((prev) => {
                    const index = colorList.indexOf(prev);
                    if (index === 0 || index === -1) {
                        return prev;
                    }
                    return colorList[index - 1];
                });
            }
            if (e.key === "o") {
                setCurrentColor((prev) => {
                    const index = colorList.indexOf(prev);
                    if (index === colorList.length - 1 || index === -1) {
                        return prev;
                    }
                    return colorList[index + 1];
                });
            }
        };
        window.addEventListener("keydown", cb);
        return () => {
            window.removeEventListener("keydown", cb);
        };
    }, [colorList, currentColor, setCurrentColor, toggleActive]);

    // Set values from config this is only revalidated on focus change
    // see useUser
    useEffect(() => {
        if (!isLoadingConfig && config) {
            // Set current color
            // Add it to the list if it's not there
            setColorList((prev) => {
                if (!prev.includes(config.pen_color)) {
                    return [...prev, config.pen_color];
                }
                return prev;
            });
            _setCurrentColor(config.pen_color);

            //Size
            _setCurrentSize(config.pen_size);

            //Smoothing
            _setCurrentSmoothing(config.pen_smoothing);
            return;
        }
        // Update the config when the text styles change
    }, [colorList, config, isLoadingConfig]);

    const ret: DoodleToolContextI = {
        active,
        toggleActive,
        activeSubMenu: active_sm,
        toggleSubMenu,

        currentColor,
        setCurrentColor,
        colorList,
        setColorList,
        currentSize,
        setCurrentSize,
        currentSmoothing,
        setCurrentSmoothing,
    };

    return (
        <DoodleToolContext.Provider value={ret}>
            {children}
        </DoodleToolContext.Provider>
    );
}

/** Button to activate the tool
 */
function Button() {
    const { active, toggleActive, toggleSubMenu, activeSubMenu, currentColor } =
        useContext(DoodleToolContext);

    return (
        <div
            className={stylesTB.toolBtn + " " + stylesTB.doodleBtn}
            data-active={active}
            data-submenu={activeSubMenu}
            onContextMenu={(e) => {
                e.preventDefault();
                toggleSubMenu();
            }}
        >
            <button
                onClick={toggleActive}
                className={"btn"}
                style={{ ["--pen-color" as string]: currentColor }}
            >
                {active ? <PenFilled /> : <PenEmpty />}
            </button>
            <div
                onClick={toggleSubMenu}
                className={stylesTB.arrow}
                data-active={activeSubMenu}
            >
                {activeSubMenu ? "<" : ">"}
            </div>
        </div>
    );
}
function SubMenu() {
    return (
        <>
            <ColorSelector />
            <SizeSelector />
            <SmoothingSelector />
        </>
    );
}
function Info() {
    return (
        <>
            <div className={stylesTB.header}>Pen</div>
            <p>Draw doodles on the page.</p>

            <table>
                <tbody>
                    <tr>
                        <td>
                            <kbd>p</kbd>
                        </td>
                        <td>Toggle tool</td>
                    </tr>
                    <tr>
                        <td>
                            <kbd>i</kbd> <kbd>o</kbd>
                        </td>
                        <td>Change color</td>
                    </tr>
                </tbody>
            </table>
        </>
    );
}

export const DoodleTool = {
    Button,
    SubMenu,
    Info,
};
