import { useCallback, useContext, useEffect, useRef, useState } from "react";

import { DoodleSnip } from "snips/general/doodle";

import { PagesContext } from "components/viewer/page";

export function useDrawing(
    active: boolean,
    lineWidth: number,
    color: string,
    smoothing: number
) {
    const {
        containerRef,
        overlayCanvasRef,
        workers,
        overlayWorkers,
        pageDataRef,
        numPages,
    } = useContext(PagesContext);
    const activeSnipRef = useRef<DoodleSnip>(null);
    const [targetOverlayCheck, setTargetOverlayCheck] =
        useState<boolean>(false);

    const render_snip = useCallback(
        (ctx: CanvasRenderingContext2D) => {
            if (activeSnipRef.current) {
                activeSnipRef.current.render(ctx);
            }
        },
        [activeSnipRef]
    );

    let pointerMoveHandler: (event: PointerEvent) => void = useCallback(() => {
        return;
    }, []);
    /* Commit the drawn doodle to the page */
    const pointerUpHandler = useCallback(() => {
        if (!activeSnipRef.current) return;
        const active_snip = activeSnipRef.current;

        // find index
        let idx;
        pageDataRef.current.forEach((page, i) => {
            if (page.id === active_snip.page_id) idx = i;
        });
        // Add snip to page
        workers[idx].insertSnip(active_snip.to_data()).then(() => {
            overlayWorkers[idx].remove(render_snip);
            activeSnipRef.current = null;
        });
        document.removeEventListener("pointermove", pointerMoveHandler);
        document.removeEventListener("pointerup", pointerUpHandler);
    }, [pointerMoveHandler, overlayWorkers, pageDataRef, render_snip, workers]);

    pointerMoveHandler = useCallback(
        (event: MouseEvent) => {
            if (!activeSnipRef.current) return;
            const active_snip = activeSnipRef.current;

            let idx;
            pageDataRef.current.forEach((page, i) => {
                if (page.id === active_snip.page_id) idx = i;
            });

            const overlayCanvas = overlayCanvasRef.current.get(idx);
            // Already drawing
            if (active_snip) {
                // Move mouse out of canvas while drawing
                console.log("test");
                if (event.target != overlayCanvas) {
                    event.preventDefault();
                    /** Interpolate new mouse position with previous mouse position
                     * This helps to draw a line to the edge of the canvas
                     * We have to do this since the reference frame changes if we have change the target
                     */
                    // Get previous mouse position
                    const previous_edge = active_snip.get_edge(-1);
                    // Translate previous edge
                    const translation = overlayWorkers[
                        idx
                    ].canvasCoords_2_dbCoords([
                        event.movementX,
                        event.movementY,
                    ]);
                    // Add translation to previous edge
                    const new_edge = [
                        previous_edge[0] + translation[0],
                        previous_edge[1] + translation[1],
                    ];
                    // round to 2 decimal places
                    new_edge[0] = Math.round(new_edge[0] * 100) / 100;
                    new_edge[1] = Math.round(new_edge[1] * 100) / 100;
                    // Add new edge to snip
                    active_snip.add_edge(new_edge);

                    /** Force pointerup i.e. snip save if pointer is released from canvas
                     */
                    return pointerUpHandler();
                } else {
                    const new_edge = overlayWorkers[
                        idx
                    ].canvasCoords_2_dbCoords([
                        Math.max(event.offsetX, 0),
                        Math.max(event.offsetY, 0),
                    ]);
                    // round to 2 decimal places
                    new_edge[0] = Math.round(new_edge[0] * 100) / 100;
                    new_edge[1] = Math.round(new_edge[1] * 100) / 100;
                    // Add edge
                    active_snip.add_edge(new_edge);

                    overlayWorkers[idx].render();

                    // Send preview event to other clients
                    overlayWorkers[idx].sendPreviewData(active_snip.to_data());
                }
            }
            // Pointer move into canvas
            else if (
                event.target === overlayCanvasRef.current.get(idx) &&
                event.buttons === 1 &&
                !targetOverlayCheck
            ) {
                setTargetOverlayCheck(true);
                // Current position
                const current_edge = overlayWorkers[
                    idx
                ].canvasCoords_2_dbCoords([event.offsetX, event.offsetY]);
                const translation = overlayWorkers[idx].canvasCoords_2_dbCoords(
                    [event.movementX, event.movementY]
                );
                const prev_edge = [
                    current_edge[0] - translation[0],
                    current_edge[1] - translation[1],
                ];
                // round to 2 decimal places
                prev_edge[0] = Math.round(prev_edge[0] * 100) / 100;
                prev_edge[1] = Math.round(prev_edge[1] * 100) / 100;
                current_edge[0] = Math.round(current_edge[0] * 100) / 100;
                current_edge[1] = Math.round(current_edge[1] * 100) / 100;

                const new_snip = new DoodleSnip({
                    edges: [prev_edge, current_edge],
                    colour: color,
                    lineWidth,
                    smoothing,
                    page_id: pageDataRef.current.get(idx).id,
                    book_id: pageDataRef.current.get(idx).book_id,
                });
                activeSnipRef.current = new_snip;
            }

            if (event.target != overlayCanvasRef.current.get(idx)) {
                setTargetOverlayCheck(false);
            } else {
                setTargetOverlayCheck(true);
            }
        },
        [
            color,
            lineWidth,
            pointerUpHandler,
            overlayCanvasRef,
            overlayWorkers,
            pageDataRef,
            smoothing,
            targetOverlayCheck,
        ]
    );

    const pointerDownHandler = useCallback(
        (event: PointerEvent, pageIdx: number) => {
            console.log(event);
            const new_edge = overlayWorkers[pageIdx].canvasCoords_2_dbCoords([
                event.offsetX,
                event.offsetY,
            ]);

            // round to 2 decimal places
            new_edge[0] = Math.round(new_edge[0] * 100) / 100;
            new_edge[1] = Math.round(new_edge[1] * 100) / 100;
            activeSnipRef.current = new DoodleSnip({
                edges: [new_edge],
                colour: color,
                lineWidth,
                smoothing,
                page_id: pageDataRef.current.get(pageIdx).id,
                book_id: pageDataRef.current.get(pageIdx).book_id,
            });
            overlayWorkers[pageIdx].add(render_snip);

            // Add pointermove and pointerup listeners
            document.addEventListener("pointermove", pointerMoveHandler);
            document.addEventListener("pointerup", pointerUpHandler);
        },
        [
            color,
            lineWidth,
            pointerMoveHandler,
            pointerUpHandler,
            overlayWorkers,
            pageDataRef,
            render_snip,
            smoothing,
        ]
    );

    // Prevent default touch events (zooming, panning )
    const touchStartHandler = useCallback(
        (event: TouchEvent) => {
            event.preventDefault();
        },
        []
    );

    /** Setup event listeners
     *
     */
    useEffect(() => {
        if (!containerRef.current || !active) return;
        const containers = Array.from(containerRef.current.values());
        const helpers_pointerdown = containers.map((container, idx) => {
            return (event: PointerEvent) => pointerDownHandler(event, idx);
        });
        const helpers_touchstart = containers.map(() => {
            return (event: TouchEvent) => touchStartHandler(event);
        });

        containers.forEach((container, idx) => {
            container.addEventListener("touchstart", helpers_touchstart[idx]);
            container.addEventListener("pointerdown", helpers_pointerdown[idx]);
        });

        return () => {
            containers.forEach((container, idx) => {
                container.removeEventListener(
                    "pointerdown",
                    helpers_pointerdown[idx]
                );
                container.removeEventListener(
                    "touchstart",
                    helpers_touchstart[idx]
                );
            });
        };
    }, [
        active,
        containerRef,
        pointerDownHandler,
        pointerMoveHandler,
        pointerUpHandler,
        overlayWorkers,
        numPages,
        touchStartHandler,
    ]);
}
