import { MutableRefObject, useContext, useRef } from "react";

import { DoodleToolContext } from "../doodle.tool";

import styles from "./styles.module.scss";

/** Color selector
 * This component is used to select color for doodle tool,
 *
 */
export function SizeSelector() {
    const rangeInputRef = useRef<HTMLInputElement>(null);
    const numberInputRef = useRef<HTMLInputElement>(null);

    // We use the tool context here, see in doodleToolContext
    const { currentSize, setCurrentSize } = useContext(DoodleToolContext);

    /* range input / slider, inputs suck thus we have to put them into their
        own containers...
    */
    return (
        <>
            <label>Size</label>
            <div className={styles.sizeSelector}>
                <div className="range-container">
                    <input
                        ref={rangeInputRef}
                        type="range"
                        min={1}
                        max={25}
                        step={1}
                        value={currentSize.toFixed(0)}
                        onInput={(e) => {
                            const current_val = (e.target as HTMLInputElement)
                                .valueAsNumber;

                            numberInputRef.current.value = String(current_val);

                            setCurrentSize(current_val);
                            setThumbSize(current_val, rangeInputRef);
                        }}
                    />
                </div>
                <div className="number-container">
                    <button
                        className="input-group-text"
                        onClick={() => {
                            setCurrentSize(4);
                            setThumbSize(4, rangeInputRef);
                        }}
                    >
                        <i className="bi bi-arrow-counterclockwise"></i>
                    </button>
                    <input
                        ref={numberInputRef}
                        type="number"
                        min={1}
                        max={25}
                        step={1}
                        value={currentSize.toFixed(0)}
                        onInput={(e) => {
                            const current_val = (e.target as HTMLInputElement)
                                .valueAsNumber;

                            rangeInputRef.current.value = String(current_val);

                            setCurrentSize(current_val);
                            setThumbSize(current_val, rangeInputRef);
                        }}
                    />
                </div>
            </div>
        </>
    );
}

/** Color selector
 * This component is used to select color for doodle tool,
 *
 */
export function ColorSelector() {
    // Invisible color selector input
    const inputRef = useRef<HTMLInputElement>(null);

    // We use the tool context here, see in doodleToolContext
    const { currentColor, setCurrentColor, colorList, setColorList } =
        useContext(DoodleToolContext);

    function getColorDiv(color: string) {
        const new_div = (
            <div
                key={color}
                data-active={color === currentColor}
                style={{ backgroundColor: color }}
                onClick={() => {
                    setCurrentColor(color);
                }}
            ></div>
        );
        return new_div;
    }
    const color_divs = colorList.map((color) => {
        return getColorDiv(color);
    });

    return (
        <>
            <label>Color</label>
            <div className={styles.colorSelector}>
                {color_divs}
                <div
                    className={styles.addIcon}
                    onClick={() => {
                        inputRef.current.click();
                    }}
                >
                    <input
                        ref={inputRef}
                        type="color"
                        style={{ display: "none" }}
                        onChange={(e) => {
                            const color = e.target.value;
                            setColorList([...colorList, color]);
                        }}
                    />
                </div>
            </div>
        </>
    );
}

/** Smoothing selector
 * This component is used to select smoothing for doodle tool,
 */
export function SmoothingSelector() {
    const rangeInputRef = useRef<HTMLInputElement>(null);
    const numberInputRef = useRef<HTMLInputElement>(null);

    // We use the tool context here, see in doodleToolContext
    const { currentSmoothing, setCurrentSmoothing } =
        useContext(DoodleToolContext);

    // Invisible color selector input
    return (
        <>
            <label>Line smoothing</label>
            <div className={styles.sizeSelector}>
                <div className="range-container">
                    <input
                        ref={rangeInputRef}
                        type="range"
                        min={0}
                        max={1}
                        step={0.01}
                        value={parseFloat(currentSmoothing.toFixed(2))}
                        onInput={(e) => {
                            const current_val = (e.target as HTMLInputElement)
                                .valueAsNumber;

                            numberInputRef.current.value = String(current_val);

                            setCurrentSmoothing(current_val);
                            setThumbSize(current_val, rangeInputRef);
                        }}
                    />
                </div>
                <div className="number-container">
                    <button
                        className="input-group-text"
                        onClick={() => {
                            setCurrentSmoothing(0);
                            setThumbSize(0, rangeInputRef);
                        }}
                    >
                        <i className="bi bi-arrow-counterclockwise"></i>
                    </button>
                    <input
                        ref={numberInputRef}
                        type="number"
                        min={0}
                        max={1}
                        step={0.01}
                        value={parseFloat(currentSmoothing.toFixed(2))}
                        onInput={(e) => {
                            const current_val = (e.target as HTMLInputElement)
                                .valueAsNumber;

                            rangeInputRef.current.value = String(current_val);

                            setCurrentSmoothing(current_val);
                            setThumbSize(current_val, rangeInputRef);
                        }}
                    />
                </div>
            </div>
        </>
    );
}

/** Change the thumb size of
 * the range slider
 */
const min_size_thumb = 0.7;
const max_size_thumb = 1.2;
function setThumbSize(size: number, ref: MutableRefObject<HTMLInputElement>) {
    const scaled_size = (size - 1) / (25 - 1);
    let size_thumb =
        scaled_size * (max_size_thumb - min_size_thumb) + min_size_thumb;
    if (size_thumb > max_size_thumb) {
        size_thumb = max_size_thumb;
    }
    ref.current.style.setProperty("--size-thumb", size_thumb + "em");
}
