import {
    createContext,
    ReactNode,
    useContext,
    useEffect,
    useState,
} from "react";

import { PagesContext } from "../page";
import { SidebarContext } from "../sidebar";
import { OverlayWorker } from "../worker/overlayWorker";
import { PageWorker } from "../worker/pageWorker";
import { InteractionEventMap } from "../worker/pageWorker/interactionEvents";
import { useBaseToolContext } from "./baseToolContext";
import { Button } from "./interaction/button";
import { get_response_handlers } from "./interaction/handle_response";
import { snip_to_overlay } from "./interaction/overlay";
import { InteractionToolContextI, ToolComponents } from "./typing";

import { BaseSnip } from "snips/general/base";

export const InteractionToolContext =
    createContext<InteractionToolContextI>(undefined);

export function InteractionToolContextProvider({
    children,
}: {
    children: React.ReactNode;
}) {
    // Default tool (undefined)
    const { active, toggleActive } = useBaseToolContext(undefined);
    const { activePageId } = useContext(SidebarContext);
    const {
        containerRef,
        overlayCanvasRef,
        workers,
        overlayWorkers,
        numPages,
    } = useContext(PagesContext);
    const [activeSnip, setActiveSnip] = useState<BaseSnip>();
    const [activePageIdx, setActivePageIdx] = useState<number>(0);

    // Build overlay snip
    const overlay: ReactNode = snip_to_overlay(activeSnip, activePageIdx);

    const ret: InteractionToolContextI = {
        active,
        toggleActive,
        activeSubMenu: false,
        toggleSubMenu: () => {},
        overlay: {
            dom: overlay,
            page_idx: activePageIdx,
        },
    };

    // Forward on_* to snips which are clicked on or hovered on
    // This is done by the workers and their might be a callback
    // from the worker or the worker might change the display

    useEffect(() => {
        if (!active) return;
        if (!overlayCanvasRef.current) return;
        if (!workers) return;

        const canvases = Array.from(overlayCanvasRef.current.values());
        const containers = Array.from(containerRef.current.values());

        const events = interactionevents_factory(
            canvases,
            containers,
            workers,
            overlayWorkers,
            setActiveSnip,
            setActivePageIdx
        ); // See events below

        events.forEach((event) => {
            canvases[event.page_idx].addEventListener(
                event.type,
                event.callback
            );
        });

        return () => {
            events.forEach((event) => {
                canvases[event.page_idx].removeEventListener(
                    event.type,
                    event.callback
                );
            });
        };
    }, [
        active,
        containerRef,
        numPages,
        overlayCanvasRef,
        overlayWorkers,
        workers,
    ]);

    useEffect(() => {
        setActiveSnip(undefined);
    }, [activePageId]);

    return (
        <InteractionToolContext.Provider value={ret}>
            {children}
        </InteractionToolContext.Provider>
    );
}

interface EventItem<K extends keyof InteractionEventMap> {
    type: K; // event type (string)
    page_idx: number; // page index
    callback: (event: Event) => void; // event callback (function)
}

/** Create all event callbacks
 * in theory we can add more here and it should work fine. For now I have added
 * only pointerdown and onhover.
 *
 * @param canvases
 */
function interactionevents_factory(
    canvases: HTMLCanvasElement[],
    containers: HTMLDivElement[],
    workers: PageWorker[],
    overlayWorkers: OverlayWorker[],
    setActiveSnip: (snip?: BaseSnip) => void,
    setActivePageIdx: (page_idx: number) => void
): EventItem<keyof InteractionEventMap>[] {
    const ret: EventItem<keyof InteractionEventMap>[] = [];

    // Register all simple events
    const events = [
        "pointerdown",
        "pointerup",
        "pointerenter",
        "pointerleave",
        "pointermove",
        "dblclick",
    ];

    for (const event of events) {
        canvases.forEach((canvas, idx) => {
            ret.push({
                type: event as keyof InteractionEventMap,
                page_idx: idx,
                callback: (event: PointerEvent) => {
                    // Forward message to worker
                    const coords = overlayWorkers[idx].canvasCoords_2_dbCoords([
                        event.offsetX,
                        event.offsetY,
                    ]) as [number, number];
                    coords[0] = Math.round(coords[0] * 100) / 100;
                    coords[1] = Math.round(coords[1] * 100) / 100;

                    const response = workers[idx].interactionEvent({
                        // event
                        type: event.type as keyof InteractionEventMap,
                        data: {
                            coords,
                        },
                    });

                    response
                        .then(get_response_handlers)
                        // Run the handlers
                        .then((handlers) => {
                            if (!handlers) return;
                            handlers.forEach((handler) => {
                                handler({
                                    canvas: canvases[idx],
                                    container: containers[idx],
                                    page_idx: idx,
                                    setActiveSnip,
                                });
                                setActivePageIdx(idx);
                            });
                        })
                        // Catch errors
                        .catch((e) => {
                            console.error("Error in interaction event", e);
                        });

                    //
                },
            });
        });
    }

    return ret;
}

export const InteractionTool: ToolComponents = {
    Button,
    SubMenu: () => null,
    Info: () => null,
};
