import {
    ReactNode,
    useCallback,
    useContext,
    useEffect,
    useRef,
    useState,
} from "react";

import { ArraySnip } from "snips/general/array";
import { BaseSnip } from "snips/general/base";
import { TextSnip } from "snips/general/text";
import { MotorsSnip } from "snips/uprp/spec/motors";

import { PagesContext } from "components/viewer/page";

import styles from "../text/styles.module.scss";

export interface IOverlayProps extends React.HTMLAttributes<HTMLDivElement> {
    // X position of the overlay (canvas coords)
    x: number;
    // Y position of the overlay (canvas coords)
    y: number;
    // Width of the overlay (canvas coords)
    width: number;
    // Height of the overlay (canvas coords)
    height: number;
    // Page index
    page_idx: number;
    // Children
    children?: ReactNode;
    callback?: (scale: number) => void;
    className?: string;
}

export function Overlay({
    x,
    y,
    width,
    height,
    page_idx = 0,
    children = null,
    callback = null,
    className = null,
    ...props
}: IOverlayProps) {
    const overlayRef = useRef<HTMLDivElement>(null);
    const { draggableRef, overlayWorkers } = useContext(PagesContext);
    const [overlayStyle, setOverlayStyle] = useState({});
    /* If snip changes update overlay this should
     * automatically align the overlay to the snip
     */
    const moveOverlay = useCallback(() => {
        const overlay = overlayRef.current;
        if (!overlay) return;
        const draggable = draggableRef.current.get(page_idx);
        const scale = draggable.clientHeight / 2000;
        const x_o = scale * x;
        const y_o = scale * y;
        // Set position
        // 5px Padding see ../text/styles.module.scss
        setOverlayStyle({
            left: x_o + draggable.offsetLeft - 5 + "px",
            top: y_o + draggable.offsetTop - 5 + "px",
            width: scale * width + 10 + "px",
            height: scale * height + 10 + "px",
        });
        if (callback) callback(scale);
    }, [callback, draggableRef, height, page_idx, width, x, y]);

    useEffect(() => {
        const observer = new ResizeObserver(moveOverlay);
        observer.observe(draggableRef.current.get(page_idx));
        moveOverlay();
        return () => {
            observer.disconnect();
        };
    }, [draggableRef, moveOverlay, page_idx, overlayWorkers]);

    return (
        <div
            ref={overlayRef}
            className={
                className ? styles.overlay + " " + className : styles.overlay
            }
            style={overlayStyle}
            data-active={false}
            {...props}
        >
            {children}
        </div>
    );
}

export interface IOverlayTextProps extends IOverlayProps {
    // Content of the overlay
    text: string;
    // Styling for content
    textStyle?: { fontFamily?: string; lineHeight?: number; fontSize?: number };
    // Props for textarea element
    textProps?: React.ComponentPropsWithRef<"textarea">;
}

export function OverlayText({
    x,
    y,
    width,
    height,
    text,
    page_idx = 0,
    textStyle = {},
    children = null,
    callback = null,
    textProps = {},
    ...props
}: IOverlayTextProps) {
    const [textStyleParsed, setTextStyleParsed] = useState({});

    /** Parse text to same format as in snip
     * sometimes new line and tabs are not parsed correctly
     */
    if (text) {
        text = text
            .replaceAll(/\t/g, "        ")
            .replaceAll(/\n/g, "\\n")
            .replaceAll(/\r/g, "")
            .replaceAll(/\\n/g, String.fromCharCode(13, 10));
    } else {
        text = "";
    }
    //If content is a textarea set its font size
    const cb = useCallback(
        (scale) => {
            if (text && textStyle) {
                // Parse font to pixels
                //Apply styles to temp element
                const el = document.createElement("span");

                el.style.setProperty(
                    "font",
                    textStyle.fontSize * 2.35 * scale +
                        "px " +
                        textStyle.fontFamily
                );

                /** Setting the lineheight in css is different from the
                 * lineheight in the canvas. Weirdly css does not allow
                 * to define a text baseline. The lineheight is defined
                 * as the distance between the baselines of two lines but it
                 * always centers the text. The canvas on the other hand
                 * defines the lineheight as the distance between the
                 * baselines of two lines but it always aligns the text
                 * to the top.
                 *
                 * css: lineheight
                 * --
                 * spacing
                 * font
                 * spacing
                 * --
                 *
                 * canvas: lineheight
                 * --
                 * font
                 * spacing
                 * --
                 *
                 * Therefore we need to dynmaically adjust the top
                 * position of the container to align both systems.
                 *
                 */
                let top = 0;
                let letterSpacing = null;
                if (textStyle.lineHeight) {
                    el.style.setProperty(
                        "line-height",
                        textStyle.lineHeight * 2.35 * scale + "px"
                    );

                    const spacing =
                        (parseFloat(el.style.lineHeight) -
                            parseFloat(el.style.fontSize)) /
                        2;
                    top = -spacing - 0.07 * textStyle.fontSize * 2.35 * scale;
                }
                console.log(el);

                // Somehow the font size is not parsed correctly
                // Might need to adjust other fonts in the future
                if (textStyle.fontFamily == "PlexMono") {
                    letterSpacing = 0.04 * scale + "px";
                }

                setTextStyleParsed({
                    font: el.style.font,
                    width: scale * width + 10 + "px",
                    height: scale * height + "px",
                    top: top + "px",
                    letterSpacing: letterSpacing,
                });

                // Call callback
                if (callback) callback(scale);
            }
        },
        [callback, height, text, textStyle, width]
    );

    return (
        <Overlay
            x={x}
            y={y}
            width={width}
            height={height}
            page_idx={page_idx}
            callback={cb}
            {...props}
        >
            <textarea
                style={textStyleParsed}
                value={text}
                {...textProps}
            ></textarea>
            {children}
        </Overlay>
    );
}

/** From a snip create a static overlay component */
export function snip_to_overlay(
    snip: BaseSnip,
    page_idx = 0,
    xy_overwrite: [number, number] = null
): ReactNode {
    let overlay: ReactNode = null;
    if (snip) {
        // Text snip(s) (Motor is also a textsnip but does not inherit it)
        if (snip instanceof TextSnip || snip instanceof MotorsSnip) {
            const textStyles = {
                //snip coords
                fontFamily: snip.fontFamily,
                fontSize: snip.fontSize,
                lineHeight: snip.lineHeight,
            };
            overlay = (
                <OverlayText
                    x={xy_overwrite ? xy_overwrite[0] : snip.x}
                    y={xy_overwrite ? xy_overwrite[1] : snip.y}
                    width={snip.width}
                    height={snip.height}
                    page_idx={page_idx}
                    key={snip.id}
                    text={snip.text}
                    textStyle={textStyles}
                    textProps={{ readOnly: true }}
                />
            );
        } else if (snip instanceof ArraySnip) {
            let sum = 0;
            overlay = (
                <>
                    {snip.snips.map((s) => {
                        overlay = snip_to_overlay(
                            s,
                            page_idx,
                            snip.legacy
                                ? [snip.x + s.x, snip.y + s.y + sum]
                                : null
                        );
                        if (snip.legacy) {
                            sum += s.height;
                        }
                        return overlay;
                    })}
                </>
            );
        }
    }
    return overlay;
}
