import { createContext, useContext, useEffect } from "react";

import { PagesContext } from "../page";
import { useBaseToolContext } from "./baseToolContext";
import { useDraggableHandlers } from "./movePage/draggable";
import { useResizable } from "./movePage/resizable";
import { SubMenu } from "./movePage/submenu";
import { MovePageToolContextI } from "./typing";

import styles from "../toolbar/toolbar.module.scss";

export const MovePageToolContext =
    createContext<MovePageToolContextI>(undefined);

export function MovePageToolContextProvider({
    children,
}: {
    children: React.ReactNode;
}) {
    const { draggableRef, wrapperRef, containerRef, numPages } =
        useContext(PagesContext);
    const { active, toggleActive, activeSubMenu, toggleSubMenu } =
        useBaseToolContext("movePage");

    // Dragging of the page inside the wrapper
    const {
        pointerDownHandler: pointerDownHandlerDrag,
        pointeMoveHandler: pointerMoveHandlerDrag,
        pointerUpHandler: pointerUpHandlerDrag,
    } = useDraggableHandlers(
        draggableRef,
        wrapperRef,
        containerRef,
        numPages,
        active
    );

    // Resizing of the page inside the wrapper
    const {
        fitPage,
        centerPage,
        zoom,
        setZoom,
        touchStartHandler: touchStartHandlerResize,
        touchMoveHandler: touchMoveHandlerResize,
        touchEndHandler: touchEndHandlerResize,
    } = useResizable(draggableRef, wrapperRef, containerRef, numPages, active);

    // Keyboard shortcuts
    useEffect(() => {
        const cb = (e: KeyboardEvent) => {
            const skip_tags = ["INPUT", "TEXTAREA"];
            if (skip_tags.includes(document.activeElement.tagName)) {
                return;
            }
            if (e.key === "m") {
                console.log("m pressed");
                toggleActive();
            }
        };
        window.addEventListener("keydown", cb);
        return () => {
            window.removeEventListener("keydown", cb);
        };
    }, [toggleActive]);

    useEffect(() => {
        if (!containerRef.current || !active) return;
        const containers = Array.from(containerRef.current.values());

        // We have to make the distinction between mouse and touch events here
        // otherwise the touch events will be triggered twice and gestures
        // will be very janky.

        const helpers_pointerdown = [];
        const helpers_touchstart = [];
        const helpers_touchend = [];
        const helpers_touchmove = [];
        const touchIdentifierMap = {};
        containers.forEach((wrapper, idx) => {
            helpers_pointerdown.push((e: PointerEvent) => {
                // Add to cache
                if (e.pointerType === "touch") {
                    return;
                }
                pointerDownHandlerDrag(e, idx);
            });

            helpers_touchstart.push((e: TouchEvent) => {
                e.preventDefault();

                // Add to cac
                touchStartHandlerResize(e);
                for (let i = 0; i < e.touches.length; i++) {
                    const touch = e.touches[i];
                    touchIdentifierMap[touch.identifier] = touch;
                }
            });

            helpers_touchmove.push((e: TouchEvent) => {
                const eP = new PointerEvent("pointerdown", {
                    pointerId: e.touches[0].identifier,
                    pointerType: "touch",
                    clientX: e.touches[0].clientX,
                    clientY: e.touches[0].clientY,
                });
                if (e.touches.length > 1) {
                    // cancel the drag if there are multiple touches
                    pointerUpHandlerDrag();

                    // Start the resize
                    touchMoveHandlerResize(e, idx);
                    return;
                }

                pointerDownHandlerDrag(eP, idx);
                pointerMoveHandlerDrag(eP);
            });

            helpers_touchend.push((e: TouchEvent) => {
                touchEndHandlerResize();
                for (let i = 0; i < e.changedTouches.length; i++) {
                    const touch = e.changedTouches[i];
                    delete touchIdentifierMap[touch.identifier];
                }
            });
        });

        // Add the event listeners
        containers.forEach((container, idx) => {
            container.addEventListener("pointerdown", helpers_pointerdown[idx]);
            container.addEventListener("touchstart", helpers_touchstart[idx]);
            container.addEventListener("touchmove", helpers_touchmove[idx]);
            container.addEventListener("touchend", helpers_touchend[idx]);
        });

        return () => {
            // Remove the event listeners
            containers.forEach((container, idx) => {
                container.removeEventListener(
                    "pointerdown",
                    helpers_pointerdown[idx]
                );
                container.removeEventListener(
                    "touchstart",
                    helpers_touchstart[idx]
                );
                container.removeEventListener(
                    "touchmove",
                    helpers_touchmove[idx]
                );
                container.removeEventListener(
                    "touchend",
                    helpers_touchend[idx]
                );
            });
        };
    }, [
        active,
        pointerDownHandlerDrag,
        touchStartHandlerResize,
        containerRef,
        pointerMoveHandlerDrag,
        pointerUpHandlerDrag,
        touchMoveHandlerResize,
        touchEndHandlerResize,
        numPages,
    ]);

    const ret: MovePageToolContextI = {
        active,
        toggleActive,
        activeSubMenu,
        toggleSubMenu,
        fitPage,
        centerPage,
        zoom,
        setZoom,
    };

    return (
        <MovePageToolContext.Provider value={ret}>
            {children}
        </MovePageToolContext.Provider>
    );
}

/** Button to activate the tool
 * TODO: Remove inline svg
 */
function Button() {
    const { active, toggleActive, toggleSubMenu, activeSubMenu } =
        useContext(MovePageToolContext);

    return (
        <div
            className={styles.toolBtn}
            data-active={active}
            data-submenu={activeSubMenu}
            onContextMenu={(e) => {
                e.preventDefault();
                toggleSubMenu();
            }}
        >
            <button className="btn" onClick={toggleActive}>
                <svg
                    className="bi bi-file-plus"
                    version="1.1"
                    viewBox="0 0 16.002 16.002"
                    xmlns="http://www.w3.org/2000/svg"
                    fill="currentColor"
                >
                    <g transform="translate(.00089378 .00089378)">
                        <path
                            d="m7.646 0.146a0.5 0.5 0 0 1 0.708 0l2 2a0.50063 0.50063 0 0 1-0.708 0.708l-1.146-1.147v3.793a0.5 0.5 0 0 1-1 0v-3.793l-1.146 1.147a0.50063 0.50063 0 1 1-0.708-0.708zm0.354 9.854a0.5 0.5 0 0 1 0.5 0.5v3.793l1.146-1.147a0.50063 0.50063 0 0 1 0.708 0.708l-2 2a0.5 0.5 0 0 1-0.708 0l-2-2a0.50063 0.50063 0 0 1 0.708-0.708l1.146 1.147v-3.793a0.5 0.5 0 0 1 0.5-0.5zm-7.854-1.646a0.5 0.5 0 0 1 0-0.708l2-2a0.50063 0.50063 0 1 1 0.708 0.708l-1.147 1.146h3.793a0.5 0.5 0 0 1 0 1h-3.793l1.147 1.146a0.50063 0.50063 0 0 1-0.708 0.708zm9.854-0.354a0.5 0.5 0 0 1 0.5-0.5h3.793l-1.147-1.146a0.50063 0.50063 0 0 1 0.708-0.708l2 2a0.5 0.5 0 0 1 0 0.708l-2 2a0.50063 0.50063 0 0 1-0.708-0.708l1.147-1.146h-3.793a0.5 0.5 0 0 1-0.5-0.5z"
                            fillRule="evenodd"
                        />
                    </g>
                </svg>
            </button>
            <div
                onClick={toggleSubMenu}
                className={styles.arrow}
                data-active={activeSubMenu}
            >
                {activeSubMenu ? "<" : ">"}
            </div>
        </div>
    );
}

function Info() {
    return (
        <>
            <div className={styles.header}>Move &amp; Resize</div>
            <p>
                Move the page by dragging it with the mouse and zoom in and out
                with the mouse wheel.
            </p>

            <table>
                <tbody>
                    <tr>
                        <td>
                            <kbd>m</kbd>
                        </td>
                        <td>Toggle tool</td>
                    </tr>
                    <tr>
                        <td>
                            <kbd>LMB²</kbd>
                        </td>
                        <td>Center page</td>
                    </tr>
                    <tr>
                        <td>
                            <kbd>MMB</kbd>
                        </td>
                        <td>Fit page</td>
                    </tr>
                </tbody>
            </table>
        </>
    );
}

export const MovePageTool = {
    Button,
    SubMenu,
    Info,
};
