import { MutableRefObject, useCallback, useEffect, useRef, useState } from "react";

import { EleMap } from "components/viewer/page";

/** Custom drag behavior using three
 * divs:
 *  - draggableRef: The element that will be dragged
 *  - wrapperRef: The element that will be used to drag the draggableRef
 *  - containerRef: The element that will contain the draggableRef
 *
 * Features:
 *  - Autoscroll when dragging off the container current view
 *
 */
export function useDraggableHandlers(
    draggableRef: MutableRefObject<EleMap<HTMLDivElement>>,
    wrapperRef: MutableRefObject<EleMap<HTMLDivElement>>,
    containerRef: MutableRefObject<EleMap<HTMLDivElement>>,
    numPages: number,
    active = true
) {
    const isPointerDown = useRef(false);
    const elemRectRef = useRef({ x: 0, y: 0, w: 0, h: 0 });
    const moveRef = useRef({ x: 0, y: 0, idx: undefined });

    const pointeMoveHandler = useCallback(
        (e: PointerEvent) => {
            if (!isPointerDown.current || !moveRef.current) return;
            const pageIdx = moveRef.current.idx;
            const draggElem = Array.from(draggableRef.current.values())[
                pageIdx
            ];
            const wrapperElem = Array.from(wrapperRef.current.values())[
                pageIdx
            ];
            const containerElem = Array.from(containerRef.current.values())[
                pageIdx
            ];

            const diff = moveRef.current;
            const elemRect = elemRectRef.current;

            // Calc new positions of draggable element
            let newY = e.clientY - diff.y,
                newX = e.clientX - diff.x;

            /** Check if element is inside container
             * view completely and if not, enable
             * sc
             */
            setScrollVisibility(
                { ...elemRect, x: newX, y: newY },
                containerElem
            );

            /** Check if element is out of bounds
             * I.e. hit scroll limits
             */
            check_bounds(
                { ...elemRect, x: newX, y: newY },
                wrapperElem.scrollWidth,
                wrapperElem.scrollHeight,
                // Top hit
                () => {
                    newY = 0;
                },
                // Left hit
                () => {
                    newX = 0;
                },
                // Bottom hit
                () => {
                    newY = wrapperElem.scrollHeight - elemRect.h;
                },
                // Right hit
                () => {
                    newX = wrapperElem.scrollWidth - elemRect.w;
                }
            );

            // Move element
            moveElm(draggElem, newX, newY);
        },
        [containerRef, draggableRef, wrapperRef]
    );
    const pointerUpHandler = useCallback(() => {
        isPointerDown.current = false;
        const wrappers = wrapperRef.current;
        // Styles
        wrappers.forEach((wrapper) => {
            wrapper.style.cursor = "grab";
            wrapper.style.removeProperty("user-select");
        });

    }, [wrapperRef]);


    const [attachhandler, setAttachHandler] = useState(false);

    const pointerDownHandler = useCallback(
        (e: PointerEvent, pageIdx: number) => {
            isPointerDown.current = true;
            setAttachHandler(true);
            // Change the cursor and prevent user from selecting the text
            const wrappers = wrapperRef.current;
            wrappers.forEach((wrapper) => {
                wrapper.style.cursor = "grabbing";
                wrapper.style.userSelect = "none";
            });

            const elem = Array.from(draggableRef.current.values())[pageIdx];

            // Get initial position of pointer
            const pointerX = e.clientX;
            const pointerY = e.clientY;

            // Get draggable element position
            const elemX = elem.offsetLeft;
            const elemY = elem.offsetTop;

            //Get dimensions of draggable element
            const elemRect = elemRectRef.current;
            elemRect.w = elem.offsetWidth;
            elemRect.h = elem.offsetHeight;

            // Get diff between pointer and draggable element
            const diff = moveRef.current;
            diff.x = pointerX - elemX;
            diff.y = pointerY - elemY;
            diff.idx = pageIdx;
        },
        [draggableRef, wrapperRef]
    );

    useEffect(() => {
        if (!wrapperRef.current) return;
        const elems = wrapperRef.current;
        if (!active) {
            setAttachHandler(false);
            elems.forEach((elem) => (elem.style.cursor = "default"));
            return;
        } else {
            elems.forEach((elem) => (elem.style.cursor = "grab"));
        }
    }, [wrapperRef, active]);

    useEffect(() => {
        if (attachhandler) {
            const wrappers = wrapperRef.current;
            document.addEventListener("pointerup", pointerUpHandler);
            document.addEventListener("pointercancel", pointerUpHandler);
            document.addEventListener("pointermove", pointeMoveHandler);
            return () => {
                // Why is this not working?
                wrappers.forEach((wrapper) => {
                    wrapper.style.cursor = "default";
                });

                document.removeEventListener("pointermove", pointeMoveHandler);
                document.removeEventListener("pointerup", pointerUpHandler);
                document.removeEventListener("pointercancel", pointerUpHandler);
            }
        }
    }, [attachhandler, pointeMoveHandler, pointerUpHandler, wrapperRef]);



    /**
     * 
    useEffect(() => {
        if (!wrapperRef.current || !active) return;
        const wrappers = Array.from(wrapperRef.current.values());
        
        const helper_mousedown = wrappers.map((wrapper, idx) => {
            return (e: PointerEvent) => {
                pointerDownHandler(e, idx);
            };
        });
        const helper_touchstart = wrappers.map((wrapper, idx) => {
            return (e: TouchEvent) => {
                if (e.touches.length > 1) return;
                console.log("touchstart");
                e.preventDefault();
                e.stopPropagation();
                const pointerEvent = new PointerEvent("pointerdown", {
                    pointerId: e.changedTouches[0].identifier,
                    clientX: e.changedTouches[0].clientX,
                    clientY: e.changedTouches[0].clientY,
                });
                pointerDownHandler(pointerEvent, idx);
            };
        });
 
        wrappers.forEach((wrapper, idx) => {
            wrapper.addEventListener("mousedown", helper_mousedown[idx]);
            wrapper.addEventListener("touchstart", helper_touchstart[idx]);
        });
 
        // Remove event listeners on cleanup
        return () => {
            wrappers.forEach((wrapper, idx) => {
                wrapper.removeEventListener("mousedown", helper_mousedown[idx]);
                wrapper.removeEventListener("touchstart", helper_touchstart[idx]);
            });
            // Remove event listeners just in case
            window.removeEventListener("pointermove", pointeMoveHandler);
            window.removeEventListener("pointerup", pointerUpHandler);
            window.removeEventListener("pointercancel", pointerUpHandler);
        };
    }, [
        wrapperRef,
        pointerDownHandler,
        pointeMoveHandler,
        pointerUpHandler,
        active,
        numPages,
    ]);
    */

    return {
        pointerDownHandler,
        pointeMoveHandler,
        pointerUpHandler,
    }
}

interface EleRect {
    x: number;
    y: number;
    w: number;
    h: number;
}
/** Check if inside bounds (assumes top left is 0,0 for both coordinate systems)
 * @param {EleRect} rect - The element rect
 * @param {number} bw - The bounds width
 * @param {number} bh - The bounds height
 */
export function check_bounds(
    { x, y, w, h }: EleRect,
    bw: number,
    bh: number,
    topCB?: () => void,
    leftCB?: () => void,
    bottomCB?: () => void,
    rightCB?: () => void,
    insideCB?: () => void
) {
    const y2 = y + h; // bottom
    const x2 = x + w; // right

    if (isInside(x, y, x2, y2, bw, bh)) {
        // top hit
        if (topCB && y < 0) {
            topCB();
        }
        // left hit
        if (leftCB && x < 0) {
            leftCB();
        }
        // bottom hit
        if (bottomCB && y2 > bh) {
            bottomCB();
        }
        // right hit
        if (rightCB && x2 > bw) {
            rightCB();
        }
    } else {
        insideCB && insideCB();
    }
}

export function isInside(x, y, x2, y2, bw, bh) {
    return x >= 0 && y >= 0 && x2 <= bw && y2 <= bh;
}

// Disables or shows the scrollbars
export function setScrollVisibility(
    { x, y, w, h }: EleRect,
    container: HTMLDivElement
) {
    if (
        isInside(
            x - container.scrollLeft,
            y - container.scrollTop,
            x + w - container.scrollLeft,
            y + h - container.scrollTop,
            container.offsetWidth,
            container.offsetHeight
        )
    ) {
        // Disable scrolling
        container.style.overflow = "hidden";
    } else {
        // Enable scrolling
        container.style.overflow = "scroll";
    }
}

export function moveElm(elm: HTMLDivElement, xPos: number, yPos: number) {
    elm.style.left = xPos + "px";
    elm.style.top = yPos + "px";
}
