import useUser from "lib/useUser";
import { MutableRefObject, useCallback, useEffect, useRef, useState } from "react";

import { isInside, moveElm, setScrollVisibility } from "./draggable";

import { EleMap } from "components/viewer/page";

/** Adds resizeability to container
 * this automatically centers the element if it is inside the
 * container and fits.
 *
 */
export function useResizable(
    draggableRef: MutableRefObject<EleMap<HTMLDivElement>>,
    wrapperRef: MutableRefObject<EleMap<HTMLDivElement>>,
    containerRef: MutableRefObject<EleMap<HTMLDivElement>>,
    numPages: number,
    active = true
) {

    const { config, isLoadingConfig, mutateConfig } = useUser();
    const containerSizeRef = useRef<
        {
            scrollLeft: number;
            scrollTop: number;
            offsetWidth: number;
            offsetHeight: number;
        }[]
    >();


    /** Everything zooooooooom
     */
    const [zoom, __setZoom] = useState<number[]>(Array(numPages).fill(1));
    // Resize the page when the zoom changes
    const _setZoom = useCallback(
        (zoom: number, idx: number) => {
            const draggable = draggableRef.current.get(idx);
            const draggable_rect = draggable.getBoundingClientRect();
            const nHeight =
                (containerRef.current.get(idx).clientHeight - 16) * zoom;
            const nWidth =
                nHeight * (draggable_rect.width / draggable_rect.height);

            resize(draggable, nWidth, nHeight);
            __setZoom((prev) => {
                prev[idx] = parseFloat(zoom.toFixed(2));
                return [...prev];
            });

        },
        [containerRef, draggableRef]
    );
    const setZoom = useCallback(
        (zoom: number, idx: number) => {
            _setZoom(zoom, idx);
            if (idx == 0) mutateConfig({ zoom1: zoom });
            if (idx == 1) mutateConfig({ zoom2: zoom });
        },
        [_setZoom, mutateConfig]
    );

    // observe changes in layout and recalculate zoom
    useEffect(() => {
        if (!draggableRef.current || !containerRef.current) return;
        // Add resize observer to update zoom when window is resized
        const draggables = Array.from(draggableRef.current.values());
        const containers = Array.from(containerRef.current.values());

        const observers = draggables.map((draggable, idx) => {
            return new ResizeObserver(() => {
                const client_rect = draggable.getBoundingClientRect();
                const zoom =
                    client_rect.height / (containers[idx].clientHeight - 16);
                __setZoom((prev) => {
                    prev[idx] = parseFloat(zoom.toFixed(2));
                    return [...prev];
                });
            });
        });

        observers.forEach((observer, idx) => {
            observer.observe(draggables[idx]);
        });

        return () => {
            observers.forEach((observer) => {
                observer.disconnect();
            });
        };
    }, [_setZoom, containerRef, draggableRef]);

    // Set initial values from config only on initial draw
    // as it is confusing otherwise
    const firstRender = useRef(true);
    useEffect(() => {
        if (!containerRef.current) return;
        if (!isLoadingConfig && config && firstRender.current) {
            firstRender.current = false;
            _setZoom(config.zoom1, 0);
            return;
        }
        if (numPages > 1) _setZoom(config.zoom2, 1);
        // Update the config when the text styles change
    }, [config, containerRef, isLoadingConfig, numPages, _setZoom]);

    /** Center the draggable element
     * inside the wrapper
     * */
    const centerPage = useCallback(
        (idx) => {
            const container = containerRef.current.get(idx);
            const wrapper = wrapperRef.current.get(idx);
            const elem = draggableRef.current.get(idx);
            _centerElement(elem, wrapper, container);
        },
        [draggableRef, wrapperRef, containerRef]
    );

    /** Fit the draggable element into the container
     * keeping the aspect ratio. Also centers the element.
     * Is Basically a reset function
     */
    const fitPage = useCallback(
        (idx) => {
            const elem = draggableRef.current.get(idx);
            const container = containerRef.current.get(idx);
            _fitElement(elem, container);
            __setZoom((prev) => {
                prev[idx] = 1;
                return [...prev];
            });
            centerPage(idx);
        },
        [draggableRef, containerRef, centerPage]
    );

    useEffect(() => {
        // Attach resize observer to each container
        if (!containerRef.current) return;
        const containers = Array.from(containerRef.current.values());

        containerSizeRef.current = containers.map((container) => {
            return {
                scrollLeft: container.scrollLeft,
                scrollTop: container.scrollTop,
                offsetWidth: container.offsetWidth,
                offsetHeight: container.offsetHeight,
            };
        });

        // Create resize observer for each container
        const resizeObservers = containers.map((container, idx) => {
            const resizeObserver = new ResizeObserver(() => {
                const preContainerSize = containerSizeRef.current[idx];
                //If the element was inside the viewport previously and is now outside, fit it
                const elem = draggableRef.current.get(idx);


                const x = elem.offsetLeft;
                const y = elem.offsetTop;
                const w = elem.offsetWidth;
                const h = elem.offsetHeight;
                if (
                    isInside(
                        x - preContainerSize.scrollLeft,
                        y - preContainerSize.scrollTop,
                        x + w - preContainerSize.scrollLeft,
                        y + h - preContainerSize.scrollTop,
                        preContainerSize.offsetWidth,
                        preContainerSize.offsetHeight
                    )
                ) {
                    fitPage(idx);
                }

                containerSizeRef.current[idx] = {
                    scrollLeft: container.scrollLeft,
                    scrollTop: container.scrollTop,
                    offsetWidth: container.offsetWidth,
                    offsetHeight: container.offsetHeight,
                };
            });
            return resizeObserver;
        });

        // Observe each container
        resizeObservers.forEach((resizeObserver, idx) => {
            resizeObserver.observe(containers[idx]);
        });

        return () => {
            resizeObservers.forEach((resizeObserver) => {
                resizeObserver.disconnect();
            });
        };
    }, [containerRef, draggableRef, fitPage, numPages]);


    const wheelHandler = useCallback(
        (e: WheelEvent, idx: number) => {
            e.preventDefault();
            const draggable = draggableRef.current.get(idx);
            const draggable_rect = draggable.getBoundingClientRect();

            const w = draggable_rect.width;
            const h = draggable_rect.height;
            const offset = e.deltaY * -0.0005

            // Add 0.05% of width/height every scroll
            // This might need adjustment
            let newWidth = w + w * offset;
            let newHeight = h + h * offset;

            // Min value is 0.25 * width/height
            const container = containerRef.current.get(idx);
            const container_rect = container.getBoundingClientRect();

            const computed_zoom = newHeight / (container_rect.height - 16);
            setZoom(computed_zoom, idx);
            if (computed_zoom < 0.24 && offset < 0) {
                newHeight = 0.25 * (container_rect.height - 16);
                newWidth = newHeight * (w / h);

            }
            if (computed_zoom > 4.0 && offset > 0) {
                newHeight = 4.0 * (container_rect.height - 16);
                newWidth = newHeight * (w / h);
            }


            resize(draggable, newWidth, newHeight);
        },
        [containerRef, draggableRef, setZoom]
    );

    const tpCache = useRef<Touch[]>([]);

    const touchMoveHandler = useCallback(
        (e: TouchEvent, idx: number) => {
            if (e.targetTouches.length !== 2 || e.changedTouches.length !== 2) return;
            const draggable = draggableRef.current.get(idx);
            const draggable_rect = draggable.getBoundingClientRect();

            // Check if the two target touches are the same ones that started
            // the 2-touch

            const point1_i = tpCache.current.findLastIndex(
                (tp) => tp.identifier === e.targetTouches[0].identifier,
            );
            const point2_i = tpCache.current.findLastIndex(
                (tp) => tp.identifier === e.targetTouches[1].identifier,
            );


            if (point1_i >= 0 && point2_i >= 0) {
                const point1 = tpCache.current[point1_i];
                const point2 = tpCache.current[point2_i];
                // Calculate the difference between the previous and current
                // distance between the two points
                const diff_old = Math.hypot(
                    point1.clientX - point2.clientX,
                    point1.clientY - point2.clientY
                );
                const diff_new = Math.hypot(
                    e.targetTouches[0].clientX - e.targetTouches[1].clientX,
                    e.targetTouches[0].clientY - e.targetTouches[1].clientY
                );

                let offset = (diff_new - diff_old);
                if (Math.abs(offset) < draggable_rect.height * 0.05) return;
                offset = offset * 0.000075;
                const w = draggable_rect.width;
                const h = draggable_rect.height;

                // Add 0.05% of width/height every scroll
                // This might need adjustment
                let newWidth = w + w * offset;
                let newHeight = h + h * offset;

                // Min value is 0.25 * width/height
                const container = containerRef.current.get(idx);
                const container_rect = container.getBoundingClientRect();

                const computed_zoom = newHeight / (container_rect.height - 16);
                setZoom(computed_zoom, idx);
                if (computed_zoom < 0.24 && offset < 0) {
                    newHeight = 0.25 * (container_rect.height - 16);
                    newWidth = newHeight * (w / h);

                }
                if (computed_zoom > 4.0 && offset > 0) {
                    newHeight = 4.0 * (container_rect.height - 16);
                    newWidth = newHeight * (w / h);
                }


                resize(draggable, newWidth, newHeight);

            } else {
                tpCache.current = [];
            }
        }, [containerRef, draggableRef, setZoom]
    );



    const touchStartHandler = useCallback(
        (e: TouchEvent) => {
            if (e.targetTouches.length != 2) return;
            for (const touch of e.targetTouches) {
                tpCache.current.push(touch);
            }
        },
        []
    );

    const touchEndHandler = useCallback(
        () => {
            tpCache.current = [];
        },
        []
    );




    /** Setup event listeners
     * - Wheel event for resizing or pinch gesture
     * - Auxclick event for fitting the element
     * - Double click event for centering the element
     * */
    useEffect(() => {
        if (!containerRef.current || !active) return;
        const containers = Array.from(containerRef.current.values());
        const helper_fkts = containers.map((container, idx) => {
            return (e: WheelEvent) => wheelHandler(e, idx);
        });

        containers.forEach((container, idx) => {
            container.addEventListener("wheel", helper_fkts[idx]);
        });

        return () => {
            containers.forEach((container, idx) => {
                container.removeEventListener("wheel", helper_fkts[idx]);
            });
        };
    }, [active, wheelHandler, containerRef, numPages, touchStartHandler]);

    useEffect(() => {
        if (!containerRef.current || !active) return;
        const containers = Array.from(containerRef.current.values());
        const helper_fkts = containers.map((container, idx) => {
            return () => fitPage(idx);
        });
        containers.forEach((container, idx) => {
            container.addEventListener("auxclick", helper_fkts[idx]);
        });
        return () => {
            containers.forEach((container, idx) => {
                container.removeEventListener("auxclick", helper_fkts[idx]);
            });
        };
    }, [active, containerRef, fitPage, numPages]);

    useEffect(() => {
        if (!containerRef.current || !active) return;
        const containers = Array.from(containerRef.current.values());
        const helper_fkts = containers.map((container, idx) => {
            return () => centerPage(idx);
        });
        containers.forEach((container, idx) => {
            container.addEventListener("dblclick", helper_fkts[idx]);
        });
        return () => {
            containers.forEach((container, idx) => {
                container.removeEventListener("dblclick", helper_fkts[idx]);
            });
        };
    }, [active, centerPage, containerRef, numPages]);




    return { fitPage, centerPage, zoom, setZoom, touchStartHandler, touchMoveHandler, touchEndHandler };
}

export function resize(elem: HTMLDivElement, width: number, height: number) {
    const scroll_rect = elem.getBoundingClientRect();
    const old_width = scroll_rect.width;
    const old_height = scroll_rect.height;

    // Resize element
    elem.style.width = width + "px";
    elem.style.height = height + "px";

    // Move element by half the difference
    const newX = elem.offsetLeft + (old_width - width) / 2;
    const newY = elem.offsetTop + (old_height - height) / 2;

    moveElm(elem, newX, newY);
}


export function resize_via_zoom(elem: HTMLDivElement, zoom: number) {
    const scroll_rect = elem.getBoundingClientRect();
    const old_width = scroll_rect.width;
    const old_height = scroll_rect.height;

    const new_height = (old_height - 16) * zoom;
    const new_width = new_height * (old_width / old_height);
    resize(elem, new_width, new_height);
}



export function _fitElement(elem: HTMLDivElement, container: HTMLDivElement) {
    const draggWidth = elem.offsetWidth;
    const draggHeight = elem.offsetHeight;

    const containerWidth = container.clientWidth - 16;
    const containerHeight = container.clientHeight - 16;

    // First check if height is fitting if width is adjusted
    const factor = containerWidth / draggWidth;
    const newHeight = draggHeight * factor;
    if (newHeight <= containerHeight) {
        // Height is fitting, adjust width
        resize(elem, containerWidth, newHeight);
    } else {
        // Height is not fitting, adjust height
        const factor = containerHeight / draggHeight;
        const newWidth = draggWidth * factor;
        resize(elem, newWidth, containerHeight);
    }
}

export function _centerElement(
    elem: HTMLDivElement,
    wrapper: HTMLDivElement,
    container: HTMLDivElement
) {
    const draggWidth = elem.offsetWidth;
    const draggHeight = elem.offsetHeight;

    const newX = wrapper.scrollWidth / 2 - elem.offsetWidth / 2;
    const newY = wrapper.scrollHeight / 2 - elem.offsetHeight / 2;

    // Scroll center with and height
    if (draggHeight > container.clientHeight) {
        container.scrollTop = newY - 8; // Small padding
    } else {
        container.scrollTop =
            newY - container.clientHeight / 2 + draggHeight / 2;
    }

    if (draggWidth > container.clientWidth) {
        container.scrollLeft = newX - 8; // Small padding
    } else {
        container.scrollLeft =
            newX - container.clientWidth / 2 + draggWidth / 2;
    }
    moveElm(elem, newX, newY);
    setScrollVisibility(
        { x: newX, y: newY, w: draggWidth, h: draggHeight },
        container
    );
}
