import {
    createContext,
    ReactNode,
    useCallback,
    useContext,
    useEffect,
    useState,
} from "react";

import { PagesContext } from "../page";
import { SidebarContext } from "../sidebar";
import { ToolBarContext } from "../toolbar";
import Button from "./snipEditor/button";
import { CutOverlay } from "./snipEditor/cut";
import Info from "./snipEditor/info";
import { ResizeOverlay } from "./snipEditor/resize";
import SubMenu from "./snipEditor/submenu";
import { Overlay, SnipEditorSubTool, SnipEditorToolContextI } from "./typing";

import { BaseSnip } from "snips/general/base";
import { ImageSnip } from "snips/general/image";

export const SnipEditorToolContext =
    createContext<SnipEditorToolContextI>(undefined);

export function SnipEditorToolContextProvider({
    children,
}: {
    children: ReactNode;
}) {
    const { activeTool, setActiveTool, activeSubMenu, setActiveSubMenu } =
        useContext(ToolBarContext);

    const { activeSnip } = useContext(SidebarContext);

    const active = activeTool === "snipEditor";
    const toggleActive = useCallback(
        () => setActiveTool("snipEditor"),
        [setActiveTool]
    );
    const active_sm = activeSubMenu === "snipEditor";
    const toggleSubMenu = useCallback(
        () => setActiveSubMenu("snipEditor"),
        [setActiveSubMenu]
    );

    useEffect(() => {
        if (!activeSnip && active_sm) {
            toggleSubMenu();
        }
    }, [activeSnip, active_sm, toggleSubMenu]);

    /** Automatically toggle the tool if a snip is selected and tool is not
     * already active and vice versa.
     */
    useEffect(() => {
        if (activeSnip && !active) {
            toggleActive();
        } else if (!activeSnip && active) {
            toggleActive();
        }
    }, [activeSnip, active, toggleActive]);

    /** The snip editor tool is effectively a nested toolbar which
     * is only shown if editor is active. It has a number of sub tools
     * which have there own state.
     */
    const [activeSubTool, _setActiveSubTool] =
        useState<SnipEditorSubTool>("resize");
    const setActiveSubTool = useCallback((tool: SnipEditorSubTool) => {
        _setActiveSubTool((prev) => {
            if (prev === tool) {
                return "resize";
            }
            return tool;
        });
    }, []);

    const [activeSubSubMenu, _setActiveSubSubMenu] =
        useState<SnipEditorSubTool>(undefined);
    const setActiveSubSubMenu = useCallback(
        (tool: SnipEditorSubTool) => {
            _setActiveSubSubMenu((prev) => {
                if (!prev) {
                    toggleSubMenu();
                }
                if (prev === tool) {
                    toggleSubMenu();
                    return undefined;
                }
                return tool;
            });
        },
        [toggleSubMenu]
    );
    useEffect(() => {
        if (!active_sm) {
            _setActiveSubSubMenu(undefined);
        }
    }, [active_sm, setActiveSubSubMenu]);

    /** The snipeditor tool is able to add an overlay to the wrapper
     * which can for instance be used to resize the snip.
     */
    const [overlay, setOverlay] = useState<Overlay>(null);

    useRenderSnip();
    const { snipRect, setSnipRect } = useSnipRect();
    const { cutoutRect, setCutoutRect } = useCutoutRect();

    const setOverlayByTool = useCallback((tool: SnipEditorSubTool) => {
        switch (tool) {
            case "resize":
                setOverlay({
                    dom: <ResizeOverlay />,
                    page_idx: 0,
                });
                break;
            case "cut":
                setOverlay({
                    dom: <CutOverlay />,
                    page_idx: 0,
                });

                break;
            default:
                setOverlay(undefined);
                break;
        }
    }, []);

    useEffect(() => {
        if (!active) {
            setOverlay(undefined);
        } else {
            setOverlayByTool(activeSubTool);
        }
    }, [active, activeSubTool, setOverlayByTool]);

    const ret: SnipEditorToolContextI = {
        active,
        toggleActive,
        activeSubMenu: active_sm,
        toggleSubMenu,

        activeSubTool,
        setActiveSubTool,
        activeSubSubMenu,
        setActiveSubSubMenu,

        overlay,
        setOverlay,

        snipRect,
        setSnipRect,
        cutoutRect,
        setCutoutRect,
    };

    return (
        <SnipEditorToolContext.Provider value={ret}>
            {children}
        </SnipEditorToolContext.Provider>
    );
}

const useRenderSnip = () => {
    const { overlayWorkers } = useContext(PagesContext);

    const { activeSnip } = useContext(SidebarContext);

    const render = (snip: BaseSnip) => {
        if (!snip) return;
        return (ctx: CanvasRenderingContext2D) => {
            snip.overlay_mode = true;
            snip.render(ctx);
            snip.overlay_mode = false;
        };
    };

    // Render if the snip changes
    useEffect(() => {
        if (!overlayWorkers || overlayWorkers.length < 1 || !activeSnip) return;
        const render_fkt = render(activeSnip);
        overlayWorkers[0].add(render_fkt);
        return () => {
            overlayWorkers[0].remove(render_fkt);
            overlayWorkers[0].render();
        };
    }, [activeSnip, overlayWorkers]);
};

interface SnipRect {
    x: number;
    y: number;
    width: number;
    height: number;
    aspectLocked: boolean;
    rotation: number;
    mirror: boolean;
}

/** Helper hook to initialize
 * the snipRect state with the activeSnip
 * and update the activeSnip with the snipRect state
 */
const useSnipRect = () => {
    const { overlayWorkers, numPages } = useContext(PagesContext);

    const { activeSnip } = useContext(SidebarContext);

    /** Width and height setter and getter to pass values from components to the activeSnip
     */
    const [snipRect, setSnipRect] = useState<SnipRect>({
        x: 0,
        y: 0,
        width: 0,
        height: 0,
        aspectLocked: true,
        rotation: 0,
        mirror: false,
    });

    const updateSnip = useCallback(
        (
            x: number,
            y: number,
            width: number,
            height: number,
            rotation: number,
            mirror: boolean
        ) => {
            if (activeSnip) {
                activeSnip.x = x;
                activeSnip.y = y;
                if (activeSnip instanceof ImageSnip) {
                    activeSnip.width = width;
                    activeSnip.height = height;
                    activeSnip.rot = `${rotation}${mirror ? "m" : ""}`;
                }
                if (overlayWorkers.length > 0) overlayWorkers[0].render();
            }
        },
        [overlayWorkers, activeSnip]
    );

    /** Update rect on snip change */
    useEffect(() => {
        if (!activeSnip) return;

        let rot = 0;
        let mirror = false;
        if (activeSnip instanceof ImageSnip) {
            rot = Number(activeSnip.rot.replace("m", ""));
            mirror = activeSnip.rot.includes("m");
        }

        setSnipRect({
            x: activeSnip.x,
            y: activeSnip.y,
            width: activeSnip.width,
            height: activeSnip.height,
            aspectLocked: true,
            rotation: rot,
            mirror: mirror,
        });
    }, [activeSnip, numPages]);

    /** Update snip on rect change */
    useEffect(() => {
        if (!activeSnip) return;
        updateSnip(
            snipRect.x,
            snipRect.y,
            snipRect.width,
            snipRect.height,
            snipRect.rotation,
            snipRect.mirror
        );
    }, [snipRect, activeSnip, updateSnip]);

    return { snipRect, setSnipRect };
};

/** Helper hook for the cutout
 * tool to get the cutout rect.
 *
 * The cutout is more or less implemented
 * similar to the drawimage slicing see
 * https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/drawImage
 */
const useCutoutRect = () => {
    const { overlayWorkers } = useContext(PagesContext);
    const { activeSnip } = useContext(SidebarContext);

    // this defines the sub rect of the snip which
    // is shown in the cutout
    const [cutoutRect, setCutoutRect] = useState({
        sx: 0,
        sy: 0,
        swidth: 0,
        sheight: 0,
    });

    const updateSnip = useCallback(
        (sx, sy, swidth, sheight) => {
            if (activeSnip && activeSnip instanceof ImageSnip) {
                activeSnip.sx = sx;
                activeSnip.sy = sy;
                activeSnip.swidth = swidth;
                activeSnip.sheight = sheight;
                if (overlayWorkers.length > 0) overlayWorkers[0].render();
            }
        },
        [overlayWorkers, activeSnip]
    );

    /** Update on snip change */
    useEffect(() => {
        if (activeSnip && activeSnip instanceof ImageSnip) {
            setCutoutRect({
                sx: activeSnip.sx,
                sy: activeSnip.sy,
                swidth: activeSnip.swidth,
                sheight: activeSnip.sheight,
            });
        }
    }, [activeSnip]);

    /** Update snip on rect change */
    useEffect(() => {
        if (!activeSnip) return;
        updateSnip(
            cutoutRect.sx,
            cutoutRect.sy,
            cutoutRect.swidth,
            cutoutRect.sheight
        );
    }, [cutoutRect, activeSnip, updateSnip]);
    return { cutoutRect, setCutoutRect };
};

export const SnipEditorTool = {
    Button,
    SubMenu,
    Info,
};
