import { useCallback, useContext, useEffect, useRef, useState } from "react";

import { SnipEditorToolContext } from "../snipEditor.tool";
import { set_rect } from "./resize";

import { ImageSnip } from "snips/general/image";

import { PagesContext } from "components/viewer/page";
import { SidebarContext } from "components/viewer/sidebar";

import styles from "./styles.module.scss";

export function CutOverlay() {
    const overlayRef = useRef<HTMLDivElement>(null);
    const handleRefTop = useRef<HTMLDivElement>(null);
    const handleRefBottom = useRef<HTMLDivElement>(null);
    const handleRefLeft = useRef<HTMLDivElement>(null);
    const handleRefRight = useRef<HTMLDivElement>(null);
    const moveRef = useRef<HTMLDivElement>(null);
    const cutoutRef = useRef<HTMLDivElement>(null);

    const { draggableRef } = useContext(PagesContext);

    const { activeSnip } = useContext(SidebarContext);

    const { snipRect, cutoutRect, setCutoutRect } = useContext(
        SnipEditorToolContext
    );

    // state to save the current mouse positions between move events
    const [mouse, setMouse] = useState<{
        prev_x: number;
        prev_y: number;
        current_x: number;
        current_y: number;
        direction: string;
    }>(undefined);

    /** If snip changes update the overlay this
     * should automatically happen when the snip changes
     */

    /** If the cutout changes update the overlay(cutout)
     * this should automatically happen when the cutout changes
     */
    useEffect(() => {
        if (
            !activeSnip ||
            !(activeSnip instanceof ImageSnip) ||
            !draggableRef.current
        )
            return;
        const overlay = overlayRef.current;
        const draggable = draggableRef.current.get(0);

        const scale = draggable.clientHeight / 2000;
        const x_o = scale * snipRect.x;
        const y_o = scale * snipRect.y;
        const w_o = scale * snipRect.width;
        const h_o = scale * snipRect.height;

        set_rect(
            overlay,
            x_o + draggable.offsetLeft,
            y_o + draggable.offsetTop,
            w_o,
            h_o
        );

        const cutout = cutoutRef.current;

        const scale_h = overlayRef.current.clientHeight / activeSnip.img_height;
        const scale_w = overlayRef.current.clientWidth / activeSnip.img_width;
        const sx = cutoutRect.sx * scale_w;
        const sy = cutoutRect.sy * scale_h;
        const sw = cutoutRect.swidth * scale_w;
        const sh = cutoutRect.sheight * scale_h;

        set_rect(cutout, sx, sy, sw, sh);
    }, [
        activeSnip,
        cutoutRect,
        draggableRef,
        snipRect.height,
        snipRect.width,
        snipRect.x,
        snipRect.y,
    ]);

    /** If we move the mouse we adjust the cutout
     * as this is a hook it will automatically
     * update when the mouse changes
     */
    useEffect(() => {
        if (!activeSnip || !mouse) return;
        if (!(activeSnip instanceof ImageSnip)) return;

        const delta = [
            mouse.current_x - mouse.prev_x,
            mouse.current_y - mouse.prev_y,
        ] as [number, number];
        if (delta[0] === 0 && delta[1] === 0) {
            return;
        }
        const scale_h = overlayRef.current.clientHeight / activeSnip.img_height;
        const scale_w = overlayRef.current.clientWidth / activeSnip.img_width;

        const delta_w = delta[0] / scale_w;
        const delta_h = delta[1] / scale_h;

        setCutoutRect((prev) => {
            const new_cutout = prev;

            // Keep track of the new cutout difference
            // This could use some refactoring but is working

            const max_width = activeSnip.img_width;
            const max_height = activeSnip.img_height;

            const collision = {
                left: prev.sx + delta_w < 0,
                top: prev.sy + delta_h < 0,
                right: prev.sx + prev.swidth + delta_w > max_width,
                bottom: prev.sy + prev.sheight + delta_h > max_height,
            };
            switch (mouse.direction) {
                case "move":
                    // check all bounds
                    if (collision.left) {
                        new_cutout.sx = 0;
                    } else if (collision.right) {
                        new_cutout.sx = max_width - prev.swidth;
                    } else {
                        new_cutout.sx += delta_w;
                    }
                    if (collision.top) {
                        new_cutout.sy = 0;
                    } else if (collision.bottom) {
                        new_cutout.sy = max_height - prev.sheight;
                    } else {
                        new_cutout.sy += delta_h;
                    }
                    break;
                case "top":
                    new_cutout.sy += delta_h;
                    new_cutout.sheight -= delta_h;
                    break;
                case "left":
                    new_cutout.sx += delta_w;
                    new_cutout.swidth -= delta_w;
                    break;
                case "right":
                    new_cutout.swidth += delta_w;
                    break;
                case "bottom":
                    new_cutout.sheight += delta_h;
                    break;
                default:
                    break;
            }

            // Boundaries
            if (new_cutout.sheight < 1) {
                new_cutout.sheight = 1;
            }
            if (new_cutout.swidth < 1) {
                new_cutout.swidth = 1;
            }
            if (new_cutout.sy < 0) {
                new_cutout.sy = 0;
            }
            if (new_cutout.sx < 0) {
                new_cutout.sx = 0;
            }
            if (new_cutout.sy + new_cutout.sheight > max_height) {
                new_cutout.sheight = max_height - new_cutout.sy;
            }
            if (new_cutout.sx + new_cutout.swidth > max_width) {
                new_cutout.swidth = max_width - new_cutout.sx;
            }

            return { ...new_cutout };
        });
    }, [activeSnip, mouse, setCutoutRect]);

    /** The mouse move handler are defined in the following
     *
     * These edit the mouseposition state which in response triggers and
     * update of the overlay (done here) and the snip (done in the snipEditor.tool.tsx)
     */
    const mouseMove = useCallback((e) => {
        e.preventDefault();
        // compute difference between current and previous mouse position
        setMouse((prev) => {
            return {
                prev_x: prev.current_x,
                prev_y: prev.current_y,
                current_x: e.clientX,
                current_y: e.clientY,
                direction: prev.direction,
            };
        });
    }, []);

    const mouseUp = useCallback(
        (e) => {
            e.preventDefault();
            window.removeEventListener("mousemove", mouseMove);
            window.removeEventListener("mouseup", mouseUp);
        },
        [mouseMove]
    );

    const mouseDown = useCallback(
        (e, direction) => {
            e.preventDefault();
            setMouse({
                prev_x: e.clientX,
                prev_y: e.clientY,
                current_x: e.clientX,
                current_y: e.clientY,
                direction: direction,
            });

            // Add event listeners to the window
            window.addEventListener("mousemove", mouseMove);
            window.addEventListener("mouseup", mouseUp);
        },
        [mouseMove, mouseUp]
    );

    /** Add event listeners to the handles
     * such that we can catch the mousemove events
     */
    useEffect(() => {
        if (
            !handleRefTop.current ||
            !handleRefBottom.current ||
            !handleRefLeft.current ||
            !handleRefRight.current ||
            !moveRef.current
        )
            return;

        const handles = [
            handleRefTop.current,
            handleRefBottom.current,
            handleRefLeft.current,
            handleRefRight.current,
            moveRef.current,
        ];

        // Helper functions to be able to remove the event listeners
        const helper_functions = handles.map((handle) => {
            return (e) => {
                mouseDown(e, handle.dataset.position);
            };
        });

        for (let i = 0; i < handles.length; i++) {
            handles[i].addEventListener("mousedown", helper_functions[i]);
        }

        return () => {
            setMouse(undefined);

            for (let i = 0; i < handles.length; i++) {
                handles[i].removeEventListener(
                    "mousedown",
                    helper_functions[i]
                );
            }
            // Remove just in case
            window.removeEventListener("mousemove", mouseMove);
            window.removeEventListener("mouseup", mouseUp);
        };
    }, [
        handleRefTop,
        handleRefBottom,
        handleRefLeft,
        handleRefRight,
        moveRef,
        mouseDown,
        mouseMove,
        mouseUp,
    ]);

    /** Some special binds:
     * - reset the cutout to the full image on double click
     */
    useEffect(() => {
        if (
            !cutoutRef.current ||
            !activeSnip ||
            !(activeSnip instanceof ImageSnip)
        )
            return;
        const cutout = cutoutRef.current;
        const reset = () => {
            setCutoutRect({
                sx: 0,
                sy: 0,
                swidth: activeSnip.img_width,
                sheight: activeSnip.img_height,
            });
        };
        cutout.addEventListener("dblclick", reset);
        return () => {
            cutout.removeEventListener("dblclick", reset);
        };
    }, [activeSnip, cutoutRef, setCutoutRect]);

    return (
        <>
            <div ref={overlayRef} className={styles.cutWrapper}>
                <div
                    ref={cutoutRef}
                    className={styles.cutOverlay}
                    data-rotation={snipRect.rotation}
                    data-mirror={snipRect.mirror}
                >
                    <div
                        ref={moveRef}
                        className={styles.move}
                        data-position="move"
                    ></div>
                    <div
                        ref={handleRefTop}
                        className={styles.handle}
                        data-position="top"
                    />
                    <div
                        ref={handleRefBottom}
                        className={styles.handle}
                        data-position="bottom"
                    />
                    <div
                        ref={handleRefLeft}
                        className={styles.handle}
                        data-position="left"
                    />
                    <div
                        ref={handleRefRight}
                        className={styles.handle}
                        data-position="right"
                    />
                </div>
            </div>
        </>
    );
}
