import { useCallback, useContext, useEffect, useRef, useState } from "react";

import { SnipEditorToolContext } from "../snipEditor.tool";

import { ImageSnip } from "snips/general/image";

import { PagesContext } from "components/viewer/page";
import { SidebarContext } from "components/viewer/sidebar";

import styles from "./styles.module.scss";

export function ResizeOverlay() {
    const overlayRef = useRef<HTMLDivElement>(null);
    const handleRefTL = useRef<HTMLDivElement>(null);
    const handleRefTR = useRef<HTMLDivElement>(null);
    const handleRefBL = useRef<HTMLDivElement>(null);
    const handleRefBR = useRef<HTMLDivElement>(null);
    const moveRef = useRef<HTMLDivElement>(null);

    const { overlayWorkers, draggableRef } = useContext(PagesContext);

    const { activeSnip } = useContext(SidebarContext);

    const { snipRect, setSnipRect } = useContext(SnipEditorToolContext);

    // state to save the current mouse positions between move events
    const [mouse, setMouse] = useState<{
        prev_x: number;
        prev_y: number;
        current_x: number;
        current_y: number;
        direction: string;
    }>(undefined);

    const resize = useCallback(() => {
        const overlay = overlayRef.current;
        const draggable = draggableRef.current.get(0);
        const scale = draggable.clientHeight / 2000;
        const x_o = scale * snipRect.x;
        const y_o = scale * snipRect.y;
        const w_o = scale * snipRect.width;
        const h_o = scale * snipRect.height;

        // Resize overlay
        set_rect(
            overlay,
            x_o + draggable.offsetLeft,
            y_o + draggable.offsetTop,
            w_o,
            h_o
        );
    }, [draggableRef, snipRect.height, snipRect.width, snipRect.x, snipRect.y]);

    /* If snip changes update overlay this should
     * automatically align the overlay to the snip
     */
    useEffect(() => {
        /** Also add resize observer if the page changes to resize the overlay
         * accordingly
         */
        const observer = new ResizeObserver(resize);
        observer.observe(draggableRef.current.get(0));
        resize();

        return () => observer.disconnect();
    }, [draggableRef, overlayWorkers, resize]);

    /**  If the mouse delta changes update the snip rect
     *  this should automatically happen when the mouse
     * moves
     * */
    useEffect(() => {
        if (!mouse) return;
        const delta = [
            mouse.current_x - mouse.prev_x,
            mouse.current_y - mouse.prev_y,
        ] as [number, number];
        if (delta[0] == 0 && delta[1] == 0) return;

        const scale = draggableRef.current.get(0).clientHeight / 2000;
        const delta_trans = [delta[0] / scale, delta[1] / scale] as [
            number,
            number
        ];

        setSnipRect((snipRect) => {
            // Check if aspect ratio is locked
            if (snipRect.aspectLocked && mouse.direction !== "move") {
                // If locked, only allow changes with locked aspect ratio
                const aR = snipRect.width / snipRect.height;
                if (mouse.direction === "top-left") {
                    delta_trans[0] = delta_trans[1] * aR;
                } else if (mouse.direction === "top-right") {
                    delta_trans[0] = -delta_trans[1] * aR;
                } else if (mouse.direction === "bottom-left") {
                    delta_trans[0] = -delta_trans[1] * aR;
                } else if (mouse.direction === "bottom-right") {
                    delta_trans[0] = delta_trans[1] * aR;
                } else {
                    return snipRect;
                }
            }

            // Get total values from diff
            return {
                ...snipRect,
                ...diff_to_total(
                    snipRect,
                    { x: delta_trans[0], y: delta_trans[1] },
                    mouse.direction
                ),
            };
        });
    }, [draggableRef, mouse, setSnipRect]);

    /** The mouse move handler are defined in the following
     *
     * These edit the mouseposition state which in response triggers and
     * update of the overlay (done here) and the snip (done in the snipEditor.tool.tsx)
     */
    const mouseMove = useCallback((e) => {
        e.preventDefault();
        // Compute difference between initial and current mouse position
        setMouse((prev) => {
            return {
                prev_x: prev.current_x,
                prev_y: prev.current_y,
                current_x: e.clientX,
                current_y: e.clientY,
                direction: prev.direction,
            };
        });
    }, []);

    const mouseUp = useCallback(
        (e) => {
            e.preventDefault();
            // Remove event listeners
            window.removeEventListener("pointermove", mouseMove);
            window.removeEventListener("pointerup", mouseUp);
        },
        [mouseMove]
    );

    const mouseDown = useCallback(
        (e, direction) => {
            e.preventDefault();
            // Set initial mouse position
            setMouse({
                prev_x: e.clientX,
                prev_y: e.clientY,
                current_x: e.clientX,
                current_y: e.clientY,
                direction: direction,
            });

            // Add event listeners
            window.addEventListener("pointermove", mouseMove);
            window.addEventListener("pointerup", mouseUp);
        },
        [mouseMove, mouseUp]
    );

    /** Add event listeners to the handles
     * such that we can catch the mousemove events
     */
    useEffect(() => {
        const handles = [
            moveRef.current,
            handleRefTL.current,
            handleRefTR.current,
            handleRefBL.current,
            handleRefBR.current,
        ].filter((e) => e !== null);

        // Helper functions to be able to remove the event listeners
        const helper_functions = handles.map((handle) => {
            return (e) => {
                mouseDown(e, handle.dataset.position);
            };
        });

        for (let i = 0; i < handles.length; i++) {
            handles[i].addEventListener("pointerdown", helper_functions[i]);
        }
        return () => {
            setMouse(undefined);

            for (let i = 0; i < handles.length; i++) {
                handles[i].removeEventListener(
                    "pointerdown",
                    helper_functions[i]
                );
            }
            window.removeEventListener("pointermove", mouseMove);
            window.removeEventListener("pointerup", mouseUp);
        };
    }, [
        handleRefTL,
        handleRefTR,
        handleRefBL,
        handleRefBR,
        moveRef,
        mouseDown,
        mouseMove,
        mouseUp,
    ]);

    /** Some special binds:
     * - reset the resize
     */
    useEffect(() => {
        if (!overlayRef.current || !activeSnip) return;
        const overlay = overlayRef.current;
        const reset = () => {
            setSnipRect((prev) => {
                return {
                    ...prev,
                    x: 0,
                    y: 0,
                    width:
                        activeSnip instanceof ImageSnip
                            ? activeSnip.img_width
                            : prev.width,
                    height:
                        activeSnip instanceof ImageSnip
                            ? activeSnip.img_height
                            : prev.height,
                };
            });
        };
        overlay.addEventListener("dblclick", reset);
        return () => {
            overlay.removeEventListener("dblclick", reset);
        };
    }, [activeSnip, overlayRef, setSnipRect]);

    return (
        <div
            ref={overlayRef}
            className={styles.resizeOverlay}
            data-rotation={snipRect.rotation}
            data-mirror={snipRect.mirror}
        >
            <div
                ref={moveRef}
                className={styles.move}
                data-position="move"
            ></div>
            {activeSnip instanceof ImageSnip && (
                <>
                    <div
                        ref={handleRefTL}
                        className={styles.handle}
                        data-position="top-left"
                    ></div>
                    <div
                        ref={handleRefTR}
                        className={styles.handle}
                        data-position="top-right"
                    ></div>
                    <div
                        ref={handleRefBL}
                        className={styles.handle}
                        data-position="bottom-left"
                    ></div>
                    <div
                        ref={handleRefBR}
                        className={styles.handle}
                        data-position="bottom-right"
                    ></div>
                </>
            )}
        </div>
    );
}

/** Sets the size an element
 *
 * @param ele The element to set the size of
 * @param x The x coordinate of the element
 * @param y The y coordinate of the element
 * @param width The width of the element
 * @param height The height of the element
 */
export const set_rect = (ele, x, y, width, height) => {
    // Expects canvas coordinates
    if (width <= 1 || height <= 1) {
        return;
    }
    ele.style.left = x + "px";
    ele.style.top = y + "px";
    ele.style.width = width + "px";
    ele.style.height = height + "px";
};

export const resize_rect = (ele, diffX, diffY, diffWidth, diffHeight) => {
    // As we do not want to resize into negative space, we need to check if the
    // element is already at the edge of the canvas and if the resize would
    // move it into negative space. If so, we need to adjust the resize
    // accordingly.
    const newWidth = ele.clientWidth + diffWidth;
    const newHeight = ele.clientHeight + diffHeight;
    if (newWidth <= 1) {
        return;
    }
    if (newHeight <= 1) {
        return;
    }

    ele.style.left = ele.offsetLeft + diffX + "px";
    ele.style.top = ele.offsetTop + diffY + "px";
    ele.style.width = newWidth + "px";
    ele.style.height = newHeight + "px";
};

interface Rect {
    x: number;
    y: number;
    width: number;
    height: number;
}
interface Diff {
    x: number;
    y: number;
}

const diff_to_total = (prev: Rect, diff: Diff, direction) => {
    switch (direction) {
        case "move":
            return {
                x: prev.x + diff.x,
                y: prev.y + diff.y,
                width: prev.width,
                height: prev.height,
            };
        case "top-left":
            return {
                x: prev.x + diff.x,
                y: prev.y + diff.y,
                width: prev.width - diff.x,
                height: prev.height - diff.y,
            };
        case "top-right":
            return {
                x: prev.x,
                y: prev.y + diff.y,
                width: prev.width + diff.x,
                height: prev.height - diff.y,
            };
        case "bottom-left":
            return {
                x: prev.x + diff.x,
                y: prev.y,
                width: prev.width - diff.x,
                height: prev.height + diff.y,
            };
        case "bottom-right":
            return {
                x: prev.x,
                y: prev.y,
                width: prev.width + diff.x,
                height: prev.height + diff.y,
            };
    }
};
