import { useCallback, useContext, useEffect, useState } from "react";

import { SnipEditorToolContext } from "../snipEditor.tool";

import styles from "./styles.module.scss";

export default function SubMenu() {
    return (
        <div>
            <Resize_WH />
            <Rotate />
        </div>
    );
}

/** Buttons to resize width and
 * height of the snip.
 */
function Resize_WH() {
    const { snipRect, setSnipRect } = useContext(SnipEditorToolContext);

    // local state for nice displaying
    const [inputState, setInputState] = useState({
        width: snipRect.width.toFixed(0),
        height: snipRect.height.toFixed(0),
    });

    const widthHandler = useCallback(
        (e) => {
            // Check if input is valid
            if (!isNaN(Number(e.target.value)) && Number(e.target.value) > 0) {
                setSnipRect((prev) => {
                    const valid_rect: {
                        width: number;
                        height?: number;
                    } = {
                        width: Number(e.target.value),
                    };
                    // Add width if aspect locked
                    if (prev.aspectLocked) {
                        valid_rect.height =
                            (valid_rect.width / prev.width) * prev.height;
                    }
                    setInputState((prev) => {
                        return {
                            ...prev,
                            height: valid_rect.height.toFixed(0),
                            width: valid_rect.width.toFixed(0),
                        };
                    });
                    return {
                        ...prev,
                        ...valid_rect,
                    };
                });
            } else {
                setInputState((prev) => {
                    return {
                        ...prev,
                        width: e.target.value,
                    };
                });
            }
        },
        [setInputState, setSnipRect]
    );

    const heightHandler = useCallback(
        (e) => {
            // Check if input is valid
            if (!isNaN(Number(e.target.value)) && Number(e.target.value) > 0) {
                setSnipRect((prev) => {
                    const valid_rect: {
                        height: number;
                        width?: number;
                    } = {
                        height: Number(e.target.value),
                    };
                    // Add width if aspect locked
                    if (prev.aspectLocked) {
                        valid_rect.width =
                            (valid_rect.height / prev.height) * prev.width;
                    }
                    setInputState((prev) => {
                        return {
                            ...prev,
                            height: valid_rect.height.toFixed(0),
                            width: valid_rect.width.toFixed(0),
                        };
                    });
                    return {
                        ...prev,
                        ...valid_rect,
                    };
                });
            } else {
                setInputState((prev) => {
                    return {
                        ...prev,
                        height: e.target.value,
                    };
                });
            }
        },
        [setInputState, setSnipRect]
    );

    useEffect(() => {
        setInputState((prev) => {
            return {
                ...prev,
                width: snipRect.width.toFixed(0),
                height: snipRect.height.toFixed(0),
            };
        });
    }, [snipRect]);

    return (
        <div className={styles.wrapperWH}>
            <div>
                <label>Width</label>
                <div className="input-container">
                    <input
                        className="form-control"
                        type="number"
                        value={inputState.width}
                        onChange={widthHandler}
                    />
                </div>
                <label>Height</label>
                <div className="input-container">
                    <input
                        className="form-control"
                        type="number"
                        value={inputState.height}
                        onChange={heightHandler}
                    />
                </div>
            </div>

            <div className={styles.linked}>
                <button
                    onClick={() => {
                        setSnipRect((prev) => {
                            return {
                                ...prev,
                                aspectLocked: !prev.aspectLocked,
                            };
                        });
                    }}
                >
                    <i
                        className={
                            snipRect.aspectLocked
                                ? "bi bi-lock"
                                : "bi bi-unlock"
                        }
                    />
                </button>
                {/*Draw lines with css*/}
                <div className="linked-line" id="linked-line-0"></div>
                <div className="linked-line" id="linked-line-1"></div>
                <div className="linked-line" id="linked-line-2"></div>
            </div>
        </div>
    );
}

function Rotate() {
    const { setSnipRect } = useContext(SnipEditorToolContext);

    const [rotation, _setRotation] = useState<number>(0);
    const [mirror, setMirror] = useState<boolean>(false);

    // Helper function to restrict rotation to [0-360)
    function setRotation(rot: number) {
        let newRot;
        if (rot < 0) {
            newRot = rot + 360;
        } else if (rot >= 360) {
            newRot = rot - 360;
        } else {
            newRot = rot;
        }
        _setRotation(newRot);
    }

    useEffect(() => {
        setSnipRect((prev) => {
            return {
                ...prev,
                rotation: rotation,
                mirror: mirror,
            };
        });
    }, [rotation, mirror, setSnipRect]);

    return (
        <div className={styles.wrapperWH}>
            <div>
                <label>Rotation</label>
                <div className="input-container">
                    <button
                        className="btn"
                        onClick={() => {
                            setRotation(rotation - 90);
                        }}
                    >
                        ↺
                    </button>
                    <button
                        className="btn"
                        style={{ marginLeft: "0.25rem" }}
                        onClick={() => {
                            setRotation(rotation + 90);
                        }}
                    >
                        ↻
                    </button>
                    <button
                        className="btn"
                        style={{ marginLeft: "0.25rem" }}
                        onClick={() => {
                            setMirror(!mirror);
                        }}
                    >
                        ⇆
                    </button>
                </div>
                <label>
                    Warning{" "}
                    <i
                        className="bi bi-exclamation-triangle"
                        style={{ display: "inline-block" }}
                    />
                </label>
                <div style={{ fontSize: "0.75rem" }}>
                    Rotation and mirroring break the resizing and cropping!
                    Please use with caution.
                </div>
            </div>
        </div>
    );
}
