import useUser from "lib/useUser";
import {
    createContext,
    SetStateAction,
    useCallback,
    useContext,
    useEffect,
    useState,
} from "react";

import { ToolBarContext } from "../toolbar";
import Button from "./text/button";
import Info from "./text/info";
import useTextboxOverlay from "./text/overlay";
import SubMenu from "./text/submenu";
import { TextStyles, TextToolContextI, ToolComponents } from "./typing";

import { AllowedFonts } from "database/fonts";
import { TextSnip } from "snips/general/text";

export const TextToolContext = createContext<TextToolContextI>(undefined);

export function TextToolContextProvider({
    children,
}: {
    children: React.ReactNode;
}) {
    const { config, isLoadingConfig, isErrorConfig, mutateConfig } = useUser();
    const [initialLoad, setInitialLoad] = useState<boolean>(false);

    const { activeTool, setActiveTool, activeSubMenu, setActiveSubMenu } =
        useContext(ToolBarContext);
    const active = activeTool === "text";
    const toggleActive = useCallback(
        () => setActiveTool("text"),
        [setActiveTool]
    );
    const active_sm = activeSubMenu === "text";
    const toggleSubMenu = useCallback(
        () => setActiveSubMenu("text"),
        [setActiveSubMenu]
    );

    const [text, setText] = useState<string>("");

    const [textStyles, _setTextStyles] = useState<TextStyles>({
        fontFamily: "Arial" as AllowedFonts,
        fontSize: 12,
        lineHeight: 1.25,
        lineWrap: 400,
    });

    const [activeSnip, setActiveSnip] = useState<TextSnip>(null);
    const [activePageIdx, setActivePageIdx] = useState<number>(0);

    /** The text tool allows to type in a
     * overlaying text box. The text box is
     * resizable and movable. The text box
     * can than be placed similar to a
     * snip
     */
    const { overlay } = useTextboxOverlay(
        active,
        text,
        setText,
        textStyles,
        _setTextStyles,
        activeSnip,
        setActiveSnip,
        activePageIdx,
        setActivePageIdx
    );

    /** Update the global config when the text styles change
     * this is one directional and only updates the config
     * */
    const setTextStyles = useCallback(
        (newStyles: SetStateAction<TextStyles>) => {
            _setTextStyles((prev) => {
                const newStyles_ =
                    typeof newStyles === "function"
                        ? newStyles(prev)
                        : newStyles;
                // Update the config when the text styles change
                mutateConfig({
                    text_font: newStyles_.fontFamily,
                    text_fontSize: newStyles_.fontSize,
                    text_lineHeight: newStyles_.lineHeight,
                    text_lineWrap: newStyles_.lineWrap,
                });
                return newStyles_;
            });
        },
        [mutateConfig]
    );

    /** Set initial text styles with user config */
    useEffect(() => {
        if (!isLoadingConfig && !isErrorConfig && config && !initialLoad) {
            setInitialLoad(true);
            _setTextStyles(() => ({
                fontFamily: config.text_font as AllowedFonts,
                fontSize: config.text_fontSize,
                lineHeight: config.text_lineHeight,
                lineWrap: config.text_lineWrap,
            }));
        }
    }, [config, isLoadingConfig, isErrorConfig, initialLoad]);

    const ret: TextToolContextI = {
        active,
        toggleActive,
        activeSubMenu: active_sm,
        toggleSubMenu,
        overlay,

        textStyles,
        setTextStyles,
        activeSnip,
        setActiveSnip,
    };

    return (
        <TextToolContext.Provider value={ret}>
            {children}
        </TextToolContext.Provider>
    );
}

export const TextTool: ToolComponents = {
    Button,
    SubMenu,
    Info,
};
