import { useMouseDelta } from "lib/useMouseDelta";
import {
    Dispatch,
    SetStateAction,
    useCallback,
    useContext,
    useEffect,
    useRef,
    useState,
} from "react";

import { IOverlayTextProps, OverlayText } from "../interaction/overlay";
import { type TextStyles } from "../typing";

import { TextSnip } from "snips/general/text";

import { PagesContext } from "components/viewer/page";

import styles from "./styles.module.scss";

interface SnipRect {
    x: number;
    y: number;
    width: number;
    height: number;
}

export default function useTextboxOverlay(
    active: boolean,
    text: string,
    setText: Dispatch<SetStateAction<string>>,
    textStyles: TextStyles,
    setTextStyles: Dispatch<SetStateAction<TextStyles>>,
    activeSnip: TextSnip,
    setActiveSnip: Dispatch<SetStateAction<TextSnip>>,
    activePageIdx: number,
    setActivePageIdx: Dispatch<SetStateAction<number>>
) {
    const { overlayCanvasRef, numPages, containerRef, overlayWorkers } =
        useContext(PagesContext);

    /** State to store the currently active text snip
     * position and dimensions.
     *
     * These values are in the database coordinate system!
     * Need to be converted for the overlay!
     */
    const [snipRect, setSnipRect] = useState<SnipRect>({
        x: 0,
        y: 0,
        width: 0,
        height: 0,
    });

    /** Update the rect if the snip changes */
    useEffect(() => {
        if (!activeSnip) return;
        setSnipRect({
            x: activeSnip.x,
            y: activeSnip.y,
            width: activeSnip.lineWrap,
            height: activeSnip.height,
        });
        setText(activeSnip.text);
    }, [activeSnip, setText]);

    /** Update the snip on changes */
    useEffect(() => {
        if (!activeSnip) return;
        activeSnip.x = snipRect.x;
        activeSnip.y = snipRect.y;
        overlayWorkers[activePageIdx].render();
    }, [activePageIdx, activeSnip, overlayWorkers, snipRect]);

    /** Update the text styles on changes */
    useEffect(() => {
        if (!activeSnip) return;
        activeSnip.fontFamily = textStyles.fontFamily;
        activeSnip.fontSize = textStyles.fontSize;
        activeSnip.lineHeight = textStyles.lineHeight * textStyles.fontSize;
        activeSnip.lineWrap = textStyles.lineWrap;
        activeSnip.text = text;

        //Recompute height
        setSnipRect((prev) => {
            console.log(activeSnip.height);
            return { ...prev, height: activeSnip.height };
        });

        overlayWorkers[activePageIdx].render();
    }, [activeSnip, textStyles, overlayWorkers, activePageIdx, text]);

    /** Render whenever font or rect changes
     *
     */
    const render = (snip: TextSnip) => {
        if (!snip) return;
        return (ctx: CanvasRenderingContext2D) => {
            snip.overlay_mode = true;
            snip.render(ctx);
            snip.overlay_mode = false;
        };
    };

    // Render if the snip changes
    useEffect(() => {
        if (!overlayWorkers || !activeSnip || !active) return;
        const render_fkt = render(activeSnip);
        overlayWorkers[activePageIdx].add(render_fkt);
        overlayWorkers[activePageIdx].render();

        return () => {
            overlayWorkers[activePageIdx].remove(render_fkt);
            overlayWorkers[activePageIdx].render();
        };
    }, [active, activePageIdx, activeSnip, overlayWorkers]);

    const mouseDown = useCallback(
        (e: MouseEvent, page_idx: number) => {
            if (!active) return;
            if (!overlayCanvasRef.current) return;

            // target not in canvas
            if (
                !Array.from(overlayCanvasRef.current.values()).includes(
                    e.target as HTMLCanvasElement
                )
            )
                return;

            const canvas = overlayCanvasRef.current.get(page_idx);
            // Transform to database coordinates system
            const scale = canvas.clientHeight / 2000;
            const x_db = e.offsetX / scale;
            const y_db = e.offsetY / scale;
            setActivePageIdx(page_idx);

            if (activeSnip) {
                // Set the snip to the new position
                setSnipRect((p) => {
                    return { ...p, x: x_db, y: y_db };
                });

                overlayWorkers[page_idx].render();
            } else {
                const textSnip = new TextSnip({
                    x: x_db,
                    y: y_db,
                });
                setActiveSnip(textSnip);
            }
        },
        [
            active,
            activeSnip,
            overlayCanvasRef,
            overlayWorkers,
            setActivePageIdx,
            setActiveSnip,
        ]
    );

    useEffect(() => {
        //mousedown on draggable
        if (!containerRef.current) return;
        const canvas = Array.from(overlayCanvasRef.current.values());
        const MDS = [];

        canvas.forEach((draggable, idx) => {
            MDS.push((e: MouseEvent) => mouseDown(e, idx));
            draggable.addEventListener("mousedown", MDS[idx]);
        });
        return () => {
            canvas.forEach((draggable, idx) => {
                draggable.removeEventListener("mousedown", MDS[idx]);
            });
        };
    }, [containerRef, mouseDown, numPages, overlayCanvasRef]);

    return {
        overlay: activeSnip
            ? {
                  dom: (
                      <OverlayTextTool
                          snipRect={snipRect}
                          setSnipRect={setSnipRect}
                          textStyle={textStyles}
                          setTextStyle={setTextStyles}
                          page_idx={activePageIdx}
                          text={text}
                          onInput={(
                              e: React.ChangeEvent<HTMLTextAreaElement>
                          ) => {
                              const text = e.currentTarget.value;
                              setText(text);
                              activeSnip.text = text;
                              overlayWorkers[activePageIdx].render();
                          }}
                      />
                  ),
                  page_idx: activePageIdx,
              }
            : null,
    };
}

interface IOverlayTextToolProps
    extends Omit<
        IOverlayTextProps,
        "textProps" | "x" | "y" | "width" | "height" | "onInput" | "textStyle"
    > {
    snipRect: SnipRect;
    setSnipRect: Dispatch<SetStateAction<SnipRect>>;
    textStyle: TextStyles;
    setTextStyle: Dispatch<SetStateAction<TextStyles>>;
    onInput?: (e: React.ChangeEvent<HTMLTextAreaElement>) => void;
}

function OverlayTextTool({
    // texttooloverlay
    snipRect,
    setSnipRect,
    textStyle,
    setTextStyle,
    onInput = undefined,
    // overlay
    text,
    page_idx = 0,
    children = null,
    callback = null,
}: IOverlayTextToolProps) {
    const { draggableRef } = useContext(PagesContext);
    const textRef = useRef<HTMLTextAreaElement>(null);
    const resizeHandleRef = useRef<HTMLDivElement>(null);
    const moveHandleRef = useRef<HTMLDivElement>(null);

    /** Use mouse move handler to get mouse delta between move events
     * and use it to move the overlay.
     */
    const { mouse: mouseMove } = useMouseDelta(moveHandleRef);
    const { mouse: mouseResize } = useMouseDelta(resizeHandleRef);

    /** State for the width/linewrap of the snip
     * default is 400 px, min values is 10 px
     * in database coordinates using snipRect
     */

    /** Update the position of the overlay via the move handle */
    useEffect(() => {
        if (!mouseMove) return;
        const delta = [
            mouseMove.current_x - mouseMove.prev_x,
            mouseMove.current_y - mouseMove.prev_y,
        ];
        if (delta[0] == 0 && delta[1] == 0) return;

        const scale = draggableRef.current.get(page_idx).clientHeight / 2000;
        const delta_trans = [delta[0] / scale, delta[1] / scale];

        setSnipRect((prev) => {
            return {
                ...prev,
                x: prev.x + delta_trans[0],
                y: prev.y + delta_trans[1],
            };
        });
    }, [mouseMove, draggableRef, setSnipRect, page_idx]);

    /** Update width of the textbox (also lineWrap) */
    useEffect(() => {
        if (!mouseResize) return;
        const delta = [
            mouseResize.current_x - mouseResize.prev_x,
            mouseResize.current_y - mouseResize.prev_y,
        ];
        if (delta[0] == 0 && delta[1] == 0) return;

        const scale = draggableRef.current.get(page_idx).clientHeight / 2000;

        setTextStyle((prev_) => {
            setSnipRect((prev) => {
                return {
                    ...prev,
                    width: prev_.lineWrap + delta[0] / scale,
                };
            });

            return {
                ...prev_,
                lineWrap: Math.max(10, prev_.lineWrap + delta[0] / scale),
            };
        });
    }, [draggableRef, mouseResize, page_idx, setSnipRect, setTextStyle]);

    return (
        <OverlayText
            x={snipRect.x}
            y={snipRect.y}
            width={snipRect.width}
            height={snipRect.height}
            text={text}
            page_idx={page_idx}
            className={styles.dynamic}
            textProps={{
                ref: textRef,
                onInput: onInput,
                spellCheck: false,
            }}
            textStyle={{
                ...textStyle,
                lineHeight: textStyle.lineHeight * textStyle.fontSize,
            }}
            callback={callback}
        >
            {/** The text input via overlaytext */}
            {/** The resize handle on right side */}
            <div
                ref={resizeHandleRef}
                className={styles.resizeHandle}
                data-position="right"
            ></div>
            {/** The move handle */}
            <div ref={moveHandleRef} className={styles.moveHandle}>
                <i className="bi bi-arrows-move"></i>
            </div>
            {children}
        </OverlayText>
    );
}
