import { useContext, useState } from "react";

import { TextToolContext } from "../text.tool";

import { AllowedFonts, FONT_URLS } from "database/fonts";

import stylesDoodle from "../doodle/styles.module.scss";
import styles from "./styles.module.scss";

export default function SubMenu() {
    return (
        <>
            <FontSelector />
            <FontSizeSelector />
            <LineHeightSelector />
            <LineWrapSelector />
        </>
    );
}

/** Font selector
 * This component is used to select a font for the text tool.
 * Basically a dropdown menu with all the allowed Fonts.
 */
function FontSelector() {
    const { textStyles, setTextStyles } = useContext(TextToolContext);

    return (
        <>
            <label>Font</label>
            <div
                className={styles.fontSelector}
                style={{ paddingBottom: "0.25rem" }}
            >
                <select
                    value={textStyles.font}
                    onChange={(e) => {
                        setTextStyles((prev) => ({
                            ...prev,
                            font: e.target.value as AllowedFonts,
                        }));
                    }}
                >
                    {
                        // Generate the options from the allowed fonts
                        FONT_URLS.map((font) => (
                            <option key={font[1]} value={font[1]}>
                                {font[1]}
                            </option>
                        ))
                    }
                </select>
            </div>
        </>
    );
}

/** Font size selector
 * This component is used to select a font size for the text tool.
 */
function FontSizeSelector() {
    const { textStyles, setTextStyles } = useContext(TextToolContext);

    return (
        <>
            <label>Font size</label>
            <div
                className={stylesDoodle.sizeSelector}
                style={{ paddingBottom: "0.25rem" }}
            >
                <div className="number-container">
                    <button
                        className="input-group-text"
                        onClick={() => {
                            setTextStyles((prev) => ({
                                ...prev,
                                fontSize: 12,
                            }));
                        }}
                    >
                        <i className="bi bi-arrow-counterclockwise"></i>
                    </button>
                    <input
                        type="number"
                        min={1}
                        max={25}
                        step={1}
                        value={textStyles.fontSize.toFixed(0)}
                        onInput={(e) => {
                            const current_val = (e.target as HTMLInputElement)
                                .valueAsNumber;
                            setTextStyles((prev) => ({
                                ...prev,
                                fontSize: current_val,
                            }));
                        }}
                    />
                </div>
            </div>
        </>
    );
}

/** Line height selector */
function LineHeightSelector() {
    const { textStyles, setTextStyles } = useContext(TextToolContext);

    const [inputState, setInputState] = useState(textStyles.lineHeight);

    return (
        <>
            <label>Line Height</label>
            <div
                className={stylesDoodle.sizeSelector}
                style={{ paddingBottom: "0.25rem" }}
            >
                <div className="number-container">
                    <button
                        className="input-group-text"
                        onClick={() => {
                            setInputState(1.25);

                            setTextStyles((prev) => ({
                                ...prev,
                                lineHeight: 1.25,
                            }));
                        }}
                    >
                        <i className="bi bi-arrow-counterclockwise"></i>
                    </button>
                    <input
                        type="number"
                        min={0.1}
                        step={0.01}
                        value={inputState}
                        onInput={(e) => {
                            const current_val = (e.target as HTMLInputElement)
                                .valueAsNumber;

                            setInputState(current_val);
                            if (!isNaN(current_val)) {
                                setTextStyles((prev) => ({
                                    ...prev,
                                    lineHeight: current_val,
                                }));
                            }
                        }}
                    />
                </div>
            </div>
        </>
    );
}

/** Line wrap selector */
function LineWrapSelector() {
    const { textStyles, setTextStyles } = useContext(TextToolContext);

    return (
        <>
            <label>Line Wrap</label>
            <div className={stylesDoodle.sizeSelector}>
                <div className="number-container">
                    <button
                        className="input-group-text"
                        onClick={() => {
                            setTextStyles((prev) => ({
                                ...prev,
                                lineWrap: 400,
                            }));
                        }}
                    >
                        <i className="bi bi-arrow-counterclockwise"></i>
                    </button>
                    <input
                        type="number"
                        min={10}
                        step={1}
                        value={textStyles.lineWrap.toFixed(0)}
                        onInput={(e) => {
                            const current_val = (e.target as HTMLInputElement)
                                .valueAsNumber;
                            setTextStyles((prev) => ({
                                ...prev,
                                lineWrap: current_val,
                            }));
                        }}
                    />
                </div>
            </div>
        </>
    );
}
