/** This file provides some typing hint for
 * tools and their components.
 */

export type Tool =
    | "movePage"
    | "doodle"
    | "snipEditor"
    | "text"
    | "interaction";

export interface ToolComponents {
    Button: () => JSX.Element;
    SubMenu: () => JSX.Element;
    Info: () => JSX.Element;
}
type Overlay = {
    dom: ReactNode;
    page_idx: number;
};

/** -------------------------------
 * Different Tool Context typings
 *
 * This basically defines the context
 * that is provided to the tools components. I.e.
 * its shared state and functionality.
 * ------------------------------*/

export interface ToolContext {
    /** The active state defines whether a tool is currently active or not.
     */
    active: boolean;
    toggleActive: () => void;

    /** The active sub menu defines whether a sub menu is currently active or not. I.e. right click menu for the tool (small arrow). This might do nothing depending on the tool.
     */
    activeSubMenu: boolean;
    toggleSubMenu: () => void;
}

export interface DoodleToolContextI extends ToolContext {
    /** Current color defines the by the user currently active color
     * for the doodle tool. Should be in hex format, some abbreviations
     * also work but I wouldnt be surprised if this is browser dependent.
     */
    currentColor: string;
    setCurrentColor: (color: string) => void;

    /** The color list defines the colors that are available to the user.
     * A color can be added to the list by the user, but the list should
     * always contain at least one color.
     */
    colorList: string[];
    setColorList: (colorList: string[]) => void;

    /** The current size defines the size of the brush. It is a number
     * between 1 and 25. This is basically converted into points internally
     * but scaled but the current page size.
     */
    currentSize: number;
    setCurrentSize: (size: number) => void;

    /** The smoothing parameter defines how much the line is smoothed. Takes
     * values between 0 and 1. 0 means no smoothing, 1 means maximum smoothing.
     */
    currentSmoothing: number;
    setCurrentSmoothing: (smoothing: number) => void;
}

export interface MovePageToolContextI extends ToolContext {
    /** Fits the current page to the viewport size given its idx. The
     * idx is the index of the viewport in the document. I.e. if only
     * one page is active, idx should be 0.
     */
    fitPage: (idx) => void;

    /** Centers the current page in the viewport given its idx. The
     * idx is the index of the viewport in the document. I.e. if only
     *  one page is active, idx should be 0.
     */
    centerPage: (idx) => void;

    /** The zoom describes the current zoom level of the page. It is
     * an array as their can be multiple pages in the document. The
     * idx is the index of the viewport in the document.
     */
    zoom: number[];
    setZoom: (idx, zoom) => void;
}

export interface TextStyles {
    /** The font family of the text.
     * this is extracted via the sql_to_ts script on
     * startup from the database. I.e. if a new font is added
     * to the database, the server needs to be restarted for it
     * to be available.
     */
    fontFamily: AllowedFonts;

    /** The font size of the text. */
    fontSize: number;

    /** The line height of the text. */
    lineHeight: number;

    /** The line wrap of the text. I.e. how many points
     * are allowed in one line before the text is wrapped.
     */
    lineWrap: number;
}

export interface TextToolContextI extends ToolContext {
    /** The text styles define the current text styles
     * that are applied to the text if it is created.
     */
    textStyles: TextStyles;
    setTextStyles: Dispatch<SetStateAction<TextStyles>>;

    /** The currently active text snip. Usefull in some
     * cases, e.g. when the user wants to edit the text
     * snip before it is finally placed.
     */
    activeSnip: TextSnip;
    setActiveSnip: Dispatch<SetStateAction<TextSnip>>;

    /**  The text tool allows to type in a
     * overlaying text box. The text box is
     * resizable and movable. The text box
     * can than be placed similar to a
     * snip.
     */
    overlay?: Overlay;
}

/** Subtools for the snipeditor tool that allow
 * to cut, move and resize the snip.
 */
export type SnipEditorSubTool = "move" | "cut" | "resize";

export interface SnipEditorToolContextI extends ToolContext {
    /**The currently active subtool, there is
     * always one active subtool if the tool
     * is active.
     */
    activeSubTool: SnipEditorSubTool;
    setActiveSubTool: (tool: SnipEditorSubTool) => void;

    /** Similar to the subtool, each subtool might also
     * have an active submenu.
     */
    activeSubSubMenu: SnipEditorSubTool;
    setActiveSubSubMenu: (tool: SnipEditorSubTool) => void;

    /** The overlay which allows to move, resize
     * , mirror or cut the snip in an interactive
     *  way. This is represented by a div.
     */
    overlay: Overlay;
    setOverlay: (overlay: Overlay) => void;

    // There is a need for more documentation here :)
    snipRect: SnipRect;
    setSnipRect: Dispatch<SetStateAction<SnipRect>>;
    cutoutRect: {
        sx: number;
        sy: number;
        swidth: number;
        sheight: number;
    };
    setCutoutRect: Dispatch<
        SetStateAction<{
            sx: number;
            sy: number;
            swidth: number;
            sheight: number;
        }>
    >;
}

export interface InteractionToolContextI extends ToolContext {
    /**  The interaction tool allows copy text via a
     * overlaying text box.
     */
    overlay?: Overlay;
}
