/**
 * README:
 * -------
 *
 * The overlay "worker" is used in the main thread to render
 * the overlay canvas. This is done to avoid the overhead of
 * creating a webworker for the overlay.
 *
 * Thus it is not a real worker but a facade to the overlay
 * rendering functions.
 */

import { MutableRefObject, useEffect, useState } from "react";

import { type EleMap } from "../page";

import { BookSocket, getSocket } from "controller/booksocket";
import { PageData } from "controller/page";
import { BaseData, BaseView, SnipData } from "snips/general/base";
import { get_snip_from_data } from "snips/get_snip_from_data";

const debug = (...args: unknown[]) => {
    console.debug("[OverlayWorker]", ...args);
};

export class OverlayWorker {
    private canvas: HTMLCanvasElement;
    private ctx: CanvasRenderingContext2D;

    private stack = new Array<(ctx: CanvasRenderingContext2D) => void>();

    private pageData: PageData | null = null;
    private socket: BookSocket | null = null;

    constructor(canvas: HTMLCanvasElement) {
        this.canvas = canvas;
        this.ctx = canvas.getContext("2d", {
            desynchronized: true,
        });
        debug("Created");

        // Listen for visibility change
        document.addEventListener("visibilitychange", () => {
            if (document.visibilityState === "visible") {
                this.render();
            }
        });
    }

    public setupPage(_page: PageData) {
        // Check if page is already setup
        if (this.pageData && this.pageData.id === _page.id) {
            return;
        }
        debug("Setup page:");
        if (this.pageData && this.socket) {
            // Leave room
            this.socket.emit("page:leave", this.pageData.id);
        }

        this.pageData = _page;

        // Clear stack from previews
        for (const renderfn of this.preview2render.values()) {
            this.remove(renderfn);
        }
        this.preview2render.clear();
        this.ctx.clearRect(0, 0, 1400, 2000);
        this.render();

        // Set socket in async function
        (async () => {
            this.socket = await getSocket(this.pageData.book_id);
            //join room
            await new Promise<void>((resolve, reject) => {
                this.socket.emit(
                    "page:join",
                    this.pageData.id,
                    (success: boolean) => {
                        if (success) {
                            resolve();
                        } else {
                            reject();
                        }
                    }
                );
            });
            this.socket.on("snip:preview", this.receivePreviewData.bind(this));
        })();
    }

    public sendPreviewData(snipData: SnipData<BaseData, BaseView>) {
        if (!this.socket) return;
        this.socket.volatile.emit("snip:preview", snipData);
    }

    private preview2render = new Map<
        number,
        (ctx: CanvasRenderingContext2D) => void
    >();
    public receivePreviewData(snipData: SnipData<BaseData, BaseView>) {
        if (!this.pageData || snipData.page_id !== this.pageData.id) return;

        // Check if snip is already rendered and remove it
        if (this.preview2render.has(snipData.id)) {
            const renderfn = this.preview2render.get(snipData.id);
            this.remove(renderfn);
        }

        // Add to stack
        const snip = get_snip_from_data(snipData);
        this.preview2render.set(snipData.id, (ctx) => {
            snip.render(ctx);
        });
        this.add(this.preview2render.get(snipData.id));
    }

    public resize(width: number, height: number) {
        // Normally the canvas should always be 1400x2000
        this.canvas.width = width;
        this.canvas.height = height;

        const max_db = [1400, 2000];
        const scale = height / max_db[1];
        this.ctx.setTransform(scale, 0, 0, scale, 0, 0);
    }

    public render() {
        this.ctx.clearRect(0, 0, 1400, 2000);
        if (this.stack.length === 0) return;
        this.stack.forEach((fn) => fn(this.ctx));
    }

    public add(fn: (ctx: CanvasRenderingContext2D) => void) {
        this.stack.push(fn);
        this.render();
        debug("Added", this.stack.length);
    }
    public remove = (fn: (ctx: CanvasRenderingContext2D) => void) => {
        const index = this.stack.indexOf(fn);
        if (index !== -1) {
            this.stack.splice(index, 1);
        }
        //this.render();
        debug("Removed", this.stack.length);
    };

    public terminate() {
        this.ctx.clearRect(0, 0, 1400, 2000);
        debug("Terminated");
    }

    /** Transform a point from canvas coordinates to
     * transformed coordinates (1400x2000).
     */
    public canvasCoords_2_dbCoords(xy: [number, number]) {
        const transform = this.ctx.getTransform().invertSelf();
        const x = transform.a * xy[0] + transform.c * xy[1] + transform.e;
        const y = transform.b * xy[0] + transform.d * xy[1] + transform.f;
        return [x, y];
    }
    public dbCoords_2_canvasCoords(xy: [number, number]) {
        const scale = this.canvas.height / 2000;
        const x = scale * xy[0];
        const y = scale * xy[1];
        return [x, y];
    }

    public getTransform() {
        return this.ctx.getTransform();
    }
}

/** React hook to use overlay render worker
 * in a react component.
 */
export function useOverlayWorker(
    canvasRef: MutableRefObject<EleMap<HTMLCanvasElement>>,
    num_pages: number
) {
    const [overlayWorkers, setWorkers] = useState<OverlayWorker[]>([]);
    useEffect(() => {
        if (!canvasRef.current) return;
        console.log("Creating overlayworker(s)");

        setWorkers((prev_workers) => {
            // Scale up/down the number of workers
            if (prev_workers) {
                //Scale down
                for (let i = prev_workers.length; i > num_pages; i--) {
                    console.log("Terminating overlayworker");
                    prev_workers.pop().terminate();
                }
                // Scale up
                for (let i = 0; i < num_pages; i++) {
                    if (prev_workers[i] === undefined) {
                        prev_workers[i] = new OverlayWorker(
                            canvasRef.current.get(i)
                        );
                    }
                }
                return [...prev_workers];
            }
            // Create new workers
            else {
                const new_workers = [];
                for (let i = 0; i < num_pages; i++) {
                    new_workers[i] = new OverlayWorker(
                        canvasRef.current.get(i)
                    );
                }
                return new_workers;
            }
        });
    }, [canvasRef, num_pages]);

    return {
        overlayWorkers,
    };
}
