/**
 * README:
 * -------
 *
 * This is the entry point to create a render worker.
 *
 * Allows to offload the render tasks to a webworker using
 * offscreen canvas functions.
 *
 * Communication has to be done via the message interface to and from
 * the main thread. And also via typical socket communication.
 *
 */
import EventEmitter from "events";
import { MutableRefObject, useEffect, useState } from "react";

import { Message, Response } from "./pageWorker/messaging";

import { type PageData } from "controller/page";
import { BaseData, BaseView, SnipData } from "snips/general/base";

/** Facade to start a render worker and interact with it
 * via the message interface. This class should be used in the
 * main thread.
 *
 * This should be used to offload expensive render tasks to a
 * webworker. For instance if rendering a page with images.
 *
 */
export class PageWorker {
    private canvas: HTMLCanvasElement;
    private worker: Worker;

    /** Emit received messages as events */
    public events = new EventEmitter();

    constructor(canvas: HTMLCanvasElement) {
        // Create offscreen canvas
        this.canvas = canvas;
        this.createWorker();

        // Listen for visibility change
        document.addEventListener("visibilitychange", () => {
            if (document.visibilityState === "visible") {
                this.render();
            }
        });

        //Forward received messages as events
        this.worker.addEventListener("message", (e: MessageEvent<Response>) => {
            if (!e.data.event_id) return;
            this.events.emit(e.data.event_id.toString(), e.data);
        });
    }
    /** Render a page given it's
     * serialized data.
     */
    public setupPage(page: PageData) {
        const msg: Message<"setupPage"> = {
            type: "setupPage",
            data: { page },
            event_id: generate_event_id(),
        };
        this.worker.postMessage(msg);
    }

    /** Place a already inserted snip to the currently active page of the
     * render worker. For instance if a new snip is placed.
     */
    public placeOnPage(snip: SnipData<BaseData, BaseView>) {
        //Identifier for messaging random int
        const event_id = generate_event_id();

        return new Promise<void>((resolve) => {
            // Send Message and wait for response
            this.events.once(event_id.toString(), () => {
                resolve();
            });

            const msg: Message<"placeOnPage"> = {
                type: "placeOnPage",
                data: { snip },
                event_id,
            };
            this.worker.postMessage(msg);
        });
    }

    /** Insert a snip to the database and render it
     */
    public async insertSnip(snip: SnipData<BaseData, BaseView>) {
        //Identifier for messaging random int
        const event_id = generate_event_id();

        return new Promise<void>((resolve) => {
            // Send Message and wait for response
            this.events.once(event_id.toString(), () => {
                resolve();
            });

            const msg: Message<"insertSnip"> = {
                type: "insertSnip",
                data: { snip },
                event_id,
            };
            this.worker.postMessage(msg);
        });
    }

    /** Forward interaction Events
     *  to the render worker.
     */
    public interactionEvent<
        K extends keyof InteractionEventMap,
        I extends keyof InteractionEventResponseMap
    >(event: InteractionEventData<K>) {
        //Identifier for messaging random int
        const event_id = generate_event_id();

        return new Promise<InteractionEventResponseData<I>[]>((resolve) => {
            // Send Message and wait for response
            this.events.once(
                event_id.toString(),
                (e: Response<"interactionEvent", I>) => {
                    resolve(e.response);
                }
            );
            const msg: Message<"interactionEvent"> = {
                type: "interactionEvent",
                data: { event },
                event_id,
            };
            this.worker.postMessage(msg);
        });
    }

    /** Render the current page
     *
     */
    public async render() {
        const event_id = generate_event_id();
        return new Promise<void>((resolve) => {
            // Send Message and wait for response
            this.events.once(event_id.toString(), () => {
                resolve();
            });
            this.worker.postMessage({ type: "render", event_id });
        });
    }

    /** Force render a page */
    public renderPage() {
        const msg: Message<"render"> = {
            type: "render",
            data: null,
            event_id: generate_event_id(),
        };
        this.worker.postMessage(msg);
    }

    public resize(width: number, height: number) {
        const msg: Message<"resize"> = {
            type: "resize",
            data: { width, height },
            event_id: generate_event_id(),
        };
        this.worker.postMessage(msg);
    }

    private createWorker() {
        // Create worker
        this.worker = new Worker(
            new URL("./pageWorker/worker.ts", import.meta.url)
        );
        // Send init message
        const url_params = new URLSearchParams(window.location.search);
        const offscreen = this.canvas.transferControlToOffscreen();
        this.worker.postMessage(
            {
                type: "init",
                data: {
                    canvas: offscreen,
                    url_params: url_params.toString(),
                },
                event_id: generate_event_id(),
            },
            [offscreen]
        );
        return;
    }

    public terminate() {
        console.log("Terminating worker");
        this.worker.postMessage({ type: "shutdown" });
        this.worker.terminate();
    }
}

import { type EleMap } from "../page";
import {
    InteractionEventData,
    InteractionEventMap,
    InteractionEventResponseData,
    InteractionEventResponseMap,
} from "./pageWorker/interactionEvents";

/** React hook to use a render worker in a react component.
 */
export function usePageWorker(
    canvasRef: MutableRefObject<EleMap<HTMLCanvasElement>>,
    num_pages: number
) {
    const [workers, setWorkers] = useState<PageWorker[]>();

    useEffect(() => {
        if (!canvasRef.current) return;
        console.log("Creating worker(s)");

        setWorkers((prev_workers) => {
            // Scale up/down the number of workers
            if (prev_workers) {
                //Scale down
                for (let i = prev_workers.length; i > num_pages; i--) {
                    prev_workers.pop().terminate();
                }
                // Scale up
                for (let i = 0; i < num_pages; i++) {
                    if (prev_workers[i] === undefined) {
                        prev_workers[i] = new PageWorker(
                            canvasRef.current.get(i)
                        );
                    }
                }
                return [...prev_workers];
            }
            // Create new workers
            else {
                const new_workers = [];
                for (let i = 0; i < num_pages; i++) {
                    new_workers[i] = new PageWorker(canvasRef.current.get(i));
                }
                return new_workers;
            }
        });
    }, [canvasRef, num_pages]);

    return {
        workers,
        setWorkers,
    };
}

function generate_event_id() {
    return Math.floor(Math.random() * 100000);
}
