import {
    InteractionEventData,
    InteractionEventMap,
    InteractionEventResponseData,
    InteractionEventResponseMap,
} from "./interactionEvents";
import { Message, MessageMap, Response, ResponseMap } from "./messaging";

import { Page, PageData } from "controller/page";
import { FONT_URLS } from "database/fonts";
import { BaseData, BaseSnip, BaseView, SnipData } from "snips/general/base";
import { get_snip_from_data } from "snips/get_snip_from_data";

/** If you want a font to be available we have to add it to the database and
 * public folder*/

for (const [url, name] of FONT_URLS) {
    const font = new FontFace(name, "url(" + url.href + ")");
    font.load().then((loaded_face) => {
        globalThis.fonts.add(loaded_face);
    });
}

const debug = (...args: unknown[]) => {
    console.debug("[PageWorker]", ...args);
};

declare global {
    // eslint-disable-next-line no-var
    var worker: PageWorker;
}

class PageWorker {
    private canvas: OffscreenCanvas;
    private ctx: OffscreenCanvasRenderingContext2D;

    private page: Page | null = null;

    constructor(canvas: OffscreenCanvas) {
        this.canvas = canvas;
        this.ctx = canvas.getContext("2d", {
            desynchronized: true,
        }) as OffscreenCanvasRenderingContext2D;
    }

    public async setupPage(pageData: PageData) {
        // Check if page is already set
        if (this.page && this.page.id === pageData.id) {
            return;
        }
        // Leave old page
        if (this.page) {
            this.page.socket.emit("page:leave", this.page.id);
        }
        debug("Setup page:");
        this.page = new Page(pageData);
        await this.page.init();
        debug("Page initialized");
        await this.page.loadAllSnips();
        debug("Snips loaded");
        await this.page.allImagesPrepared();
        debug("Images prepared");

        //Resize and render
        this.resize(this.canvas.width, this.canvas.height);
        this.render();
        globalThis.fonts.ready.then(() => {
            this.render();
        });

        // Subscribe to changes
        const l = ["insert", "remove", "placed"];
        l.forEach((event) => {
            this.page.events.addListener(event, () => {
                console.log("Get Event:", event);
                this.render();
            });
        });
    }

    public render(): Promise<void> {
        if (!this.page) {
            debug("No page set");
            return;
        }
        const saved_transform = this.ctx.getTransform();
        const draw = new Promise<void>((resolve) => {
            const transform = this.ctx.getTransform().invertSelf();
            // Transform width and height to database coordinates
            // to clear the canvas
            const t_width =
                transform.a * this.canvas.width +
                transform.b * this.canvas.height +
                transform.e;
            const t_height =
                transform.c * this.canvas.width +
                transform.d * this.canvas.height +
                transform.f;

            this.ctx.clearRect(0, 0, t_width, t_height);

            this.page.render(this.ctx);
            debug("Rendered");
            resolve();
        });
        draw.then(() => {
            this.ctx.setTransform(saved_transform);
        });
        return draw;
    }

    public placeOnPage(snipData: SnipData<BaseData, BaseView>): Promise<void> {
        if (!this.page) {
            debug("No page set");
            return;
        }

        const place = new Promise<void>((resolve) => {
            debug("Place on page");
            const snip = get_snip_from_data(snipData);
            // only render new snip
            this.page.add_snip(snip);
            const saved_transform = this.ctx.getTransform();
            snip.render(this.ctx);
            this.ctx.setTransform(saved_transform);
            resolve();

            // Insert snip into database
            snip.socket = this.page.socket;
            snip.place_on_page(this.page.id);
        });
        return place;
    }

    public insertSnip(snipData: SnipData<BaseData, BaseView>): Promise<void> {
        if (!this.page) {
            debug("No page set");
            return;
        }

        const insert = new Promise<void>((resolve) => {
            debug("Insert snip", snipData);

            const snip = get_snip_from_data(snipData);
            snip.render(this.ctx);
            this.page.add_snip(snip);
            resolve();

            // Insert snip into database
            snip.socket = this.page.socket;
            snip.insert();
        });
        return insert;
    }

    private snip_under_cursor: BaseSnip = null;
    public interaction(
        event: InteractionEventData<keyof InteractionEventMap>
    ): InteractionEventResponseData<keyof InteractionEventResponseMap>[] {
        if (!this.page) {
            debug("No page set");
            return;
        }

        const responses: InteractionEventResponseData<
            keyof InteractionEventResponseMap
        >[] = [];
        let snip = null;
        let move = false;

        // Base move events
        switch (event.type) {
            case "pointerenter":
            case "pointermove":
                snip = this.page.get_snip_at(event.data.coords);
                move = true;
                break;
            case "pointerleave":
                snip = this.page.get_snip_at(event.data.coords);
                move = true;
                break;
            case "dblclick":
            case "pointerdown":
                snip = this.page.get_snip_at(event.data.coords);
                break;
            default:
                break;
        }

        // enter/move/leave events from page to snip
        if (move && snip && this.snip_under_cursor !== snip) {
            if (this.snip_under_cursor) {
                responses.push({
                    type: "on_pointerleave",
                    data: this.snip_under_cursor.to_data(),
                });
            }
            responses.push({
                type: "on_pointerenter",
                data: snip.to_data(),
            });
            this.snip_under_cursor = snip;
        } else if (move && !snip && this.snip_under_cursor) {
            responses.push({
                type: "on_pointerleave",
                data: this.snip_under_cursor.to_data(),
            });
            this.snip_under_cursor = null;
        }

        // dblclick event
        if (event.type === "dblclick" && snip) {
            responses.push({
                type: "on_dblclick",
                data: snip.to_data(),
            });
        }

        if (event.type === "pointerdown" && snip) {
            responses.push({
                type: "on_pointerdown",
                data: snip.to_data(),
            });
        }

        this.handle_worker_response(responses).catch((err) => {
            console.error("Error handling worker response", err);
        });

        return responses;
    }

    async handle_worker_response(
        responses: InteractionEventResponseData<
            keyof InteractionEventResponseMap
        >[]
    ): Promise<void> {
        return new Promise<void>((resolve) => {
            let render = false;
            for (const response of responses) {
                const snip = this.page.get_snip_by_id(response.data.id);
                if (
                    snip &&
                    typeof snip[response.type + "_worker"] === "function"
                ) {
                    snip[response.type + "_worker"]();
                    render = true;
                }
            }
            if (render) {
                this.render();
            }
            resolve();
        });
    }

    public resize(width: number, height: number) {
        this.canvas.width = width;
        this.canvas.height = height;

        //Render again
        const max_db = [1400, 2000];
        const scale = this.canvas.height / max_db[1];
        this.ctx.setTransform(scale, 0, 0, scale, 0, 0);
    }

    public shutdown() {
        debug("Shutting down");
        if (this.page) {
            this.page.socket.emit("page:leave", this.page.id);
            this.page.socket.close();
        }
    }
}

/** MESSAGEING */

function send_response<K extends keyof ResponseMap>(
    type: K,
    event_id: number,
    response: ResponseMap[K]
) {
    globalThis.postMessage({
        type,
        event_id,
        response,
    } as Response);
}

self.onmessage = async (event: MessageEvent<Message>) => {
    let response: ResponseMap[keyof ResponseMap] = null;
    let promise: Promise<ResponseMap[keyof ResponseMap]> = null;

    const event_id = event.data.event_id;
    const event_type = event.data.type;
    const event_data = event.data.data;
    switch (event_type) {
        case "interactionEvent":
            response = globalThis.worker.interaction(
                (event_data as MessageMap["interactionEvent"]).event
            );
            break;
        case "init":
            globalThis.worker = new PageWorker(
                (event_data as MessageMap["init"]).canvas
            );
            globalThis.url_params = new URLSearchParams(
                (event_data as MessageMap["init"]).url_params
            );
            break;
        case "shutdown":
            globalThis.worker.shutdown();
            break;
        case "setupPage":
            await globalThis.worker.setupPage(
                (event_data as MessageMap["setupPage"]).page
            );
            break;
        case "render":
            promise = globalThis.worker.render();
            break;
        case "resize":
            globalThis.worker.resize(
                (event_data as MessageMap["resize"]).width,
                (event_data as MessageMap["resize"]).height
            );
            break;
        case "placeOnPage":
            promise = globalThis.worker.placeOnPage(
                (event_data as MessageMap["placeOnPage"]).snip
            );
            break;
        case "insertSnip":
            promise = globalThis.worker.insertSnip(
                (event_data as MessageMap["insertSnip"]).snip
            );
            break;
        default:
            debug("Message could not be parsed", event.data);
            break;
    }

    // Send response
    if (promise) {
        promise.then((response) => {
            send_response(event_type, event_id, response);
        });
    } else {
        send_response(event_type, event_id, response);
    }
};
