import EventEmitter from "events";

import { type BookSocket, getSocket } from "./booksocket";
import { PageData } from "./page";

import { background_types, booksData, pagesData } from "database/sql.typings";
import { BaseData, BaseView, SnipData } from "snips/general/base";
export type BookData = booksData;

const log = (...args: unknown[]) => console.log("[Book]", ...args);

export class Book implements Required<BookData> {
    // Data
    public id: number;
    public title: string;
    public comment: string;
    public created: Date;
    public finished: Date | null;
    public last_updated: Date;
    public owner_user_id: number;
    public owner_group_id: number;
    public cover_page_id: number | null;
    public default_background_type_id: number;

    // Client side only vars
    private socket?: BookSocket;
    public pages: PageData[] = [];

    /** Event emitter for all changes to the book.
     *
     * Events:
     * - insertPage: A new page was added to the book
     * - deletePage: A page was deleted
     * - updatePreview: The preview of a page was updated
     * - newUnplacedSnip: A new snip was added to the book but not placed on a page yet
     */
    public events = new EventEmitter();

    constructor({
        id,
        title,
        comment,
        created = new Date(),
        finished = null,
        last_updated = new Date(),
        owner_user_id = null,
        owner_group_id = null,
        cover_page_id = null,
        default_background_type_id = background_types["blank"],
    }: BookData) {
        log("Book constructor");
        this.id = id;
        this.title = title;
        this.comment = comment;
        // Parse dates just to be sure
        this.created = new Date(created);
        this.finished = finished ? new Date(finished) : null;
        this.last_updated = new Date(last_updated);

        // Set owner
        this.owner_user_id = owner_user_id;
        this.owner_group_id = owner_group_id;

        // Set cover page
        this.cover_page_id = cover_page_id;
        this.default_background_type_id = default_background_type_id;
    }

    public async init(): Promise<void> {
        log("Init book");
        this.socket = await getSocket(this.id);
        await this.subscribeToChanges();
    }

    static from_data(data: BookData): Book {
        return new Book(data);
    }

    public to_data(): Required<BookData> {
        return {
            id: this.id,
            title: this.title,
            comment: this.comment,
            created: this.created,
            finished: this.finished,
            last_updated: this.last_updated,
            owner_user_id: this.owner_user_id,
            owner_group_id: this.owner_group_id,
            cover_page_id: this.cover_page_id,
            default_background_type_id: this.default_background_type_id,
        };
    }

    /** Load all pages of the book.
     */
    public async loadPages(): Promise<void> {
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }

        // Get pageData and previews for all pages
        const pageData = await new Promise<pagesData[]>((resolve, reject) => {
            this.socket?.emit("book:getPages", (pages: pagesData[]) => {
                if (!pages) {
                    reject("Could not get pages");
                }
                resolve(pages);
            });
        });
        // Set pages
        this.pages = pageData;
    }

    /** Get all unplaced snips of the book.
     * Unplaced snips are snips that are not placed on a page yet.
     * They are shown in the snip list.
     * @returns An array of snip data
     */
    public async getUnplacedSnips(): Promise<SnipData<BaseData, BaseView>[]> {
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }

        // Get all unplaced snips
        const unplacedSnips = await new Promise<SnipData<BaseData, BaseView>[]>(
            (resolve, reject) => {
                this.socket?.emit(
                    "book:getUnplacedSnips",
                    (snips: SnipData<BaseData, BaseView>[]) => {
                        if (!snips) {
                            reject("Could not get unplaced snips");
                        }
                        resolve(snips);
                    }
                );
            }
        );
        return unplacedSnips;
    }

    /** Create a page using
     * the socket connection.
     *
     * @returns The page created
     */
    public async createPage(): Promise<PageData> {
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }

        // Create a new page
        let newPage: PageData = {
            id: null,
            page_number: null,
            book_id: this.id,
            created: new Date(),
            last_updated: new Date(),
            background_type_id: this.default_background_type_id,
        };

        // Send the new page to the server
        newPage = await new Promise((resolve, reject) => {
            this.socket.emit(
                "book:insertPage",
                newPage,
                (newPage: pagesData) => {
                    if (!newPage) {
                        reject("Could not create page");
                    }
                    resolve(newPage);
                }
            );
        });
        this.pages.push(newPage);
        this.events.emit("insertPage", newPage as PageData);
        return newPage as PageData;
    }

    /** Get previews of all pages
     * of the book.
     */
    public async getPreviews(): Promise<string[]> {
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }

        // Get all previews images
        const previews = await new Promise<string[]>((resolve, reject) => {
            this.socket?.emit("book:getPreviews", (previews: string[]) => {
                if (!previews) {
                    reject("Could not get previews");
                }
                resolve(previews);
            });
        });
        return previews;
    }

    /** Get a single preview image
     * of a page.
     */
    public async getPreview(pageId: number): Promise<string> {
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }
        return await new Promise((resolve, reject) => {
            this.socket?.emit("page:getPreview", pageId, (preview: string) => {
                if (!preview) {
                    reject("Could not get preview");
                }
                resolve(preview);
            });
        });
    }

    private async subscribeToChanges(): Promise<void> {
        log("Subscribe to changes");
        // New page added
        this.socket?.on("book:insertPage", (data: PageData) => {
            this.pages.push(data);
            this.events.emit("insertPage", data);
        });

        // Delete a page
        this.socket?.on("book:deletePage", (page_id: number) => {
            this.pages = this.pages.filter((page) => page.id !== page_id);
            this.events.emit("deletePage", page_id);
        });

        // Update preview of a page
        this.socket?.on(
            "book:updatePreview",
            (page_id: number, img: string) => {
                this.events.emit("updatePreview", page_id, img);
            }
        );

        // Update on insert of not yet placed snip
        this.socket?.on("book:insertedIntoQueue", (data) => {
            console.log("book:insertedIntoQueue", data);
            this.events.emit("book:insertedIntoQueue", data);
        });

        this.socket.on("snip:placed", (page_id, snip_id) => {
            this.events.emit("snip:placed", page_id, snip_id);
        });
    }
}
