import { io, Socket } from "socket.io-client";

// Only single instance of each book is allowed
const instances: Map<string, BookSocket> = new Map();

export class BookSocket {
    public book_id: string;

    public socket: Socket;

    /* Wrapper functions for socket.io */
    public on(event: string, callback: (...args: unknown[]) => void): void {
        this.socket.on(event, callback);
    }
    public emit(event: string, ...args: unknown[]): void {
        this.socket.emit(event, ...args);
    }
    public once(event: string, callback: (...args: unknown[]) => void): void {
        this.socket.once(event, callback);
    }
    public off(event: string, callback: (...args: unknown[]) => void): void {
        this.socket.off(event, callback);
    }
    public get volatile(): Socket {
        return this.socket.volatile;
    }
    public get connected(): boolean {
        return this.socket.connected;
    }
    public get disconnected(): boolean {
        return this.socket.disconnected;
    }

    /** Each book can only have one socket connection
     * This function returns the existing socket if it exists
     * or creates a new one if it does not exist.
     */
    public static async get(book_id: number | string): Promise<BookSocket> {
        if (instances.has(String(book_id))) {
            return instances.get(String(book_id));
        }
        const inst = new BookSocket(book_id);
        await inst.connect();
        instances.set(String(book_id), inst);
        return inst;
    }

    // Can not be called from outside
    private constructor(book_id: number | string) {
        this.book_id = String(book_id);
    }

    /** Connect to the socket
     * this is normally called by the constructor
     */
    public async connect(): Promise<void> {
        const token = await fetchToken(this.book_id);
        const socket = io("/book-" + this.book_id, {
            auth: {
                token: token,
            },
        });
        socket.on("error", (err) => {
            console.log(err);
        });
        socket.on("connect_error", (err) => {
            console.log(err);
        });
        socket.on("connect", () => {
            console.log("Connected to server via websocket");
        });
        this.socket = socket;
    }
    /** Close the socket */
    public close(): void {
        this.socket.close();
    }
}

export async function getSocket(
    book_id: number | string
): Promise<BookSocket | null> {
    if (isBrowser() || isWorker()) {
        return await BookSocket.get(book_id);
    }
    return null;
}

async function fetchToken(book_id: string): Promise<string> {

    const header_adds = {}
    let token_url = null;
    // In workers the url is parsed via globalThis
    if (isWorker()) {
        token_url = globalThis.url_params.get("token");
    } else {
        const this_url = new URL(window.location.href);
        token_url = this_url.searchParams.get("token");
    }

    //Check if url has token
    if (token_url) {
        header_adds["Authorization"] = "Bearer " + token_url;
    }

    const { token } = await fetch(`/api/books/${book_id}/socket_token`, {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
            ...header_adds
        },
        body: JSON.stringify({ book_id }),
    }).then((res) => {
        if (res.status === 200) {
            return res.json();
        } else {
            console.error("Could not establish secure connection");
            throw new Error("Could not fetch token");
        }
    });
    return token;
}

export function isBrowser(): boolean {
    return typeof window !== "undefined";
}

export function isWorker(): boolean {
    return typeof importScripts === "function";
}


