import EventEmitter from "events";

import { render_background_by_id } from "./backgrounds/render_background_by_id";
import { type BookSocket, getSocket } from "./booksocket";

import { pagesData } from "database/sql.typings";
import { ArraySnip } from "snips/general/array";
import type {
    BaseData,
    BaseSnip,
    BaseView,
    SnipData,
} from "snips/general/base";
import { ImageSnip } from "snips/general/image";
import { get_snip_from_data } from "snips/get_snip_from_data";
export type PageData = pagesData;

/** A static page is a page to render a predefined page
 * no dynamic content is used. See Page for dynamic content
 * using a socket connection.
 */
export class LocalPage implements Required<PageData> {
    public id: number;
    public book_id: number;
    public page_number: number;
    public readonly created: Date;
    public readonly last_updated: Date | null;
    public background_type_id: number | null;

    protected snips: BaseSnip[];

    /** Create a new page object
     *
     * This allows your to access the render function of the page object.
     *
     * @param id The id of the page
     * @param book_id The id of the book
     * @param page_number The page number
     * @param created The date the page was created
     * @param last_updated The date the page was last updated
     * @param background_type_id The id of the background type
     * @param snips The snips on the page can be none if not loaded
     */
    constructor({
        id,
        book_id,
        page_number,
        created = new Date(),
        last_updated = new Date(),
        background_type_id = null,
        snips = [],
    }: PageData & { snips?: BaseSnip[] }) {
        this.id = id;
        this.book_id = book_id;
        this.page_number = page_number;
        this.last_updated = last_updated;
        this.created = created;
        this.background_type_id = background_type_id;
        this.snips = snips;
    }

    static from_data(data: PageData): LocalPage {
        return new LocalPage({
            ...data,
        });
    }

    public to_data(): Required<PageData> {
        return {
            id: this.id,
            book_id: this.book_id,
            page_number: this.page_number,
            created: this.created,
            last_updated: this.last_updated,
            background_type_id: this.background_type_id,
        };
    }

    public set_snips(snips: BaseSnip[]): void {
        this.snips = snips;
    }

    public add_snip(snip: BaseSnip): void {
        /** If array snip remove all
         * nested snips and add them to the arraysnip
         */
        if (snip instanceof ArraySnip) {
            snip.snips.forEach((s) => {
                this.remove_snip(s.id);
            });
        }

        /** Get all array snips and check if they contain the snip
         */
        const array_snips = this.snips.filter(
            (s) => s instanceof ArraySnip
        ) as ArraySnip[];
        for (const array_snip of array_snips) {
            for (let s of array_snip.snips) {
                if (s.id === snip.id) {
                    s = snip;
                    return;
                }
            }
        }

        this.snips.push(snip);
    }

    public get_snip_at(coords: [number, number]): BaseSnip | null {
        // Check if snips are loaded
        if (this.snips === undefined) {
            return null;
        }
        // Check if snip is at coords
        let snips = [];
        if (this.snips.length > 25) {
            // If more than 25 snips filter out doodle snips
            snips = this.snips.filter((snip) => snip.type != "doodle");
        } else {
            snips = this.snips;
        }
        // Reverse order as the last snip is on top
        snips = snips.toReversed();

        const snip = snips.find((s) => {
            const poly = s.polygon;
            return ptinpoly(coords[0], coords[1], poly);
        });
        if (snip) {
            return snip;
        }
        return null;
    }

    public get_snip_by_id(snip_id: number): BaseSnip | null {
        const snip = this.snips.find((s) => s.id === snip_id);
        if (snip) {
            return snip;
        }
        return null;
    }

    public update_snip(snip: BaseSnip): void {
        const index = this.snips.findIndex((s) => s.id === snip.id);
        if (index !== -1) {
            this.snips[index] = snip;
        } else {
            this.add_snip(snip);
        }
    }

    public remove_snip(snip_id): boolean {
        const index = this.snips.findIndex((s) => s.id === snip_id);
        if (index !== -1) {
            this.snips.splice(index, 1);
            return true;
        }
        return false;
    }

    public render(
        ctx: CanvasRenderingContext2D | OffscreenCanvasRenderingContext2D
    ): void {
        // Render background
        render_background_by_id(this.background_type_id, ctx);
        // Render snips
        if (this.snips !== undefined) {
            this.snips.forEach((snip) => {
                snip.render(ctx);
            });
        }
    }
}

/** A page is able to render a page and retrieve data from the server
 * using a socket connection.
 */
export class Page extends LocalPage {
    public socket?: BookSocket;

    /** Event emitter for all changes to the page.
     *
     * The idea is to e.g. subscribe to the change event and
     * rerender the page.
     */
    public events = new EventEmitter();

    constructor({
        id,
        book_id,
        page_number,
        created = new Date(),
        last_updated = new Date(),
        background_type_id = null,
        snips = [],
    }: PageData & { snips?: BaseSnip[] }) {
        super({
            id,
            book_id,
            page_number,
            created,
            last_updated,
            background_type_id,
            snips,
        });
    }

    async init(): Promise<void> {
        this.socket = await getSocket(this.book_id);
        await this.joinRoom();
        await this.subscribeToChanges();
    }

    /** Get the preview image of the page */
    public async getPreviewImageBuffer(): Promise<Buffer> {
        return await new Promise<Buffer>((resolve) => {
            this.socket.emit(
                "page:getPreviewImage",
                this.id,
                (buffer: Buffer) => {
                    resolve(buffer);
                }
            );
        });
    }

    /** Join room for the page */
    private async joinRoom(): Promise<void> {
        if (!this.socket) {
            return;
        }
        return await new Promise<void>((resolve, reject) => {
            this.socket.emit("page:join", this.id, (success: boolean) => {
                if (success) {
                    resolve();
                } else {
                    reject();
                }
            });
        });
    }

    /** Use socket to load all snips */
    public async loadAllSnips(): Promise<void> {
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }
        await new Promise<void>((resolve) => {
            this.socket.emit(
                "page:getSnips",
                this.id,
                (snips: SnipData<BaseData, BaseView>[]) => {
                    this.set_snips(
                        snips.map((snip) => get_snip_from_data(snip))
                    );
                    resolve();
                }
            );
        });
    }

    /** Images need
     * to be loaded before they can be rendered.
     */
    public async allImagesPrepared(): Promise<boolean> {
        return await new Promise<boolean>((resolve) => {
            const image_snips = this.snips.filter(
                (snip) => snip.type === "image"
            ) as ImageSnip[];
            if (image_snips.length === 0) {
                resolve(true);
            }
            let loaded = 0;
            //Check if ready is set to true
            image_snips.forEach((snip) => {
                if (snip.ready) {
                    loaded++;
                } else {
                    snip.events.on("ready", () => {
                        loaded++;
                        if (loaded === image_snips.length) {
                            resolve(true);
                        }
                    });
                }
            });
        });
    }

    /** Subscribe to changes */
    private subscribeToChanges(): void {
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }
        this.socket.on(
            "snip:insert",
            (newSnip: SnipData<BaseData, BaseView>) => {
                // Check if snip is on this page
                if (newSnip.page_id === this.id) {
                    this.add_snip(get_snip_from_data(newSnip));
                    this.events.emit("insert", newSnip);
                }
            }
        );

        this.socket.on(
            "snip:update",
            (updatedSnip: SnipData<BaseData, BaseView>) => {
                // Check if snip is on this page
                if (updatedSnip.page_id === this.id) {
                    this.update_snip(get_snip_from_data(updatedSnip));
                    this.events.emit("update", updatedSnip);
                }
            }
        );

        this.socket.on("snip:remove", (snipId: number) => {
            // Check if snip is on this page
            const success = this.remove_snip(snipId);
            if (success) {
                this.events.emit("remove", snipId);
            }
        });

        this.socket.on("snip:placed", (pageId: number, snipId: number) => {
            // Check if snip is on this page
            if (pageId === this.id) {
                // Check if array snip and if so, update the snip
                const snip = this.snips.find((s) => s.id === snipId);
                if (snip && (snip as ArraySnip).snips) {
                    // find all snips in array and remove them
                    const array_snip = snip as ArraySnip;
                    const snips_to_remove = array_snip.snips.map((s) => s.id);
                    // find in snips
                    this.snips = this.snips.filter(
                        (s) => !snips_to_remove.includes(s.id)
                    );
                }
                this.events.emit("placed", snipId);
            }
        });
    }
    static from_data(data: PageData): Page {
        return new Page({
            ...data,
        });
    }
}

// poly is an array of [ x, y ] elements
function ptinpoly(x: number, y: number, poly: [number, number][]) {
    let c = false;
    for (let l = poly.length, i = 0, j = l - 1; i < l; j = i++) {
        const xj = poly[j][0],
            yj = poly[j][1],
            xi = poly[i][0],
            yi = poly[i][1];
        const where = (yi - yj) * (x - xi) - (xi - xj) * (y - yi);
        if (yj < yi) {
            if (y >= yj && y < yi) {
                if (where == 0) return true; // point on the line
                if (where > 0) {
                    if (y == yj) {
                        // ray intersects vertex
                        if (y > poly[j == 0 ? l - 1 : j - 1][1]) {
                            c = !c;
                        }
                    } else {
                        c = !c;
                    }
                }
            }
        } else if (yi < yj) {
            if (y > yi && y <= yj) {
                if (where == 0) return true; // point on the line
                if (where < 0) {
                    if (y == yj) {
                        // ray intersects vertex
                        if (y < poly[j == 0 ? l - 1 : j - 1][1]) {
                            c = !c;
                        }
                    } else {
                        c = !c;
                    }
                }
            }
        } else if (y == yi && ((x >= xj && x <= xi) || (x >= xi && x <= xj))) {
            return true; // point on horizontal edge
        }
    }
    return c;
}
