/* eslint-disable no-unused-vars */
/* Implements Base service interface which can be adapted for different database implementations */

import {
    aclData,
    books_resolvedData,
    booksData,
    groupsData,
    pagesData,
    user_config_resolvedData,
    usersData,
} from "database/sql.typings";
import { BaseData, BaseView, SnipData } from "snips/general/base";

/* Pass types from the sql to the controller */
export interface PermissionACLData extends aclData {
    entity_type: Entity;
    resource_type: Resource;
}
export type UserData = Omit<usersData, "hashedPassword">;
export type GroupData = groupsData;
export type PageData = pagesData;
export type BookData = booksData;
export type BookDataResolved = books_resolvedData;
export type UserConfigData = user_config_resolvedData;


import { BookOwner } from "pages/api/books/[id]/edit/get_owner";

interface BaseService<T> {
    /** Get a single row from the database
     *
     * @param id
     */
    getById(id: number): Promise<T>;
    // Get multiple rows by id
    getByIds(ids: number[]): Promise<T[]>;
    insert(object: T): Promise<T>;
    update(object: T, id?: number): Promise<T>;
    search(query: string): Promise<T[]>;
    delete(id: number): Promise<boolean>;
}

// Extend functionality for specific database tables
export interface UsersServiceInterface extends BaseService<UserData> {
    login(email: string, password: string): Promise<boolean>;
    verify(email: string): Promise<boolean>;
    getByEmail(email: string): Promise<UserData>;
    signup(email: string, password: string): Promise<UserData>;
    isAdmin(user_id: number): Promise<boolean>;
    search(search: string): Promise<UserData[]>;
    getPasswordHash(email: string): Promise<string>;
    updatePassword(email: string, password: string): Promise<boolean>;
}

export interface GroupsServiceInterface extends BaseService<GroupData> {
    getAllByUserId(user_id: number): Promise<GroupData[]>;
    getAllIdsByUserId(user_id: number): Promise<number[]>;
}

export interface PageServiceInterface extends BaseService<PageData> {
    getByBookId(book_id: number): Promise<PageData[]>;
    getFirstByBookId(book_id: number): Promise<PageData>;
}

export interface BookServiceInterface extends BaseService<BookData> {
    getAll(): Promise<BookData[]>;
    getByIdsResolved(ids: number[]): Promise<BookDataResolved[]>;
    setOwner(
        entity_id: number,
        entity_type: Entity | "user" | "group",
        book_id: number
    ): Promise<boolean>;
    setOwnerByName(entity_name: string, book_id: number): Promise<BookOwner>;
}

export interface SnipServiceInterface
    extends BaseService<SnipData<BaseData, BaseView>> {
    getByPageId(page_id: number): Promise<SnipData<BaseData, BaseView>[]>;
    getNotPlacedByBookId(
        book_id: number
    ): Promise<SnipData<BaseData, BaseView>[]>;
}


export interface ConfigServiceInterface {
    getByUserId(user_id: number): Promise<UserConfigData>;
    updateByUserId(user_id: number, config: Partial<UserConfigData>): Promise<boolean>;
}


// As the permission table has two primary keys we cant use the typical
// BaseService
export const RESOURCE_BOOK = 1;
export const RESOURCE_INSTRUMENT = 2;
export type Resource = typeof RESOURCE_BOOK | typeof RESOURCE_INSTRUMENT;
export const ENTITY_USER = 1;
export const ENTITY_GROUP = 2;
export const ENTITY_TOKEN = -10;
export type Entity =
    | typeof ENTITY_USER
    | typeof ENTITY_GROUP
    | typeof ENTITY_TOKEN;

export interface PermissionServiceInterface {
    getACLById(acl_id: number): Promise<PermissionACLData>;
    getACLByEntityAndResource(
        entity_id: number,
        resource_id: number,
        entity_type: Entity,
        resource_type: Resource
    ): Promise<PermissionACLData>;
    getACLByEntityAndResourceReduced(
        entity_id: number,
        resource_id: number,
        entity_type: Entity,
        resource_type: Resource
    ): Promise<PermissionACLData>;
    getACLByUserBook(
        user_id: number,
        book_id: number
    ): Promise<PermissionACLData[]>;
    getACLByUserBooks(
        user_id: number,
        book_ids: number[]
    ): Promise<PermissionACLData[]>;
    getACLByUserBookReduced(
        user_id: number,
        book_id: number
    ): Promise<PermissionACLData>;
    getACLByEntity(
        entity_id: number,
        entity_type?: Entity
    ): Promise<PermissionACLData[]>;
    setACLById(
        id: number,
        pRead: boolean,
        pWrite: boolean,
        pDelete: boolean,
        pACL: boolean
    ): Promise<boolean>;
    setACLByEntityAndResource(
        entity_id: number,
        entity_type: Entity,
        resource_id: number,
        resource_type: Resource,
        pRead: boolean,
        pWrite: boolean,
        pDelete: boolean,
        pACL: boolean
    ): Promise<boolean>;
}

// Main interface for database service
export interface ServiceInterface {
    user: UsersServiceInterface;
    group: GroupsServiceInterface;
    page: PageServiceInterface;
    book: BookServiceInterface;
    snip: SnipServiceInterface;
    permission: PermissionServiceInterface;
    config: ConfigServiceInterface;
}
