import { BookOwner } from "pages/api/books/[id]/edit/get_owner";

import { BookData, BookDataResolved } from "../service.typings";
import { BookServiceInterface, Entity } from "../service.typings";

import { pool_data as pool, pool_permissions } from "database/sql.connection";

const book_default: BookData = {
    id: -1,
    title: "default_book",
    comment: "",
    created: new Date(),
    finished: null,
    last_updated: new Date(),
    owner_user_id: null,
    owner_group_id: null,
    cover_page_id: null,
};

async function getAll(): Promise<BookData[]> {
    const sql = "SELECT * FROM books";
    const books: BookData[] = await pool
        .query(sql)
        .then((res) => {
            return res as BookData[];
        })
        .catch((err) => {
            console.error(err);
            return [];
        });
    return books;
}

/** Get a single lab-book give its MARIADB database id
 *
 * @param book_id
 */
async function getById(book_id: number): Promise<BookData> {
    if (isNaN(book_id)) {
        return book_default;
    }

    const sql = "SELECT * FROM books WHERE id = ?";
    const bookData: BookData = await pool
        .query(sql, [book_id])
        .then((res) => {
            if (res.length != 1) {
                throw new Error("Book not found");
            }
            // As the database structure matches our BookData interface
            // we can just return the first row
            return res[0];
        })
        .catch((err) => {
            console.error(err);
            return book_default;
        });
    return bookData;
}


/** Get multiple books and resolve their owner
 * @param book_ids
 */
async function getByIdsResolved(
    book_ids: number[]
): Promise<BookDataResolved[]> {
    if (book_ids.length == 0) {
        return [];
    }

    const sql = "SELECT * FROM books_resolved WHERE id IN (?)";

    const books: BookDataResolved[] = await pool
        .query(sql, [book_ids])
        .then((res) => {
            return res as BookDataResolved[];
        })
        .catch((err) => {
            console.error(err);
            return [];
        });
    return books;
}

/** Get multiple lab-books given their MARIADB database ids
 *
 * @param book_ids
 */
async function getByIds(book_ids: number[]): Promise<BookData[]> {
    if (book_ids.length == 0) {
        return [];
    }

    const sql = "SELECT * FROM books WHERE id IN (?)";

    const books: BookData[] = await pool
        .query(sql, [book_ids])
        .then((res) => {
            return res as BookData[];
        })
        .catch((err) => {
            console.error(err);
            return [];
        });
    return books;
}

/** Insert a new lab-book into the database, returns updated
 * book data, i.e. with new id and updated dates.
 *
 * @param bookData
 */
async function insert(bookData: BookData): Promise<BookData> {
    const sql = `
        INSERT INTO books 
            (title, comment, owner_user_id, owner_group_id, cover_page_id ) 
        VALUES 
            (?, ?, ?, ?, ?);
    `;
    const bookData_updated: BookData = await pool
        .query(sql, [
            bookData.title,
            bookData.comment ? bookData.comment : null,
            bookData.owner_user_id ? bookData.owner_user_id : null,
            bookData.owner_group_id ? bookData.owner_group_id : null,
            bookData.cover_page_id ? bookData.cover_page_id : null,
        ])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Insert failed status: " + res.warningStatus);
            }
            return res.insertId;
        })
        .then((id) => {
            return this.getById(id);
        })
        .catch((err) => {
            console.error(err);
            return book_default;
        });
    return bookData_updated;
}

/** Update a lab-book in the database, returns true if update was a success.
 *
 * @param bookData
 */
async function update(bookData: BookData): Promise<BookData> {
    const sql = `
        UPDATE books SET 
            title=?, comment=?, owner_user_id=?, owner_group_id=?, cover_page_id=?
            WHERE id = ?;
        `;

    const book = await pool
        .query(sql, [
            bookData.title,
            bookData.comment,
            bookData.owner_user_id,
            bookData.owner_group_id,
            bookData.cover_page_id,
            bookData.id,
        ])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Update failed status: " + res.warningStatus);
            }
            return this.getById(bookData.id);
        })
        .catch((err) => {
            console.error(err);
            return book_default;
        });
    return book;
}

async function search(search: string): Promise<BookData[]> {
    const sql = `
    SELECT * FROM books 
        WHERE title LIKE ? OR
        id LIKE ? OR
        comment LIKE ? OR
        created LIKE ? OR
        finished LIKE ? OR
        last_updated LIKE ?
    `;

    const books: BookData[] = await pool
        .query(sql, [
            `%${search}%`,
            `%${search}%`,
            `%${search}%`,
            `%${search}%`,
            `%${search}%`,
            `%${search}%`,
        ])
        .then((rows) => {
            return rows as BookData[];
        })
        .catch((err) => {
            console.error(err);
            return [];
        });
    return books;
}

/** Delete a lab-book from the database, returns true if deletion was a success.
 *
 * @param id
 * @returns
 */
async function _delete(id: number): Promise<boolean> {
    const sql = "DELETE FROM `books` WHERE id = ?";

    const success = await pool
        .query(sql, [id])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Delete failed status: " + res.warningStatus);
            }
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return success;
}

/** Set owner of book
 *
 */
async function setOwner(
    entity_id: number,
    entity_type: Entity | "user" | "group",
    book_id: number
): Promise<boolean> {
    let sql;
    if (entity_type == "user" || entity_type == 1) {
        sql =
            "UPDATE `books` SET `owner_user_id`=?, `owner_group_id`=NULL WHERE id = ?";
    } else if (entity_type == "group" || entity_type == 2) {
        sql =
            "UPDATE `books` SET `owner_group_id`=?, `owner_user_id`=NULL WHERE id = ?";
    } else {
        return false;
    }
    // The perms table is updated by a trigger
    const success = await pool
        .query(sql, [entity_id, book_id])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error(
                    "Set owner failed status: " + res.warningStatus
                );
            }
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return success;
}

async function setOwnerByName(
    entity_name: string,
    book_id: number
): Promise<BookOwner> {
    const sql_get_entity = `
        SELECT 
            id,
            "user" as entity_type,
            email as entity_name
        FROM users WHERE email = ?
        UNION
        SELECT
            id,
            "group" as entity_type,
            name as entity_name
        FROM groups WHERE name = ?
    `;
    const entity: BookOwner = await pool_permissions
        .query(sql_get_entity, [entity_name, entity_name])
        .then((res) => {
            return res[0] as BookOwner;
        })
        .catch((err) => {
            console.error(err);
            return null;
        });

    if (entity == null) {
        return null;
    }

    // Transfer to new owner
    const success = await setOwner(entity.id, entity.entity_type, book_id);
    if (!success) {
        return null;
    }
    return entity;
}

const BookService: BookServiceInterface = {
    getAll,
    getById,
    getByIds,
    getByIdsResolved,
    setOwner,
    setOwnerByName,
    search,
    insert,
    update,
    delete: _delete,
};
export default BookService;
