import { GroupsServiceInterface } from "../service.typings";

import { pool_permissions as pool } from "database/sql.connection";
import { groupsData as GroupData } from "database/sql.typings";

const group_default: GroupData = {
    id: -1,
    name: "",
    description: "",
    created: new Date(),
    last_updated: new Date(),
};

/** Get a single group object given its MARIADB id
 *
 * @param snip_id
 */
async function getById(id: number): Promise<GroupData> {
    const sql = "SELECT * FROM groups WHERE id = ?";

    const groupData: GroupData = await pool
        .query(sql, [id])
        .then((res) => {
            if (res.length != 1) {
                throw new Error("Group not found");
            }
            // As the database structure matches our GroupData interface
            // we can just return the first row
            return res[0];
        })
        .catch((err) => {
            console.error(err);
            return {
                id: -1,
                name: "",
                description: undefined,
                created: new Date(),
                last_updated: new Date(),
            };
        });
    return groupData;
}

/** Get multiple groups given their MARIADB id
 *
 * @param snip_ids
 */
async function getByIds(ids: number[]): Promise<GroupData[]> {
    const sql = "SELECT * FROM groups WHERE id IN (?)";

    const groups: GroupData[] = await pool
        .query(sql, [ids])
        .then((res) => {
            return res;
        })
        .catch((err) => {
            console.error(err);
            return [];
        });

    return groups;
}

async function getAllByUserId(user_id: number): Promise<GroupData[]> {
    console.log("[Database][groups] Getting groups for user: " + user_id);
    throw new Error("Method not implemented.");
}

async function getAllIdsByUserId(user_id: number): Promise<number[]> {
    // Use junction table to get groups of user
    const sql = "SELECT group_id FROM users_groups WHERE user_id = ?";

    const group_ids: number[] = await pool
        .query(sql, [user_id])
        .then((res) => {
            return res.map((row) => row.group_id);
        })
        .catch((err) => {
            console.error(err);
            return [];
        });
    return group_ids;
}

/** Create a new group in database
 *
 * @param group
 */
async function insert(group: GroupData): Promise<GroupData> {
    const sql = `INSERT INTO \`groups\` 
            (\`name\`, \`description\`) VALUES 
            (?, ?)`;

    const groupData_inserted: GroupData = await pool
        .query(sql, [group.name, group.description])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Insert failed");
            }
            return res.insertId;
        })
        .then((id) => {
            return this.getById(id);
        })
        .catch((err) => {
            console.error(err);
            return group_default;
        });
    return groupData_inserted;
}

/** Update group data in database
 *
 * @param group
 */
async function update(group: GroupData, id?: number): Promise<GroupData> {
    const sql = "UPDATE `groups` SET name = ?, description = ? WHERE id = ?";

    const groupData_updated = await pool
        .query(sql, [group.name, group.description, id])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Update failed" + res.warningStatus);
            }
            return this.getById(group.id);
        })
        .catch((err) => {
            console.error(err);
            return group_default;
        });
    return groupData_updated;
}

/** Delete group from database
 *
 */
async function _delete(id: number): Promise<boolean> {
    const sql = "DELETE FROM `groups` WHERE id = ?";

    const success = await pool
        .query(sql, [id])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Delete failed status:" + res.warningStatus);
            }
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return success;
}

async function search(query: string): Promise<GroupData[]> {
    const sql =
        "SELECT * FROM `groups` WHERE name LIKE ? OR description LIKE ?";

    const groups: GroupData[] = await pool
        .query(sql, [`%${query}%`, `%${query}%`])
        .then((res) => {
            return res;
        })
        .catch((err) => {
            console.error(err);
            return [];
        });
    return groups;
}

/** Check typing */
const GroupsService: GroupsServiceInterface = {
    getById,
    getByIds,
    search,
    getAllByUserId,
    getAllIdsByUserId,
    insert,
    update,
    delete: _delete,
};
export default GroupsService;
