import { PageServiceInterface } from "../service.typings";

import { pool_data as pool } from "database/sql.connection";
import { pagesData as PageData } from "database/sql.typings";

const page_default: PageData = {
    id: -1,
    book_id: -1,
    page_number: -1,
    created: new Date(),
    last_updated: new Date(),
    background_type_id: 1,
};

/** Get a single page give its MARIADB database id
 *
 * @param page_id
 */
async function getById(page_id: number): Promise<PageData> {
    const sql = "SELECT * FROM pagesView WHERE id = ?";

    const pageData: PageData = await pool
        .query(sql, [page_id])
        .then((res) => {
            if (res.length != 1) {
                throw new Error("Page not found " + page_id);
            }
            // As the database structure matches our PageData interface
            // we can just return the first row
            return res[0];
        })
        .catch((err) => {
            console.error(err);
            return page_default;
        });
    return pageData;
}

/** Get multiple pages given their MARIADB database ids
 *
 * @param page_ids
 */
async function getByIds(page_ids: number[]): Promise<PageData[]> {
    const sql = "SELECT * FROM pagesView WHERE id IN (?)";

    const pages: PageData[] = await pool.query(sql, [page_ids]).catch((err) => {
        console.error(err);
        return [];
    });
    return pages;
}

/** Get all pages in a book
 *
 * @param book_id
 */
async function getByBookId(book_id: number): Promise<PageData[]> {
    const sql = "SELECT * FROM pagesView WHERE book_id = ?";

    const pages = await pool
        .query(sql, [book_id])
        .then((res) => {
            if (res.length == 0) {
                throw new Error("No pages found");
            }
            // should only be the same format as PageData
            return res;
        })
        .catch((err) => {
            console.error(err);
            return [];
        });
    return pages;
}

/** Insert a new page into the database, returns the updated
 * page object.
 *
 * @param page
 */
async function insert(pageData: PageData): Promise<PageData> {
    // Count number of pages in book
    const sql2 = "SELECT COUNT(*) AS count FROM pagesView WHERE book_id = ?";

    const num_pages = await pool.query(sql2, [pageData.book_id]).then((res) => {
        if (res.length == 0) {
            return 0;
        }
        return res[0].count;
    });

    // Insert page
    const sql = "INSERT INTO `pages` (`book_id`, `page_number`) VALUES (?, ?)";

    const page_inserted = await pool
        .query(sql, [pageData.book_id, num_pages])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Insert failed");
            }
            return res.insertId;
        })
        .then(async (id) => {
            return await this.getById(id);
        });
    return page_inserted;
}

/** Update a page in the database, returns true if update was a success.
 *
 * @param page
 */
async function update(pageData: PageData): Promise<PageData> {
    const sql =
        "UPDATE `pages` SET `book_id` = ?, `page_number` = ?, `created`=? WHERE `id` = ?";

    const pageUpdated = await pool
        .query(sql, [
            pageData.book_id,
            pageData.page_number,
            pageData.created,
            pageData.id,
        ])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Update failed status");
            }
            return this.getById(pageData.id);
        })
        .catch((err) => {
            console.error(err);
            return page_default;
        });
    return pageUpdated;
}

/** Search for pages in a book
 *
 */
async function search(search: string): Promise<PageData[]> {
    console.log("searching for " + search);
    throw new Error("Not implemented");
}

/** Get first page from book
 * @param book_id
 */
async function getFirstByBookId(book_id: number): Promise<PageData> {
    const sql =
        "SELECT * FROM pagesView WHERE book_id = ? ORDER BY page_number ASC LIMIT 1";

    const page = await pool
        .query(sql, [book_id])
        .then((res) => {
            if (res.length != 1) {
                throw new Error("No pages found");
            }
            // should only be the same format as PageData
            return res[0];
        })
        .catch((err) => {
            console.error(err);
            return page_default;
        });
    return page;
}

/** Delete a page from the database, returns true if delete was a success.
 *
 * .. warning::
 *      This could be a problem since the page numbers are not updated!
 *
 * @param page_id
 */
async function _delete(page_id: number): Promise<boolean> {
    const sql = "DELETE FROM `pages` WHERE `id` = ?";

    const success = await pool
        .query(sql, [page_id])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Delete failed");
            }
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });

    //Update page number
    //TODO
    return success;
}

const PageService: PageServiceInterface = {
    getById,
    getByIds,
    getByBookId,
    search,
    insert,
    update,
    getFirstByBookId,
    delete: _delete,
};

export default PageService;
