import {
    Entity,
    ENTITY_GROUP,
    ENTITY_USER,
    PermissionACLData,
    PermissionServiceInterface,
    Resource,
    RESOURCE_BOOK,
} from "../service.typings";

import { pool_permissions as pool } from "database/sql.connection";
import { pool_data } from "database/sql.connection";

export const acl_access_denied: PermissionACLData = {
    id: -1,
    entity_id: -1,
    resource_id: -1,
    entity_type: ENTITY_USER,
    resource_type: RESOURCE_BOOK,
    pRead: false,
    pWrite: false,
    pDelete: false,
    pACL: false,
};

/** Get a single access control list entry given its MARIADB database id
 *
 * @param acl_id
 * @returns
 */
async function getACLById(acl_id: number): Promise<PermissionACLData> {
    const sql = "SELECT * FROM acl WHERE id = ?";

    const aclData: PermissionACLData = await pool
        .query(sql, [acl_id])
        .then((res) => {
            if (res.length != 1) {
                throw new Error("ACL not found " + acl_id);
            }
            // As the database structure matches our PermissionACLData interface
            // we can just return the first row
            return res[0];
        })
        .catch((err) => {
            console.error(err);
            return acl_access_denied;
        });
    return aclData;
}

/** Get a single access control list entry by the given entity and resource combo
 *
 * @param entity_id
 * @param resource_id
 * @param entity_type
 * @param resource_type
 */
async function getACLByEntityAndResourceReduced(
    entity_id: number,
    resource_id: number,
    entity_type: Entity,
    resource_type: Resource
): Promise<PermissionACLData> {
    if (entity_type == ENTITY_USER && resource_type == RESOURCE_BOOK) {
        return await getACLByUserBookReduced(entity_id, resource_id);
    }
    const sql = `
    SELECT * FROM acl WHERE entity_id = ? AND resource_id = ? AND entity_type = ? AND resource_type = ?`;

    const aclData: PermissionACLData = await pool
        .query(sql, [entity_id, resource_id, entity_type, resource_type])
        .then((res) => {
            if (res.length != 1) {
                throw new Error(
                    "ACL not found " + entity_id + " " + resource_id
                );
            }
            // unique key constraint on entity_id, resource_id, entity_type, resource_type
            // we can just return the first row
            return res[0];
        })
        .catch((err) => {
            console.error(err);
            // Overwrite defaults and return no access
            return {
                ...acl_access_denied,
                resource_type: resource_type,
                entity_type: entity_type,
                resource_id: resource_id,
                entity_id: entity_id,
            };
        });
    return aclData;
}

/** Get ACL by entity, does not solve permissions for groups!
 */
async function getACLByEntityAndResource(
    entity_id: number,
    resource_id: number,
    entity_type: Entity,
    resource_type: Resource
): Promise<PermissionACLData> {
    const sql = `
            SELECT * FROM acl WHERE 
                entity_id = ? AND entity_type = ? AND
                resource_id = ? AND
                resource_type = ?`;

    const aclData: PermissionACLData = await pool
        .query(sql, [entity_id, entity_type, resource_id, resource_type])
        .then((res) => {
            if (res.length < 1) {
                throw new Error(
                    "ACL not found " +
                    entity_id +
                    " " +
                    resource_id +
                    ". Getting default one"
                );
            }
            // unique key constraint on entity_id, resource_id, entity_type, resource_type
            // we can just return the first row
            return res[0];
        })
        .catch((err) => {
            console.error(err);
            // Overwrite defaults and return no access
            return {
                ...acl_access_denied,
                resource_type: resource_type,
                entity_type: entity_type,
                resource_id: resource_id,
                entity_id: entity_id,
            };
        });
    return aclData;
}

/** Get user book acl
 * This solves for all the users groups and returns the highest permission!
 *
 * @param user_id
 * @param book_id
 */
async function getACLByUserBook(
    user_id: number,
    book_id: number
): Promise<PermissionACLData[]> {
    // A bit more complex query which also checks for the users groups
    const sql = `SELECT * FROM acl WHERE 
        ((entity_id = ? AND entity_type = ${ENTITY_USER}) OR 
        (entity_id IN (SELECT group_id FROM users_groups WHERE user_id = ?)) AND
        entity_type = ${ENTITY_GROUP}) AND resource_id = ? AND resource_type = ${RESOURCE_BOOK}`;

    let aclData: PermissionACLData[] = await pool
        .query(sql, [user_id, user_id, book_id])
        .then((res) => {
            if (res.length < 1) {
                return [acl_access_denied];
            }
            return res;
        })
        .catch((err) => {
            console.error(err);
            // Overwrite defaults and return no access
            return [];
        });


    // Check if book is finished
    const sql_finished = `SELECT finished FROM books WHERE id = ?`;
    const finished = await pool_data
        .query(sql_finished, [book_id])
        .then((res) => {
            if (res.length != 1) {
                throw new Error("Book not found " + book_id);
            }
            return res[0].finished;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });


    if (finished) {
        // If book is finished, we remove all write permissions
        aclData = aclData.map((acl) => {
            return {
                ...acl,
                pWrite: false,
            };
        });
    }

    return aclData;
}



/** Get user books acl */
async function getACLByUserBooks(
    user_id: number,
    book_ids: number[]
): Promise<PermissionACLData[]> {
    // A bit more complex query which also checks for the users groups
    const sql = `SELECT * FROM acl WHERE
        ((entity_id = ? AND entity_type = ${ENTITY_USER}) OR
        (entity_id IN (SELECT group_id FROM users_groups WHERE user_id = ?)) AND
        entity_type = ${ENTITY_GROUP}) AND resource_id IN (?) AND resource_type = ${RESOURCE_BOOK}`;

    let aclData: PermissionACLData[] = await pool
        .query(sql, [user_id, user_id, book_ids])
        .then((res) => {
            if (res.length != book_ids.length) {
                throw new Error("ACL not found " + user_id + " " + book_ids);
            }
            return res;
        }).catch((err) => {
            console.error(err);
            // Overwrite defaults and return no access
            return [];
        });

    // Check if book is finished
    const sql_finished = `SELECT finished FROM books WHERE id IN (?)`;
    const finished = await pool_data
        .query(sql_finished, [book_ids])
        .then((res) => {
            if (res.length != book_ids.length) {
                throw new Error("Book not found " + book_ids);
            }
            return res.map((r) => r.finished);
        })
        .catch((err) => {
            console.error(err);
            return [];
        });

    // If book is finished, we remove all write permissions
    aclData = aclData.map((acl, i) => {
        if (finished[i]) {
            return {
                ...acl,
                pWrite: false,
            };
        }
        return acl;
    });
    return aclData;
}

async function getACLByUserBookReduced(
    user_id: number,
    book_id: number
): Promise<PermissionACLData> {
    const all = await getACLByUserBook(user_id, book_id);
    const reduced = all.reduce(
        (prev, cur) => {
            return {
                id: -2,
                entity_id: user_id,
                entity_type: ENTITY_USER,
                resource_id: cur.resource_id,
                resource_type: RESOURCE_BOOK,
                pRead: prev.pRead || cur.pRead,
                pWrite: prev.pWrite || cur.pWrite,
                pDelete: prev.pDelete || cur.pDelete,
                pACL: prev.pACL || cur.pACL,
            };
        },
        {
            ...acl_access_denied,
            entity_id: user_id,
            entity_type: ENTITY_USER,
            resource_id: book_id,
            resource_type: RESOURCE_BOOK,
        }
    );

    // Convert pRead, pWrite, pDelete, pACL to boolean
    reduced.pRead = reduced.pRead ? true : false;
    reduced.pWrite = reduced.pWrite ? true : false;
    reduced.pDelete = reduced.pDelete ? true : false;
    reduced.pACL = reduced.pACL ? true : false;

    return reduced;
}

/** Get all allowed (pRead==true) resources of a user/group.
 * This is used to get all books a user can read. At also solves for groups.
 *
 * @param entity_id
 * @param entity_type
 * @return book_ids
 */
async function getACLByEntity(
    entity_id: number,
    entity_type: Entity = ENTITY_USER
): Promise<PermissionACLData[]> {
    // Get all books of a user/group which have read access
    let sql: string;
    let data = [];

    switch (entity_type) {
        case ENTITY_USER:
            sql = `
            SELECT * FROM acl 
                WHERE (
                    entity_id = ? AND entity_type = ${ENTITY_USER} AND pRead = 1
                ) OR (
                    entity_id IN(
                        SELECT group_id FROM users_groups WHERE user_id = ?
                    ) AND entity_type = ${ENTITY_GROUP} AND pRead = 1
                ); 
            `;
            data = [entity_id, entity_id];
            break;
        case ENTITY_GROUP:
            sql = `
            SELECT * FROM acl
                WHERE entity_id = ? AND entity_type = ${ENTITY_GROUP} AND pRead = 1;
            `;
            data = [entity_id];
            break;
        default:
            throw new Error("Invalid entity type");
    }

    const acl_pRead: PermissionACLData[] = await pool
        .query(sql, data)
        .then((res) => {
            return res;
        })
        .catch((err) => {
            console.error(err);
            // Return empty list (no books)
            return [];
        });

    return acl_pRead;
}

async function setACLById(
    id: number,
    pRead: boolean,
    pWrite: boolean,
    pDelete: boolean,
    pACL: boolean
): Promise<boolean> {
    const sql = `UPDATE acl SET pRead = ?, pWrite = ?, pDelete = ?, pACL = ? WHERE id = ?`;

    const success: boolean = await pool
        .query(sql, [pRead, pWrite, pDelete, pACL, id])
        .then((res) => {
            if (res.affectedRows != 1) {
                throw new Error("ACL not found " + id);
            }
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return success;
}

async function setACLByEntityAndResource(
    entity_id: number,
    entity_type: Entity,
    resource_id: number,
    resource_type: Resource,
    pRead: boolean,
    pWrite: boolean,
    pDelete: boolean,
    pACL: boolean
): Promise<boolean> {
    // Insert and updated on duplicate key

    const sql = `
    INSERT INTO acl (entity_id, entity_type, resource_id, resource_type, pRead, pWrite, pDelete, pACL) 
    VALUES (?, ?, ?, ?, ?, ?, ?, ?) 
    ON DUPLICATE KEY UPDATE pRead = ?, pWrite = ?, pDelete = ?, pACL = ?`;

    const success: boolean = await pool
        .query(sql, [
            entity_id,
            entity_type,
            resource_id,
            resource_type,
            pRead,
            pWrite,
            pDelete,
            pACL,
            pRead,
            pWrite,
            pDelete,
            pACL,
        ])
        .then((res) => {
            if (res.warningStatus > 1) {
                throw new Error("Insert/Update of acl failed");
            }
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return success;
}

/** Check typing */
const PermissionService: PermissionServiceInterface = {
    getACLById,
    getACLByEntityAndResource,
    getACLByEntityAndResourceReduced,
    getACLByUserBook,
    getACLByUserBooks,
    getACLByUserBookReduced,
    getACLByEntity,
    setACLById,
    setACLByEntityAndResource,
};
export default PermissionService;
