/* eslint-disable @typescript-eslint/no-explicit-any */
import { createHash } from "crypto";
import { hasOwnProperty } from "lib/utils";

import { pool_data as pool } from "../../sql.connection";
import { SnipServiceInterface } from "../service.typings";

import { ArrayData } from "snips/general/array";
import { BaseData, BaseView, SnipData } from "snips/general/base";
import { BlobData, ImageData } from "snips/general/image";

const snip_default: SnipData<BaseData, BaseView> = {
    id: null,
    page_id: null,
    book_id: null,
    data: {
        type: "base",
    },
    view: {
        x: 0,
        y: 0,
    },
    type: "base",
    created: new Date(),
    last_updated: new Date(),
};

/** Get a single snip object given its MARIADB id
 *
 * @param snip_id
 */
async function getById(snip_id: number): Promise<SnipData<BaseData, BaseView>> {
    const sql = "SELECT * FROM snips WHERE id = ?";

    let snipData = await pool
        .query(sql, [snip_id])
        .then((res) => {
            if (res.length != 1) {
                throw new Error("Snip not found");
            }
            return res[0];
        })
        .catch((err) => {
            console.error(err);
            return snip_default;
        });

    snipData = await parseSnips([snipData]);
    return snipData[0];
}

/** Get multiple snips given their MARIADB id
 *
 * @param snip_ids
 */
async function getByIds(
    snip_ids: number[]
): Promise<SnipData<BaseData, BaseView>[]> {
    const sql = "SELECT * FROM snips WHERE id IN (?)";

    let snips = await pool
        .query(sql, [snip_ids])
        .then((res) => {
            return res;
        })
        .catch((err) => {
            console.error(err);
            return [];
        });



    // We still need to parse the json from the database
    // This should run in parallel
    snips = parseSnips(snips);
    return snips;
}
async function getByPageId(
    page_id: number
): Promise<SnipData<BaseData, BaseView>[]> {
    // This is highly inefficient: uncommented for now as it produces a lot of lag
    // on page switch
    //const sql = "SELECT id,type,data_json,view_json FROM (SELECT t1.id,t1.type,t1.data_json,t1.view_json,t2.foreign_id FROM (SELECT * FROM snips WHERE page_id=? AND type!='array') AS t1 LEFT JOIN (SELECT snips.id,t2.* FROM snips, JSON_TABLE( snips.data_json, '$' COLUMNS(NESTED PATH '$.snip_ids[*]' COLUMNS(foreign_id INT PATH '$')) ) AS t2 WHERE snips.page_id=? AND snips.type='array') AS t2 ON t1.id = t2.foreign_id WHERE t2.id IS NULL UNION SELECT id,type,data_json,view_json,NULL FROM snips WHERE page_id=? AND type='array') AS ttt; ";
    const sql = "SELECT * FROM snips WHERE page_id = ?";

    let snips = await pool
        .query(sql, [page_id, page_id, page_id])
        .then((res) => {
            return res;
        })
        .catch((err) => {
            console.error(err);
            return [];
        });

    // We still need to parse the json from the database
    // This should run in parallel
    snips = await parseSnips(snips);

    return snips;
}

/** Get all snips which are not yet placed on a page but still have a book ID set
 *
 * @param book_id
 * @param show_hidden
 * @returns
 */
async function getNotPlacedByBookId(
    book_id: number,
    show_hidden = false
): Promise<SnipData<BaseData, BaseView>[]> {
    let sql = "SELECT * FROM `snips` WHERE `book_id` = ? AND `page_id` IS NULL";
    if (!show_hidden) {
        sql += " AND `hide` = 0";
    }

    let snips = await pool
        .query(sql, [book_id])
        .then((res) => {
            // should only be the same format as SnipData
            return res;
        })
        .catch((err) => {
            console.error(err);
            return [];
        });

    // We still need to parse the json from the database
    snips = await parseSnips(snips);
    return snips;
}

/** Insert a new snip  */
async function insert(
    snipData: SnipData<BaseData, BaseView>
): Promise<SnipData<BaseData, BaseView>> {
    // Set id to null to prevent sql errors
    snipData.id = null;

    // Check if snip has blob and insert
    let blob_id: number | null = null;
    if (hasOwnProperty(snipData.data, "blob")) {
        (snipData.data as ImageData).blob.snip_id = null;
        (snipData.data as ImageData).blob = await insertBlob(
            (snipData.data as ImageData).blob
        );
        blob_id = (snipData.data as ImageData).blob.id;
    }

    // Check if array snip
    let snips_before;
    if (hasOwnProperty(snipData.data, "snips")) {
        const snipd_ids = await Promise.all(
            (snipData.data as ArrayData).snips.map(async (snipd) => {
                snipd = await insert(snipd);
                return snipd.id;
            })
        );
        //Update ids of snips
        (snipData.data as ArrayData).snips.forEach((snipd, index) => {
            snipd.id = snipd_ids[index];
        });

        // Set ids field for db
        (snipData.data as any).snip_ids = snipd_ids;

        snips_before = (snipData.data as any).snips;
        delete (snipData.data as any).snips;
    }

    const snip_sql =
        "INSERT INTO `snips` (`page_id`, `book_id`, `type`, `data_json`, `view_json`, `blob_id`) VALUES (?, ?, ?, ?, ?, ?)";

    //Stringify json data
    const data_json = await encodeSnipData(snipData.data);
    const view_json = JSON.stringify(snipData.view);

    // Insert data
    const snip_id = await pool
        .query(snip_sql, [
            snipData.page_id,
            snipData.book_id,
            snipData.type,
            data_json,
            view_json,
            blob_id,
        ])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Insert failed");
            }
            return res.insertId;
        })
        .catch((err) => {
            console.error(err);
            return snip_default;
        });

    snipData.id = snip_id;
    if (snips_before) {
        (snipData.data as any).snips = snips_before;
    }
    return snipData;
}

async function update(
    snipData: SnipData<BaseData, BaseView>
): Promise<SnipData<BaseData, BaseView>> {
    // Get data strings from snip
    const data_json = snipData.data;
    const view_json = snipData.view;

    // Replace snips with their id if we have a nested snip
    if (hasOwnProperty(data_json, "snips")) {
        //Force update on nested snips
        (data_json as ArrayData).snips.forEach(async (s) => {
            await update(s);
        });
        //Add ids to datajson
        (data_json as any)["snip_ids"] = (data_json as ArrayData).snips.map(
            (s) => {
                return s.id;
            }
        );
        // Delete old tag
        delete (data_json as any)["snips"];
    }

    // Update blob if present
    if (hasOwnProperty(data_json, "blob")) {
        const blob = (data_json as ImageData).blob;
        await updateBlob(blob);
        (data_json as any)["blobid"] = blob.id;
        delete (data_json as any)["blob"];
    }

    // Upload
    const sql_data =
        "UPDATE `snips` SET `page_id`=?,`book_id`=?,`type`=?,`data_json`=?,`view_json`=? WHERE `id`=?";

    const success = await pool
        .query(sql_data, [
            snipData.page_id,
            snipData.book_id,
            snipData.type,
            JSON.stringify(data_json),
            JSON.stringify(view_json),
            snipData.id,
        ])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Update failed");
            }
            return getById(snipData.id);
        })
        .catch((err) => {
            console.error(err);
            return snip_default;
        });
    return success;
}

/** Delete a snip  */
async function _delete(snip_id: number): Promise<boolean> {
    const snip_data = await getById(snip_id);
    // Delete blob if there is one
    if (hasOwnProperty(snip_data.data, "blob")) {
        const blob = (snip_data.data as ImageData).blob;
        await deleteBlob(blob.id);
    }
    //Delete snip
    const sql = "DELETE FROM `snips` WHERE `id`=?";

    const success = await pool
        .query(sql, [snip_id])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Delete failed");
            }
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return success;
}

async function encodeSnipData(data: BaseData): Promise<string> {
    // Create copy
    const data_c = { ...data };

    // Check if blob
    if (hasOwnProperty(data, "blob")) {
        (data_c as any)["blobid"] = (data_c as ImageData).blob.id;
        delete (data_c as any)["blob"];
    }

    return JSON.stringify(data_c);
}

async function parseSnips(
    snips: any[]
): Promise<SnipData<BaseData, BaseView>[]> {
    snips = await Promise.all(
        snips.map(async (item) => {
            try {
                item.data = await parseSnipData(item.data_json, snips);
                item.view = await parseSnipView(item.view_json);
                delete item.data_json;
                delete item.view_json;

                // Check for blob and get it
                if (item.blob_id != null) {
                    item.data.blob = await getBlob(item.blob_id);
                }
            } catch (err) {
                console.log(
                    "[snip.service] Could not parse snip data, may need manual fix"
                );
                console.error(item);
                console.error(err);
                item.data = { type: "base" };
                item.view = [{ x: 0, y: 0 }];
            }
            return item;
        })
    );
    // Remove nested array snips, is way faster than doing it using sql
    let to_remove: number[] = [];
    for (const snip of snips) {
        if (hasOwnProperty(snip.data, "snip_ids")) {
            to_remove = to_remove.concat(snip.data.snip_ids);
        }
    }
    snips = snips.filter((snip) => {
        return !to_remove.includes(snip.id);
    });
    return snips;
}

/** Parses snip data if an snips array is given the array snips
 * are not retrieved from the database again but filtered from the snips
 *
 * @param data_string
 * @param snips
 * @returns
 */
async function parseSnipData(data_string: string, snips?: any[]): Promise<any> {
    const data = JSON.parse(data_string);

    // Check for array
    if (hasOwnProperty(data, "snip_ids")) {
        data.snips = [];
        // Check if we have the snips already
        for (const id of data.snip_ids) {
            const found = snips.find((snip) => snip.id == id);
            if (found) {
                data.snips.push(found);
            } else {
                data.snips.push(await getById(id));
            }
        }
    }

    //Parse to int
    if (hasOwnProperty(data, "width")) {
        data.width = parseInt(data.width);
    }
    if (hasOwnProperty(data, "height")) {
        data.height = parseInt(data.height);
    }
    return data;
}

async function parseSnipView(view_string: string | null): Promise<any> {
    // Parse to int
    if (view_string == null) {
        return "";
    }
    const view = JSON.parse(view_string);
    if (hasOwnProperty(view, "rx")) {
        view.rx = parseInt(view.rx);
    }
    if (hasOwnProperty(view, "ry")) {
        view.ry = parseInt(view.ry);
    }
    if (hasOwnProperty(view, "rw")) {
        view.rw = parseInt(view.rw);
    }
    if (hasOwnProperty(view, "rh")) {
        view.rh = parseInt(view.rh);
    }
    if (hasOwnProperty(view, "width")) {
        view.width = parseInt(view.width);
    }
    return view;
}

/** Insert a new blob
 *
 * @param blob Blob to insert
 */
async function insertBlob(blob: BlobData): Promise<BlobData> {
    const sql =
        "INSERT INTO `blobs` (`mime`, `size`, `data`, `hash`) VALUES (?, ?, FROM_BASE64(?), ?)";


    // Compute hash if not given (md5)
    if (!blob.hash) {
        const hash = createHash("md5");
        hash.update(blob.data);
        blob.hash = hash.digest("hex");
    }

    // compute size if not given in bytes
    if (!blob.size) {
        blob.size = Buffer.byteLength(blob.data, "base64");
    }


    const blob_id = await pool
        .query(sql, [blob.mime, blob.size, blob.data, blob.hash])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Insert failed");
            }
            return res.insertId;
        });
    blob.id = blob_id;

    return blob;
}

/** Return blob from mysql database via id
 *
 * @param blob_id
 * @returns
 */
async function getBlob(blob_id: number): Promise<BlobData> {
    const sql = "SELECT * FROM blobs WHERE id = ?";

    const blob = await pool
        .query(sql, [blob_id])
        .then((res) => {
            if (res.length != 1) {
                return Error("Blob not found");
            }

            return res[0];
        })
        .catch((err) => {
            console.error(err);
            throw err;
        });

    //Convert blob data to base64
    blob.data = Buffer.from(blob.data).toString("base64");
    return blob;
}

async function updateBlob(blob: BlobData): Promise<boolean> {
    const sql_data =
        "UPDATE `blobs` SET `mime`=?,`size`=?,`data`=FROM_BASE64(?),`hash`=? WHERE `id`=?";

    const success = await pool
        .query(sql_data, [blob.mime, blob.size, blob.data, blob.hash, blob.id])
        .then(() => {
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return success;
}
async function deleteBlob(blob_id: number | null): Promise<boolean> {
    if (blob_id == null) {
        return true;
    }
    const sql = "DELETE FROM `blobs` WHERE `id`=?";

    const success = await pool
        .query(sql, [blob_id])
        .then(() => {
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return success;
}

async function search(search: string): Promise<any[]> {
    console.log("Search: " + search);
    throw new Error("Not implemented");
}

const SnipService: SnipServiceInterface = {
    getById,
    getByIds,
    getByPageId,
    getNotPlacedByBookId,
    search,
    insert,
    update,
    delete: _delete,
};
export default SnipService;
