import { compare, hash } from "bcrypt";

import { UsersServiceInterface } from "../service.typings";
import GroupsService from "./groups.service";

import { pool_permissions as pool } from "database/sql.connection";
import { usersData as UserData } from "database/sql.typings";

const user_default: UserData = {
    id: -1,
    email: "",
    emailValidated: false,
    created: new Date(),
    last_updated: new Date(),
    hashedPassword: "",
};

/** Tries to match a users email with
 * the hashed password in the database
 */
async function login(email: string, password: string): Promise<boolean> {
    const sql = "SELECT hashedPassword FROM users WHERE email = ?";

    const validated = await pool
        .query(sql, [email])
        .then((res) => {
            if (res.length > 0) {
                // Compare hashedPassword with password
                return compare(password, res[0].hashedPassword);
            } else {
                return false;
            }
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return validated;
}

async function verify(email: string): Promise<boolean> {
    //Check if user is already verified
    const user = await getByEmail(email);

    if (user.emailValidated) {
        return false;
    }
    // Update user
    const sql = "UPDATE users SET `emailValidated` = ? WHERE `email` = ?";

    const verified = await pool
        .query(sql, [true, email])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Update failed");
            }
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return verified;
}

async function getPasswordHash(email: string): Promise<string> {
    const sql = "SELECT hashedPassword FROM users WHERE email = ?";

    const hash = await pool
        .query(sql, [email])
        .then((res) => {
            if (res.length > 0) {
                return res[0].hashedPassword;
            } else {
                return "";
            }
        })
        .catch((err) => {
            console.error(err);
            return "";
        });
    return hash;
}

// Returns undefined on error
async function signup(
    email: string,
    password: string
): Promise<UserData | undefined> {
    const user = await getByEmail(email);
    if (user) {
        console.log("[Database][users] User already exists");
        return undefined;
    }
    // hash password to store in db
    const hashedPassword = await hash(password, 15);

    const sql =
        "INSERT INTO users (`email`, `hashedPassword`, `emailValidated`) VALUES (?, ?, ?)";

    const u = await pool
        .query(sql, [email, hashedPassword, false])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Insert failed");
            }
            return res.insertId;
        })
        .then((id) => {
            return getById(id);
        })
        .catch((err) => {
            console.error(err);
            return undefined;
        });
    return u;
}

async function updatePassword(
    email: string,
    password: string
): Promise<boolean> {
    // hash password to store in db
    const hashedPassword = await hash(password, 15);

    const sql = "UPDATE users SET `hashedPassword` = ? WHERE `email` = ?";
    const updated = await pool
        .query(sql, [hashedPassword, email])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Update failed");
            }
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return updated;
}

async function getByEmail(email: string): Promise<UserData> {
    const sql = "SELECT * FROM users WHERE email = ?";

    const user = await pool
        .query(sql, [email])
        .then((res) => {
            if (res.length != 1) {
                throw new Error("User not found");
            }
            return res[0];
        })
        .catch((err) => {
            console.error(err);
            return undefined;
        });
    // Parse user groups and delete hashedPassword as we dont want to transmit it
    if (user) {
        delete user.hashedPassword;
        user.emailValidated = user.emailValidated == 1;
    }
    return user;
}
async function getById(id: number): Promise<UserData | undefined> {
    const sql = "SELECT * FROM users WHERE id = ?";

    const user = await pool
        .query(sql, [id])
        .then((res) => {
            if (res.length != 1) {
                throw new Error("User not found");
            }
            return res[0];
        })
        .catch((err) => {
            console.error(err);
            return undefined;
        });
    // Parse user groups and delete hashedPassword as we dont want to transmit it
    if (user) {
        delete user.hashedPassword;
        user.emailValidated = user.emailValidated == 1;
    }
    return user;
}
async function getByIds(ids: number[]): Promise<UserData[]> {
    const sql = "SELECT * FROM users WHERE id IN (?)";

    const users = await pool
        .query(sql, [ids])
        .then((res) => {
            return res;
        })
        .catch((err) => {
            console.error(err);
            return [];
        });
    for (const user of users) {
        delete user.hashedPassword;
        user.emailValidated = user.emailValidated == 1;
    }
    return users as UserData[];
}

async function isAdmin(user_id: number) {
    const groups = await GroupsService.getAllIdsByUserId(user_id);
    if (groups.includes(1)) {
        return true;
    }
    return false;
}

// Insert function just insert a new user, the password is not set tho!
async function insert(object: UserData): Promise<UserData | undefined> {
    //Stringify group ids
    const sql = "INSERT INTO user (`email`,  `emailValidated`) VALUES (?, ?)";

    const user = await pool
        .query(sql, [object.email, object.emailValidated])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Insert failed");
            }
            return res.insertId;
        })
        .then((id) => {
            return getById(id);
        })
        .catch((err) => {
            console.error(err);
            return undefined;
        });
    return user;
}

async function update(object: UserData): Promise<UserData> {
    const sql =
        "UPDATE users SET `email` = ?, `emailValidated` = ? WHERE `id` = ?";

    const user = await pool
        .query(sql, [object.email, object.emailValidated, object.id])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Update failed");
            }
            return getById(object.id);
        })
        .catch((err) => {
            console.error(err);
            return user_default;
        });
    return user;
}

async function _delete(id: number): Promise<boolean> {
    const sql = "DELETE FROM users WHERE id = ?";

    const deleted = await pool
        .query(sql, [id])
        .then((res) => {
            if (res.warningStatus > 0) {
                throw new Error("Delete failed");
            }
            return true;
        })
        .catch((err) => {
            console.error(err);
            return false;
        });
    return deleted;
}

async function search(search: string) {
    const sql = `
    SELECT * FROM users
        WHERE email LIKE ? OR
        id LIKE ? OR
        emailValidated LIKE ? OR
        created LIKE ?
    `;

    const users = await pool
        .query(sql, [
            `%${search}%`,
            `%${search}%`,
            `%${search}%`,
            `%${search}%`,
        ])
        .then((res) => {
            return res;
        })
        .catch((err) => {
            console.error(err);
            return [];
        });
    for (const user of users) {
        delete user.hashedPassword;
        user.emailValidated = user.emailValidated == 1;
    }
    return users;
}

const UserService: UsersServiceInterface = {
    login,
    signup,
    verify,
    isAdmin,
    search,
    getByEmail,
    getPasswordHash,
    getById,
    getByIds,
    insert,
    update,
    updatePassword,
    delete: _delete,
};
export default UserService;
