import {
    GetServerSidePropsContext,
    GetServerSidePropsResult,
    NextApiRequest,
    NextApiResponse,
    PreviewData,
} from "next/types";
import { service } from "pages/api/_service";
import { ParsedUrlQuery } from "querystring";

import { hashToken } from "./hash_token";

import {
    ENTITY_TOKEN,
    PermissionACLData,
    RESOURCE_BOOK,
} from "database/services/service.typings";
import { pool_permissions } from "database/sql.connection";

/** This function can be used
 * to secure a books page using SSR
 * books/[id]/? route.
 *
 * This attaches the perms object of the book and current user
 * to the params of the page.
 */
export function withAuthBookIdSsr<
    P extends { [key: string]: unknown } = { [key: string]: unknown },
    Q extends ParsedUrlQuery = ParsedUrlQuery,
    D extends PreviewData = PreviewData
>(
    handler: (
        context: GetServerSidePropsContext<Q, D>,
        perms: PermissionACLData & { token_used: boolean, expires_at?: Date }
    ) => GetServerSidePropsResult<P> | Promise<GetServerSidePropsResult<P>>
) {
    return async (ctx: GetServerSidePropsContext<Q, D>) => {
        const { req, query } = ctx;
        const book_id = query.id as string;
        const user_id = req.session.user_id;


        // Check if token is present
        const url = new URL(req.url, "http://localhost");
        const token = url.searchParams.get("token");
        if (book_id && token) {
            const perms = await verify_with_token(token, book_id);
            return await handler(ctx, perms);
        }

        // First check for user
        if (book_id && user_id) {
            const perms = await service.permission.getACLByUserBookReduced(
                user_id,
                parseInt(book_id)
            ) as PermissionACLData & { token_used: boolean, expires_at?: Date };
            perms.token_used = false;
            return await handler(ctx, perms);
        }

        return {
            redirect: {
                permanent: false,
                destination: "/unauthorized",
            },
        };
    };
}

export declare type NextApiHandlerWithPerms<T = unknown> = (
    req: NextApiRequest,
    res: NextApiResponse<T>,
    perms: PermissionACLData
) => unknown | Promise<unknown>;

/** This function can be used
 * to secure a api routes
 * api/books/[id]/? route.
 *
 * This attaches the perms object of the book and current user
 * to the params of the page.
 */
export function withAuthBookId(handler: NextApiHandlerWithPerms) {
    return async (req: NextApiRequest, res: NextApiResponse) => {
        const { query } = req;
        const book_id = query.id as string;

        // Get the users permission for this book
        if (book_id) {
            const token = req.headers.authorization?.split(" ")[1];
            const perms = await bookAuth(req.session.user_id, book_id, token);
            return await handler(req, res, perms);
        } else {
            res.status(400).json({
                error: "Bad Request",
                details: { id: "Missing!" },
            });
            return;
        }
    };
}


/** This function can be used to secure any api route
 * api/? route.
 */
export async function bookAuth(user_id: number, book_id: string | number, token?: string) {
    let perms;
    if (token) {
        perms = await verify_with_token(token, book_id);
    } else {
        perms = await service.permission.getACLByUserBookReduced(
            user_id,
            parseInt(book_id as string)
        );
    }
    return perms;
}


async function verify_with_token(
    token: string,
    book_id: string | number,
): Promise<PermissionACLData & { token_used: boolean, expires_at?: Date }> {
    const hashed_token = hashToken(token);
    const { found, type, expires_at } = await pool_permissions
        .query("SELECT * FROM tokens WHERE book_id = ? AND hashed_token = ?", [
            book_id,
            hashed_token,
        ])
        .then((res) => {
            if (res.length > 0) {
                return {
                    found: true,
                    type: res[0].type,
                    expires_at: res[0].expires_at as Date
                }
            } else {
                return {
                    found: false,
                    type: null,
                    expires_at: null
                }
            }
        })
        .catch((err) => {
            console.error(err);
            return {
                found: false,
                type: null,
                expires_at: null
            }
        });

    const perms: PermissionACLData & { token_used: boolean, expires_at?: Date }
        = {
        entity_id: -1,
        entity_type: ENTITY_TOKEN,
        resource_id: parseInt(book_id as string),
        resource_type: RESOURCE_BOOK,
        pACL: false,
        pDelete: false,
        pRead: false,
        pWrite: false,
        token_used: true,
        expires_at: null
    }
    if (!found) {
        return perms;
    }

    if (expires_at) {
        // Should work as it is valid in the database
        perms.expires_at = new Date(expires_at)
        if (perms.expires_at < new Date()) {
            return perms;
        }
    }

    switch (type) {
        case "api":
            perms.pRead = true;
            perms.pWrite = true;
            perms.pDelete = true;
            break;
        case "ui":
            perms.pRead = true;
            break;
    }

    return perms;
}
