import {
    Dispatch,
    SetStateAction,
    useCallback,
    useEffect,
    useState,
} from "react";
import useSWR, { KeyedMutator } from "swr";

import { fetcher } from "./fetcher";
import { useDebounce } from "./utils";

import { UserConfigData, UserData } from "database/services/service.typings";

interface UserHook {
    user: UserData | undefined;
    isLoadingUser: boolean;
    isErrorUser: boolean;
    mutateUser: KeyedMutator<UserData | undefined>;
    isLoggedIn: boolean;

    config: UserConfigData | undefined;
    isLoadingConfig: boolean;
    isErrorConfig: boolean;
    mutateConfig: Dispatch<SetStateAction<Partial<UserConfigData>>>;
}

export default function useUser() {
    const {
        data: user,
        mutate: mutateUser,
        error: errorUser,
    } = useSWR<UserData>("/api/account/user", fetcher, {
        revalidateIfStale: true,
        revalidateOnFocus: false,
        revalidateOnReconnect: false,
        fallbackData: {
            created: new Date(),
            email: "",
            emailValidated: false,
            id: -1,
            last_updated: new Date(),
        },
    });

    const {
        data: config,
        error: errorConfig,
        mutate: mutateConfig,
    } = useSWR<UserConfigData>("/api/account/config", fetcher, {
        revalidateIfStale: true,
        revalidateOnFocus: true,
        revalidateOnReconnect: true,
        fallbackData: {
            id: -1,
            last_updated: new Date(),
            pen_color: "#000000",
            pen_size: 4,
            pen_smoothing: 0.0,
            text_font_id: 1,
            text_font: "Arial",
            text_fontColor: "#000000",
            text_fontSize: 12,
            text_lineHeight: 1.25,
            text_lineWrap: 400,
            user_id: -1,
            zoom1: 1,
            zoom2: 1,
        },
    });

    /** Debounce server update to prevent api spam
     */
    const [upd, _setUpd] = useState<Partial<UserConfigData> | undefined>(
        undefined
    );
    const setUpd = useCallback(
        (newUpd: Partial<UserConfigData>) => {
            _setUpd(newUpd);
            //Update local config
            const newConfig = { ...config, ...newUpd };
            mutateConfig(newConfig, { populateCache: true, revalidate: false });
        },
        [config, mutateConfig]
    );

    const debounceUpdate = useDebounce(upd, 2000); // wait for 2 seconds before update database
    useEffect(() => {
        if (debounceUpdate && user.id > 0) {
            updateServer(debounceUpdate, config).then(() => {
                setUpd(undefined);
            });
        }
    }, [debounceUpdate, config, mutateConfig, setUpd, user]);

    const isLoadingUser = !errorUser && !user;

    const ret: UserHook = {
        user,
        isLoadingUser,
        isErrorUser: errorUser,
        mutateUser,
        isLoggedIn: !isLoadingUser && !errorUser && user.id > 0,

        config: config,
        isLoadingConfig: !errorConfig && !config,
        isErrorConfig: errorConfig,
        mutateConfig: setUpd,
    };
    return ret;
}

async function updateServer(
    configUpd: Partial<UserConfigData>,
    oldConfig: UserConfigData
): Promise<UserConfigData> {
    return await fetch("/api/account/config", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify(configUpd),
    })
        .then((res) => res.json())
        .then((res) => {
            if (res.error) {
                console.error(res);
                throw new Error(res.error);
            }

            return { ...oldConfig, ...res };
        })
        .catch((err) => {
            console.error(err);
            return oldConfig;
        });
}

/** THIS API SHOULD NOT BE USED FOR CRITICAL APPLICATION PARTS
 *  AS A RESPONSE CAN BE MANIPULATED BY THE USER EASILY!
 *
 * Only use this API to show or hide certain parts of the UI
 *
 */
export function useIsAdmin() {
    const {
        data: isAdmin,
        mutate: mutateIsAdmin,
        error,
    } = useSWR<boolean>("/api/authentication/isAdmin", fetcher);

    return {
        isAdmin: isAdmin,
        isLoading: !error && !isAdmin,
        isError: error,
        mutateIsAdmin,
    };
}
