import { getIronSession } from "iron-session/edge";
import { sessionOptions } from "lib/session";
import type { NextRequest } from "next/server";
import { NextResponse } from "next/server";

// Define logger with prefix
const logger = (...args) => console.log(`[Middleware NextJS]`, ...args);

export async function middleware(req: NextRequest) {
    logger(req.nextUrl.pathname);

    const base = req.url;

    // Rewrites for login and signup
    if (req.nextUrl.pathname === "/login") {
        return NextResponse.rewrite(new URL("/account/login", base));
    } else if (req.nextUrl.pathname === "/signup") {
        return NextResponse.rewrite(new URL("/account/signup", base));
    }
    // Handle admin routes
    if (
        req.nextUrl.pathname.startsWith("/admin") ||
        req.nextUrl.pathname.startsWith("/api/admin")
    ) {
        return await middleware_admin(req);
    }

    // Check if route starts with /api or not
    if (req.nextUrl.pathname.startsWith("/api")) {
        return await middleware_api(req);
    } else {
        return await middleware_ui(req);
    }
}

/** Middleware function that checks whether the user is authenticated or has
 * provided a valid token in the header. If neither is true, returns a
 * 401 Unauthorized response. Otherwise, passes the request to the next
 * middleware in the chain.
 *
 * Keep in mind this does not verify the token, it only checks if it exists.
 *
 */
async function middleware_api(req: NextRequest): Promise<NextResponse> {
    // There are two ways to access the api
    // 1. The user is logged in and has a session
    // 2. The user provides a token in the header

    const response = NextResponse.next();
    const base = req.url;
    const session = await getIronSession(req, response, sessionOptions);
    const { user_id, email_verified } = session;
    const valid_user = user_id && email_verified;
    const token = req.headers.get("authorization")?.split(" ")[1];

    // Neither a valid user nor a token is provided
    if (!valid_user && !token) {
        return NextResponse.rewrite(new URL(`/api/unauthorized`, base));
    }

    // If the user is logged in and verified
    return response;
}

/** Middleware function that checks whether the user is logged in and verified.
 * If the user is not logged in, redirects the user to the login page with the
 * current URL as the redirect parameter. If the user is logged in but not
 * verified, rewrites the URL to the account verification page.
 *
 */
async function middleware_ui(req: NextRequest): Promise<NextResponse> {
    // Check if the user is logged in
    const response = NextResponse.next();
    const base = req.url;
    const session = await getIronSession(req, response, sessionOptions);
    const { user_id, email_verified } = session;
    //Rewrite for index page
    if ((req.nextUrl.pathname === "/" && !user_id)) {
        return NextResponse.redirect(new URL(`/frontpage`, base));
    } else if (req.nextUrl.pathname === "/" && user_id) {
        return NextResponse.redirect(new URL(`/books`, base));
    }

    // Special check for the books route if token is in header
    if (req.nextUrl.pathname.startsWith("/books/")) {
        const token = req.nextUrl.searchParams.get("token");
        if (token) {
            // Token is checked in the api route
            return response;
        }
    }

    // No user_id means the user is not logged in
    if (!user_id) {
        const redirect = req.nextUrl.pathname;
        let url_string = `/account/login`;
        if (redirect !== "/") {
            url_string += `?redirect=${redirect}`;
        }
        return NextResponse.redirect(new URL(url_string, base));
    }

    // No email_verified means the user is not verified
    if (!email_verified) {
        return NextResponse.rewrite(new URL(`/account/verify`, base));
    }

    // If the user is logged in and verified
    return response;
}

async function middleware_admin(req: NextRequest): Promise<NextResponse> {
    // Check if the user is logged in and is an admin
    const response = NextResponse.next();
    const base = req.url;
    const session = await getIronSession(req, response, sessionOptions);
    const { user_id, email_verified, is_admin } = session;

    // No user_id means the user is not logged in
    if (!user_id) {
        const redirect = req.nextUrl.pathname;
        let url_string = `/account/login`;
        if (redirect !== "/") {
            url_string += `?redirect=${redirect}`;
        }
        return NextResponse.redirect(new URL(url_string, base));
    }

    if (!email_verified) {
        return NextResponse.rewrite(new URL(`/account/verify`, base));
    }

    if (!is_admin) {
        return NextResponse.redirect(new URL("/unauthorized", base));
    }
}

/** The middleware matches all
 * routes that are not in the opt_out array
 * "account/login",
 * "account/signup",
 * "account/resetPassword",
 * "account/forgotPassword",
 * "account/verify",
 * "api/account/login",
 * "api/account/signup",
 * "api/account/resetPassword",
 * "api/unauthorized",
 * "api/authentication/verify",
 * "api/authentication/resendVerify",
 * "/docs"
 * "privacy"
 * "imprint"
 * "/unauthorized"
 * "_next/static",
 * "_next/image",
 * "favicon.ico",
 * "frontpage",
 * "/img"
 * "token_expired",
 */

export const config = {
    matcher: [
        "/((?!frontpage|account/login|account/signup|account/resetPassword|account/verify|account/forgotPassword|api/authentication/verify|api/authentication/resendVerify|api/account/login|api/account/signup|api/account/resetPassword|api/unauthorized|unauthorized|docs|img|token_expired|privacy|imprint|_next/static|_next/image|favicon.ico).*)",
    ],
};
