import { config } from "config";
import { verify } from "jsonwebtoken";
import Link from "next/link";
import { service } from "pages/api/_service";
import { useState } from "react";

import { getLayout } from "./login";

import { ConfirmPasswordInput, PasswordInput } from "components/forms/Form";
import { validatePW } from "components/forms/validation";

import styles from "components/forms/Form.module.scss";

export default function ResetPassword({ isValidToken, token, email }) {
    const [password, setPassword] = useState("");
    const [confirmPassword, setConfirmPassword] = useState("");
    const [success, setSuccess] = useState(false);
    const [error, setError] = useState("");

    const handleSubmit = async (event) => {
        event.preventDefault();

        if (password !== confirmPassword) {
            setError("Passwords do not match");
            return;
        }

        const response = await fetch("/api/account/resetPassword", {
            method: "POST",
            headers: {
                "Content-Type": "application/json",
            },
            body: JSON.stringify({
                token,
                email,
                password,
            }),
        }).then((res) => res.json());

        if (response.success) {
            setSuccess(true);
            setPassword("");
            setConfirmPassword("");
        } else {
            setError(response.error);
        }
    };
    const pw_valid = validatePW(password);

    if (!isValidToken) {
        return (
            <div className={styles.form}>
                <div className={styles.header}>Invalid Token!</div>
                <p>The token you provided is invalid. Please try again.</p>
            </div>
        );
    }

    if (success) {
        return (
            <div className={styles.form}>
                <div className={styles.header}>
                    Password successfully reset for {email}
                </div>
                <div className="d-flex align-items-center text-center flex-column">
                    <Link href="/account/login">
                        <button
                            className="btn btn-primary btn-block gradient-custom-2 "
                            style={{ width: "90%" }}
                        >
                            Log in
                        </button>
                    </Link>
                </div>
            </div>
        );
    }

    return (
        <div className="w-100 h-100 d-flex justify-content-center align-items-center">
            <div className={styles.form}>
                {error && <p>{error}</p>}
                <form onSubmit={handleSubmit}>
                    <fieldset>
                        <div className={styles.header}>Reset your password</div>

                        <PasswordInput
                            isValid={pw_valid}
                            isInvalid={!pw_valid}
                            onInput={(e) => setPassword(e.target.value)}
                        />
                        <ConfirmPasswordInput
                            isValid={password == confirmPassword && pw_valid}
                            isInvalid={password != confirmPassword || !pw_valid}
                            onInput={(e) => setConfirmPassword(e.target.value)}
                        />

                        <div className="d-flex align-items-center text-center mb-3 flex-column">
                            <button
                                className="btn btn-primary btn-block fa-lg gradient-custom-2 mb-2"
                                style={{ width: "90%" }}
                                type="submit"
                            >
                                Reset Password
                            </button>
                        </div>
                    </fieldset>
                </form>
            </div>
        </div>
    );
}

ResetPassword.getLayout = getLayout;

export async function getServerSideProps({ query }) {
    let isValidToken = false;
    if (!query) {
        return {
            props: {
                isValidToken,
                token: "",
                email: "",
            },
        };
    }
    const { token, email } = query;
    if (!token || !email) {
        return {
            props: {
                isValidToken,
                token: "",
                email: "",
            },
        };
    }
    try {
        const pw_hash = await service.user.getPasswordHash(email);
        const payload = verify(token, config.EMAIL.TOKEN_SECRET + pw_hash) as {
            email: string;
            type: string;
        };
        if (payload.email == email && payload.type === "pw_reset") {
            isValidToken = true;
        }
    } catch (err) {
        isValidToken = false;
    }

    return {
        props: {
            isValidToken,
            token,
            email,
        },
    };
}
