import { withSessionRoute } from "lib/withSession";
import { NextApiRequest, NextApiResponse } from "next/types";

import { ApiError, service } from "../_service";

import { FONT_URLS } from "database/fonts";
import { UserConfigData } from "database/services/service.typings";
import { pool_permissions as pool } from "database/sql.connection";

export default withSessionRoute(userRoute);

//Gets user via session cookie
async function userRoute(
    req: NextApiRequest,
    res: NextApiResponse<UserConfigData | Partial<UserConfigData> | ApiError>
) {
    if (req.method == "GET") {
        return configGet(req, res);
    } else if (req.method == "POST") {
        return configPost(req, res);
    } else {
        res.status(405).json({ error: "Method not allowed" });
    }

}


interface PostApiRequest extends NextApiRequest {
    body: Partial<UserConfigData>
}


async function configGet(
    req: NextApiRequest,
    res: NextApiResponse<UserConfigData | ApiError>
) {
    const config = await service.config.getByUserId(req.session.user_id);
    res.json(config);
}

async function configPost(
    req: PostApiRequest,
    res: NextApiResponse<Partial<UserConfigData> | ApiError>
) {
    // Check if any updates are given
    if (Object.keys(req.body).length === 0) {
        res.status(400).json({
            error: "Bad Request",
            details: "No updates given",
        });
        return;
    }

    // Validate
    const { valid, errors, config } = await validateConfig(req.body);
    if (Object.keys(config).length === 0) {
        res.status(400).json({
            error: "Bad Request",
            details: "No valid updates" + (errors.length > 0 ? ": " + errors.join(", ") : ""),
        });
        return;
    }
    if (!valid) {
        res.status(400).json({
            error: "Bad Request",
            details: errors,
        });
        return;
    }

    // Update
    const success = await service.config.updateByUserId(req.session.user_id, config).catch((err) => {
        res.status(500).json({
            error: "Internal Server Error",
            details: err,
        });
        return false;
    });

    if (success) {
        res.status(200).json(config);
    }
}


type validateConfigResult = {
    valid: boolean,
    errors: string[],
    config: Partial<UserConfigData>
}

async function validateConfig(config: Partial<UserConfigData>): Promise<validateConfigResult> {
    // Iterate all keys and validate
    const errors: string[] = [];
    let valid = true;

    // Remove all keys that are not allowed
    const allowedKeys = [
        ["zoom1", "number"],
        ["zoom2", "number"],
        ["text_lineWrap", "number"],
        ["text_lineHeight", "number"],
        ["text_fontSize", "number"],
        ["text_fontColor", "string"],
        ["text_font", "string"],
        ["pen_size", "number"],
        ["pen_color", "string"],
        ["pen_smoothing", "number"]
    ];

    // Remove all keys that are not allowed
    for (const key in config) {
        if (!allowedKeys.find((allowedKey) => allowedKey[0] == key)) {
            errors.push("key " + key + " is not allowed");
            delete config[key];
        }
    }

    // Check if exists and is of correct type
    for (const [key, type] of allowedKeys) {
        if (config[key] && typeof config[key] !== type) {
            valid = false;
            errors.push("typeof " + key + " must be a " + type + ", found: " + typeof config[key]);
        }
    }

    // Special case for pen_color (hex)
    if (config.pen_color && !config.pen_color.match(/^#[0-9A-F]{6}$/i)) {
        valid = false;
        errors.push("pen_color must be a hex color, found: " + config.pen_color);
    }

    // Special case for text_fontColor (hex)
    if (config.text_fontColor && !config.text_fontColor.match(/^#[0-9A-F]{6}$/i)) {
        valid = false;
        errors.push("text_fontColor must be a hex color, found: " + config.text_fontColor);
    }

    // Special case for text_font (string)
    if (config.text_font && typeof config.text_font == "string") {
        //check if font is in FONT_URLS
        const found = FONT_URLS.find((font) => font[1] == config.text_font);

        if (!found) {
            valid = false;
            errors.push("text_font must be a valid font, found: " + config.text_font + " (not in (" + FONT_URLS.map((font) => font[1]).join(", ") + "))");
        } else {
            // Get id and delete font name
            await pool.query("SELECT id FROM fonts WHERE name = ?", [config.text_font]).then((result) => {
                config.text_font_id = result[0].id;
                delete config["text_font"];
            }).catch(() => {
                valid = false;
                errors.push("text_font must be a valid font, found: " + config.text_font + " (not in (" + FONT_URLS.map((font) => font[1]).join(", ") + "))");
            });
        }
    }

    // Special case for pen_smoothing (number) [0-1]
    if (config.pen_smoothing && typeof config.pen_smoothing == "number") {
        if (config.pen_smoothing < 0 || config.pen_smoothing > 1) {
            valid = false;
            errors.push("pen_smoothing must be a number between 0 and 1, found: " + config.pen_smoothing);
        }
    }


    return { valid, errors, config };
}
