import { execFile } from "child_process";
import { config } from "config";
import { verify } from "jsonwebtoken";
import { NextApiRequest, NextApiResponse } from "next";

import { ApiError, service } from "../_service";

/** This api can do two thing
 *
 * 1. Send a reset password email to the user
 * 2. Reset the password of the user if the token is valid
 */
export default async function resetPassword(
    req: NextApiRequest,
    res: NextApiResponse<ApiError | { success: boolean; details: string }>
) {
    // Check if post request
    if (req.method !== "POST") {
        res.status(405).json({ error: "Method not allowed" });
        return;
    }

    // >>> Rate limiting >>>
    // Rate limiting minute
    const requestsLastMinute = requestsLastX(req, 60);

    if (requestsLastMinute > 2) {
        res.status(429).json({
            error: "Too Many Requests",
            details: "Too many requests in the last minute",
        });
        return;
    }

    // Rate limiting hour
    const requestsLastHour = requestsLastX(req, 3600);

    if (requestsLastHour > 5) {
        res.status(429).json({
            error: "Too Many Requests",
            details: "Too many requests in the last hour",
        });
        return;
    }

    // Rate last 5 seconds
    const lastRequestTime = lastRequest(req);
    if (lastRequestTime) {
        const timeSinceLastRequest =
            new Date().getTime() - lastRequestTime.getTime();
        if (timeSinceLastRequest < 5000) {
            res.status(429).json({
                error: "Too Many Requests",
                details: "Please wait 5s between each request",
            });
            return;
        }
    }

    // Delete all requests older than 1 hour
    ipRequests.forEach((value, key) => {
        const now = new Date().getTime();
        ipRequests.set(
            key,
            value.filter((timestamp) => timestamp.getTime() > now - 3600 * 1000)
        );
    });
    // Add request to ipRequests
    const clientIp = req.socket.remoteAddress;
    if (ipRequests.has(clientIp)) {
        ipRequests.get(clientIp).push(new Date());
    } else {
        ipRequests.set(clientIp, [new Date()]);
    }
    // <<< Rate limiting <<<

    /** Two possible inputs
     * 1. email only
     * 2. email, token, password
     *
     * 1. Send email with token
     * 2. Check token and reset password
     */
    const { email, token, password } = req.body;

    // 1. send email with token

    // Get email from request
    if (!email) {
        res.status(400).json({
            error: "Bad Request",
            details: "Email is required",
        });
        return;
    }

    if (!token && !password) {
        const success = await service.user
            .getByEmail(email)
            .then(async (userData) => {
                const pw_hash = await service.user.getPasswordHash(email);
                send_password_reset_email(userData.email, pw_hash);
                return true;
            })
            .catch((err) => {
                console.error(err);
                return false;
            });
        res.json({
            success,
            details: "Email sent if it exists",
        });
        return;
    } else {
        // 2. Check token and reset password
        if (!token || !password) {
            res.status(400).json({
                error: "Bad Request",
                details: "Token and password are required",
            });
            return;
        }

        // Validate token
        const success = await service.user
            .getByEmail(email)
            .then(async (userData) => {
                const pw_hash = await service.user.getPasswordHash(email);
                const payload = verify(
                    token,
                    config.EMAIL.TOKEN_SECRET + pw_hash
                ) as {
                    email: string;
                    type: string;
                };
                let success = false;
                if (payload.email == email && payload.type === "pw_reset") {
                    // valid!
                    success = await service.user.updatePassword(
                        userData.email,
                        password
                    );
                }
                return success;
            })
            .catch((err) => {
                console.error(err);
                return false;
            });
        res.json({
            success,
            details: "Password reset",
        });
    }
}

// Track IP addresses and their request counts
const ipRequests = new Map<string, Date[]>();

/** Get requests in the last x seconds
 * @param req
 * @param x number of seconds
 *
 * @returns number of requests in the last x seconds
 */
function requestsLastX(req: NextApiRequest, x: number) {
    // Get ip  address
    // Get the client's IP address
    const clientIp = req.socket.remoteAddress;

    // Get the current time
    const now = new Date().getTime();

    // Get the number of requests made by this IP address in the last hour
    const requestsThisHour = ipRequests.has(clientIp)
        ? ipRequests
              .get(clientIp)
              .filter((timestamp) => timestamp.getTime() > now - x * 1000)
              .length
        : 0;

    return requestsThisHour;
}

/** Get last timestamp of client
 * @param req
 */
function lastRequest(req: NextApiRequest) {
    // Get ip  address
    // Get the client's IP address
    const clientIp = req.socket.remoteAddress;

    // Get the last request made by this IP address
    const lastRequest = ipRequests.has(clientIp)
        ? ipRequests.get(clientIp).at(-1)
        : undefined;

    return lastRequest;
}

function send_password_reset_email(email: string, pw_hash: string) {
    const filename = "lib/email/send_pw_reset_email.ts";
    // Execute with ts-node
    execFile(
        "ts-node",
        [filename, email, pw_hash],
        { cwd: "app/" },
        (err, stdout) => {
            if (err) {
                console.error(err);
                return;
            }
            console.log(stdout);
        }
    );
}
