import { NextApiRequest, NextApiResponse } from "next/types";

import { ApiError, service } from "../_service";

import {
    Entity,
    ENTITY_GROUP,
    ENTITY_USER,
    PermissionACLData,
    Resource,
    RESOURCE_BOOK,
} from "database/services/service.typings";

/** Get or set acl for entity and resource
 *
 */
export default async function acl(
    req: NextApiRequest,
    res: NextApiResponse<PermissionACLData[] | ApiError>
) {
    //Check method
    if (req.method !== "POST") {
        res.status(405).json({
            error: "Method Not Allowed",
            details: "Only POST requests are allowed!",
        });
        return;
    }
    // Check if parameters are given
    const { entity_id, resource_id, entity_type, resource_type } = req.body;
    if (!entity_id || !resource_id || !entity_type || !resource_type) {
        res.status(400).json({
            error: "Bad Request",
            details:
                "Missing parameters entity_id, resource_id, entity_type and/or resource_type!",
        });
        return;
    }
    // Parse entity_id and resource_id to integer
    const entity_id_int = parseInt(entity_id);
    const resource_id_int = parseInt(resource_id);
    // Check if entity_id and resource_id are valid integers
    if (isNaN(entity_id_int) || isNaN(resource_id_int)) {
        res.status(400).json({
            error: "Bad Request",
            details: "entity_id and resource_id must be valid integers!",
        });
        return;
    }
    //  Parse entity_type to int
    let entity_type_int: Entity;
    switch (entity_type) {
        case "user":
            entity_type_int = ENTITY_USER;
            break;
        case "group":
            entity_type_int = ENTITY_GROUP;
            break;
        default:
            res.status(400).json({
                error: "Bad Request",
                details: "entity_type must be in [ user, group ]!",
            });
            return;
    }

    // Parse resource_type to int
    let resource_type_int: Resource;
    switch (resource_type) {
        case "book":
            resource_type_int = RESOURCE_BOOK;
            break;
        default:
            res.status(400).json({
                error: "Bad Request",
                details: "resource_type must be in [ book ]!",
            });
            return;
    }

    // Get acl
    const aclData = await service.permission.getACLByEntityAndResource(
        entity_id_int,
        resource_id_int,
        entity_type_int,
        resource_type_int
    );

    // Return acl data
    res.status(200).json([aclData]);
}
