import { NextApiRequest, NextApiResponse } from "next/types";

import { ApiError, service } from "../_service";

import {
    BookData,
    GroupData,
    UserData,
} from "database/services/service.typings";

// Allowed search types for the admin api
const allowedTypes = ["user", "group", "book"] as const;
export type AllowedTypes = typeof allowedTypes[number];

// Allowed search results for the admin api
export type AllowedReturn = UserData | GroupData | BookData;

// Defines the search query
export interface SearchQuery {
    query: string;
    type: AllowedTypes;
}

export interface SearchResults {
    type: AllowedTypes;
    data: Array<AllowedReturn>;
}

export default async function searchForEntity(
    req: NextApiRequest,
    res: NextApiResponse<SearchResults | ApiError>
) {
    // Check if parameters are given
    const { search, type } = req.body;
    if (!search || !type) {
        res.status(400).json({
            error: "Bad Request",
            details: "Missing parameters search and/or type!",
        });
        return;
    }

    // Parse search to string
    const search_query = String(search);

    let results: SearchResults;

    switch (type) {
        case "user":
            results = {
                type: "user",
                data: await service.user.search(search_query),
            };
            break;
        case "group":
            results = {
                type: "group",
                data: await service.group.search(search_query),
            };
            break;
        case "book":
            results = {
                type: "book",
                data: await service.book.search(search_query),
            };
            break;
        default:
            res.status(400).json({
                error: "Bad Request",
                details: "Type must be in [" + allowedTypes.join(", ") + "]!",
            });
            return;
    }

    res.json(results);
}
