import { NextApiRequest, NextApiResponse } from "next/types";

import { ApiError, service } from "../_service";

import {
    Entity,
    ENTITY_GROUP,
    ENTITY_USER,
    PermissionACLData,
    Resource,
    RESOURCE_BOOK,
} from "database/services/service.typings";

const parseResourceType = (resource_type: string | Entity): Resource => {
    switch (resource_type) {
        case "book":
            return RESOURCE_BOOK;
        case RESOURCE_BOOK:
            return resource_type;
        default:
            throw new Error("resource_type must be in [ book ]!");
    }
};

const parseEntityType = (entity_type: string | Entity): Entity => {
    // Parse string or enume type to enum type
    switch (entity_type) {
        case "user":
            return ENTITY_USER;
        case "group":
            return ENTITY_GROUP;
        case ENTITY_USER:
        case ENTITY_GROUP:
            return entity_type;
        default:
            throw new Error("entity_type must be in [ user, group ]!");
    }
};

export const parseBoolean = (value: number | string | boolean): boolean => {
    if (typeof value === "boolean") {
        return value;
    }
    if (typeof value === "string") {
        return value === "true";
    }
    if (typeof value === "number") {
        return value === 1;
    }
    throw new Error("value must be a boolean, string or number!");
};

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export function parseACL(acl: Array<any>): PermissionACLData[] {
    const acl_parsed = acl.map((acl) => {
        let { id } = acl;
        const { entity_id, resource_id, entity_type, resource_type } = acl;

        if (!entity_id || !resource_id || !entity_type || !resource_type) {
            throw new Error(
                "Missing parameters entity_id, resource_id, entity_type and/or resource_type!"
            );
        }

        if (!id || isNaN(id)) {
            id = -1;
        }

        // Parse entity_id and resource_id to integer
        const id_int = parseInt(id);
        const entity_id_int = parseInt(entity_id);
        const resource_id_int = parseInt(resource_id);
        // Check if entity_id and resource_id are valid integers
        if (isNaN(entity_id_int) || isNaN(resource_id_int)) {
            throw new Error(
                "entity_id and resource_id must be valid integers!"
            );
        }
        //  Parse entity_type to int
        const entity_type_int = parseEntityType(entity_type);
        //  Parse resource_type to int
        const resource_type_int = parseResourceType(resource_type);

        // Parse permissions i.e. parse to boolean
        const { pRead, pWrite, pDelete, pACL } = acl;
        const pRead_bool = parseBoolean(pRead);
        const pWrite_bool = parseBoolean(pWrite);
        const pDelete_bool = parseBoolean(pDelete);
        const pACL_bool = parseBoolean(pACL);

        return {
            id: id_int,
            entity_id: entity_id_int,
            resource_id: resource_id_int,
            entity_type: entity_type_int,
            resource_type: resource_type_int,
            pRead: pRead_bool,
            pWrite: pWrite_bool,
            pDelete: pDelete_bool,
            pACL: pACL_bool,
        };
    });
    return acl_parsed;
}

export default async function acl(
    req: NextApiRequest,
    res: NextApiResponse<boolean | ApiError>
) {
    //Check method
    if (req.method !== "POST") {
        res.status(405).json({
            error: "Method Not Allowed",
            details: "Only POST requests are allowed!",
        });
        return;
    }

    // Get ACL
    const { acl } = req.body;

    // Check if valid ACL
    if (!acl && !Array.isArray(acl)) {
        res.status(400).json({
            error: "Bad Request",
            details: "Missing parameter acl!",
        });
        return;
    }

    // Parse ACL
    try {
        const acl_parsed = parseACL(acl);
        // Set ACL
        const successes = acl_parsed.map(async (acl) => {
            const {
                entity_id,
                resource_id,
                entity_type,
                resource_type,
                pRead,
                pWrite,
                pDelete,
                pACL,
            } = acl;
            return await service.permission.setACLByEntityAndResource(
                entity_id,
                entity_type,
                resource_id,
                resource_type,
                pRead,
                pWrite,
                pDelete,
                pACL
            );
        });

        if (successes.every((success) => success)) {
            res.status(200).json(true);
        } else {
            res.status(500).json({
                error: "Internal Server Error",
                details: "Failed to set ACL!",
            });
        }
    } catch (error) {
        res.status(400).json({
            error: "Bad Request",
            details: error.message,
        });
    }
}
