import { NextApiRequest, NextApiResponse } from "next/types";

import { ApiError } from "../_service";
import { parseBoolean } from "./set_acl";

import { pool_permissions } from "database/sql.connection";

import { UsersGroups } from "components/admin_panel/utils";

// eslint-disable-next-line @typescript-eslint/no-explicit-any
const parseUserGroups = (user_groups: Array<any>): UsersGroups[] => {
    const user_groups_parsed = user_groups.map((user_group) => {
        const { group_id, user_id } = user_group;
        if (!group_id || !user_id) {
            throw new Error("Missing parameters group_id and/or user_id!");
        }
        const group_id_int = parseInt(group_id);
        const user_id_int = parseInt(user_id);
        if (isNaN(group_id_int) || isNaN(user_id_int)) {
            throw new Error("group_id and user_id must be valid integers!");
        }

        // Parse member
        const member = parseBoolean(user_group.member);

        return {
            group_id: group_id_int,
            user_id: user_id_int,
            member,
        };
    });
    return user_groups_parsed;
};

export default async function users_groups(
    req: NextApiRequest,
    res: NextApiResponse<boolean | ApiError>
) {
    if (req.method !== "POST") {
        res.status(405).json({
            error: "Method Not Allowed",
            details: "Only POST requests are allowed!",
        });
        return;
    }

    const { users_groups } = req.body;

    if (!users_groups || !Array.isArray(users_groups)) {
        res.status(400).json({
            error: "Bad Request",
            details: "Missing parameter users_groups!",
        });
        return;
    }

    try {
        const users_groups_parsed = parseUserGroups(users_groups);

        // Add
        const users_groups_add = users_groups_parsed
            .filter((user_group) => user_group.member)
            .map((ug) => {
                return `(${ug.group_id}, ${ug.user_id})`;
            });

        const users_groups_remove = users_groups_parsed
            .filter((user_group) => !user_group.member)
            .map((ug) => {
                return `(${ug.group_id}, ${ug.user_id})`;
            });

        // Remove all with member is false and add all with member is true
        // to junction table
        const sql_add = `
            INSERT INTO users_groups (group_id, user_id)
            VALUES ${users_groups_add.join(", ")}
            ON DUPLICATE KEY UPDATE group_id = group_id;
        `;
        const sql_remove = `
            DELETE FROM users_groups WHERE (group_id, user_id)
            IN (${users_groups_remove.join(", ")});
        `;

        if (users_groups_add.length > 0) {
            await pool_permissions.query(sql_add).then((res) => {
                if (res.warningStatus > 1) {
                    throw new Error("None found");
                }
                // As the database structure matches our BookData interface
                // we can just return the first row
                return true;
            });
        }
        if (users_groups_remove.length > 0) {
            await pool_permissions.query(sql_remove).then((res) => {
                if (res.warningStatus > 1) {
                    throw new Error("None found");
                }
                // As the database structure matches our BookData interface
                // we can just return the first row
                return true;
            });
        }

        res.status(200).json(true);
        return;
    } catch (err) {
        res.status(500).json({
            error: "Internal Server Error",
            details: err.message,
        });
        return;
    }
}
