import { config } from "config";
import { JwtPayload, verify } from "jsonwebtoken";
import { withSessionRoute } from "lib/withSession";
import { NextApiRequest, NextApiResponse } from "next/types";

import { ApiError, service } from "../_service";

export default withSessionRoute(verifyRoute);

async function verifyRoute(
    req: NextApiRequest,
    res: NextApiResponse<ApiError | VerifyPayload>
) {
    // Check if post request
    if (req.method !== "POST") {
        res.status(405).json({ error: "Method not allowed" });
        return;
    }

    const { token } = await req.body;

    if (!token) {
        res.status(400).json({
            error: "Bad Request",
            details: "Token is required",
        });
        return;
    }

    // decode token
    const payload = await verify_token(token);

    // Update session cookie if token is valid
    if (payload.success && req.session.user_id) {
        req.session.email_verified = true;
        await req.session.save();
    }

    res.json(payload);
}

export interface VerifyPayload {
    success: boolean;
    error?: string;
    email?: string;
}

/**Verifies a token and returns the corresponding
 * email address of the user.
 *
 * @param token The token to verify
 * @returns The email address of the user
 * @throws Error if token is invalid
 */
export async function verify_token(token: string): Promise<VerifyPayload> {
    const email = new Promise<string>((resolve, reject) => {
        // verify a token and get the email address
        try {
            const decoded = verify(token, config.EMAIL.TOKEN_SECRET);
            resolve((decoded as JwtPayload).email);
        } catch (err) {
            reject("Token is invalid");
        }
    });

    // check if email is in db
    const payload = new Promise<VerifyPayload>((resolve,) => {
        email
            .then(async (email) => {
                return {
                    success: await service.user.verify(email),
                    email: email,
                };
            })
            .then(({ success, email }) => {
                if (success) {
                    resolve({ success: true, email });
                } else {
                    throw "Email already verified";
                }
            })
            .catch((err) => {
                resolve({ success: false, error: err });
            });
    });

    return payload;
}
