import { withSessionRoute } from "lib/withSession";
import { NextApiRequest, NextApiResponse } from "next/types";

import { service } from "../_service";

import { RESOURCE_BOOK } from "database/services/service.typings";
import { pool_data } from "database/sql.connection";

/** Returns fileMap for the books browser on the main
 * page.
 */

export default withSessionRoute(fileMapRoute);

async function fileMapRoute(req: NextApiRequest, res: NextApiResponse) {
    const fileData = await getFileMapByUser(req.session.user_id).catch((e) => {
        console.error(e);
        res.status(500).json({
            error: "Internal server error",
            details: e,
        });
        return;
    });
    if (!fileData) return;

    // Transform array to dict and a
    const fileMap: FileMap = {};
    fileData.forEach((f) => {
        if (!f.parentId) {
            f.parentId = "root";
        }
        fileMap[f.id] = f;
    });
    // Add root folder
    fileMap["root"] = {
        id: "root",
        name: "Home",
        isDir: true,
        parentId: null,
        numPages: null,
        created: null,
        last_updated: null,
        finished: null,
    };

    res.status(200).json(fileMap);
}

export async function getFileMapByUser(user_id: number): Promise<FileData[]> {
    //Get books for user, i.e. all books the user has access to
    const acl = await service.permission.getACLByEntity(user_id);
    const books_with_pRead = acl.filter(
        (a) => a.resource_type == RESOURCE_BOOK && a.pRead
    );
    const book_ids = books_with_pRead.map((a) => a.resource_id);
    if (book_ids.length == 0) {
        return [];
    }

    // Get book and folder data
    const sql = `
        SELECT 
            CAST(B.id AS CHAR) as id,
            B.title as name,
            B.comment,
        
            false as isDir,
            CASE
                WHEN BF.browserFolder_id IS NOT NULL THEN CONCAT('folder_', CAST(BF.browserFolder_id AS CHAR))
                ELSE NULL
            END AS parentId,

            getNumPages(B.id) as numPages,
            B.created,
            B.last_updated,
            B.finished,
            B.owner_name as ownerName
        FROM books_resolved as B
        LEFT JOIN browserFolders_books as BF ON BF.book_id = B.id AND BF.user_id = ? WHERE B.id IN (?) 
        UNION
        SELECT
            CONCAT('folder_', CAST(F.id AS CHAR)) as id,
            F.name,
            F.comment,
            true as isDir,
            F.parent_id as parentId,

            NULL as numPages,
            F.created,
            NULL as last_updated,
            NULL as finished,
            U.email as ownerName
        FROM browserFolders as F
        LEFT JOIN snip_perms.users as U ON U.id = F.user_id
        WHERE F.user_id = ?;
    `;

    // Attach acl to books
    const file_data = await pool_data
        .query(sql, [user_id, book_ids, user_id])
        .then((r) => r as FileData[]);

    // Attach acl to books
    const file_data_with_acl = file_data.map((f) => {
        const acl = books_with_pRead.find((a) => a.resource_id == Number(f.id));
        if (acl) {
            f.perms = {
                pRead: acl.pRead,
                pWrite: acl.pWrite,
                pDelete: acl.pDelete,
                pACL: acl.pACL,
            };
        }
        return f;
    });
    return file_data_with_acl;
}
