
import { withSessionRoute } from "lib/withSession";
import type { NextApiRequest, NextApiResponse } from "next/types";

import { ApiError } from "../_service";
import { validateItemId } from "./move";

import { pool_data } from "database/sql.connection";


interface DeleteFolderRequest extends NextApiRequest {
    body: {
        folderID: number;
    };
}
export default withSessionRoute(deleteFolder);

async function deleteFolder(
    req: DeleteFolderRequest,
    res: NextApiResponse<{ success: boolean } | ApiError>
) {
    if (req.method !== "POST") {
        res.status(405).json({ error: "Method Not Allowed" });
        return;
    }

    // Check if user is logged in
    if (!req.session.user_id) {
        res.status(401).json({ error: "Unauthorized" });
        return;
    }

    const { folderID } = await req.body;

    /** Validate
     */
    try {
        await validateItemId("folder", folderID, req.session.user_id);
    } catch (error) {
        res.status(400).json({ error: error.message });
        return;
    }

    // Delete folder
    try {
        await deleteFolderFromDatabase(folderID);
    } catch (error) {
        res.status(500).json({ error: error.message });
        return;
    }
    res.status(200).json({ success: true });
    return;
}

/** Deletes an folder from the database and also moves all 
 * items in the folder to the parent folder.
 */
async function deleteFolderFromDatabase(folderID: number) {
    console.log("Deleting folder", folderID);


    // Move first
    const sql_folders = `
        UPDATE browserFolders SET parent_id = (
            SELECT parent_id FROM browserFolders WHERE id = ?
        ) WHERE parent_id = ?;
    `;
    const sql_books_del_null = `
        DELETE FROM browserFolders_books
        WHERE browserFolder_id = ?
        AND (
            SELECT parent_id FROM browserFolders WHERE id = ?
        ) IS NULL;
    `;
    const sql_books = `
        UPDATE browserFolders_books SET browserFolder_id = (
            SELECT parent_id FROM browserFolders WHERE id = ?
        ) WHERE browserFolder_id = ?;
    `;


    await pool_data.query(sql_folders, [folderID, folderID]);
    await pool_data.query(sql_books_del_null, [folderID, folderID]);
    await pool_data.query(sql_books, [folderID, folderID]);
    // Delete folder
    const sql2 = `
        DELETE FROM browserFolders WHERE id = ?;
    `;
    await pool_data.query(sql2, [folderID]);
}


