import { withSessionRoute } from "lib/withSession";
import type { NextApiRequest, NextApiResponse } from "next/types";

import { ApiError } from "../_service";

import { pool_data } from "database/sql.connection";

interface MoveItemRequest extends NextApiRequest {
    body: {
        type: "folder" | "book";
        itemID: number;
        moveToID: number | "root";
    };
}

export default withSessionRoute(moveItem);

async function moveItem(
    req: MoveItemRequest,
    res: NextApiResponse<{ success: boolean } | ApiError>
) {

    if (req.method !== "POST") {
        res.status(405).json({ error: "Method Not Allowed" });
        return;
    }

    // Check if user is logged in
    if (!req.session.user_id) {
        res.status(401).json({ error: "Unauthorized" });
        return;
    }

    const { type, itemID, moveToID } = await req.body;


    /** Validate
     */
    try {
        await validateItemId(type, itemID, req.session.user_id);
        if (moveToID !== "root")
            await validateItemId("folder", moveToID, req.session.user_id);
    } catch (error) {
        res.status(400).json({ error: error.message });
        return;
    }

    // Move item
    try {
        await moveItemToFolder(type, itemID, moveToID, req.session.user_id);
    }
    catch (error) {
        res.status(500).json({ error: error.message });
        return;
    }
    res.status(200).json({ success: true });
    return;
}

export async function validateItemId(type: "folder" | "book", itemID: number, user_id: number) {


    if (type === "folder") {
        const sql = "SELECT 1 FROM browserFolders WHERE id = ? AND user_id = ?";
        const folder = await pool_data.query(sql, [itemID, user_id]);
        if (!folder) {
            throw new Error(`Folder does not exist`);
        }
        return;
    } else if (type === "book") {
        const sql = "SELECT 1 FROM books WHERE id = ?";
        const book = await pool_data.query(sql, [itemID]);
        if (!book) {
            throw new Error(`Book does not exist`);
        }
        return;
    } else {
        throw new Error(`Invalid type, must be 'folder' or 'book'.`);
    }
}

async function moveItemToFolder(type: "folder" | "book", itemID: number, moveToID: number | "root", user_id: number) {

    if (moveToID === "root") {
        moveToID = null;
    }
    if (type === "folder") {
        const sql = "UPDATE browserFolders SET parent_id = ? WHERE id = ?";
        await pool_data.query(sql, [moveToID, itemID]);
        return;
    } else if (type === "book" && moveToID !== null) {
        const sql = "INSERT INTO browserFolders_books (user_id, browserFolder_id, book_id) VALUES (?, ?, ?) ON DUPLICATE KEY UPDATE browserFolder_id = ?";
        await pool_data.query(sql, [user_id, moveToID, itemID, moveToID]);
        return;
    } else if (type === "book" && moveToID === null) {
        const sql = "DELETE FROM browserFolders_books WHERE user_id = ? AND book_id = ?";
        await pool_data.query(sql, [user_id, itemID]);
        return;
    } else {
        throw new Error(`Invalid type, must be 'folder' or 'book'.`);
    }
}

