import { withSessionRoute } from "lib/withSession";
import type { NextApiRequest, NextApiResponse } from "next/types";

import { ApiError } from "../_service";

import { pool_data } from "database/sql.connection";

interface NewFolderRequest extends NextApiRequest {
    body: {
        name: string;
        comment?: string;
        parentID?: string;
    };
}


export default withSessionRoute(createFolder);


async function createFolder(
    req: NewFolderRequest,
    res: NextApiResponse<{ id: number } | ApiError>
) {

    if (req.method !== "POST") {
        res.status(405).json({ error: "Method Not Allowed" });
        return;
    }

    // Check if user is logged in
    if (!req.session.user_id) {
        res.status(401).json({ error: "Unauthorized" });
        return;
    }

    let { name, comment, parentID } = await req.body;
    name = name || null;
    comment = comment || null;
    parentID = parentID || null;

    /** Validate name 
     */
    try {
        await validateName(name, req.session.user_id);
    } catch (error) {
        res.status(400).json({ error: error.message });
        return;
    }


    /** Validate parentID
     */
    try {
        await validateParent(parentID, req.session.user_id);
    } catch (error) {
        res.status(400).json({ error: error.message });
        return;
    }

    // Insert folder
    const sql = `INSERT INTO browserFolders (name, comment, parent_id, user_id) VALUES (?, ?, ?, ?) RETURNING id`;
    const id = await pool_data.query<{ id: number }>(sql, [name, comment, parentID, req.session.user_id]).catch((error) => {
        console.log(error);
        return null;
    });
    if (!id) {
        res.status(500).json({ error: "Internal Server Error" });
        return;
    }
    res.status(200).json({ id: id });
}


const validateName = async (name: string, user_id: number): Promise<boolean> => {
    // Check if name is empty
    if (!name) {
        throw new Error("Folder name is not set!");
    }
    if (name.length > 2048) {
        throw new Error("Folder name is too long, < 2048 chars!");
    }
    // Check if name exists
    const sql = `SELECT 1 FROM browserFolders WHERE name = ? AND user_id = ?`;
    const folder = await pool_data.query(sql, [name, user_id]).then((result) => {
        return result.length > 0;
    });

    if (folder) {
        throw new Error("Folder name already exists, choose another name!");
    }

    return;
}


export const validateParent = async (parentID: string, user_id: number) => {
    // Check if parentID is empty
    if (!parentID) {
        return;
    }

    // Check if parentID exists and belongs to user
    const sql = `SELECT 1 FROM browserFolders WHERE id = ? AND user_id = ?`;
    const parentFolder = await pool_data.query(sql, [parentID, user_id]);
    if (!parentFolder) {
        throw new Error("Parent folder name does not exist");
    }

    return;
}
