import { withAuthBookId } from "lib/access_control/books";
import { withSessionRoute } from "lib/withSession";
import { NextApiRequest, NextApiResponse } from "next/types";

import { ApiError } from "../_service";

import { PermissionACLData } from "database/services/service.typings";
import { pool_data } from "database/sql.connection";

export default withSessionRoute(withAuthBookId(searchForPages));

/**
 * Search for pages by string and return a list of pageids
 *
 * Parameters:
 * -----------
 * - id: number The book id
 * - search: string The search query
 */
async function searchForPages(
    req: NextApiRequest,
    res: NextApiResponse<number[] | ApiError>,
    perms: PermissionACLData
) {
    const { query } = req;
    const book_id = query.id as string;

    const search = query.search as string;

    if (!search) {
        res.status(400).json({ error: "No search query provided" });
        return;
    }

    // Check read access
    if (!perms.pRead) {
        res.status(401).json({
            error: "Unauthorized",
            details: { id: "You do not have read access to this book!" },
        });
        return;
    }

    // Perform search
    // TODO: JOIN with snips table
    const sql = `
    SELECT id FROM pages 
        WHERE book_id = ? AND (
            page_number + 1  LIKE ? OR 
            id LIKE ? OR 
            created LIKE ? OR
            last_updated LIKE ?
        ); 
    `;

    const page_ids: number[] = await pool_data
        .query(sql, [book_id, search, search, search, search])
        .then((res) => {
            if (res.length < 1) {
                // Return empty list (no pages found)
                return [];
            }
            return (res as { id: number }[]).map((row) => row.id);
        })
        .catch((err) => {
            console.error(err);
            // Return empty list (no pages found)
            return [];
        });

    res.status(200).json(page_ids);
    return;
}
