import { BaseData, BaseSnip, BaseSnipArgs, BaseView, SnipData } from "../general/base";

export interface PlotData extends BaseData {
    xarray: number[];
    yarray: number[];
}

export interface PlotView extends BaseView {
    vwidth: number;
    vheight: number;
    xrange: [number, number];
    yrange: [number, number];
}
export interface ImageSnipArgs extends BaseSnipArgs {
    xarray: number[];
    yarray: number[];
    vwidth: number;
    vheight: number;
    xrange: [number, number];
    yrange: [number, number];
}


export class PlotSnip extends BaseSnip {
    public type = "plot";

    private xarray: number[];
    private yarray: number[];
    private vwidth: number;
    private vheight: number;
    private xrange: [number, number];
    private yrange: [number, number];

    // We can define default values here
    constructor({
        xarray,
        yarray,
        vwidth,
        vheight,
        xrange,
        yrange,
        ...baseArgs
    }: ImageSnipArgs) {
        super({ ...baseArgs });

        if (xarray.length !== yarray.length) {
            throw new Error("xarray and yarray must have the same length");
        }
        this.xarray = xarray;
        this.yarray = yarray;

        if (vwidth < 0) {
            throw new Error("width must be positive");
        }
        this.vwidth = vwidth;

        if (vheight < 0) {
            throw new Error("height must be positive");
        }
        this.vheight = vheight;

        this.xrange = xrange;
        this.yrange = yrange;

    }

    /** Construct a instance of the snip from a
     * SnipData object. This is used to create a snip
     * object from the database data dict.
     *
     * @param {SnipData<PlotData, PlotData>} data - Data dict
     */
    public from_data(data: SnipData<PlotData, PlotView>): PlotSnip {
        return new PlotSnip({
            //Plot
            xarray: data.data.xarray,
            yarray: data.data.yarray,
            vwidth: data.view.vwidth,
            vheight: data.view.vheight,
            xrange: data.view.xrange,
            yrange: data.view.yrange,
            //Base
            x: data.view.x,
            y: data.view.y,
            id: data.id,
            page_id: data.page_id,
            book_id: data.book_id,
            last_updated: data.last_updated,
            created: data.created,
        });
    }


    // As we extend the base snip we do not have to define the to_data method
    // but we have to define the view and data methods
    public view(): PlotView {
        const view = super.view() as PlotView;
        return {
            vwidth: this.vwidth,
            vheight: this.vheight,
            xrange: this.xrange,
            yrange: this.yrange,
            ...view,
        };
    }

    public data(): PlotData {
        return {
            type: this.type,
            xarray: this.xarray,
            yarray: this.yarray,
        };
    }


    /** Renders the plot on the page
     *
     * @param {CanvasRenderingContext2D} ctx - Canvas context
     */
    render(ctx: CanvasRenderingContext2D): void {
        // Render axis
        ctx.beginPath();
        ctx.moveTo(this.x, this.y);
        ctx.lineTo(this.x, this.y + this.height);
        ctx.lineTo(this.x + this.width, this.y + this.height);
        ctx.stroke();

        // Render data
        ctx.beginPath();
        for (let i = 0; i < this.xarray.length; i++) {
            const x = this.xarray[i];
            const y = this.yarray[i];
            ctx.moveTo(this.x + x, this.y + y);
            ctx.arc(this.x + x, this.y + y, 2, 0, 2 * Math.PI);
        }
        ctx.stroke();
    }

    // We can also define a number of methods to interact with the snip
    // e.g. to link external resources
    // these are not required but can be useful and are all prefixed with
    // on_* to indicate that they are event handlers
    // At the moment we support on_click, on_hover, on_mousedown and on_mouseup
    public on_click(event: MouseEvent): void {
    }
    public on_hover(event: MouseEvent): void {
    }
    public on_mousedown(event: MouseEvent): void {
    }
    public on_mouseup(event: MouseEvent): void {
    }
}