import { type BookSocket } from "controller/booksocket";

export interface BaseData {
    type: string;
}
export interface BaseView {
    x: number;
    y: number;
}

export interface SnipData<d extends BaseData, v extends BaseView> {
    id: number | null;
    page_id: number | null;
    book_id: number | null;
    data: d;
    type: string;
    view: v;
    last_updated: Date;
    created: Date;
}

export interface BaseSnipArgs {
    /** x coordinate
     * This is the vertical position where the snip will appear on the page and
     * corresponds to the upper left corner of the snip
     *
     * default: 0
     */
    x?: number;

    /** y coordinate
     * This is the horizontal  position where the snip will appear on the page and
     * corresponds to the upper left corner of the snip
     *
     * default: 0
     * */
    y?: number;

    /** Snip id
     * This is the id of the snip in the database. This is set during the first
     * save of the snip and should not be changed after that.
     * See BaseSnip.to_db()
     *
     * default: random number
     */
    id?: number;

    /** Page id
     * This is the id of the page that the snip is on. This is set during the first
     * save of the snip and should not be changed after that.
     *
     * If this is not set the snip will be a "snip_to_place" snip and will not be
     * placed on a page until it is given a page id.
     * default: null
     */
    page_id?: number;

    /** Book id
     * This is the id of the book that the snip is associated with.
     * This should be set when the snip is created and should not be changed after.
     *
     * If not given expect strange behavior.
     * default: null
     */
    book_id?: number;

    /** Last updated
     * This is the date that the snip was last updated. This is set during every change of
     * the snip via a db trigger. It should not be changed manually.
     *
     * default: new Date()
     */
    readonly last_updated?: Date;

    /** Created
     * This is the date that the snip was created. This is set during the first
     * save of the snip and should not be changed after that.
     *
     * default: new Date()
     */
    readonly created?: Date;

    /** Socket
     * This is the socket that the snip is associated with. Can be parsed if
     * the socket is already created.
     */
    socket?: BookSocket;
}

/** This class has no real features, but is used as an abstract base class by
 *  other snip classes via inheritance. It implements some features
 *  that every snip shares.
 */
export class BaseSnip {
    static description = "Base Snip";
    public type = "base";

    public readonly created: Date;
    public readonly last_updated: Date;

    public socket?: BookSocket;

    constructor({
        x = 0,
        y = 0,
        id = null,
        page_id = null,
        book_id = null,
        last_updated = new Date(),
        created = new Date(),
        socket = null,
    }: BaseSnipArgs) {
        // Parse default values
        this._x = x;
        this._y = y;
        this._id = id || -1 * Math.floor(Math.random() * 100000);
        this._page_id = page_id;
        this._book_id = book_id;
        this.last_updated = new Date(last_updated);
        this.created = new Date(created);

        if (socket) {
            this.socket = socket;
        }
    }

    /** Construct a instance of the snip from a
     * SnipData object. This is used to create a snip
     * object from the database data dict.
     *
     * @param {SnipData<BaseData, BaseView>} data - Data dict
     */
    static from_data(data: SnipData<BaseData, BaseView>): BaseSnip {
        const temp = new BaseSnip({
            x: data.view.x,
            y: data.view.y,
            id: data.id,
            page_id: data.page_id,
            book_id: data.book_id,
            last_updated: data.last_updated,
            created: data.created,
        });
        // We set the type here because this function
        // might be used by another snip class
        temp.type = data.type;
        return temp;
    }

    /** Render function, should be used by parent classes
     * This is the most important function of any snip class
     * as it is used to render the snip on the page.
     *
     * The base snip class does not have much rendering functionality
     * so it is left to the child classes to implement this.
     *
     * This class only implements the basic bounding box plotting.
     *
     * @param {CanvasRenderingContext2D} ctx - Canvas context
     */
    render(ctx: CanvasRenderingContext2D | OffscreenCanvasRenderingContext2D) {
        //Render bounding box
        if (this.show_bounding_box) {
            this.render_bounds(ctx);
        }
    }

    render_bounds(
        ctx: CanvasRenderingContext2D | OffscreenCanvasRenderingContext2D
    ) {
        //Render bounding box
        ctx.strokeStyle = "red";
        ctx.lineWidth = 4;
        ctx.strokeRect(this.x, this.y, this.width, this.height);
    }

    /** Get SnipData object from Snip, this should be used
     * to send the snip to other clients. As its footprint is
     * smaller and it is easily json serializable.
     */
    to_data(): SnipData<BaseData, BaseView> {
        return {
            id: this.id,
            page_id: this._page_id,
            book_id: this.book_id,
            data: this.data(),
            view: this.view(),
            created: this.created,
            last_updated: this.last_updated,
            type: this.type,
        };
    }

    /** Corresponds to data object in the database */
    data(): BaseData {
        return {
            type: this.type,
        };
    }
    /* Corresponds to view object in the database */
    view(): BaseView {
        return {
            x: this._x,
            y: this._y,
        };
    }

    /** There are some special modes which can be used/implemented
     * by the snip. These include overlay mode and bounding box mode.
     * The overlay mode tag is set if the snip is drawn onto the
     * overlay canvas. This is useful if snips are currently placed
     * and additional information needs to be displayed.
     */
    get height(): number {
        return 0;
    }
    get width(): number {
        return 0;
    }
    private _show_bounding_box = false;
    get show_bounding_box(): boolean {
        return this._show_bounding_box;
    }
    set show_bounding_box(value: boolean) {
        this._show_bounding_box = value;
    }
    private _overlay_mode = false;
    get overlay_mode(): boolean {
        return this._overlay_mode;
    }
    set overlay_mode(value: boolean) {
        this._overlay_mode = value;
    }

    /** Setter and getter methods for all basic parameters.
     * These can be used by child classes to implement value
     * checks or other functionality.
     *
     * For example see x and y setters for a use with an event
     * handler.
     */
    private _x: number;
    get x(): number {
        return this._x;
    }
    set x(value: number) {
        this._x = value;
    }

    private _y: number;
    get y(): number {
        return this._y;
    }
    set y(value: number) {
        this._y = value;
    }

    private _id: number;
    get id(): number {
        return this._id;
    }
    set id(value: number) {
        this._id = value;
    }

    protected _page_id: number;
    get page_id(): number {
        return this._page_id;
    }
    set page_id(val: number) {
        this._page_id = val;
    }

    private _book_id: number;
    get book_id(): number {
        return this._book_id;
    }
    set book_id(val: number) {
        this._book_id = val;
    }

    /** Returns polygon of snip
     *
     * might be overwritten by child classes
     * and can be used to check if a point is inside
     *
     */
    get polygon(): [number, number][] {
        return [
            [this.x, this.y],
            [this.x + this.width, this.y],
            [this.x + this.width, this.y + this.height],
            [this.x, this.y + this.height],
        ];
    }

    /** Setup socket connection to server
     * in the base snip this class does nothing
     */
    protected setupSocket(): void {
        return;
    }

    /** Add the snip to the database and retrieve the new id
     * from the database, overwrites the id with the db id.
     *
     * This function always creates a new object in the db,
     * call update_db if you want to update an existing object.
     * Even if the snip id exists! It will create a new object.
     *
     * Returns the id of the snip in the database
     */
    public async insert(): Promise<number> {
        return new Promise<number>((resolve) => {
            this.socket.emit("snip:insert", this.to_data(), (id: number) => {
                this.id = id;
                resolve(id);
            });
        });
    }

    /* Update a snip in the database this is done by sending
     * the snip data to the server and the server will update
     * the snip in the database. If the snip does not exist
     * in the database it will be created.
     *
     * see socket/snip/socket.ts for more info
     *
     * Returns the id of the snip in the database
     */
    public async update(): Promise<number> {
        return new Promise<number>((resolve) => {
            this.socket.emit("snip:update", this.to_data(), (id: number) => {
                resolve(id);
            });
        });
    }

    /** Remove a snip from the database, this should normally not be done
     * as it will remove the snip from all pages and books. It might also
     * be not possible to do this depending on the users permissions.
     */
    public async remove(): Promise<void> {
        return new Promise<void>((resolve) => {
            this.socket.emit("snip:remove", this.id, () => {
                resolve();
            });
        });
    }

    /** Place the snip on the page
     * This method can be used to place the snip on a given page of
     * previously no page was set.
     *
     * It updates the page_id and emits a socket event to the server
     */
    public async place_on_page(page_id: number): Promise<void> {
        this.page_id = page_id;
        await this.update();
        // None async version because the snip is already in the db
        // we just want to sync with the other clients
        this.socket.emit("snip:placed", this.page_id, this.id);
    }

    /** If a snip has a preview state, e.g. while drawing but not placed yet.
     * The snip can be send to other clients with this method.
     */
    public send_preview(): void {
        this.socket.volatile.emit("snip:preview", this.to_data());
    }

    /** Get type string, used by mainly arraySnips to display all
     * nested snips.
     */
    public get typeStr(): string {
        return this.type;
    }

    /** Get date as formatted string
     */
    public get createdStr(): string {
        return this.created.toLocaleTimeString("en-US", date_options);
    }
    public get last_updatedStr(): string {
        return this.last_updated.toLocaleTimeString("en-US", date_options);
    }

    /** Handler for interactions
     *
     * If handlers are defined these snips are interactive, the can be
     * clicked on and react to mouse events. This is very basic and
     * should be implemented by child classes.
     *
     * At the moment the following events are supported:
     * - on_pointerdown
     *  This is called when the snip is clicked on
     * - on_dblclick
     *  This is called when the snip is double clicked on
     * - on_pointerenter / on_pointerleave
     *  This is called when the mouse first enters the snip and when it leaves
     *
     * The base snip is not interactive and does not implement any
     * of these handlers as if they are defined an additional
     * transfer of the snip data is required. Which might impact
     * performance.
     *
     * The most complex interaction is implemented by the
     * textSnip which allows to copy text from the snip.
     *
     */
    public on_pointerdown: (context: EventContext) => void;
    public on_pointerenter = ({ setActiveSnip }: EventContext) => {
        // We dont wont to import react that's why we
        // use this approach to create a overlay
        // in the tool component.
        // See interaction.tool.tsx
        setActiveSnip(this);
        return;
    };
    public on_pointerleave = ({ setActiveSnip }: EventContext) => {
        setActiveSnip(undefined);
        return;
    };
    public on_dblclick: (context: EventContext) => void;

    // public on_pointerdown(event: IBaseEvent): void {}
    // public on_dblclick(event: IBaseEvent): void {}
    // public on_pointerenter(event: IBaseEvent): void {}
    // public on_pointerleave(event: IBaseEvent): void {}

    /** Further if they are subfixed with the "worker" tag. The event
     * is called in the worker thread and not in the main thread. This is
     * useful if you for instance want to change the drawing of the snip on
     * hover. After the event is called in the worker thread the page
     * is rerendered and the snip is redrawn.
     *
     */
    public on_pointerdown_worker: () => void;
    public on_pointerenter_worker: () => void;
    public on_pointerleave_worker: () => void;
    public on_dblclick_worker: () => void;
}

export interface EventContext {
    canvas: HTMLCanvasElement;
    container: HTMLDivElement;
    setActiveSnip: (snip?: BaseSnip) => void;
    page_idx: number;
}

const date_options = {
    weekday: "long",
    year: "numeric",
    month: "short",
    day: "numeric",
    hour: "2-digit",
    minute: "2-digit",
    hour12: false,
} as const;
