import EventEmitter from "events";
import { sync as probe_sync } from "probe-image-size";

import { BaseData, BaseSnip, BaseSnipArgs, BaseView, SnipData } from "./base";

import { isBrowser, isWorker } from "controller/booksocket";

/** As we use the canvas package on the server side
 * we need to check if we are in the browser or not
 * and update the respective types.
 * This is a workaround as the Image type is not
 * available on the server side.
 */
if (!isBrowser() && !isWorker()) {
    // eslint-disable-next-line @typescript-eslint/no-var-requires
    const Canvas = require("canvas");
    global.Image = Canvas.Image;

}

export interface ImageData extends BaseData {
    blob: BlobData;
}

export interface ImageView extends BaseView {
    rot: string;
    sx: number;
    sy: number;
    swidth: number;
    sheight: number;
    width: number;
    height: number;
}

export interface BlobData {
    id: number | null;
    snip_id: number | null;
    upd: Date;
    mime: string;
    size: number;
    data: string;
    hash: string;
}

export interface ImageSnipArgs extends BaseSnipArgs {
    /** The blob data of the image
     * the image width and height are probed from the image
     */
    blob: BlobData;

    /** The rotation of the image
     * At the moment we only allow increments of 90 and mirroring
     *
     * Allowed values: "0", "90", "180", "270", "m0", "m90", "m180", "m270"
     * default: "0"
     */
    rot?: string;

    /** The x-axis coordinate of the top left corner of the sub-rectangle of the source
     *  image to draw into the destination context.
     *
     * see https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/drawImage
     *
     * default: 0
     */
    sx?: number;

    /** The y-axis coordinate of the top left corner of the sub-rectangle of the source
     * image to draw into the destination context.
     *
     * see https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/drawImage
     *
     * default: 0
     */
    sy?: number;

    /** The width of the sub-rectangle of the source image to draw into the destination
     * context. If not specified, the entire rectangle from the coordinates specified by
     * rx and ry to the bottom-right corner of the image is used.
     *
     * default: img_width
     */
    swidth?: number;

    /** The height of the sub-rectangle of the source image to draw into the destination
     * context. If not specified, the entire rectangle from the coordinates specified by
     * rx and ry to the bottom-right corner of the image is used.
     *
     * default: img_height
     */
    sheight?: number;

    /** The desired width of the snip in pixels
     *
     * default: img_width
     */
    width?: number;

    /** The height of the image in pixels
     *
     * default img_height
     */
    height?: number;
}

export class ImageSnip extends BaseSnip {
    public type = "image";
    public blob: BlobData;

    bitmap: ImageBitmap | HTMLImageElement;

    // Image dimensions in pixels (probed from the image (blob))
    readonly img_width: number;
    readonly img_height: number;

    rot: string;
    sx: number;
    sy: number;
    private _swidth: number;
    private _sheight: number;
    private _width: number;
    private _height: number;

    public ready = false;
    public events = new EventEmitter();

    constructor({
        blob,
        rot = "0",
        sx = 0,
        sy = 0,
        swidth = null,
        sheight = null,
        width = null,
        height = null,
        ...baseArgs
    }: ImageSnipArgs) {
        super({ ...baseArgs });

        // Set default values
        this.blob = blob;
        this.blob.snip_id = this.id;

        // Probe image dimensions
        const { width: img_width, height: img_height } = probe_sync(
            Buffer.from(this.blob.data, "base64")
        );
        this.img_width = img_width;
        this.img_height = img_height;

        // Slicing if needed
        this.sx = sx;
        this.sy = sy;
        this.swidth = swidth || this.img_width;
        this.sheight = sheight || this.img_height;

        // Width and height
        this._width = width || this.img_width;
        this._height = height || this.img_height;

        // Legacy support TODO: find smarter solution
        if (this.created < new Date("2023-04-08T00:00:00.000Z")) {
            this._height = (this.sheight * this._width) / this.swidth;
        }

        // Rotation
        this.rot = rot;

        // Convert string to base64
        (async () => {
            if (isBrowser() || isWorker()) {
                await this.prepare_images();
            }
        })();
    }

    public async prepare_images() {
        if (isBrowser() || isWorker()) {
            const base64Response = await fetch(
                `data:${this.blob.mime};base64,${this.blob.data}`
            );
            const blobb = await base64Response.blob();
            this.bitmap = await createImageBitmap(blobb).then((bitmap) => {
                return bitmap;
            });
            this.ready = true;
            this.events.emit("ready");
        }
    }


    /** Construct a instance of the snip from a
     * SnipData object. This is used to create a snip
     * object from the database data dict.
     *
     * @param {SnipData<ImageData, ImageView>} data - Data dict
     */
    static from_data(data: SnipData<ImageData, ImageView>): ImageSnip {

        if (!data.data.blob) {
            throw new Error("Blob is null");
        }
        if (!data.data.blob.mime) {
            throw new Error("Mime is not set!");
        }


        return new ImageSnip({
            // Image
            blob: data.data.blob,
            rot: data.view.rot,
            sx: data.view.sx,
            sy: data.view.sy,
            swidth: data.view.swidth,
            sheight: data.view.sheight,
            width: data.view.width,
            height: data.view.height,
            //Base
            x: data.view.x,
            y: data.view.y,
            id: data.id,
            page_id: data.page_id,
            book_id: data.book_id,
            last_updated: data.last_updated,
            created: data.created,
        });
    }

    /** Renders the simple text to the canvas function
     *
     * @param {CanvasRenderingContext2D} ctx - Canvas context
     */
    render(ctx: CanvasRenderingContext2D): void {
        super.render(ctx);
        if (this.ready) {
            this.drawImage(ctx);
        } else {
            // Draw a placeholder
            this.drawPlaceholder(ctx);
            const saved_transform = ctx.getTransform();
            this.events.once("ready", () => {
                const save_new = ctx.getTransform();
                ctx.setTransform(saved_transform);
                this.drawImage(ctx)
                ctx.setTransform(save_new);
            });
        }
    }

    /** Create a dictionary in data standard format
     *
     * @returns {Dictionary}
     */
    data(): ImageData {
        return {
            type: "image",
            blob: this.blob,
        };
    }
    view(): ImageView {
        const view = super.view() as ImageView;
        view.rot = this.rot;
        view.sx = this.sx;
        view.sy = this.sy;
        view.swidth = this.swidth;
        view.sheight = this.sheight;
        view.width = this.width;
        view.height = this.height;
        return view;
    }

    /** Helper function to draw the image to the canvas
     *
     * @param ctx
     * @returns
     */
    private drawImage(ctx: CanvasRenderingContext2D): void {
        // Draw underneath the text with composite operation
        //let old_op = ctx.globalCompositeOperation;
        //ctx.globalCompositeOperation = "source-over";
        const transform = ctx.getTransform();
        this.rotateContext(ctx);

        // If slicing is needed we
        // are also able to draw the original image
        // this is useful for placement of the image
        // Draw full image (should only happen on the client)
        if (this.overlay_mode) {
            ctx.save();
            ctx.globalAlpha = 0.1;
            ctx.drawImage(
                this.bitmap,
                0,
                0,
                this.width,
                this.height
            );
            ctx.restore();

            //Draw rectangle around sliced image
            ctx.save();
            ctx.strokeStyle = "#2f3992";
            ctx.lineWidth = 2;
            ctx.lineDashOffset = 0;
            ctx.setLineDash([12, 4]);
            ctx.strokeRect(
                this.sx * (this.width / this.img_width) - 1,
                this.sy * (this.height / this.img_height) - 1,
                (this.swidth * this.width) / this.img_width + 2,
                (this.sheight * this.height) / this.img_height + 2
            );
            ctx.restore();
        }

        // Legacy support TODO remove
        if (this.created < new Date("2023-04-08T00:00:00.000Z")) {
            ctx.drawImage(
                this.bitmap,
                this.sx,
                this.sy,
                this.swidth,
                this.sheight,
                this.sx,
                this.sy,
                this.width,
                this.height
            );
        } else {
            ctx.drawImage(
                this.bitmap,
                // In image coordinates
                this.sx,
                this.sy,
                this.swidth,
                this.sheight,
                // In canvas coordinates
                this.sx * (this.width / this.img_width),
                this.sy * (this.height / this.img_height),
                (this.swidth * this.width) / this.img_width,
                (this.sheight * this.height) / this.img_height
            );
        }
        ctx.setTransform(transform);
    }


    private drawPlaceholder(ctx: CanvasRenderingContext2D): void {
        const transform = ctx.getTransform();
        this.rotateContext(ctx);

        ctx.fillStyle = "#2f3992";
        ctx.globalAlpha = 0.1;
        ctx.fillRect(this.sx, this.sy, this.width, this.height);
        ctx.globalAlpha = 1;
        ctx.setTransform(transform);
    }


    private rotateContext(ctx: CanvasRenderingContext2D): void {
        //TODO make rotation more general
        const x = this.x;
        const y = this.y;
        switch (this.rot) {
            default:
            case "0":
                ctx.translate(x, y);
                break;
            case "90":
                ctx.translate(this.height + x, y);
                ctx.rotate((1 * Math.PI) / 2);
                break;
            case "180":
                ctx.translate(this.width + x, this.height + y);
                ctx.rotate((2 * Math.PI) / 2);
                ctx.scale(1, 1);
                break;
            case "270":
                ctx.translate(x, this.width + y);
                ctx.rotate((3 * Math.PI) / 2);
                ctx.scale(1, 1);
                break;
            case "0m":
                ctx.translate(this.width + x, y);
                ctx.scale(-1, 1);
                break;
            case "90m":
                ctx.translate(x + this.height, y + this.width);
                ctx.rotate((1 * Math.PI) / 2);
                ctx.scale(-1, 1);
                break;
            case "180m":
                ctx.translate(x, this.height + y);
                ctx.rotate((2 * Math.PI) / 2);
                ctx.scale(-1, 1);
                break;
            case "270m":
                ctx.translate(x + this.height, y + this.width);
                ctx.rotate((3 * Math.PI) / 2);
                ctx.scale(1, -1);
                break;
        }
    }


    //Callback can be overwritten if one wants to plot the image inside an array snip
    public callbackLoaded: (ctx: CanvasRenderingContext2D) => void =
        this.drawImage;

    get width(): number {
        return this._width;
    }
    set width(width: number) {
        this._width = width;
    }
    get height(): number {
        return this._height;
    }
    set height(height: number) {
        this._height = height;
    }
    get sheight(): number {
        return this._sheight;
    }
    set sheight(sheight: number) {
        this._sheight = sheight;
    }
    get swidth(): number {
        return this._swidth;
    }
    set swidth(swidth: number) {
        this._swidth = swidth;
    }
    get id(): number {
        return super.id;
    }
    set id(id: number) {
        super.id = id;
        //Also set blob snip_id
        this.blob.snip_id = id;
    }
}
