import {
    BaseData,
    BaseSnip,
    BaseSnipArgs,
    BaseView,
    EventContext,
    SnipData,
} from "./base";

import { isBrowser } from "controller/booksocket";

export const allowedFontFamilies = [
    "PlexMono",
    "Caveat",
    "Arial",
    "Times New Roman",
    "Sans-serif",
    "Serif",
];

export interface TextData extends BaseData {
    text: string;
    version: number;
}

export interface TextView extends BaseView {
    size: number;
    lheight: number;
    font: string;
    wrap: number;
    colour: string;
}

export interface TextSnipArgs extends BaseSnipArgs {
    /** The text to be displayed
     *
     * default: ""
     */
    text?: string;

    /** The font family to use for the text, this
     * must be available on the client and the server
     * for the text to be displayed correctly.
     *
     * default: "PlexMono"
     */
    fontFamily?: string;

    /** The font size of the text
     *
     * See https://gitlab.gwdg.de/irp/snip/-/blob/legacy/public/assets/js/common.js#L1159 for defaults
     *
     * default: 12
     */
    fontSize?: number;

    /** The line height of the text in pixels
     * (i.e. the distance between the baseline of
     * two lines of text)
     *
     * default: fontSize*1.25
     */
    lineHeight?: number;

    /** The colour of the text to be displayed
     *
     * default: "black"
     */
    colour?: string;

    /** The number of characters to wrap the text at
     *
     * This should not be used if possible as it is
     * kinda buggy and not very useful atm.
     *
     * default: 78
     */
    lineWrap?: number;

    /** Version number as used by the database
     *
     * Changelog:
     * 2 - Linewrap is in pixels instead of characters
     */
    version?: number;
}

/** Wrap text
 * keeps all defined line breaks but adds
 * newlines at the defined wrap points
 * i.e. this.lineWrap
 *
 * http://web.archive.org/web/20200925112638/https://j11y.io/snippets/wordwrap-for-javascript/
 *
 * @param {string} text The text to be wrapped
 * @param {boolean} cut If the cut is set to TRUE, the string is always
 * wrapped at or before the specified width. So if you have a word that is larger than the given width,
 * it is broken apart. When FALSE the function does not split the word even if the width is smaller than the word width.
 * @param {number} lineWrap The width of the line in characters.
 */
export function wrap_text(text: string, lineWrap: number, cut = true): string {
    //?
    text = text.replaceAll(/\t/g, "        ").replaceAll(/\n/g, "\\n");

    //TODO measure test here!
    if (lineWrap < 0 || text.length == 0) {
        return text;
    }
    const regex: string =
        ".{1," +
        lineWrap +
        "}(s|$)" +
        (cut ? "|.{" + lineWrap + "}|.+$" : "|S+?(s|$)");
    const match = text.match(RegExp(regex, "g"))?.join("\\n");

    return match;
}

export class TextSnip extends BaseSnip {
    public type = "text";

    public fontFamily: string;
    public fontSize: number;
    public colour: string;
    public lineHeight: number;
    public version: number;

    constructor({
        text = "",
        fontFamily = "PlexMono",
        fontSize = 12,
        colour = "black",
        lineHeight = undefined,
        lineWrap = undefined,
        version = 2,
        ...baseArgs
    }: TextSnipArgs) {
        super({ ...baseArgs });

        this.version = version;

        this.fontFamily = fontFamily;
        this.fontSize = fontSize;
        this.colour = colour;
        this.lineHeight = lineHeight || fontSize * 1.25;
        // Set default vars
        if (version < 2) {
            this._lineWrap = lineWrap || 78;
            this._text = wrap_text(text, this._lineWrap);
        } else {
            this._lineWrap = lineWrap || 400;
            this._text = this.formatText(text);
        }
        this._height =
            this.lineHeight * 2.35 * this.getNumLines() +
            0.25 * this.fontSize * 2.35;
    }

    /** Construct a instance of the snip from a
     * SnipData object. This is used to create a snip
     * object from the database data dict.
     *
     * @param {SnipData<TextData, TextView>} data - Data dict
     */
    static from_data(data: SnipData<TextData, TextView>): TextSnip {
        return new TextSnip({
            // Text
            text: data.data.text,
            fontFamily: data.view.font,
            fontSize: data.view.size,
            colour: data.view.colour,
            lineHeight: data.view.lheight,
            lineWrap: data.view.wrap,
            version: data.data.version,
            //Base
            x: data.view.x,
            y: data.view.y,
            id: data.id,
            page_id: data.page_id,
            book_id: data.book_id,
            last_updated: data.last_updated,
            created: data.created,
        });
    }

    /** Renders the simple text to the canvas function
     *
     * @param {CanvasRenderingContext2D} ctx - Canvas context
     */
    render(ctx: CanvasRenderingContext2D): void {
        // Set font and style
        // 2480/1400  (TODO put into database update script)
        ctx.font = this.fontSize * 2.35 + "px " + this.fontFamily;
        ctx.fillStyle = this.colour;
        ctx.textAlign = "left";
        ctx.textBaseline = "top";

        // Split text in lines
        // For now only newline is supported
        // TODO: lineWrap is not implemented
        const line = this.text.split("\\n");
        let y = this.y;

        // Draw text
        line.forEach((text: string) => {
            ctx.fillText(text, this.x, y);
            y += this.lineHeight * 2.35;
        });

        super.render(ctx);
    }

    /** Create a json string for mysql from a text snip
     *
     * @returns {Dictionary}
     */
    data(): TextData {
        return {
            type: "text",
            text: this.text,
            version: this.version,
        };
    }
    view(): TextView {
        const view = super.view() as TextView;
        view.font = this.fontFamily;
        view.size = this.fontSize;
        view.colour = this.colour;
        view.lheight = this.lineHeight;
        view.wrap = this.lineWrap;
        return view;
    }

    public getNumLines(): number {
        return this.text.split("\\n").length;
    }

    private _height: number;
    get height(): number {
        return (
            this.lineHeight * 2.35 * this.getNumLines() +
            0.25 * this.fontSize * 2.35
        );
    }
    get width(): number {
        if (this.version < 2) {
            return getWidth(this.text, this.fontSize, this.fontFamily);
        }
        const canvas = new OffscreenCanvas(1, 1);
        const ctx = canvas.getContext(
            "2d"
        ) as OffscreenCanvasRenderingContext2D;
        ctx.font = this.fontSize * 2.35 + "px " + this.fontFamily;
        let maxWidth = 0;
        const lines = this.text.split("\\n");
        for (let i = 0; i < lines.length; i++) {
            maxWidth = Math.max(maxWidth, ctx.measureText(lines[i]).width);
        }

        return maxWidth;
    }

    private _lineWrap: number;
    get lineWrap(): number {
        return this._lineWrap;
    }
    set lineWrap(lineWrap: number) {
        this._lineWrap = lineWrap;
        if (this.version < 2) {
            this.text = wrap_text(this.text, this.lineWrap);
        } else {
            this.text = this.formatText(this.text);
        }
    }

    private _text: string;
    get text(): string {
        return this._text;
    }
    set text(text: string) {
        // format text
        if (this.version < 2) {
            this._text = wrap_text(text, this.lineWrap);
        } else {
            this._text = this.formatText(text);
        }
    }
    /** Formats the text to a usable format
     * - Replaces tabs with 8 spaces
     * - Replaces newlines with \n
     * - Calculate width and further split text if
     *  lineWrap is set and text width is too long
     *
     */
    formatText(inputString): string {
        // Replace tabs with 8 spaces
        inputString = inputString.replaceAll(/\t/g, "        ");
        // Replace newlines with \n
        inputString = inputString.replaceAll(/\n/g, "\\n");

        // split text in lines
        const lines = inputString.split("\\n");

        // Measure text width
        const font = this.fontSize * 2.35 + "px " + this.fontFamily;
        for (let i = 0; i < lines.length; i++) {
            const o = get_overlap_index(lines[i], this.lineWrap, font);
            if (o !== undefined && o > 0) {
                // Split line and add to next line
                if (lines[i + 1] === undefined) lines[i + 1] = "";
                lines[i + 1] = lines[i].substring(o) + lines[i + 1];
                lines[i] = lines[i].substring(0, o);
            }
            if (o == 0) {
                // each character on a new line
                lines[i] = lines[i].split("").join("\\n");
            }
        }

        // Join lines
        return lines.join("\\n");
    }

    /** Interaction events */
    public on_pointerenter = ({ setActiveSnip }: EventContext) => {
        // We dont wont to import react that's why we
        // use this approach to create a overlay
        // See interaction.tool.tsx
        setActiveSnip(this);

        return;
    };
    public on_pointerleave = ({ setActiveSnip }: EventContext) => {
        setActiveSnip(undefined);
        return;
    };
}

/** Returns index of character which
 * overlaps the lineWrap limit.
 *
 * Returns undefined if no character overlaps
 */
function get_overlap_index(str: string, maxWidth: number, font: string) {
    // Check if offscreen canvas is supported
    if (!("OffscreenCanvas" in globalThis)) {
        return undefined;
    }
    if (maxWidth == -1) {
        return undefined;
    }

    const canvas = new OffscreenCanvas(1, 1);
    const ctx = canvas.getContext("2d") as OffscreenCanvasRenderingContext2D;
    ctx.font = font;

    let width = 0;
    for (let c = 0; c < str.length; c++) {
        const char = str.charAt(c);
        width += ctx.measureText(char).width;
        if (width > maxWidth) {
            return c;
        }
    }
    return undefined;
}

/**
 * @deprecated
 */
export function getWidth(
    text: string,
    fontSize: number,
    fontFamily: string
): number {
    if (isBrowser()) {
        const canvas = new OffscreenCanvas(1, 1);
        const ctx = canvas.getContext(
            "2d"
        ) as OffscreenCanvasRenderingContext2D;
        ctx.font = fontSize * 2.35 + "px " + fontFamily;
        let width = 0;
        text.split("\\n").forEach((line: string) => {
            const _width = ctx?.measureText(line).width || 0;
            if (_width > width) width = _width;
        });
        return width;
    } else {
        return 0;
    }
}
