import { randomUUID } from "crypto";
import { existsSync, mkdirSync, readFileSync, writeFileSync } from "fs";
import { service } from "pages/api/_service";
import { PDFDocument } from "pdf-lib";

import { BookData } from "database/services/service.typings";

import { PageNumberLocation, SelectedPage } from "components/forms/DownloadPdfForm";

/** Create a pdf and write it to "/usr/src/app/database/pdf"
 * 
 * @param bookData 
 * @param selectedPages 
 * @returns path to the pdf
 */
export async function download_book_as_pdf(
    bookData: BookData,
    selectedPages: SelectedPage[] = [-1],
    pageNumberLocation: PageNumberLocation
): Promise<string> {
    //Timing


    let pages = await service.page.getByBookId(bookData.id);
    const path = "/usr/src/app/database/preview_images/" + bookData.id + "/";

    // Select pages which should be included in the pdf
    if (selectedPages[0] !== -1) {
        pages = pages.filter((page) => {
            for (const selectedPage of selectedPages) {
                // Range
                if (Array.isArray(selectedPage)) {
                    if (
                        page.page_number + 1 >= selectedPage[0] &&
                        page.page_number + 1 <= selectedPage[1]
                    ) {
                        return true;
                    }
                }
                // Single page
                else {
                    if (page.page_number + 1 === selectedPage) {
                        return true;
                    }
                }
            }
            return false;
        });
    }


    // First check if all pdfs which have to be merged exist in the folder
    pages.forEach((page) => {
        if (!existsSync(path + page.id + ".pdf")) {
            throw new Error("Page " + page.id + " does not exist as pdf yet");
        }
    });

    // get all pdf buffer which need to be combined
    // these are already created with every update to the page
    const pdfsToMerge = pages.map((page) => {
        return readFileSync(path + page.id + ".pdf");
    });

    const mergedPdf = await PDFDocument.create();
    for (const pdfBytes of pdfsToMerge) {
        const pdf = await PDFDocument.load(pdfBytes);
        // Get the form containing all the fields
        const form = pdf.getForm();
        form.flatten();

        const copiedPages = await mergedPdf.copyPages(
            pdf,
            pdf.getPageIndices()
        );
        copiedPages.forEach((page) => {
            mergedPdf.addPage(page);
        });
    }

    // Set metadata
    mergedPdf.setTitle(bookData.title.toString());
    mergedPdf.setCreationDate(new Date(bookData.created));
    mergedPdf.setModificationDate(new Date(bookData.last_updated));
    mergedPdf.setCreator("Snip Digital Lab-book");
    mergedPdf.setProducer("Snip Digital Lab-book");


    // Place page numbers
    if (pageNumberLocation !== "none") {
        const pagesPdf = mergedPdf.getPages();
        const font = await mergedPdf.embedFont("Helvetica");
        const fontSize = 24;

        for (let i = 0; i < pagesPdf.length; i++) {
            const { width, height } = pagesPdf[i].getSize();
            let pos: [number, number] = [0, 0];

            const text = `${pages[i].page_number + 1}`;

            const textWidth = font.widthOfTextAtSize(text, fontSize);
            const textHeight = font.heightAtSize(fontSize);

            switch (pageNumberLocation) {

                case "tl": // top left
                    pos = [fontSize, height - fontSize - textHeight];
                    break;
                case "tr": // top right
                    pos = [width - textWidth - fontSize, height - fontSize - textHeight];
                    break;
                case "bl": // bottom left
                    pos = [fontSize, fontSize];
                    break;
                case "br": // bottom right
                default:
                    pos = [width - textWidth - fontSize, fontSize];
                    break;
            }

            pagesPdf[i].drawText(`${pages[i].page_number + 1}`, {
                x: pos[0],
                y: pos[1],
                size: fontSize,
                font: font,
            });
        }

    }

    const buf = await mergedPdf.save(); // Uint8Array

    //Save in temp folder
    const save_path = "/usr/src/app/database/pdf";

    // Generate a uuid for the filename to prevent collisions
    // Normally the file is deleted after the download
    const filename = randomUUID();

    if (!existsSync(save_path)) {
        mkdirSync(save_path);
    }

    writeFileSync(save_path + "/" + filename + ".pdf", buf, {
        flag: "w+",
    });

    return save_path + "/" + filename + ".pdf";
}
