import { service } from "pages/api/_service";
import { page_preview } from "socket/page/preview";

import { log, SocketData } from "../socket";

import { pool_data } from "database/sql.connection";
import { pagesData as PageData } from "database/sql.typings";
import { BaseData, BaseView, SnipData } from "snips/general/base";

export function setupBookPrefix({ socket, book_nsp }: SocketData) {
    /** Called on book init to get all page data
     *  - this does not transfer snip data
     */
    socket.on(
        "book:getPages",
        async (callback: (pages: PageData[]) => void) => {
            // Check if user has permission to view the page
            if (!(socket.perms.pRead == true)) {
                return socket.emit(
                    "error",
                    "You do not have access to view this book"
                );
            }
            log("[Book: " + socket.book_id + "] book:getPages");

            const pages = await service.page.getByBookId(socket.book_id);
            callback(pages);
        }
    );

    /** Called on book init to get all not yet on pages placed snips
     */
    socket.on(
        "book:getUnplacedSnips",
        (callback: (snipsData: SnipData<BaseData, BaseView>[]) => void) => {
            // Check if user has permission to view the page
            if (!(socket.perms.pRead == true)) {
                return socket.emit(
                    "error",
                    "You do not have access to view this book"
                );
            }
            log("[Book: " + socket.book_id + "] book:getUnplacedSnips");
            getUnplacedSnips(socket.book_id, callback);
        }
    );

    socket.on(
        "book:getPreviews",
        async (callback: (images: string[]) => void) => {
            // Check if user has permission to view the page
            if (!(socket.perms.pRead == true)) {
                return socket.emit(
                    "error",
                    "You do not have access to view this book"
                );
            }
            log("[Book: " + socket.book_id + "] book:getPreviews");

            // Get all pages ids in the book
            const sql = "SELECT id FROM pages WHERE book_id = ?";
            const pageids: number[] = await pool_data
                .query(sql, [socket.book_id])
                .then((res) => {
                    return res.map((row) => row.id);
                })
                .catch((err) => {
                    console.log(err);
                    return [];
                });

            // Get all previews
            const promises = pageids.map((pageid) => {
                return page_preview(pageid, socket.book_id, false);
            });

            const previews = await Promise.all(promises);

            callback(previews);
        }
    );

    /** Insert a new page into the database
     */
    socket.on(
        "book:insertPage",
        async (pageData: PageData, callback: (pageData: PageData) => void) => {
            // Check if user has permission to write to the page
            if (!(socket.perms.pWrite == true)) {
                return socket.emit(
                    "error",
                    "You do not have access to write to this book. Could be finished!"
                );
            }
            log("[Book: " + socket.book_id + "] book:insertPage");
            // Add the page to the database
            const inserted_page = await service.page.insert(pageData);
            callback(inserted_page);

            // Emit to clients
            socket.broadcast.emit("book:insertPage", inserted_page);

            // Render previews
            const img = await page_preview(
                inserted_page.id,
                socket.book_id,
                false
            );
            book_nsp.emit("book:updatePreview", inserted_page.id, img);
        }
    );
}

/** Get all not yet placed snips for a book
 */
async function getUnplacedSnips(
    book_id: number,
    callback: (snipsData: SnipData<BaseData, BaseView>[]) => void
) {
    // Get all not placed snips by book id
    const snipsData = await service.snip.getNotPlacedByBookId(book_id);

    // Map the snips to data only for smaller transfer size
    callback(snipsData);
}
