import { Canvas, createCanvas, loadImage, registerFont } from "canvas";
import { service } from "pages/api/_service";
import sharp from "sharp";

import { MessageInterface } from "./process_pool";

import { LocalPage } from "controller/page";
import { ArraySnip } from "snips/general/array";
import { ImageSnip } from "snips/general/image";
import { get_snip_from_data } from "snips/get_snip_from_data";

// We need to load fonts for the page
registerFont(
    __dirname + "/../../public/fonts/Inter-VariableFont_slnt,wght.ttf",
    { family: "Inter" }
);
registerFont(__dirname + "/../../public/fonts/PlexMono-Regular.otf", {
    family: "PlexMono",
});




ImageSnip.prototype.prepare_images = async function () {

    let src;
    // Webp support for node-canvas
    if (this.blob.mime === "image/webp") {
        const png_buffer = await sharp(Buffer.from(this.blob.data, "base64"))
            .toFormat('png')
            .toBuffer();
        src = `data:image/png;base64,${png_buffer.toString('base64')}`;
    } else {
        src = `data:${this.blob.mime};base64,${this.blob.data}`;
    }
    this.bitmap = await loadImage(src);
    this.ready = true;
};



/** Render a page preview to image and pdf
 *
 * @param page_id
 * @returns {Promise<string>} Base64 encoded jpeg image
 */
export async function render_page_preview(
    page_id: number
): Promise<MessageInterface> {
    // Get page and all snips
    const pageData = await service.page.getById(page_id);
    const page = LocalPage.from_data(pageData);
    const snipsData = await service.snip.getByPageId(page_id);
    const snips = snipsData.map((snipData) => {
        return get_snip_from_data(snipData);
    });

    // Prepare snip images
    const promises = [];
    for (const snip of snips) {
        if (snip instanceof ImageSnip) {
            promises.push(snip.prepare_images());
        }
        if (snip instanceof ArraySnip) {
            for (const child_snip of snip.snips) {
                if (child_snip instanceof ImageSnip) {
                    promises.push(child_snip.prepare_images());
                }
            }
        }
    }

    await Promise.all(promises);
    page.set_snips(snips);


    // Create a canvas
    const canvases: Canvas[] = [];
    canvases.push(createCanvas(240, 350));
    canvases.push(createCanvas(1400, 2000, "pdf"));

    // Render the page (background has to be rendered white since we're not using transparency)
    canvases.forEach((canvas) => {
        const ctx = canvas.getContext("2d");
        const transform = ctx.getTransform().invertSelf();
        const width =
            transform.a * canvas.width +
            transform.b * canvas.height +
            transform.e;
        const height =
            transform.c * canvas.width +
            transform.d * canvas.height +
            transform.f;
        ctx.fillStyle = "white";
        ctx.fillRect(0, 0, width, height);

        const max_db = [1400, 2000];
        const scale = canvas.height / max_db[1];
        ctx.scale(scale, scale);
        page.render(ctx as unknown as CanvasRenderingContext2D);
    });

    // Create jpeg-encoded image
    const buf_dataurl = await canvases[0].toDataURL("image/jpeg", 0.9);
    const buf_pdf = await canvases[1].toBuffer("application/pdf");

    return { id: page_id, dataurl: buf_dataurl, pdf: buf_pdf };
}

// simple process that can render a page and return the image as string
process.on("message", async (id: number) => {
    //console.debug("[RenderChildProcess] rendering page " + id);
    try {

        const message = await render_page_preview(id);
        //console.debug("[RenderChildProcess] rendering page " + id + " done");
        process.send?.(message);
    }
    catch (e) {
        // print error with stack trace
        console.error(e);
    }
});
