import { ChildProcess, fork, ForkOptions, Serializable } from "child_process";
import { config } from "config";
// Get env variable
interface QueueItem {
    id: number;
    callback: (buffers: { dataurl: string; pdf: Buffer }) => void;
}

class Worker {
    private filename: string;
    private args: string[];
    private options: ForkOptions;
    private child: ChildProcess;

    constructor(filename: string, args: string[], options: ForkOptions) {
        this.filename = filename;
        this.args = args;
        this.options = options;

        this.child = this.create_child();
    }

    /** Create child process */
    create_child(): ChildProcess {
        const child = fork(this.filename, this.args, this.options);
        //Setup exit and error
        child.on("exit", this.exit.bind(this));
        child.on("error", this.error.bind(this));
        child.on("message", this.message.bind(this));
        return child;
    }

    /** Send message to child */
    send(message: Serializable) {
        this.child.send(message);
    }

    /** Close child process */
    close() {
        this.child.kill();
    }

    /* Error listener, add callbacks via onerror function */
    error_callbacks: ((err: Error) => void)[] = [];
    error(err: Error) {
        console.error("[Render Pool] Worker error:" + err);
        this.error_callbacks.forEach((callback) => {
            callback(err);
        });
        //Restart the worker
        this.child.kill();
        this.child = this.create_child();
    }
    onerror(callback: (err: Error) => void) {
        this.error_callbacks.push(callback);
    }

    /* Exit listener */
    exit_callbacks: ((code: number, signal: string) => void)[] = [];
    exit(code: number, signal: string) {
        console.log(
            "[RenderPool] Exiting with code",
            code,
            "and signal",
            signal
        );
        this.exit_callbacks.forEach((callback) => {
            callback(code, signal);
        });
        // Restart the worker
        this.child = this.create_child();
    }

    /** Message listener */
    message_callbacks: ((message: Serializable) => void)[] = [];
    message(message: Serializable) {

        this.message_callbacks.forEach((callback) => {
            callback(message);
        });

    }
    onmessage(callback: (message: Serializable) => void) {
        this.message_callbacks.push(callback);
    }
}

/** Render items we receive from the child process after it
 * has finished rendering a job
 */
export interface MessageInterface {
    id: number;
    dataurl: string;
    pdf: Buffer;
}

export class RenderPool {
    // Singleton
    static instance: RenderPool;

    //Worker and queue
    private workers: Worker[] = [];
    private queue: QueueItem[] = [];

    private activeRenders: QueueItem[] = [];

    constructor(
        __filename: string,
        args: string[],
        options: ForkOptions,
        max_workers: number = parseInt(config.NUM_RENDER_PROCESSES)
    ) {
        if (RenderPool.instance) {
            return RenderPool.instance;
        }
        for (let i = 0; i < max_workers; i++) {
            const worker = new Worker(__filename, args, options);
            worker.onmessage((message: MessageInterface) => {
                this.callback_task(message, worker);
            });
            this.workers.push(worker);
        }
        console.log(
            "[RenderPool] Created with",
            this.workers.length,
            "workers"
        );
        RenderPool.instance = this;
    }

    /** Called once a worker finishes a job */
    callback_task(message: MessageInterface, worker: Worker) {
        // Delete from active renders
        const index = this.activeRenders.findIndex((item) => {
            return item.id == message.id;
        });
        if (index == -1) {
            console.error("[RenderPool] Could not find item in active renders");
        }
        const queueItem = this.activeRenders.splice(index, 1)[0];

        // Add next task to worker
        this.checkQueue(worker);

        // Run callback
        queueItem.callback({ dataurl: message.dataurl, pdf: message.pdf });
    }

    /** Check queue */
    checkQueue(worker: Worker) {
        const item = this.queue.shift();
        if (item) {
            this.activeRenders.push(item);
            worker.send(item.id);
        }
    }

    /** Add an item to a queue and run it if their is space
     *
     * @param id Unique id for the item
     * @param callback Callback function to run once the item is finished
     */
    addToQueue(
        id: number,
        callback: (buffers: { dataurl: string; pdf: Buffer }) => void
    ) {
        this.queue.push({ id, callback: callback });
        if (this.activeRenders.length < this.workers.length) {
            //TODO: load balancing
            this.checkQueue(this.workers[this.activeRenders.length]);
        }
    }
    render_page(
        id: number,
        callback: (buffers: { dataurl: string; pdf: Buffer }) => void
    ) {
        this.addToQueue(id, callback);
    }
}
