import { page_preview } from "socket/page/preview";

import { log, SocketData } from "../socket";

import { SQLService } from "database/services";
import { pool_data } from "database/sql.connection";
import { BaseData, BaseView, SnipData } from "snips/general/base";

export function setupSnipPrefix({ socket, book_nsp }: SocketData) {
    socket.on(
        "snip:insert",
        async (
            snip_data: SnipData<BaseData, BaseView>,
            callback: (id: number | null) => void
        ) => {
            // Check if user has permission to write to the page
            if (!(socket.perms.pWrite === true)) {
                return socket.emit(
                    "error",
                    "You do not have access to write to this book. Could be finished!"
                );
            }

            return await insert({ socket, book_nsp }, snip_data, callback);
        }
    );

    socket.on(
        "snip:update",
        async (
            snip_data: SnipData<BaseData, BaseView>,
            callback: (id: number | null) => void
        ) => {
            // Check if user has permission to write to the page
            if (!(socket.perms.pWrite === true)) {
                return socket.emit(
                    "error",
                    "You do not have access to write to this book. Could be finished!"
                );
            }

            return await update({ socket, book_nsp }, snip_data, callback);
        }
    );

    // Remove snip from db
    socket.on(
        "snip:remove",
        async (snip_id: number, callback: (success: boolean) => void) => {
            // Check if user has permission to delete from the page
            if (!(socket.perms.pDelete === true)) {
                callback(false);
                return socket.emit(
                    "error",
                    "You do not have access to delete from this book."
                );
            }
            console.log(
                "[Socket][Book: " + socket.book_id + "] snip:remove " + snip_id
            );
            await SQLService.snip.delete(snip_id);
            callback(true);
            // Remove placed snip
            book_nsp.emit("snip:remove", snip_id);
        }
    );

    socket.on(
        "snip:preview",
        async (SnipData: SnipData<BaseData, BaseView>) => {
            // Check if user has permission to write to the page
            if (!(socket.perms.pWrite === true)) {
                return socket.emit(
                    "error",
                    "You do not have access to write to this book. Could be finished!"
                );
            }

            socket
                .to("page-" + String(SnipData.page_id))
                .volatile
                .emit("snip:preview", SnipData);
        }
    );

    // Just a way to forward the event to the page namespace
    // and sync the placement of the snip
    socket.on("snip:placed", async (page_id: number, snip_id: number) => {
        // Check if user has permission to write to the page
        if (!(socket.perms.pWrite === true)) {
            return socket.emit(
                "error",
                "You do not have access to write to this book."
            );
        }

        console.log(
            "[Socket][Book: " + socket.book_id + "] snip:placed " + snip_id
        );

        // Remove placed snip
        book_nsp.emit("snip:placed", page_id, snip_id);
    });
}

async function insert(
    { socket, book_nsp }: SocketData,
    snip_data: SnipData<BaseData, BaseView>,
    callback: (id: number | null) => void
) {
    const page_id = snip_data.page_id;
    log("[Book: " + socket.book_id + "][Page: " + page_id + "] snip:insert");

    // Sanity check for page_id
    if (page_id === null) {
        console.error("[Socket][Book: " + socket.book_id + "] Page id is null");
        return;
    }

    const old_id = snip_data.id;

    // Sanity check page id and book id
    snip_data.book_id = socket.book_id;

    const snip_data_new = await SQLService.snip.insert(snip_data);

    // Remove from preview stack if old_id is defined
    if (old_id && old_id !== -1) {
        book_nsp
            .to("page-" + String(page_id))
            .emit("snip:preview:remove", old_id);
    }

    // Send snip to all other clients and insert into book/page
    socket.to("page-" + String(page_id)).emit("snip:insert", snip_data_new);

    // Send id back to initial client
    callback(snip_data.id);

    // Trigger page preview update
    const img = await page_preview(page_id, socket.book_id, true);
    book_nsp.emit("book:updatePreview", page_id, img);
}

async function update(
    { socket, book_nsp }: SocketData,
    snip_data: SnipData<BaseData, BaseView>,
    callback: (id: number | null) => void
) {
    const page_id = snip_data.page_id;

    // Sanity check for page_id
    if (page_id === null) {
        console.error("[Socket][Book: " + socket.book_id + "] Page id is null");
        return;
    }

    console.log(
        "[Socket][Book: " +
        socket.book_id +
        "][Page: " +
        page_id +
        "] snip:update_db"
    );

    // Sanity check if snip is in database
    let inserted_data;
    if (!snip_data.id) {
        inserted_data = await SQLService.snip.insert(snip_data);
    } else {
        //Check if in db
        const exists = await pool_data
            .query("SELECT id FROM snips WHERE id = ?", [snip_data.id])
            .then((res) => {
                return res.length > 0;
            });
        if (!exists) {
            inserted_data = await SQLService.snip.insert(snip_data);
            book_nsp
                .to("page-" + String(page_id))
                .emit("snip:preview:remove", snip_data.id);
        } else {
            inserted_data = await SQLService.snip.update(snip_data);
        }
    }

    // Send snip to all other clients and update in book/page
    const updated_snip = await SQLService.snip.getById(inserted_data.id);
    socket.to("page-" + String(page_id)).emit("snip:update", updated_snip);

    // Send id back to initial client
    callback(updated_snip.id);

    // Trigger page preview update
    const img = await page_preview(page_id, socket.book_id, true);
    book_nsp.emit("book:updatePreview", page_id, img);
}
