/* eslint-disable no-unused-vars */
import { Namespace, Server, Socket } from "socket.io";

import { add_perms, authenticate, authenticateApi, update_perms } from "./auth";
import { setupBookPrefix } from "./book/socket";
import { setupPagesPrefix } from "./page/socket";
import { setupSnipPrefix } from "./snip/socket";

import { SQLService } from "database/services";
import { PermissionACLData } from "database/services/service.typings";
import { BaseData, BaseView, SnipData } from "snips/general/base";

// Extend socket to include book_id, user_id and perms
// This is set in the the book_authenticate middleware
declare module "socket.io" {
    interface Socket {
        book_id: number;
        user_id: number | undefined;
        auth_token: string | undefined;
        perms: PermissionACLData;
    }
}

export const log = (...args: unknown[]) => console.log("[Socket]", ...args);
export const debug = (...args: unknown[]) => console.debug("[Socket]", ...args);

export interface SocketData {
    book_nsp: Namespace;
    socket: Socket;
}

export function setupNamespaces(server: Server) {
    // Dynamic namespace creation for the books
    // All namespaces begin with "/book-"
    const book_nsp = server.of(/^\/book-/);

    // Check for rights i.e if user is allowed to access book and adds perms to socket
    book_nsp.use(authenticate);
    book_nsp.use(add_perms);

    book_nsp.on("connection", async (socket: Socket) => {
        log(
            "[Book: " +
            socket.book_id +
            "] Client (" +
            socket.id +
            ") connected"
        );

        socket.on("disconnect", () => {
            log(
                "[Book: " +
                socket.book_id +
                "] Client (" +
                socket.id +
                ") disconnected"
            );
        });

        const data: SocketData = {
            book_nsp,
            socket,
        };

        /** Setup all function calls prefixed with pages
         * /book-[book_id]/pages:[function]/
         */
        setupPagesPrefix(data);

        /** Setup all function calls prefixed with book
         * /book-[book_id]/books:[function]/
         * /book-[book_id]/book:[function]/
         */
        setupBookPrefix(data);

        /** Setup all function calls prefixed with snip
         * /book-[book_id]/snip:[function]/
         */
        setupSnipPrefix(data);

        socket.on(
            "join_page",
            async (
                page_id: number,
                callback: (snips: SnipData<BaseData, BaseView>[]) => void
            ) => {
                log("[Book: " + socket.book_id + "] join_page " + page_id);
                // Deprecated
                log("[Book: " + socket.book_id + "] join_page is deprecated");

                // Join the room for the page and keep listening for snip updates
                socket.join("page:" + page_id);

                // Send all snips back to the client
                const snips = await SQLService.snip.getByPageId(page_id);
                callback(snips);
            }
        );

        socket.on("leave_page", async (page_id: number) => {
            log("[Book: " + socket.book_id + "] leave_page " + page_id);

            // Leave the room for the page
            socket.leave("page:" + page_id);
        });
    });

    setupApiSpace(server);
}


function setupApiSpace(server: Server) {
    const api_nsp = server.of("/api");

    api_nsp.use(authenticateApi);

    api_nsp.on("connection", async (socket: Socket) => {
        log("[Api] Client (" + socket.id + ") connected");


        socket.on("api:insertedIntoQueue", async (snip_id: number[]) => {
            log("[Api] insertedIntoQueue");
            const snips = await SQLService.snip.getByIds(snip_id);
            snips.forEach((snip) => {
                server.of("/book-" + snip.book_id).emit("book:insertedIntoQueue", snip);
            })
        });

        socket.on("resetSocketPerms", async (book_id: number) => {
            log("[Api] resetSocketPerms");
            server.of("/book-" + book_id).sockets.forEach((socket) => {
                update_perms(socket);
            });
        });


        socket.on("disconnect", () => {
            log("[Api] Client (" + socket.id + ") disconnected");
        });
    });
}

