import { default as fs } from "fs";
import { pool_permissions } from "../app/database/sql.connection";
import { fontsData } from "../app/database/sql.typings";
import sqlts from "@rmp135/sql-ts";
import { config } from "../app/config";

const log = (...args: any[]) => console.log("[prebuild]", ...args);


const COMMENT = "/* This file is generated by scripts/prebuild.ts */\n/* eslint-disable */\n\n";

/** Script which is run before
 * the build command is executed.
 * 
 * At the moment, this script is used to:
 * - update the version number
 * - generate the imports for the fonts
 * - generate the typings for the sql database
*/
async function prebuild() {
    version_to_file();
    await create_fonts_import();
    await create_sql_typings();
}


function version_to_file() {
    const version = JSON.stringify(require("../package.json").version);
    const file = "./app/version.ts";
    const file_content = COMMENT + `export const LIB_VERSION = ${version};`;
    log(`Writing version ${version} to ${file}`);
    fs.writeFileSync(file, file_content);
}

async function create_fonts_import() {

    const fonts = await pool_permissions.query<fontsData[]>("SELECT * FROM fonts").then((rows) => {
        return rows;
    }).catch((err) => {
        console.error(err);
        return [];
    });


    // We need the fonts at two places, as import via css
    // and as import via js
    const file_js = "./app/database/fonts.ts";

    let file_content = COMMENT + "export const FONT_URLS: [URL,string][] = [\n";
    for (const font of fonts) {
        file_content += `\t[new URL("@fonts/${font.fname}", import.meta.url), "${font.name}"],\n`;
    }
    file_content += "];\n\n";

    file_content += "export type AllowedFonts ="
    for (const font of fonts) {
        file_content += ` "${font.name}" |`;
    }
    file_content = file_content.slice(0, -1) + ";\n";

    log(`Writing fonts mapping to ${file_js}`);
    fs.writeFileSync(file_js, file_content);


    const file_css = "./app/components/styles/fonts.scss";
    let file_content_css = COMMENT;
    for (const font of fonts) {
        file_content_css += `@font-face {\n\tfont-family: "${font.name}";\n\tsrc: url("/fonts/${font.fname}");\n}\n`;
    }

    log(`Writing fonts mapping to ${file_css}`);
    fs.writeFileSync(file_css, file_content_css);

}

async function create_sql_typings() {
    const sql_ts_config = {
        client: "mysql",
        connection: {
            host: config.MYSQL.HOST,
            user: config.MYSQL.USER,
            password: config.MYSQL.PASSWORD,
            database: "snip_data",
        },
        interfaceNameFormat: "${table}Data", // User becomes UserModel
        typeOverrides: {
            "snip_perms.acl.pRead": "boolean",
            "snip_perms.acl.pWrite": "boolean",
            "snip_perms.acl.pDelete": "boolean",
            "snip_perms.acl.pACL": "boolean",
            "snip_perms.users.emailValidated": "boolean",
        },
        typeMap: {
            tinyint: ["boolean"],
        },
        tableEnums: {
            "snip_data.background_types": {
                key: "name",
                value: "id",
            },
        },
    };
    // Get the SQL schema as typescript interfaces
    const data = await sqlts.toTypeScript(sql_ts_config);
    // Write the typescript interfaces to a file
    log("Writing SQL typings to ./app/database/sql.typings.ts");
    fs.writeFileSync("./app/database/sql.typings.ts", data);
}

(async () => {
    log("Starting prebuild");
    await prebuild();
    log("Finished prebuild");
    //quit
    process.exit(0);
}
)();
